/* $Header: d:/cvsroot/tads/html/htmlpng.h,v 1.1.1.1 1999/05/09 21:32:27 MJRoberts Exp $ */

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmlpng.h - PNG image object
Function
  
Notes
  
Modified
  11/08/97 MJRoberts  - Creation
*/

#ifndef HTMLPNG_H
#define HTMLPNG_H

/* include PNG library header */
#include <png.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif


class CHtmlPng
{
public:
    CHtmlPng();
    ~CHtmlPng();

    /* load an image from a file */
    int read_png_file(const textchar_t *filename,
                      unsigned long seekpos, unsigned long filesize,
                      int convert_256_color);

    /* get the palette - returns null for RGB images */
    HTML_color_t *get_palette() const { return palette_; }
    unsigned int get_palette_size() const { return palette_size_; }

    /* get the dimensions */
    unsigned long get_height() const { return height_; }
    unsigned long get_width() const { return width_; }

    /* get the number of bytes per row */
    unsigned long get_row_bytes() const { return row_bytes_; }

    /* get the row pointers */
    const png_byte *const *get_rows() const { return rows_; }

private:
    /* allocate memory for the image */
    void allocate_rows(png_struct *png_ptr,
                       png_info *info_ptr,
                       unsigned long height);
    
    /* delete the existing image, if any */
    void delete_image();

    /* PNG callback function to read from the file */
    static void fread_cb(png_structp, png_bytep, png_size_t);

    /* palette, if the image has been converted to a palette-index format */
    HTML_color_t *palette_;
    unsigned int palette_size_;
    
    /* pointers to the rows containing the image bytes */
    png_bytep *rows_;

    /* number of rows allocated */
    unsigned long height_;

    /* width of each row */
    unsigned long width_;

    /* bytes in each row */
    unsigned long row_bytes_;
};

#endif /* HTMLPNG_H */

