/* $Header: d:/cvsroot/tads/html/win32/tadsapp.h,v 1.1.1.1 1999/05/09 21:32:31 MJRoberts Exp $ */

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tadsapp.h - TADS Windows application object
Function
  This object provides a repository for data and methods associated
  with the application.
Notes
  
Modified
  09/16/97 MJRoberts  - Creation
*/

#ifndef TADSAPP_H
#define TADSAPP_H

#include <windows.h>

#ifndef TADSFONT_H
#include "tadsfont.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   Application object 
 */
class CTadsApp
{
public:
    /*
     *   Run the main event loop.  Returns either when *flag becomes
     *   non-zero, or when the application is terminated by some external
     *   or user action.  If the application is terminated, we'll return
     *   false; if the application should keep running, we'll return true.
     *   If accel and *accel are not null, we'll use *accel to translate
     *   accelerators and send resulting commands to the indicated window
     *   handle.  (We receive a pointer to an accelerator table, so that
     *   the window can change the accelerator table from time to time.)  
     */
    int event_loop(int *flag, const HACCEL *accel, HWND accel_win);

    /* do idle-time processing */
    void do_idle();

    /* create the application object */
    static void create_app(HANDLE inst, HANDLE pinst,
                           LPSTR cmdline, int cmdshow);

    /* get the main application object */
    static CTadsApp *get_app() { return app_; }

    /* delete the application object */
    static void delete_app();

    /* get the application instance handle */
    HANDLE get_instance() { return instance_; }

    /* 
     *   Get a font matching a given description.  If we already have a
     *   font matching the description, we'll return it and set *created
     *   to false; otherwise, we'll invoke the callback function (if
     *   provided; if not, we'll use a default creation function) to
     *   create the font, and set *created to true. 
     */
    class CTadsFont *get_font(const CTadsLOGFONT *font,
                              class CTadsFont *
                              (*create_func)(const CTadsLOGFONT *),
                              int *created);

    /*
     *   Get/set the directory for the standard file dialogs.  If the
     *   application wants to keep track of the directory last used for
     *   these dialogs without actually changing the directory, it can use
     *   these operations.  Simply set the lpstrInitialDir member of the
     *   dialog control structure to get_openfile_dir() before calling the
     *   open file function, and then call set_openfile_dir() with the
     *   resulting file to save the directory for next time. 
     */
    char *get_openfile_dir() const { return open_file_dir_; }
    void set_openfile_dir(const char *fname);

    /*
     *   Add a modeless dialog to our list.  Each modeless dialog that's
     *   running must be added to this list in order for keyboard events
     *   to be properly distributed to it.
     */
    void add_modeless(HWND win);

    /* 
     *   Remove a modeless dialog from our list.  A dialog that's been
     *   added to our list must be removed before its window is destroyed,
     *   so that we don't refer to it after it's gone.
     */
    void remove_modeless(HWND win);

    /*
     *   Set the MDI frame window.  This should be called when the MDI
     *   frame is first created to set the MDI window.  This should also
     *   be called with a null argument before the MDI window is
     *   destroyed.  
     */
    void set_mdi_win(class CTadsSyswinMdiFrame *mdi_win)
        { mdi_win_ = mdi_win; }

private:
    CTadsApp();
    CTadsApp(HANDLE inst, HANDLE pinst, LPSTR cmdline, int cmdshow);

    virtual ~CTadsApp();

private:
    /* application instance handle */
    HANDLE  instance_;

    /* the global application object */
    static CTadsApp *app_;

    /*
     *   font cache - an array of fonts, the number allocated, and the
     *   number actually in use 
     */
    class CTadsFont **fontlist_;
    int fonts_allocated_;
    int fontcount_;

    /* open-file directory */
    char *open_file_dir_;

    /* 
     *   our list of modeless dialogs: space for the window handles, the
     *   count of the number of slots allocated, and the number of slots
     *   actually in use 
     */
    HWND *modeless_dlgs_;
    int modeless_dlg_alloc_;
    int modeless_dlg_cnt_;

    /* flag indicating that OLE was initialized successfully */
    int ole_inited_ : 1;

    /* MDI frame window, if this is an MDI application */
    class CTadsSyswinMdiFrame *mdi_win_;
};


#endif /* TADSAPP_H */

