#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/tadspng.cpp,v 1.1.1.1 1999/05/09 21:32:31 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tadspng.cpp - Win32 PNG object implementation
Function
  
Notes
  
Modified
  11/08/97 MJRoberts  - Creation
*/

#include <windows.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef TADSWIN_H
#include "tadswin.h"
#endif
#ifndef HTMLPNG_H
#include "htmlpng.h"
#endif
#ifndef TADSPNG_H
#include "tadspng.h"
#endif


CTadsPng::CTadsPng()
{
}

CTadsPng::~CTadsPng()
{
}

/*
 *   Load the iamge from a PNG object 
 */
int CTadsPng::load_from_png(CHtmlPng *png)
{
    /* drop any previously loaded image */
    delete_image();

    /* create a logical palette from the PNG's palette, if it has one */
    create_palhdl(png->get_palette(), png->get_palette_size());

    /* create a DWORD-aligned pixel buffer from the PNG rows */
    if (create_pix_dword_aligned(png->get_rows(),
                                 png->get_width(),
                                 png->get_row_bytes(),
                                 png->get_height()))
        return 1;

    /* success */
    return 0;
}

/*
 *   Create a DWORD-aligned pixel buffer from an array of PNG rows 
 */
int CTadsPng::create_pix_dword_aligned(const png_byte *const *src_rows,
                                       unsigned long width_pix,
                                       unsigned long width_bytes,
                                       unsigned long height)
{
    long out_width_bytes;
    long in_width_bytes;
    long width_bits;
    long new_size;
    unsigned long row;
    OS_HUGEPTR(unsigned char) dst;
    const png_byte *const *src;
    int in_bytes_per_pixel;
    int out_bytes_per_pixel;

    /* store the width and height */
    width_ = width_pix;
    height_ = height;

    /* if there are no rows, we can't do anything */
    if (src_rows == 0)
        return 1;

    /* figure out how many bytes per pixel */
    in_width_bytes = width_bytes;
    in_bytes_per_pixel = in_width_bytes / width_pix;
    out_bytes_per_pixel = (in_bytes_per_pixel > 3 ? 3 : in_bytes_per_pixel);

    /* calculate the smallest DWORD-aligned buffer we can use */
    width_bits = width_pix * out_bytes_per_pixel * 8;
    out_width_bytes = ((width_bits + 31) / 32) * 4;

    /* calculate the size we'll need for the DWORD-aligned buffer */
    new_size = (unsigned long)out_width_bytes * (unsigned long)height;

    /* allocate our buffer */
    pix_ = (OS_HUGEPTR(unsigned char))os_alloc_huge(new_size);
    if (pix_ == 0)
        return 1;

    /* copy the rows */
    dst = os_add_huge(pix_, out_width_bytes * (height - 1));
    if (in_bytes_per_pixel == out_bytes_per_pixel)
    {
        /* the pixels are the same size, so we can copy a row at a time */
        for (src = src_rows, row = 0 ; row < height ; ++row, ++src)
        {
            /* copy this row */
            memcpy(dst, *src, in_bytes_per_pixel * width_pix);

            /* move to next row of output */
            dst = os_add_huge(dst, -out_width_bytes);
        }
    }
    else
    {
        /* 
         *   we need to adjust the size of each pixel, so copy only one
         *   pixel at a time 
         */
        for (src = src_rows, row = 0 ; row < height ; ++row, ++src)
        {
            unsigned long i;
            const png_byte *srcp;
            unsigned char *dstp;
            
            /* copy this row */
            for (i = 0, srcp = *src, dstp = dst ; i < width_pix ;
                 ++i, srcp += in_bytes_per_pixel, dstp += out_bytes_per_pixel)
            {
                /* copy this pixel */
                memcpy(dstp, srcp, out_bytes_per_pixel);
            }
            
            /* move to next row */
            dst = os_add_huge(dst, -out_width_bytes);
        }
    }

    /* remember the byte width */
    width_bytes_ = out_width_bytes;

    /* success */
    return 0;
}

