#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/w32dbgpr.cpp,v 1.1.1.1 1999/05/09 21:32:32 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  w32dbgpr.cpp - win32 debugger preferences dialog
Function
  
Notes
  
Modified
  03/14/98 MJRoberts  - Creation
*/

#include <string.h>
#include <stdlib.h>

#include <Windows.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef W32DBGPR_H
#include "w32dbgpr.h"
#endif
#ifndef TADSDLG_H
#include "tadsdlg.h"
#endif
#ifndef TADSAPP_H
#include "tadsapp.h"
#endif
#ifndef TADSCBTN_H
#include "tadscbtn.h"
#endif
#ifndef HTMLRES_H
#include "htmlres.h"
#endif
#ifndef HTMLDCFG_H
#include "htmldcfg.h"
#endif
#ifndef W32TDB_H
#include "w32tdb.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   Debugger preferences page dialog for source window options 
 */

class CHtmlDialogDebugSrcwin: public CTadsDialogPropPage
{
public:
    CHtmlDialogDebugSrcwin(CHtmlSys_dbgmain *mainwin,
                           CHtmlDebugConfig *config)
    {
        mainwin_ = mainwin;
        config_ = config;
    }
    
    void init();
    int do_command(WPARAM cmd, HWND ctl);
    int do_notify(NMHDR *nm);

private:
    /* read a popup's current value into a configuration variable */
    int read_popup(int id, const textchar_t *element);
    
    /* color settings */
    COLORREF cust_colors_[16];
    HTML_color_t text_color_;
    HTML_color_t bkg_color_;
    int use_win_colors_;

    /* font settings */
    CStringBuf font_;
    int font_size_;

    /* tab spacing */
    int tab_size_;

    /* configuration object */
    CHtmlDebugConfig *config_;

    /* main window */
    CHtmlSys_dbgmain *mainwin_;
};

/*
 *   Read a popup 
 */
/*
 *   read the contents of one of the font popups into a CStringBuf 
 */
int CHtmlDialogDebugSrcwin::read_popup(int id, const textchar_t *element)
{
    HWND ctl;
    char buf[128];
    textchar_t oldval[128];
    size_t len;
    int idx;
    int changed;

    /* get the control handle */
    ctl = GetDlgItem(handle_, id);

    /* get the selection number */
    idx = SendMessage(ctl, CB_GETCURSEL, 0, 0);

    /* get the selected string */
    len = SendMessage(ctl, CB_GETLBTEXT, (WPARAM)idx, (LPARAM)buf);

    /* get the old value from the configuration */
    if (config_->getval("src win options", element,
                        0, oldval, sizeof(oldval)))
        oldval[0] = '\0';
    
    /* note whether this is a change */
    changed = (len != get_strlen(oldval)
               || memcmp(buf, oldval, len) != 0);

    /* store it in the configuration */
    config_->setval("src win options", element, 0, buf, len);

    /* return change indication */
    return changed;
}


/*
 *   Initialize 
 */
void CHtmlDialogDebugSrcwin::init()
{
    textchar_t buf[256];
    static const textchar_t *fontsizes[] =
    {
        "4", "5", "6", "8", "9", "10", "11", "12", "14", "16", "18",
        "20", "24", "28", "36", 0
    };
    static const textchar_t *tabsizes[] =
    {
        "1", "2", "3", "4", "5", "6", "7", "8", "10", "16", 0
    };
    const textchar_t **p;
    int idx;
    
    /* get the initial color settings */
    if (config_->getval_color("src win options", "text color", &text_color_))
        text_color_ = HTML_make_color(0, 0, 0);
    if (config_->getval_color("src win options", "bkg color", &bkg_color_))
        bkg_color_ = HTML_make_color(0xff, 0xff, 0xff);

    /* get the initial use-windows-colors setting */
    if (config_->getval("src win options", "use win colors",
                        &use_win_colors_))
        use_win_colors_ = TRUE;

    /* get the custom colors from the preferences */
    if (config_->getval_bytes("src win options", "custom colors",
                              cust_colors_, sizeof(cust_colors_)))
        memset(cust_colors_, 0, sizeof(cust_colors_));

    /* get the initial font from the preferences */
    if (config_->getval("src win options", "font name", 0, buf, sizeof(buf)))
        font_.set("Courier New");
    else
        font_.set(buf);
    
    /* get the initial font size from the preferences */
    if (config_->getval("src win options", "font size", &font_size_))
        font_size_ = 10;

    /* set up the color buttons */
    controls_.add(new CColorBtnPropPage(handle_, IDC_BTN_DBGTXTCLR,
                                        &text_color_, cust_colors_, this));
    controls_.add(new CColorBtnPropPage(handle_, IDC_BTN_DBGBKGCLR,
                                        &bkg_color_, cust_colors_, this));

    /* get the tab spacing */
    if (config_->getval("src win options", "tab size", &tab_size_))
        tab_size_ = 8;

    /* set the use-window-colors checkbox */
    CheckDlgButton(handle_, IDC_CK_DBGWINCLR,
                   (use_win_colors_ ? BST_CHECKED : BST_UNCHECKED));
    
    /* 
     *   enable or disable the text and background buttons according to
     *   whether the "use windows colors" checkbox is checked 
     */
    EnableWindow(GetDlgItem(handle_, IDC_BTN_DBGTXTCLR), !use_win_colors_);
    EnableWindow(GetDlgItem(handle_, IDC_BTN_DBGBKGCLR), !use_win_colors_);

    /* set up the font combo with a list of the available fonts */
    init_font_popup(IDC_CBO_DBGFONT, FALSE, TRUE, font_.get(), TRUE);

    /* set the initial font size field */
    sprintf(buf, "%d", font_size_);
    SetWindowText(GetDlgItem(handle_, IDC_CBO_DBGFONTSIZE), buf);

    /* set the initial tab size field */
    sprintf(buf, "%d", tab_size_);
    SetWindowText(GetDlgItem(handle_, IDC_CBO_DBGTABSIZE), buf);

    /* load up the suggested font sizes list */
    for (idx = 0, p = fontsizes ; *p != 0 ; ++p, ++idx)
        SendMessage(GetDlgItem(handle_, IDC_CBO_DBGFONTSIZE),
                    CB_INSERTSTRING, idx, (LPARAM)*p);

    /* load up the suggested tab sizes list */
    for (idx = 0, p = tabsizes ; *p != 0 ; ++p, ++idx)
        SendMessage(GetDlgItem(handle_, IDC_CBO_DBGTABSIZE),
                    CB_INSERTSTRING, idx, (LPARAM)*p);
}

int CHtmlDialogDebugSrcwin::do_command(WPARAM cmd, HWND ctl)
{
    switch(LOWORD(cmd))
    {
    case IDC_CK_DBGWINCLR:
        {
            int checked;

            /* get the state of the button */
            checked = (IsDlgButtonChecked(handle_, IDC_CK_DBGWINCLR)
                       == BST_CHECKED);

            /* 
             *   enable or disable the text and background buttons
             *   accordingly 
             */
            EnableWindow(GetDlgItem(handle_, IDC_BTN_DBGTXTCLR), !checked);
            EnableWindow(GetDlgItem(handle_, IDC_BTN_DBGBKGCLR), !checked);

            /* note the state change */
            set_changed(TRUE);
        }
        return TRUE;

    case IDC_CBO_DBGFONT:
    case IDC_CBO_DBGFONTSIZE:
    case IDC_CBO_DBGTABSIZE:
        switch(HIWORD(cmd))
        {
        case CBN_EDITCHANGE:
        case CBN_SELCHANGE:
            /* note the change to the value */
            set_changed(TRUE);
            break;

        default:
            /* ignore other notifications */
            break;
        }
        return TRUE;

    default:
        /* inherit default behavior */
        return CTadsDialogPropPage::do_command(cmd, ctl);
    }
}

int CHtmlDialogDebugSrcwin::do_notify(NMHDR *nm)
{
    char buf[256];
    
    switch(nm->code)
    {
    case PSN_APPLY:
        /* save the color settings */
        config_->setval_color("src win options", "text color", text_color_);
        config_->setval_color("src win options", "bkg color", bkg_color_);
        config_->setval_bytes("src win options", "custom colors",
                              cust_colors_, sizeof(cust_colors_));

        /* save the use-windows-colors setting */
        use_win_colors_ = (IsDlgButtonChecked(handle_, IDC_CK_DBGWINCLR)
                           == BST_CHECKED);
        config_->setval("src win options", "use win colors",
                        use_win_colors_);

        /* save the font */
        read_popup(IDC_CBO_DBGFONT, "font name");

        /* save the font size */
        GetWindowText(GetDlgItem(handle_, IDC_CBO_DBGFONTSIZE),
                      buf, sizeof(buf));
        config_->setval("src win options", "font size", (int)get_atol(buf));

        /* save the tab size */
        GetWindowText(GetDlgItem(handle_, IDC_CBO_DBGTABSIZE),
                      buf, sizeof(buf));
        config_->setval("src win options", "tab size", (int)get_atol(buf));

        /* tell the main window that we've updated the preferences */
        mainwin_->on_update_srcwin_options();

        /* done */
        return TRUE;
            
    default:
        /* inherit default behavior */
        return CTadsDialogPropPage::do_notify(nm);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Run the debugger preferences dialog 
 */
void run_dbg_prefs_dlg(HWND owner, CHtmlSys_dbgmain *mainwin)
{
    PROPSHEETPAGE psp[1];
    PROPSHEETHEADER psh;
    CHtmlDialogDebugSrcwin *srcwin_dlg;
    void *ctx;

    /* create our pages */
    srcwin_dlg = new CHtmlDialogDebugSrcwin(mainwin, mainwin->get_config());

    /* set up the property page descriptor for the source window page */
    psp[0].dwSize = sizeof(PROPSHEETPAGE);
    psp[0].dwFlags = PSP_USETITLE;
    psp[0].hInstance = CTadsApp::get_app()->get_instance();
    psp[0].pszTemplate = MAKEINTRESOURCE(DLG_DEBUG_SRCWIN);
    psp[0].pszIcon = 0;
    psp[0].pszTitle = MAKEINTRESOURCE(IDS_DEBUG_SRCWIN);
    srcwin_dlg->prepare_prop_page(&psp[0]);

    /* set up the main dialog descriptor */
    psh.dwSize = sizeof(PROPSHEETHEADER);
    psh.dwFlags = PSH_USEICONID | PSH_PROPSHEETPAGE;
    psh.hwndParent = owner;
    psh.hInstance = CTadsApp::get_app()->get_instance();
    psh.pszIcon = 0;
    psh.pszCaption = (LPSTR)"Debugger Options";
    psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
    psh.nStartPage = 0;
    psh.ppsp = (LPCPROPSHEETPAGE)&psp;
    psh.pfnCallback = 0;

    /* disable windows owned by the main window before showing the dialog */
    ctx = CTadsDialog::modal_dlg_pre(mainwin->get_handle(), FALSE);

    /* run the property sheet */
    PropertySheet(&psh);

    /* re-enable windows */
    CTadsDialog::modal_dlg_post(ctx);

    /* delete the dialogs */
    delete srcwin_dlg;
}

