#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/w32tdb.cpp,v 1.1.1.1 1999/05/09 21:32:32 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  w32tdb.cpp - win32 HTML TADS debugger configuration 
Function
  
Notes
  
Modified
  01/31/98 MJRoberts  - Creation
*/

#include <ctype.h>

#include <Windows.h>
#include <commctrl.h>

/* include TADS runtime interface */
#include "trd.h"

/* include TADS debugger interface */
#include "dbg.h"

/* include TADS OS interface */
#include "os.h"

/* include TADS console I/O interface */
#include "tio.h"

#ifndef W32MAIN_H
#include "w32main.h"
#endif
#ifndef HTML_OS_H
#include "html_os.h"
#endif
#ifndef TADSWIN_H
#include "tadswin.h"
#endif
#ifndef HTMLW32_H
#include "htmlw32.h"
#endif
#ifndef W32TDB_H
#include "w32tdb.h"
#endif
#ifndef HTMLPREF_H
#include "htmlpref.h"
#endif
#ifndef TADSAPP_H
#include "tadsapp.h"
#endif
#ifndef HTMLRES_H
#include "htmlres.h"
#endif
#ifndef HTMLDBG_H
#include "htmldbg.h"
#endif
#ifndef HTMLFMT_H
#include "htmlfmt.h"
#endif
#ifndef W32EXPR_H
#include "w32expr.h"
#endif
#ifndef HTMLDCFG_H
#include "htmldcfg.h"
#endif
#ifndef W32BPDLG_H
#include "w32bpdlg.h"
#endif
#ifndef W32FNDLG_H
#include "w32fndlg.h"
#endif
#ifndef HTMLTXAR_H
#include "htmltxar.h"
#endif
#ifndef W32DBGPR_H
#include "w32dbgpr.h"
#endif
#ifndef TADSDLG_H
#include "tadsdlg.h"
#endif
#ifndef HTMLDISP_H
#include "htmldisp.h"
#endif
#ifndef FOLDSEL_H
#include "foldsel.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   Define startup configuration variables for the TADS debugger
 */
int (*w32_tadsmain)(int, char **, struct appctxdef *, char *) = tddmain;
char *w32_beforeopts = "i";
char *w32_configfile = "config.tdh";
int w32_allow_debugwin = FALSE;
int w32_always_pause_on_exit = TRUE;
char *w32_setup_reg_val_name = "Debugger Setup Done";
char *w32_usage_app_name = "htmltdb";


/* ------------------------------------------------------------------------ */
/*
 *   Static list of directories we've seen in dbgu_find_src().  Each time
 *   we find a file, we'll remember the directory where we found it, since
 *   it's likely that there will be more files in the same place.  Then,
 *   if we're asked to locate another file, we'll check in all of the
 *   places we've found files previously before bothering the user with a
 *   new dialog.  
 */
class CDbg_win32_srcdir
{
public:
    /* the directory prefix */
    CStringBuf dir_;

    /* next item in our list */
    CDbg_win32_srcdir *next_;
};

/* head of our list of directories */
static CDbg_win32_srcdir *srcdirs_ = 0;


/* ------------------------------------------------------------------------ */
/*
 *   Debugger user interface entrypoints.  The TADS code calls these
 *   entrypoints to let the debugger user interface take control at
 *   appropriate points. 
 */

/*
 *   This UI implementation allows moving the instruction pointer, so we
 *   can resume from an error. 
 */
int dbgu_err_resume(dbgcxdef *)
{
    return TRUE;
}

/*
 *   Debugger user interface initialization.  The debugger calls this
 *   routine during startup to let the user interface set itself up.  
 */
void dbguini(dbgcxdef *ctx, const char *game_filename)
{
    /* allocate the main debugger control window */
    ctx->dbgcxui = (void *)new CHtmlSys_dbgmain(ctx, game_filename);
}

/*
 *   Debugger user interface initialization, phase two.  The debugger
 *   calls this routine during startup, after the .GAM file has been
 *   loaded, to let the user interface do any additional setup that
 *   depends on the .GAM file having been loaded. 
 */
void dbguini2(dbgcxdef *ctx)
{
    /* tell the main control window to finish initialization */
    ((CHtmlSys_dbgmain *)ctx->dbgcxui)->post_load_init(ctx);
}

/*
 *   Debugger user interface termination 
 */
void dbguterm(dbgcxdef *ctx)
{
    /* close the main debugger control window */
    if (ctx->dbgcxui != 0)
    {
        ((CHtmlSys_dbgmain *)ctx->dbgcxui)->terminate_debugger();
        ctx->dbgcxui = 0;
    }

    /* delete all of the source directories we might have allocated */
    while (srcdirs_ != 0)
    {
        CDbg_win32_srcdir *nxt;

        /* remember the next item */
        nxt = srcdirs_->next_;

        /* delete this item */
        delete srcdirs_;

        /* move on to the next item */
        srcdirs_ = nxt;
    }
}

/*
 *   Debugger user interface - display an error 
 */
void dbguerr(dbgcxdef *ctx, int errno, char *msg)
{
    /* display the message in the debug log window */
    oshtml_dbg_printf("[TADS-%d: %s]\n", errno, msg);

    /* display it in the main game window as well */
    os_printf("[TADS-%d: %s]\n", errno, msg);
}

/*
 *   Callback for the folder selector dialog.  We'll look in the selected
 *   path for the desired file, and if it exists allow immediate dismissal
 *   of the dialog.  If they clicked the OK button on a folder that
 *   doesn't contain our file, we'll show a message box complaining about
 *   it and refuse to allow dismissal of the dialog.  
 */
int dbgu_find_src_cb(void *ctx, const char *path, int okbtn, HWND dlg)
{
    const char *origname;
    char buf[OSFNMAX + 128];
    size_t len;
    osfildef *fp;
    
    /* get the original name - we used this as the callback context */
    origname = (const char *)ctx;

    /* get the path */
    len = strlen(path);
    memcpy(buf, path, len);

    /* append backslash if it doesn't end in one already */
    if (len == 0 || buf[len-1] != '\\')
        buf[len++] = '\\';

    /* append the filename */
    strcpy(buf + len, origname);

    /* check to see if the file exists */
    fp = osfoprs(buf, OSFTTEXT);
    if (fp != 0)
    {
        /* 
         *   found it - close the file, and dismiss the dialog (by
         *   returning true) 
         */
        osfcls(fp);
        return TRUE;
    }

    /* 
     *   We didn't find the file - we can't dismiss the dialog yet.  If
     *   the user clicked OK, complain about it.  (Don't complain if they
     *   merely double-clicked a directory path, since they may just be
     *   navigating.)  
     */
    if (okbtn)
    {
        sprintf(buf, "This folder does not contain the file \"%s\".",
                origname);
        MessageBox(dlg, buf, "File Not Found", MB_OK | MB_ICONEXCLAMATION);
    }

    /* don't allow dismissal */
    return FALSE;
}

/*
 *   Find a source file using UI-dependent mechanisms.  If we aren't
 *   required to find the file, we'll do nothing except return success, to
 *   tell the caller to defer asking us to find the file until they really
 *   need it.  If we are required to find the file, we'll use a standard
 *   file open dialog to ask the user to locate the file. 
 */
int dbgu_find_src(char *origname, int origlen,
                  char *fullname, size_t full_len, int must_find_file)
{
    char title_buf[OSFNMAX + 128];
    char prompt_buf[OSFNMAX + 128];
    CDbg_win32_srcdir *srcdir;
    size_t len;

    /* 
     *   if we don't *really* need to find it, return success, to indicate
     *   to the caller that they should ask us again when it's actually
     *   important -- this way, we'll avoid pestering the user with a
     *   bunch of file-open requests on startup, and instead ask the user
     *   for files one at a time as they're actually needed, which is much
     *   less irritating 
     */
    if (!must_find_file)
    {
        fullname[0] = '\0';
        return TRUE;
    }

    /*
     *   They do require the file.  First, check our cached directory
     *   list, to see if it's in a directory we've already visited.  
     */
    for (srcdir = srcdirs_ ; srcdir != 0 ; srcdir = srcdir->next_)
    {
        osfildef *fp;

        /* build this name */
        strcpy(fullname, srcdir->dir_.get());
        strcat(fullname, origname);

        /* open the file to see if it exists */
        fp = osfoprs(fullname, OSFTTEXT);
        if (fp != 0)
        {
            /* this is the one - close it and return success */
            osfcls(fp);
            return TRUE;
        }
    }

    /*
     *   We couldn't find it in our cached directory list.  Put up the
     *   folder-browser dialog to ask the user to locate the file.
     */
    sprintf(title_buf, "%s - Please Locate File", origname);
    sprintf(prompt_buf, "&Path to file \"%s\":", origname);
    if (!CTadsDialogFolderSel2::
        run_select_folder(0, CTadsApp::get_app()->get_instance(),
                          prompt_buf, title_buf, fullname, full_len, 0,
                          dbgu_find_src_cb, origname))
    {
        /* they cancelled - return failure */
        fullname[0] = '\0';
        return FALSE;
    }
        
    /* if the prefix doesn't end in a backslash, add the backslash */
    len = get_strlen(fullname);
    if (len == 0 || fullname[len-1] != '\\')
        fullname[len++] = '\\';
    
    /* append the filename to the path */
    strcpy(fullname + len, origname);
    
    /* add the prefix to our list */
    srcdir = new CDbg_win32_srcdir;
    srcdir->dir_.set(fullname, len);
    srcdir->next_ = srcdirs_;
    srcdirs_ = srcdir;
    
    /* success */
    return TRUE;
}


/*
 *   Debugger user interface main command entrypoint
 */
void dbgucmd(dbgcxdef *ctx, int bphit, int err, unsigned int *exec_ofs)
{
    int restart;
    int abort;
    
    /* invoke the debugger event loop in the debugger control window */
    restart = abort = FALSE;
    if (ctx->dbgcxui != 0
        && !(((CHtmlSys_dbgmain *)ctx->dbgcxui)->
             dbg_event_loop(ctx, bphit, err, &restart, &abort, exec_ofs)))
    {
        /* 
         *   the application was terminated within the event loop - signal
         *   a TADS error so that TADS shuts itself down 
         */
        CHtmlDebugHelper::signal_quit(ctx);
    }

    /* if the event loop said to restart, do so now */
    if (restart)
        CHtmlDebugHelper::signal_restart(ctx);

    /* if the event loop said to abort, do so now */
    if (abort)
        CHtmlDebugHelper::signal_abort(ctx);
}

/*
 *   Quitting.  Enter the debugger user interface event loop as normal.
 *   If the event loop says to keep going, restart the game. 
 */
void dbguquitting(dbgcxdef *ctx)
{
    int restart;
    int abort;
    CHtmlSys_dbgmain *mainwin = (CHtmlSys_dbgmain *)ctx->dbgcxui;

    /* note that the game is inactive */
    mainwin->set_game_over(TRUE);

    /* show a message about the game quitting */
    MessageBox(0, "The game has terminated.", "TADS Debugger",
               MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);

    /* invoke the debugger event loop in the debugger control window */
    restart = abort = FALSE;
    if (ctx->dbgcxui != 0
        && !mainwin->dbg_event_loop(ctx, 0, 0, &restart, &abort, 0))
    {
        /* 
         *   they didn't restart, so let the system terminate -- all we
         *   need to do is return without signalling a restart 
         */
        return;
    }

    /* they didn't quit, so the game is running again */
    mainwin->set_game_over(FALSE);

    /* put the game in single-step mode */
    mainwin->exec_step_into();

    /* restart the game */
    CHtmlDebugHelper::signal_restart(ctx);
}

/*
 *   Turn hidden output tracking on/off 
 */
void trchid()
{
    outflushn(1);
    os_printf("---Hidden output---");
    outflushn(1);
}

void trcsho()
{
    outflushn(1);
    os_printf("---End of hidden output, normal output resumes---");
    outflushn(1);
}

/* ------------------------------------------------------------------------ */
/*
 *   Debugger main frame window implementation
 */

CHtmlSys_dbgmain::CHtmlSys_dbgmain(dbgcxdef *ctx, const char *game_filename)
{
    RECT rc;
    RECT deskrc;
    CHtmlSys_mainwin *main_win;
    CHtmlRect winpos;
    textchar_t config_filename[OSFNMAX];
    int vis;
    int i;
    CDbg_win32_srcdir *srcdir;

    /* no toolbar yet */
    toolbar_ = 0;

    /* remember the context */
    dbgctx_ = ctx;

    /* create a helper object */
    helper_ = new CHtmlDebugHelper();

    /* create my configuration object */
    config_ = new CHtmlDebugConfig();

    /* 
     *   build the configuration file name -- use the game file name, with
     *   the extension replaced by TDC (TADS debugger Configuration) 
     */
    do_strcpy(config_filename, game_filename);
    os_remext(config_filename);
    os_addext(config_filename, "TDC");
    config_filename_.set(config_filename);

    /* read the existing configuration file */
    config_->load_file(config_filename_.get());

    /* load the source directory list */
    i = config_->get_strlist_cnt("srcdir", 0);
    while (i > 0)
    {
        textchar_t buf[OSFNMAX];
        
        /* load the next source directory */
        --i;
        if (!config_->getval("srcdir", 0, i, buf, sizeof(buf)))
        {
            /* create the new entry */
            srcdir = new CDbg_win32_srcdir;
            srcdir->dir_.set(buf);

            /* add it to the list */
            srcdir->next_ = srcdirs_;
            srcdirs_ = srcdir;
        }
    }

    /* set initial configuration options */
    on_update_srcwin_options();

    /* not yet quitting, restarting, or aborting */
    quitting_ = FALSE;
    restarting_ = FALSE;
    aborting_cmd_ = FALSE;

    /* the game starts off running */
    game_over_ = FALSE;

    /* use the main window's preferences object */
    main_win = CHtmlSys_mainwin::get_main_win();
    prefs_ = (main_win != 0 ? main_win->get_prefs() : 0);

    /* read my stored position from the preferences */
    if (config_->getval("main win pos", 0, &winpos))
        w32tdb_set_default_frame_rc(&rc);
    else
        SetRect(&rc, winpos.left, winpos.top,
                winpos.right - winpos.left, winpos.bottom - winpos.top);

    /* open my window */
    create_system_window(0, TRUE, "TADS Debugger", &rc,
                         w32tdb_new_frame_syswin(this, -2));

    /* 
     *   Get the desktop window area, so that we can set up default
     *   positions for our windows at the edges of the desktop area.  This
     *   will retrieve the area of the entire Windows desktop in SDI mode,
     *   or of the MDI client area in MDI mode.  
     */
    w32tdb_get_desktop_rc(this, &deskrc);

    /* load my accelerators */
    accel_ = LoadAccelerators(CTadsApp::get_app()->get_instance(),
                              MAKEINTRESOURCE(IDR_ACCEL_DEBUGMAIN));

    /* no active window yet */
    active_dbg_win_ = 0;
    last_active_dbg_win_ = 0;
    post_menu_active_win_ = 0;

    /* create the debug log window */
    helper_->create_debuglog_win(this);

    /* create the stack window if appropriate */
    if (config_->getval("stack win vis", 0, &vis) || vis)
        helper_->create_stack_win(this);

    /* create the watch window */
    watchwin_ = new CHtmlDbgWatchWinExpr();
    watchwin_->init_panels(this, helper_);
    SetRect(&rc, deskrc.left + 2 + 300 + 2, deskrc.bottom - 2 - 150,
            300, 150);
    watchwin_->load_win_config(config_, "watch win", &rc, &vis);
    watchwin_->create_system_window(this, vis, "Watch Expressions", &rc,
                                    w32tdb_new_child_syswin(watchwin_));
    watchwin_->load_panel_config(config_, "watch win");

    /* create the local variables window */
    localwin_ = new CHtmlDbgWatchWinLocals();
    localwin_->init_panels(this, helper_);
    SetRect(&rc, deskrc.left + 2, deskrc.bottom - 2 - 150, 300, 150);
    localwin_->load_win_config(config_, "locals win", &rc, &vis);
    localwin_->create_system_window(this, vis, "Locals", &rc,
                                    w32tdb_new_child_syswin(localwin_));

    /* create the history window if appropriate */
    if (!config_->getval("hist win vis", 0, &vis) && vis)
        helper_->create_hist_win(this);

    /* create the find dialog */
    find_dlg_ = new CTadsDialogFind();

    /* no "Find" text yet (so we can't "find next") */
    find_text_[0] = '\0';

    /* load search configuration */
    if (config_->getval("search", "exact case", &search_exact_case_))
        search_exact_case_ = FALSE;
    if (config_->getval("search", "wrap", &search_wrap_))
        search_wrap_ = TRUE;
    if (config_->getval("search", "dir", &search_dir_))
        search_dir_ = 1;
    if (config_->getval("search", "start at top", &search_from_top_))
        search_from_top_ = FALSE;
    config_->getval("search", "text", 0, find_text_, sizeof(find_text_));
    find_dlg_->load_config(config_, "search dlg");
}

/*
 *   additional initialization after the .GAM file has been loaded 
 */
void CHtmlSys_dbgmain::post_load_init(dbgcxdef *ctx)
{
    int i;
    
    /* 
     *   load the file menu, now that we know the source files (references
     *   are embedded in the .GAM file, which we've now loaded) 
     */
    load_file_menu();

    /*
     *   Load the source windows that were open at the end of the last
     *   session, if any 
     */
    i = config_->get_strlist_cnt("src win list", 0);
    while (i > 0)
    {
        textchar_t buf[OSFNMAX];
        
        /* get the next window's name */
        --i;
        if (!config_->getval("src win list", 0, i, buf, sizeof(buf)))
        {
            /* open the file */
            helper_->open_text_file(ctx, this, buf);
        }
    }

    /* restore saved breakpoints */
    helper_->load_bp_config(ctx, config_);
}

/* 
 *   callback context structure for source window enumeration for saving
 *   source windows 
 */
struct enum_win_save_ctx_t
{
    int count_;
    CHtmlDebugConfig *config_;
};

CHtmlSys_dbgmain::~CHtmlSys_dbgmain()
{
    /* save the configuration */
    config_->save_file(config_filename_.get());

    /* done with the configuration object */
    delete config_;
    config_ = 0;

    /* delete my helper object */
    delete helper_;

    /* delete the "Find" dialog */
    delete find_dlg_;
}

/*
 *   window destruction 
 */
void CHtmlSys_dbgmain::do_destroy()
{
    /* remove the toolbar timer */
    if (toolbar_ != 0)
        rem_toolbar_proc(toolbar_);
    
    /* inherit default handling */
    CHtmlSys_framewin::do_destroy();
}

/*
 *   callback for enumerating open source windows and saving them 
 */
void CHtmlSys_dbgmain::enum_win_save_cb(void *ctx0, int /*idx*/,
                                        CHtmlSysWin *win,
                                        int /*line_source_id*/,
                                        HtmlDbg_wintype_t win_type)
{
    enum_win_save_ctx_t *ctx = (enum_win_save_ctx_t *)ctx0;
    textchar_t win_title[256];

    /* 
     *   don't bother saving anything for special windows, such as the
     *   stack and the call history window -- save only regular text file
     *   windows 
     */
    if (win_type != HTMLDBG_WINTYPE_SRCFILE)
        return;

    /* get the title of this window */
    GetWindowText(((CHtmlSysWin_win32 *)win)->get_frame_handle(),
                  win_title, sizeof(win_title));

    /* save it in the list */
    ctx->config_->setval("src win list", 0, ctx->count_, win_title);

    /* count it */
    ctx->count_++;
}

/*
 *   window creation
 */
void CHtmlSys_dbgmain::do_create()
{
    HMENU mnu;
    TBBUTTON buttons[30];
    int i;

    /* inherit default */
    CHtmlSys_framewin::do_create();

    /* load my menu */
    mnu = LoadMenu(CTadsApp::get_app()->get_instance(),
                   MAKEINTRESOURCE(IDR_DEBUGMAIN_MENU));
    set_win_menu(mnu, -2);

    /* 
     *   initialize the button descriptors 
     */
    i = 0;

    /* GO button */
    buttons[i].iBitmap = 0;
    buttons[i].idCommand = ID_DEBUG_GO;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* STEP INTO button */
    buttons[i].iBitmap = 1;
    buttons[i].idCommand = ID_DEBUG_STEPINTO;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* STEP OVER button */
    buttons[i].iBitmap = 2;
    buttons[i].idCommand = ID_DEBUG_STEPOVER;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* STEP OUT button */
    buttons[i].iBitmap = 14;
    buttons[i].idCommand = ID_DEBUG_STEPOUT;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* run to cursor */
    buttons[i].iBitmap = 5;
    buttons[i].idCommand = ID_DEBUG_RUNTOCURSOR;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* set next statement */
    buttons[i].iBitmap = 15;
    buttons[i].idCommand = ID_DEBUG_SETNEXTSTATEMENT;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* break execution */
    buttons[i].iBitmap = 11;
    buttons[i].idCommand = ID_DEBUG_BREAK;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* separator */
    buttons[i].iBitmap = 0;
    buttons[i].idCommand = 0;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_SEP;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* set/clear breakpoint */
    buttons[i].iBitmap = 3;
    buttons[i].idCommand = ID_DEBUG_SETCLEARBREAKPOINT;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* enable/disable breakpoint */
    buttons[i].iBitmap = 6;
    buttons[i].idCommand = ID_DEBUG_DISABLEBREAKPOINT;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* separator */
    buttons[i].iBitmap = 0;
    buttons[i].idCommand = 0;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_SEP;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* evaluate expression */
    buttons[i].iBitmap = 12;
    buttons[i].idCommand = ID_DEBUG_EVALUATE;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* separator */
    buttons[i].iBitmap = 0;
    buttons[i].idCommand = 0;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_SEP;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* show next statement */
    buttons[i].iBitmap = 13;
    buttons[i].idCommand = ID_DEBUG_SHOWNEXTSTATEMENT;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* show game window */
    buttons[i].iBitmap = 7;
    buttons[i].idCommand = ID_VIEW_GAMEWIN;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* show locals window */
    buttons[i].iBitmap = 8;
    buttons[i].idCommand = ID_VIEW_LOCALS;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* show watch window */
    buttons[i].iBitmap = 9;
    buttons[i].idCommand = ID_VIEW_WATCH;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* show stack window */
    buttons[i].iBitmap = 10;
    buttons[i].idCommand = ID_VIEW_STACK;
    buttons[i].fsState = TBSTATE_ENABLED;
    buttons[i].fsStyle = TBSTYLE_BUTTON;
    buttons[i].dwData = 0;
    buttons[i].iString = 0;
    ++i;

    /* set up the toolbar */
    toolbar_ = CreateToolbarEx(handle_,
                               WS_CHILD | TBSTYLE_TOOLTIPS | TBSTYLE_FLAT,
                               0, 1, CTadsApp::get_app()->get_instance(),
                               IDB_DEBUG_TOOLBAR, buttons, i,
                               16, 15, 16, 15, sizeof(buttons[0]));
    if (toolbar_ != 0)
    {
        /* show the toolbar window */
        ShowWindow(toolbar_, SW_SHOW);

        /* set up the toolbar to do idle status updating */
        add_toolbar_proc(toolbar_);
    }
}

/*
 *   Process a notification message.  We receive notifications from the
 *   toolbar to get tooltip text.  
 */
int CHtmlSys_dbgmain::do_notify(int /*control_id*/, int notify_code,
                                LPNMHDR nmhdr)
{
    switch(notify_code)
    {
    case TTN_NEEDTEXT:
        /* retreive tooltip text for a button */
        {
            TOOLTIPTEXT *ttt = (TOOLTIPTEXT *)nmhdr;

            /* get the strings from my application instance */
            ttt->hinst = CTadsApp::get_app()->get_instance();

            /* determine which button was pressed */
            switch(ttt->hdr.idFrom)
            {
            case ID_DEBUG_GO:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_GO);
                break;
                
            case ID_DEBUG_STEPINTO:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_STEPINTO);
                break;
                
            case ID_DEBUG_STEPOVER:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_STEPOVER);
                break;
                
            case ID_DEBUG_RUNTOCURSOR:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_RUNTOCURSOR);
                break;
                
            case ID_DEBUG_BREAK:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_BREAK);
                break;
                
            case ID_DEBUG_SETCLEARBREAKPOINT:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_SETCLEARBREAKPOINT);
                break;
                
            case ID_DEBUG_DISABLEBREAKPOINT:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_DISABLEBREAKPOINT);
                break;
                
            case ID_DEBUG_EVALUATE:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_EVALUATE);
                break;
                
            case ID_VIEW_GAMEWIN:
                ttt->lpszText = MAKEINTRESOURCE(IDS_VIEW_GAMEWIN);
                break;
                
            case ID_VIEW_WATCH:
                ttt->lpszText = MAKEINTRESOURCE(IDS_VIEW_WATCH);
                break;
                
            case ID_VIEW_LOCALS:
                ttt->lpszText = MAKEINTRESOURCE(IDS_VIEW_LOCALS);
                break;
                
            case ID_VIEW_STACK:
                ttt->lpszText = MAKEINTRESOURCE(IDS_VIEW_STACK);
                break;

            case ID_DEBUG_SHOWNEXTSTATEMENT:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_SHOWNEXTSTATEMENT);
                break;

            case ID_DEBUG_SETNEXTSTATEMENT:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_SETNEXTSTATEMENT);
                break;

            case ID_DEBUG_STEPOUT:
                ttt->lpszText = MAKEINTRESOURCE(IDS_DEBUG_STEPOUT);
                break;
                
            default:
                /* not handled */
                return FALSE;
            }
        }
        
        /* handled */
        return TRUE;

    default:
        /* ignore other notifications */
        return FALSE;
    }
}

/*
 *   close the main window 
 */
int CHtmlSys_dbgmain::do_close()
{
    /* 
     *   this will stop the debugging session -- ask the user for
     *   confirmation 
     */
    ask_quit();

    /* 
     *   Even if we're quitting, don't actually close the window for now;
     *   the quitting flag will make us shut down, so we don't need to
     *   close the window for real right now 
     */
    return FALSE;
}

/* 
 *   resize the window 
 */
void CHtmlSys_dbgmain::do_resize(int mode, int x, int y)
{
    /* inherit default */
    CHtmlSys_framewin::do_resize(mode, x, y);

    /* 
     *   Send the resize to the toolbar window, so that it can reposition
     *   itself correctly.  It doesn't matter what the parameters are (it
     *   looks at the parent window to figure out what size to use), but
     *   it's important that the window receives the WM_SIZE whenever the
     *   parent window's size changes.  
     */
    if (toolbar_ != 0)
        SendMessage(toolbar_, TB_AUTOSIZE, 0, 0);
}

/*
 *   paint the background 
 */
int CHtmlSys_dbgmain::do_erase_bkg(HDC hdc)
{
    RECT rc;
    
    /* fill in the background with button face color */
    GetClientRect(handle_, &rc);
    FillRect(hdc, &rc, GetSysColorBrush(COLOR_3DFACE));
    return TRUE;
}

/*
 *   Ask the user if they want to restart the game 
 */
int CHtmlSys_dbgmain::ask_restart()
{
    CHtmlSys_mainwin *mainwin = CHtmlSys_mainwin::get_main_win();

    /* 
     *   ask the user if they want to restart (unless the game is over, in
     *   which case we don't need to pester them with the question, since
     *   it's obvious they'd want to restart) 
     */
    if (!game_over_)
    {
        switch(MessageBox(0,
                          "Do you really want to restart the game?",
                          "TADS Debugger",
                          MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2
                          | MB_TASKMODAL))
        {
        case IDYES:
        case 0:
            /* go ahead and restart */
            break;

        default:
            /* don't restart */
            return FALSE;
        }
    }
        
    /* set the flag to indicate that we should restart the game */
    restarting_ = TRUE;
    
    /* tell the event loop to relinquish control */
    go_ = TRUE;
    
    /*
     *   If we're not in the debugger, set stepping mode, and tell the
     *   main window to abort the current command.  This will force
     *   debugger entry, at which point we'll notice that we need to
     *   restart.  
     */
    if (!in_debugger_)
    {
        /* 
         *   make sure we're in stepping mode, so that we regain control
         *   when the current command ends 
         */
        helper_->set_exec_state_step_into(dbgctx_);
        
        /* tell the main window to abort the current command */
        if (mainwin != 0)
            mainwin->abort_command();
    }
    
    /* start a new page in the main output window */
    if (mainwin != 0)
    {
        outflush();
        mainwin->clear_all_pages();
    }
    
    /* they do want to restart */
    return TRUE;
}

/*
 *   Ask the user if they want to quit.  Return true if so, false if not.  
 */
int CHtmlSys_dbgmain::ask_quit()
{
    /* 
     *   put up a message box asking about quitting (unless the game is
     *   over, in which case we can quit without asking) 
     */
    if (!game_over_)
    {
        switch(MessageBox(0,
                          "This will terminate your debugging session.  "
                          "Do you really want to quit?", "TADS Debugger",
                          MB_YESNO | MB_ICONQUESTION
                          | MB_DEFBUTTON2 | MB_TASKMODAL))
        {
        case IDYES:
        case 0:
            /* go ahead and quit */
            break;
            
        default:
            /* don't quit */
            return FALSE;
        }
    }
        
    /* set the flag to indicate that we should quit the program */
    quitting_ = TRUE;
    
    /* tell the event loop to relinquish control */
    go_ = TRUE;
    
    /*
     *   If we're not in the debugger, set stepping mode, and tell the
     *   main window to abort the current command.  This will force
     *   debugger entry, at which point we'll notice that we need to shut
     *   down.  
     */
    if (!in_debugger_)
    {
        /* 
         *   make sure we're in stepping mode, so that we regain control
         *   when the current command ends 
         */
        helper_->set_exec_state_step_into(dbgctx_);
        
        /* tell the main window to abort the current command */
        if (CHtmlSys_mainwin::get_main_win() != 0)
            CHtmlSys_mainwin::get_main_win()->abort_command();
    }
    
    /* indicate that we are quitting */
    return TRUE;
}

/*
 *   store my window position preferences 
 */
void CHtmlSys_dbgmain::set_winpos_prefs(const CHtmlRect *winpos)
{
    /* save the position in the debugger configuration */
    config_->setval("main win pos", 0, winpos);
}


/* 
 *   context structure for source file enumerator callback 
 */
struct HtmlDbg_srcenum_t
{
    /* number of items inserted into the menu so far */
    int count_;

    /* handle of the menu into which we're inserting */
    HMENU menu_;
};

/*
 *   Set up the file menu.  We'll invoke this the first time we enter the
 *   debugger (after the game is loaded) to load the source files into the
 *   quick-open file submenu.  Note that we can't do this during debugger
 *   initialization, because that happens before the game file is loaded,
 *   so we don't know any of the source files at that time.  
 */
void CHtmlSys_dbgmain::load_file_menu()
{
    HMENU mnu;
    HMENU submnu;
    HtmlDbg_srcenum_t cbctx;

    /* get the File menu (first menu), then the popup for the source files */
    mnu = GetMenu(handle_);
    submnu = GetSubMenu(mnu, 0);
    submnu = GetSubMenu(submnu, 1);

    /* load the source files into the menu */
    cbctx.count_ = 0;
    cbctx.menu_ = submnu;
    helper_->enum_source_files(dbgctx_, load_menu_cb, &cbctx);

    /* 
     *   if we added any items, remove the "no files" placeholder item
     *   that's in the menu resource -- it will still be the first item,
     *   so simply remove the first item from the menu 
     */
    if (cbctx.count_ > 0)
    {
        /* delete the placeholder */
        DeleteMenu(submnu, 0, MF_BYPOSITION);
    }
}

/* 
 *   Callback for source file enumerator - loads the current source file
 *   into the menu.
 */
void CHtmlSys_dbgmain::load_menu_cb(void *ctx0, const textchar_t *fname,
                                    int line_source_id)
{
    MENUITEMINFO info;
    HtmlDbg_srcenum_t *ctx = (HtmlDbg_srcenum_t *)ctx0;
    char caption[260];

    /*
     *   Generate the menu item caption.  Prepend a keyboard shortcut key
     *   to the name using the current item number. 
     */
    ctx->count_++;
    caption[0] = '&';
    caption[1] = ctx->count_ + (ctx->count_ < 10 ? '0' : 'a' - 10);
    caption[2] = ' ';
    do_strcpy(caption + 3, fname);
    
    /* 
     *   Set up the menu item information structure.  Generate the command
     *   ID by adding the line source ID to the base file command ID.  
     */
    info.cbSize = sizeof(info);
    info.fMask = MIIM_TYPE | MIIM_DATA | MIIM_ID;
    info.fType = MFT_STRING;
    info.wID = ID_FILE_OPENLINESRC_FIRST + line_source_id;
    info.dwTypeData = caption;
    info.cch = get_strlen(caption);

    /* insert the menu at the end of the menu */
    InsertMenuItem(ctx->menu_, ctx->count_, TRUE, &info);
}

/*
 *   Terminate the debugger 
 */
void CHtmlSys_dbgmain::terminate_debugger()
{
    enum_win_save_ctx_t cbctx;
    HANDLE hwnd;
    CHtmlRect hrc;
    CDbg_win32_srcdir *srcdir;
    int i;

    /*
     *   Before we do anything else, save the configuration.  We need to
     *   do this before destroying the main debugger window, because all
     *   of the other debugger windows are owned by the main window, hence
     *   destroying the main window will destroy all of them first. 
     */
    
    /* save the list of open windows */
    config_->clear_strlist("src win list", 0);
    cbctx.count_ = 0;
    cbctx.config_ = config_;
    helper_->enum_source_windows(&enum_win_save_cb, &cbctx);

    /* save the watch and local configurations */
    watchwin_->save_config(config_, "watch win");
    localwin_->save_config(config_, "locals win");

    /* save the visibility and position of the debug log window */
    if (helper_->get_debuglog_win() != 0)
    {
        RECT rc;
        CHtmlRect hrc;
        
        /* save its visibility */
        hwnd = ((CHtmlSysWin_win32 *)helper_->get_debuglog_win())
               ->get_frame_handle();
        config_->setval("log win vis", 0, IsWindowVisible(hwnd));

        /* get its position, adjusted to frame-relative coordinates */
        GetWindowRect(hwnd, &rc);
        w32tdb_adjust_client_pos(this, &rc);

        /* save the position */
        hrc.set(rc.left, rc.top, rc.right, rc.bottom);
        config_->setval("log win pos", 0, &hrc);
    }
    else
        config_->setval("log win vis", 0, FALSE);

    /* save the visibility and position of the stack window */
    if (helper_->get_stack_win() != 0)
    {
        hwnd = ((CHtmlSysWin_win32 *)helper_->get_stack_win())
               ->get_frame_handle();
        config_->setval("stack win vis", 0, IsWindowVisible(hwnd));
    }
    else
        config_->setval("stack win vis", 0, FALSE);

    /* save visibility and position of the call history window */
    if (helper_->get_hist_win() != 0)
    {
        hwnd = ((CHtmlSysWin_win32 *)helper_->get_hist_win())
               ->get_frame_handle();
        config_->setval("hist win vis", 0, IsWindowVisible(hwnd));
    }
    else
        config_->setval("hist win vis", 0, FALSE);

    /* save breakpoint list */
    helper_->save_bp_config(dbgctx_, config_);

    /* save expression list */
    watchwin_->save_config(config_, "watch win exprs");

    /* save the search configuration */
    config_->setval("search", "exact case", search_exact_case_);
    config_->setval("search", "wrap", search_wrap_);
    config_->setval("search", "dir", search_dir_);
    config_->setval("search", "start at top", search_from_top_);
    config_->setval("search", "text", 0, find_text_);
    find_dlg_->save_config(config_, "search dlg");

    /* save the source search directory list */
    for (i = 0, srcdir = srcdirs_ ; srcdir != 0 ; srcdir = srcdir->next_, ++i)
        config_->setval("srcdir", 0, i, srcdir->dir_.get());

    /* delete the watch window and the local variables window */
    DestroyWindow(watchwin_->get_handle());
    DestroyWindow(localwin_->get_handle());

    /* delete the main debugger window */
    DestroyWindow(handle_);
}

/*
 *   Debugger main event loop 
 */
int CHtmlSys_dbgmain::dbg_event_loop(dbgcxdef *ctx, int bphit, int err,
                                     int *restart, int *abort,
                                     unsigned int *exec_ofs)
{
    int ret;
    int old_in_dbg;
    CHtmlSysWin_win32 *win;
    
    /* remember the current debugger context */
    dbgctx_ = ctx;

    /* remember the pointer to the debugger's execution offset */
    exec_ofs_ptr_ = exec_ofs;

    /* 
     *   if the 'quitting' flag is already set, it means they closed the
     *   debugger while the game was running, and we're just now getting
     *   control - immediately return and indicate that we should
     *   terminate 
     */
    if (quitting_)
        return FALSE;

    /*
     *   If the 'restarting' flag is already set, it means they restarted
     *   the game while the debugger was inactive, and we're just now
     *   getting control - immediately return and indicate that we should
     *   restart 
     */
    if (restarting_)
    {
        /* only restart once per request - we've satisfied the request now */
        restarting_ = FALSE;

        /* tell the caller to restart */
        *restart = TRUE;

        /* make sure we re-enter the debugger as soon as the game restarts */
        helper_->set_exec_state_step_into(ctx);

        /* tell the caller to keep running */
        return TRUE;
    }

    /* remember whether we were in the debugger when we entered */
    old_in_dbg = in_debugger_;

    /* we're in the debugger */
    in_debugger_ = TRUE;

    /* 
     *   If the game window is in front, move it behind all of the
     *   debugger windows, since we probably are entering the debugger
     *   after having started reading a command (which brings the game
     *   window to the front). 
     */
    w32tdb_bring_group_to_top(get_handle());
    
    /* clear any temporary breakpoint that we had set */
    helper_->clear_temp_bp(ctx);

    /* set the title bar to indicate that the game is paused */
    SetWindowText(handle_,
                  game_over_ ? "TADS Debugger [game terminated]"
                             : "TADS Debugger [game paused]");

    /* 
     *   Display the current source file.  If we can't find one, bring the
     *   main debugger window (i.e., myself) to the front. 
     */
    win = (CHtmlSysWin_win32 *)helper_->update_for_entry(ctx, this);
    if (win != 0)
        win->bring_to_front();
    else
        bring_to_front();

    /* update the expressions window */
    update_expr_windows();

    /* check why we stopped, and display a message about it if appropriate */
    if (err != 0)
    {
        char msgbuf[1024];

        /* 
         *   we stopped due to a run-time error -- display a message box
         *   with the error text 
         */
        helper_->format_error(ctx, err, msgbuf, sizeof(msgbuf));
        MessageBox(0, msgbuf, "TADS Runtime Error",
                   MB_OK | MB_ICONEXCLAMATION | MB_TASKMODAL);
    }
    else
    {
        /*
         *   Check the breakpoint - if it's a global breakpoint, pop up a
         *   message box telling the user what happened, since it's not
         *   immediately obvious just looking at the code why we entered
         *   the debugger. 
         */
        if (helper_->is_bp_global(ctx, bphit))
        {
            textchar_t cond[HTMLDBG_MAXCOND + 50];
            size_t len;
            
            /* 
             *   it's a global breakpoint -- get the text of the
             *   condition, and display it so the user can see why we
             *   stopped 
             */
            strcpy(cond, "Stopped at global breakpoint: ");
            len = get_strlen(cond);
            if (!helper_->get_bp_cond(ctx, bphit, cond + len,
                                      sizeof(cond) - len))
            {
                /* show the message */
                MessageBox(0, cond, "TADS Debugger",
                           MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
            }
        }
    }
    
    /* process events until we start stepping */
    go_ = FALSE;
    ret = CTadsApp::get_app()->event_loop(&go_, &accel_, handle_);

    /* pop debugger status flag */
    in_debugger_ = old_in_dbg;

    /* set the title bar to indicate that the game is paused */
    SetWindowText(handle_, "TADS Debugger [game running]");

    /* 
     *   if we're quitting, force the return value to FALSE so that we
     *   terminate the game 
     */
    if (quitting_)
        ret = FALSE;

    /*
     *   if we're restarting, tell the caller to restart 
     */
    if (restarting_)
    {
        /* only restart once per request */
        restarting_ = FALSE;

        /* make sure we re-enter the debugger as soon as the game restarts */
        helper_->set_exec_state_step_into(ctx);

        /* tell the caller to do the restart */
        *restart = TRUE;
    }

    /*
     *   If the 'aborting' flag is set, tell the caller to abort the
     *   current command 
     */
    if (aborting_cmd_)
    {
        /* only abort once per request */
        aborting_cmd_ = FALSE;

        /* make sure we re-enter the debugger on the next command */
        helper_->set_exec_state_step_into(ctx);

        /* tell the caller to do the abort */
        *abort = TRUE;
    }

    /* return the result */
    return ret;
}

/*
 *   Update expression windows 
 */
void CHtmlSys_dbgmain::update_expr_windows()
{
    /* update the watch window */
    if (watchwin_ != 0)
        watchwin_->update_all();

    /* update the local variables */
    if (localwin_ != 0)
        localwin_->update_all();
}

/* 
 *   Add a new watch expression.  This will add the new expression at the
 *   end of the currently selected panel of the watch window.  
 */
void CHtmlSys_dbgmain::add_watch_expr(const textchar_t *expr)
{
    if (watchwin_ != 0)
    {
        /* add the expression */
        watchwin_->add_expr(expr, TRUE, TRUE);

        /* make sure the watch window is visible */
        ShowWindow(watchwin_->get_handle(), SW_SHOW);
    }
}

/*
 *   Get the text for a Find command 
 */
const textchar_t *CHtmlSys_dbgmain::
   get_find_text(int command_id, int *exact_case, int *start_at_top,
                 int *wrap, int *dir)
{
    /*
     *   If this is a Find command, or if we don't have any text from the
     *   last "Find" command, run the dialog. 
     */
    if (command_id == ID_EDIT_FIND || find_text_[0] == '\0')
    {
        int new_exact_case = search_exact_case_;
        int new_search_from_top = search_from_top_;
        int new_wrap = search_wrap_;
        int new_dir = search_dir_;
        
        /* run the dialog - if they cancel, return null */
        if (!find_dlg_->run_find_dlg(this, find_text_, sizeof(find_text_),
                                     &new_exact_case, &new_search_from_top,
                                     &new_wrap, &new_dir))
            return 0;

        /* they accepted the dialog -- remember the new settings */
        search_exact_case_ = new_exact_case;
        search_from_top_ = new_search_from_top;
        search_wrap_ = new_wrap;
        search_dir_ = new_dir;

        /* use the search-from-top setting from the dialog */
        *start_at_top = search_from_top_;
    }
    else
    {
        /* 
         *   continuing an old search, so don't start over at the top,
         *   regardless of the last dialog settings 
         */
        *start_at_top = FALSE;
    }

    /* 
     *   fill in the flags from the last dialog settings, whether this is
     *   a new search or a continued search 
     */
    *exact_case = search_exact_case_;
    *wrap = search_wrap_;
    *dir = search_dir_;

    /* return the text to find */
    return find_text_;
}


/*
 *   Check command status 
 */
TadsCmdStat_t CHtmlSys_dbgmain::check_command(int command_id)
{
    CHtmlDbgSubWin *active_win;

    /* if it's a source file command, enable it */
    if (command_id >= ID_FILE_OPENLINESRC_FIRST
        && command_id <= ID_FILE_OPENLINESRC_LAST)
        return TADSCMD_ENABLED;

    /* if it's a window command, enable it */
    if (command_id >= ID_WINDOW_FIRST
        && command_id <= ID_WINDOW_LAST)
        return TADSCMD_ENABLED;

    /* check the command */
    switch(command_id)
    {
    case ID_DEBUG_SHOWHIDDENOUTPUT:
        return (helper_->get_dbg_hid() ? TADSCMD_CHECKED : TADSCMD_ENABLED);
        break;

    case ID_FILE_OPEN_NONE:
    case ID_WINDOW_NONE:
        /* dummy entry for when there are no source files; always disabled */
        return TADSCMD_DISABLED;

    case ID_FILE_OPEN:
    case ID_FILE_QUIT:
    case ID_DEBUG_EDITBREAKPOINTS:
    case ID_FILE_RESTARTGAME:
    case ID_DEBUG_OPTIONS:
        /* always enabled */
        return TADSCMD_ENABLED;

    case ID_DEBUG_GO:
    case ID_DEBUG_STEPINTO:
    case ID_DEBUG_STEPOVER:
    case ID_DEBUG_STEPOUT:
        /* these only work while the debugger is active */
        return in_debugger_ ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_DEBUG_EVALUATE:
    case ID_DEBUG_ABORTCMD:
    case ID_DEBUG_SHOWNEXTSTATEMENT:
        /* 
         *   these only work when the game is running and the debugger is
         *   active 
         */
        return (in_debugger_ && !game_over_
                ? TADSCMD_ENABLED : TADSCMD_DISABLED);

    case ID_DEBUG_SETNEXTSTATEMENT:
    case ID_DEBUG_RUNTOCURSOR:
        /* 
         *   These work only while the debugger is active and the game is
         *   not over, but may not always work when the debugger is active
         *   - disallow them here if the debugger is not active; otherwise
         *   allow them for now, but proceed with additional checks.  
         */
        if (!in_debugger_ || game_over_)
            return TADSCMD_DISABLED;
        break;

    case ID_DEBUG_CLEARHISTORYLOG:
        /* 
         *   this doesn't work when the debugger isn't active; when the
         *   debugger is active, only allow it when there's a non-empty
         *   call history log 
         */
        return (in_debugger_
                && helper_->get_call_trace_len(dbgctx_) != 0
                ? TADSCMD_ENABLED : TADSCMD_DISABLED);

    case ID_DEBUG_HISTORYLOG:
        /* 
         *   this works only while the debugger is active, but check it or
         *   not depending on whether tracing is on 
         */
        if (helper_->is_call_trace_active(dbgctx_))
            return in_debugger_ ? TADSCMD_CHECKED
                                : TADSCMD_DISABLED_CHECKED;
        else
            return in_debugger_ ? TADSCMD_ENABLED
                                : TADSCMD_DISABLED;

    case ID_DEBUG_BREAK:
        /* this only works while the game is active */
        return in_debugger_ ? TADSCMD_DISABLED : TADSCMD_ENABLED;

    case ID_VIEW_STACK:
    case ID_SHOW_DEBUG_WIN:
    case ID_VIEW_WATCH:
    case ID_VIEW_TRACE:
    case ID_VIEW_LOCALS:
    case ID_VIEW_GAMEWIN:
    case ID_HELP_ABOUT:
    case ID_HELP_TOPICS:
        return TADSCMD_ENABLED;

    default:
        break;
    }

    /* 
     *   we didn't handle it -- see if the active debugger window wants to
     *   handle it 
     */
    active_win = get_active_dbg_win();
    if (active_win != 0)
    {
        TadsCmdStat_t result;

        /* run it past the active debugger window */
        result = active_win->active_check_command(command_id);

        /* if the active window handled the command, return the result */
        if (result != TADSCMD_UNKNOWN)
            return result;
    }

    /* inherit default behavior */
    return CHtmlSys_framewin::check_command(command_id);
}

/* callback context for enumerating source windows */
struct winmenu_win_enum_cb_ctx_t
{
    /* number of windows inserted into the menu so far */
    int count_;

    /* the debugger main window object */
    CHtmlSys_dbgmain *dbgmain_;

    /* handle to the Window menu */
    HMENU menuhdl_;
};

/*
 *   Run the main debugger window's menu, activating it with the given
 *   system keydown message parameters.  This can be used when a window
 *   wants to let the keyboard menu interface in its window run the main
 *   debugger window's menu.  
 */
void CHtmlSys_dbgmain::run_menu_kb_ifc(int vkey, long keydata, HANDLE srcwin)
{
    /* 
     *   remember the current active window, so that we can activate it
     *   again when we're done 
     */
    post_menu_active_win_ = srcwin;

    /*
     *   Bring the debugger main window to the top.  We need to do this
     *   before we can open its menu, because Windows will always open the
     *   active window's menu when the keyboard interface runs, regardless
     *   of which window is actually the target of the message. 
     */
    BringWindowToTop(handle_);

    /* 
     *   send myself the WM_SYSKEYDOWN message that the source window
     *   delegated to us 
     */
    PostMessage(handle_, WM_SYSKEYDOWN, vkey, keydata);
}

/*
 *   Receive notification that we've closed a menu 
 */
int CHtmlSys_dbgmain::menu_close(unsigned int /*item*/)
{
    /* 
     *   if we opened the menu from the keyboard while another window was
     *   active, make the original window active again 
     */
    if (post_menu_active_win_ != 0)
    {
        BringWindowToTop(post_menu_active_win_);
        post_menu_active_win_ = 0;
    }
    
    /* return false to proceed with system default handling */
    return FALSE;
}

/*
 *   Initialize a popup menu that's about to be displayed 
 */
void CHtmlSys_dbgmain::init_menu_popup(HMENU menuhdl, unsigned int pos,
                                       int sysmenu)
{
    winmenu_win_enum_cb_ctx_t cbctx;
    int winmenupos;
    HMENU mymenu;
    
    /* 
     *   determine if this is the Window menu; if not, or if this is the
     *   system menu, inherit default handling
     */
    mymenu = GetMenu(handle_);
    winmenupos = GetMenuItemCount(mymenu) - 2;
    if (sysmenu || menuhdl != GetSubMenu(mymenu, winmenupos))
    {
        CHtmlSys_framewin::init_menu_popup(menuhdl, pos, sysmenu);
        return;
    }

    /* erase everything currently in the menu */
    while (GetMenuItemCount(menuhdl) != 0)
        DeleteMenu(menuhdl, 0, MF_BYPOSITION);

    /* insert all of the open source windows into the menu */
    cbctx.count_ = 0;
    cbctx.dbgmain_ = this;
    cbctx.menuhdl_ = menuhdl;
    helper_->enum_source_windows(&winmenu_win_enum_cb, &cbctx);

    /* if the menu is empty, insert the dummy item for an empty menu */
    if (GetMenuItemCount(menuhdl) == 0)
    {
        MENUITEMINFO info;
        static const textchar_t *itemname = "No Windows";

        /* set up the new menu item descriptor */
        info.cbSize = sizeof(info);
        info.fMask = MIIM_TYPE | MIIM_DATA | MIIM_ID;
        info.fType = MFT_STRING;
        info.wID = ID_WINDOW_NONE;
        info.dwTypeData = (char *)itemname;
        info.cch = get_strlen(itemname);

        /* insert the new menu item */
        InsertMenuItem(menuhdl, 0, TRUE, &info);
    }

    /* inherit the default handling now that we've built the menu */
    CHtmlSys_framewin::init_menu_popup(menuhdl, pos, sysmenu);
}

void CHtmlSys_dbgmain::winmenu_win_enum_cb(void *ctx0, int idx,
                                           CHtmlSysWin *win,
                                           int /*line_source_id*/,
                                           HtmlDbg_wintype_t win_type)
{
    winmenu_win_enum_cb_ctx_t *ctx = (winmenu_win_enum_cb_ctx_t *)ctx0;
    MENUITEMINFO info;
    textchar_t win_title[260];
    char *p;

    /* 
     *   ignore everything but regular file windows (the special windows,
     *   such as the stack window and the history window, are already in
     *   the "view" menu) 
     */
    if (win_type != HTMLDBG_WINTYPE_SRCFILE)
        return;

    /* if we're on number 1 through 9, include a numeric shortcut */
    if (ctx->count_ + 1 <= 9)
    {
        sprintf(win_title, "&%d ", ctx->count_ + 1);
        p = win_title + get_strlen(win_title);
    }
    else
    {
        /* 
         *   don't include a shortcut prefix - put the title at the start
         *   of the buffer 
         */
        p = win_title;
    }

    /* get the window's title - use this as the name in the menu */
    GetWindowText(((CHtmlSysWin_win32 *)win)->get_frame_handle(),
                  p, sizeof(win_title) - (p - win_title));

    /* set up the new menu item descriptor for the window */
    info.cbSize = sizeof(info);
    info.fMask = MIIM_TYPE | MIIM_DATA | MIIM_ID;
    info.fType = MFT_STRING;
    info.wID = ID_WINDOW_FIRST + idx;
    info.dwTypeData = win_title;
    info.cch = get_strlen(win_title);

    /* insert the new menu item */
    InsertMenuItem(ctx->menuhdl_, ctx->count_, TRUE, &info);

    /* count it */
    ctx->count_++;
}

/*
 *   Process a command
 */
int CHtmlSys_dbgmain::do_command(int notify_code, int command_id, HWND ctl)
{
    CHtmlDbgSubWin *active_win;

    /* get the active debugger window */
    active_win = get_active_dbg_win();

    /* 
     *   most commands require that the debugger be active; if it's not,
     *   ignore such commands 
     */
    switch(command_id)
    {
    case ID_HELP_TOPICS:
        {
            char buf[256];
            char *p;
            char *rootname;

            /* get the executable's filename so we can get its directory */
            GetModuleFileName(CTadsApp::get_app()->get_instance(),
                              buf, sizeof(buf));

            /* find the directory part */
            for (p = rootname = buf ; *p != '\0'; ++p)
            {
                /* if it's a separator, the next char starts the root name */
                if (*p == '\\' || *p == ':' || *p == '/')
                    rootname = p + 1;
            }

            /* look for "helptdb.htm" in the executable's directory */
            strcpy(rootname, "helptdb.htm");

            /* if it's not there, try the NOTES subdirectory */
            if (GetFileAttributes(buf) == 0xffffffff)
                strcpy(rootname, "notes\\helptdb.htm");

            /* open helptdb.htm */
            if ((unsigned long)ShellExecute(0, "open", buf, 0, 0, 0) <= 32)
                MessageBox(0, "Couldn't open help.  You must have a web "
                           "browser or other HTML viewer installed to view "
                           "help.", "TADS Debugger",
                           MB_OK | MB_ICONEXCLAMATION | MB_TASKMODAL);
        }
        return TRUE;

    case ID_HELP_ABOUT:
        {
            /* run the debugger "about" dialog */
            CTadsDlg_dbgabout *dlg = new CTadsDlg_dbgabout();
            dlg->run_modal(DLG_DBGABOUT, handle_,
                           CTadsApp::get_app()->get_instance());
            delete dlg;
        }
        return TRUE;
        
    case ID_DEBUG_GO:
    case ID_DEBUG_STEPINTO:
    case ID_DEBUG_STEPOVER:
    case ID_DEBUG_STEPOUT:
    case ID_DEBUG_RUNTOCURSOR:
    case ID_DEBUG_CLEARHISTORYLOG:
    case ID_DEBUG_HISTORYLOG:
        /* these commands all require the debugger to be active */
        if (!in_debugger_)
        {
            /* provide some minimal feedback in the form of an alert sound */
            Beep(1000, 100);

            /* ignore it */
            return FALSE;
        }
        break;

    case ID_DEBUG_ABORTCMD:
    case ID_DEBUG_EVALUATE:
    case ID_DEBUG_SHOWNEXTSTATEMENT:
    case ID_DEBUG_SETNEXTSTATEMENT:
        /* 
         *   these commands require the debugger to be active and the game
         *   to be running 
         */
        if (!in_debugger_ || game_over_)
        {
            /* provide some minimal feedback in the form of an alert sound */
            Beep(1000, 100);

            /* ignore it */
            return FALSE;
        }
        break;
    }

    /*
     *   Check for a source-file open command 
     */
    if (command_id >= ID_FILE_OPENLINESRC_FIRST
        && command_id <= ID_FILE_OPENLINESRC_LAST)
    {
        CHtmlSysWin_win32 *win;
        
        /* open the line source and bring it to the front */
        win = (CHtmlSysWin_win32 *)helper_
              ->open_line_source(dbgctx_, this,
                                 command_id - ID_FILE_OPENLINESRC_FIRST);
        if (win != 0)
            win->bring_to_front();
        
        /* handled */
        return TRUE;
    }

    /*
     *   Check for Window commands 
     */
    if (command_id >= ID_WINDOW_FIRST
        && command_id <= ID_WINDOW_LAST)
    {
        CHtmlSysWin_win32 *win;
        
        /* get the correct window */
        win = (CHtmlSysWin_win32 *)helper_
              ->get_source_window_by_index(command_id - ID_WINDOW_FIRST);

        /* bring it to the front */
        if (win != 0)
            win->bring_to_front();

        /* handled */
        return TRUE;
    }
    
    /* check the command */
    switch(command_id)
    {
    case ID_DEBUG_EVALUATE:
    case ID_DEBUG_ADDTOWATCH:
        /* 
         *   if the active window wants to handle it, let it; otherwise,
         *   provide a default implementation here 
         */
        if (active_win != 0
            && active_win->active_do_command(notify_code, command_id, ctl))
        {
            /* it handled it */
            return TRUE;
        }
        else
        {
            CHtmlFormatter *formatter;
            unsigned long sel_start, sel_end;
            textchar_t buf[256];

            /* presume we won't find a selection */
            buf[0] = '\0';
            
            /* get the active window's selection range */
            if (active_win != 0
                && active_win->get_sel_range(&formatter, &sel_start,
                                             &sel_end))
            {
                CHtmlTextArray *arr;
                unsigned long len;

                /* if the selection range is nil, get the word limits */
                if (sel_start == sel_end)
                    formatter->get_word_limits(&sel_start, &sel_end,
                                               sel_start);
                
                /* 
                 *   see how long the selection is - if it's too long,
                 *   don't bother 
                 */
                arr = formatter->get_text_array();
                len = arr->get_char_count(sel_start, sel_end);
                if (len < sizeof(buf))
                {
                    textchar_t *p;
                    int all_spaces;
                    
                    /* get the selection in chunks */
                    for (p = buf ; len > 0 ; )
                    {
                        const textchar_t *chunk;
                        size_t curlen;
                        
                        /* get the next chunk */
                        chunk = arr->get_text_chunk(&sel_start, &curlen, len);
                        memcpy(p, chunk, curlen);

                        /* advance our pointers */
                        len -= curlen;
                        p += curlen;
                    }

                    /* null-terminate the string */
                    *p = '\0';

                    /* 
                     *   scan through the text, converting control
                     *   characters to spaces 
                     */
                    for (p = buf, all_spaces = TRUE ; *p != '\0' ; ++p)
                    {
                        /* if it's not printable, convert it to a space */
                        if (!isprint(*p))
                            *p = ' ';

                        /* if it's not a space, so note */
                        if (!isspace(*p))
                            all_spaces = FALSE;
                    }

                    /* 
                     *   if the buffer is nothing but spaces, consider it
                     *   to be blank 
                     */
                    if (all_spaces)
                        buf[0] = '\0';
                }
            }

            switch(command_id)
            {
            case ID_DEBUG_EVALUATE:
                /* run the dialog */
                CHtmlDbgWatchWinDlg::run_as_dialog(this, helper_, buf);
                break;

            case ID_DEBUG_ADDTOWATCH:
                add_watch_expr(buf);
                break;
            }
        }
        return TRUE;

    case ID_DEBUG_BREAK:
        /* only allow this while the game is running */
        if (!in_debugger_)
        {
            /* 
             *   put the runtime in stepping mode so that we regain
             *   control as soon as the current command ends 
             */
            helper_->set_exec_state_step_into(dbgctx_);

            /* tell the main window to abort the current command */
            if (CHtmlSys_mainwin::get_main_win() != 0)
                CHtmlSys_mainwin::get_main_win()->abort_command();
        }
        return TRUE;

    case ID_FILE_OPEN:
        {
            CHtmlSysWin_win32 *win;
            OPENFILENAME info;
            char fname[256];

            /* ask for a file to open */
            info.lStructSize = sizeof(info);
            info.hwndOwner = handle_;
            info.hInstance = CTadsApp::get_app()->get_instance();
            info.lpstrFilter = "TADS Source Files\0*.T\0"
                               "Text Files\0*.TXT\0"
                               "All Files\0*.*\0\0";
            info.lpstrCustomFilter = 0;
            info.nFilterIndex = 0;
            info.lpstrFile = fname;
            fname[0] = '\0';
            info.nMaxFile = sizeof(fname);
            info.lpstrFileTitle = 0;
            info.nMaxFileTitle = 0;
            info.lpstrInitialDir = CTadsApp::get_app()->get_openfile_dir();
            info.lpstrTitle = "Select a text file";
            info.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST
                         | OFN_NOCHANGEDIR;
            info.nFileOffset = 0;
            info.nFileExtension = 0;
            info.lpstrDefExt = 0;
            info.lCustData = 0;
            info.lpfnHook = 0;
            info.lpTemplateName = 0;
            CTadsDialog::set_filedlg_center_hook(&info);
            if (!GetOpenFileName(&info))
                return TRUE;

            /* 
             *   save the new directory as the initial directory for the
             *   next file open operation -- we do this in lieu of
             *   allowing the dialog to change the current directory,
             *   since we need to stay in the initial game directory at
             *   all times (since the game incorporates relative filename
             *   paths for source files) 
             */
            CTadsApp::get_app()->set_openfile_dir(fname);

            /* open the file */
            win = (CHtmlSysWin_win32 *)
                  helper_->open_text_file(dbgctx_, this, fname);
            if (win != 0)
                win->bring_to_front();
        }
        return TRUE;
        
    case ID_DEBUG_SHOWNEXTSTATEMENT:
        {
            CHtmlSysWin_win32 *win;

            /* tell the helper to activate the top item on the stack */
            win = (CHtmlSysWin_win32 *)helper_->activate_stack_win_line(
                0, dbgctx_, this);

            /* update expressions */
            update_expr_windows();

            /* bring the window to the front */
            if (win != 0)
                win->bring_to_front();
        }
        return TRUE;

    case ID_DEBUG_SHOWHIDDENOUTPUT:
        helper_->toggle_dbg_hid();
        return TRUE;

    case ID_FILE_QUIT:
        /* make sure they really want to quit */
        ask_quit();
        return TRUE;

    case ID_FILE_RESTARTGAME:
        /* make sure they really want to do this */
        ask_restart();
        return TRUE;

    case ID_DEBUG_OPTIONS:
        /* run the options dialog */
        run_dbg_prefs_dlg(handle_, this);
        return TRUE;

    case ID_DEBUG_GO:
        /* set the new execution state */
        helper_->set_exec_state_go(dbgctx_);

        /* tell the event loop to relinquish control */
        go_ = TRUE;
        return TRUE;

    case ID_DEBUG_ABORTCMD:
        /* set the "abort" flag */
        aborting_cmd_ = TRUE;

        /* exit the debugger event loop, so that we can signal the abort */
        go_ = TRUE;
        return TRUE;

    case ID_DEBUG_STEPINTO:
        /* set new execution state */
        helper_->set_exec_state_step_into(dbgctx_);

        /* done with debugging for now */
        go_ = TRUE;
        return TRUE;

    case ID_DEBUG_STEPOVER:
        /* set the new execution state */
        helper_->set_exec_state_step_over(dbgctx_);

        /* done with debugging for now */
        go_ = TRUE;
        return TRUE;

    case ID_DEBUG_STEPOUT:
        /* set the new execution state */
        helper_->set_exec_state_step_out(dbgctx_);

        /* done with debugging for now */
        go_ = TRUE;
        return TRUE;

    case ID_VIEW_STACK:
        /* show the stack window */
        {
            CHtmlSysWin_win32 *win;

            /* create and/or show the stack window */
            win = (CHtmlSysWin_win32 *)helper_->create_stack_win(this);

            /* make sure it's up to date */
            helper_->update_stack_window(dbgctx_, FALSE);

            /* bring it to the front */
            if (win != 0)
                win->bring_to_front();
        }
        return TRUE;

    case ID_VIEW_TRACE:
    show_trace_window:
        /* show the history trace window */
        {
            CHtmlSysWin_win32 *win;

            /* create and/or show the history trace window */
            win = (CHtmlSysWin_win32 *)helper_->create_hist_win(this);

            /* make sure it's up to date */
            helper_->update_hist_window(dbgctx_);

            /* bring it to the front */
            if (win != 0)
                win->bring_to_front();
        }
        return TRUE;

    case ID_VIEW_WATCH:
        /* show the watch window and bring it to the front */
        ShowWindow(watchwin_->get_handle(), SW_SHOW);
        BringWindowToTop(watchwin_->get_handle());
        return TRUE;

    case ID_VIEW_LOCALS:
        /* show the local variables window and bring it to the front */
        ShowWindow(localwin_->get_handle(), SW_SHOW);
        BringWindowToTop(localwin_->get_handle());
        return TRUE;
            
    case ID_SHOW_DEBUG_WIN:
        /* ask the main window to handle the request */
        if (CHtmlSys_mainwin::get_main_win() != 0)
            CHtmlSys_mainwin::get_main_win()
                ->do_command(0, ID_SHOW_DEBUG_WIN, 0);
        return TRUE;

    case ID_VIEW_GAMEWIN:
        /* bring the game window to the front */
        if (CHtmlSys_mainwin::get_main_win() != 0)
            CHtmlSys_mainwin::get_main_win()->bring_owner_to_front();
        return TRUE;

    case ID_DEBUG_EDITBREAKPOINTS:
        /* run the breakpoints dialog */
        CTadsDialogBreakpoints::run_bp_dlg(this);
        return TRUE;

    case ID_DEBUG_HISTORYLOG:
        /* reverse history logging state */
        helper_->set_call_trace_active
            (dbgctx_, !helper_->is_call_trace_active(dbgctx_));

        /* 
         *   if we just turned on call tracing, show the trace window and
         *   bring it to the front 
         */
        if (helper_->is_call_trace_active(dbgctx_))
            goto show_trace_window;

        /* handled */
        return TRUE;

    case ID_DEBUG_CLEARHISTORYLOG:
        /* clear the history log */
        helper_->clear_call_trace_log(dbgctx_);
        return TRUE;

    default:
        break;
    }

    /* 
     *   we didn't handle it -- run the command by the active debugger
     *   window, and see if it can handle the command 
     */
    if (active_win != 0
        && active_win->active_do_command(notify_code, command_id, ctl))
        return TRUE;

    /* inherit default behavior */
    return CHtmlSys_framewin::do_command(notify_code, command_id, ctl);
}

/*
 *   Receive notification that source window preferences have changed.
 *   We'll update the source windows accordingly.  
 */
void CHtmlSys_dbgmain::on_update_srcwin_options()
{
    int new_tab_size;

    /* update each window's color settings */
    helper_->enum_source_windows(&options_enum_srcwin_cb, this);

    /* update tab size if necessary */
    if (!config_->getval("src win options", "tab size", &new_tab_size)
        && helper_->get_tab_size() != new_tab_size)
    {
        /* 
         *   Tell the helper to update the tab size - this will
         *   necessitate reloading all source files, since tabs are
         *   translated at load time.  The helper will reformat all of the
         *   windows as a result, so we'll pick up any other changes to
         *   the source options.  
         */
        helper_->set_tab_size(dbgctx_, this, new_tab_size);
    }
    else
    {
        /* reformat the source windows to account for the changes */
        helper_->reformat_source_windows(dbgctx_, this);
    }
}

/*
 *   Resume execution 
 */
void CHtmlSys_dbgmain::exec_go()
{
    /* set the new execution state */
    helper_->set_exec_state_go(dbgctx_);

    /* tell the event loop to relinquish control */
    go_ = TRUE;
}

/*
 *   Resume execution for a single step 
 */
void CHtmlSys_dbgmain::exec_step_into()
{
    /* set the new execution state */
    helper_->set_exec_state_step_into(dbgctx_);

    /* tell the event loop to return control to the run-time */
    go_ = TRUE;
}

/* 
 *   enumeration callback for applying options changes to source windows 
 */
void CHtmlSys_dbgmain::options_enum_srcwin_cb(void *ctx, int /*idx*/,
                                              CHtmlSysWin *win,
                                              int /*line_source_id*/,
                                              HtmlDbg_wintype_t /*win_type*/)
{
    /* apply the color settings to this window */
    ((CHtmlSys_dbgmain *)ctx)
        ->apply_color_options((CHtmlSysWin_win32_dbgsrc *)win);
}

/*
 *   Apply the color options to a given source window
 */
void CHtmlSys_dbgmain::apply_color_options(CHtmlSysWin_win32_dbgsrc *win)
{
    HTML_color_t bkg_color, text_color;
    int use_win_colors;

    /* get the color preferences */
    if (config_->getval("src win options", "use win colors",
                        &use_win_colors))
        use_win_colors = TRUE;
    if (config_->getval_color("src win options", "bkg color", &bkg_color))
        bkg_color = HTML_make_color(0xff, 0xff, 0xff);
    if (config_->getval_color("src win options", "text color", &text_color))
        bkg_color = HTML_make_color(0, 0, 0);

    /* apply the current color settings to the window */
    win->set_src_colors(bkg_color, text_color, use_win_colors);
}

/*
 *   Get the last active window for command routing.  If there's an active
 *   debugger window, we'll return that window.  If the debugger main
 *   window is itself the active window, and the previous active debugger
 *   window is the second window in Z order, we'll return that window,
 *   since we left it explicitly to come to the debugger main window,
 *   presumably for a toolbar or menu operation. 
 */
CHtmlDbgSubWin *CHtmlSys_dbgmain::get_active_dbg_win()
{
    HWND top_owned;
    
    /* if we have an active debugger window, use it */
    if (active_dbg_win_ != 0)
        return active_dbg_win_;

    /* find the top window that I own */
    for (top_owned = GetTopWindow(0) ;
         top_owned != 0 && GetWindow(top_owned, GW_OWNER) != handle_ ;
         top_owned = GetWindow(top_owned, GW_HWNDNEXT)) ;

    /* 
     *   if the main window itself is active, and the previous window in Z
     *   order is the last active debugger window, route the command to
     *   the previous window 
     */
    if (last_active_dbg_win_ != 0
        && GetActiveWindow() == handle_
        && top_owned == last_active_dbg_win_->active_get_handle())
#if 0
        && (GetWindow(handle_, GW_HWNDNEXT)
            == last_active_dbg_win_->active_get_handle())
#endif
        return last_active_dbg_win_;

    /* there isn't an appropriate window */
    return 0;
}

/*
 *   Bring the window to the front 
 */
void CHtmlSys_dbgmain::bring_to_front()
{
    BringWindowToTop(handle_);
}

/*
 *   Set the active debugger window.  Source, stack, and other debugger
 *   windows should set themselves active when they're activated, and
 *   should clear the active window when they're deactivated.  We keep
 *   track of the active debugger window for command routine purposes. 
 */
void CHtmlSys_dbgmain::set_active_dbg_win(CHtmlDbgSubWin *win)
{
    /* remember the active window */
    active_dbg_win_ = win;
}

/*
 *   Clear the active debugger window.  Windows that register themselves
 *   upon activation with set_active_dbg_win() should call this routine
 *   when they're deactivated.  
 */
void CHtmlSys_dbgmain::clear_active_dbg_win()
{
    /* 
     *   Remember this as the previous active debugger window.  When the
     *   debugger main window is the active window, and the previously
     *   active window is the next window in window order, we'll route
     *   commands that belong to debugger windows to the old active
     *   window. 
     */
    last_active_dbg_win_ = active_dbg_win_;

    /* forget the window */
    active_dbg_win_ = 0;
}

/*
 *   Receive notification that we're closing a debugger window.  Any
 *   window that registers itself with set_active_dbg_win() on activation
 *   must call this routine when the window is destroyed, to ensure that
 *   we don't try to route any commands to the window after it's gone. 
 */
void CHtmlSys_dbgmain::on_close_dbg_win(CHtmlDbgSubWin *win)
{
    /* 
     *   if this is the old active window or the current active window,
     *   forget about it 
     */
    if (active_dbg_win_ == win)
        active_dbg_win_ = 0;
    if (last_active_dbg_win_ == win)
        last_active_dbg_win_ = 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   CHtmlDebugSysIfc_win implementation 
 */

/*
 *   create a source window 
 */
CHtmlSysWin *CHtmlSys_dbgmain::dbgsys_create_win(
    CHtmlParser *parser,  CHtmlFormatter *formatter,
    const textchar_t *win_title, HtmlDbg_wintype_t win_type)
{
    CHtmlSys_dbgsrcwin *win;
    RECT pos;
    CHtmlRect hrc;
    int vis;

    /* create the new frame window */
    switch(win_type)
    {
    case HTMLDBG_WINTYPE_SRCFILE:
        win = new CHtmlSys_dbgsrcwin(parser, prefs_, helper_, this);
        if (!config_->getval("src win pos", win_title, &hrc))
            SetRect(&pos, hrc.left, hrc.top,
                    hrc.right - hrc.left, hrc.bottom - hrc.top);
        else
            SetRect(&pos, CW_USEDEFAULT, CW_USEDEFAULT,
                    CW_USEDEFAULT, CW_USEDEFAULT);
        vis = TRUE;
        break;

    case HTMLDBG_WINTYPE_STACK:
        win = new CHtmlSys_dbgstkwin(parser, prefs_, helper_, this);
        if (!config_->getval("stack win pos", 0, &hrc))
            SetRect(&pos, hrc.left, hrc.top,
                    hrc.right - hrc.left, hrc.bottom - hrc.top);
        else
            w32tdb_set_default_stk_rc(&pos);
        vis = TRUE;
        break;

    case HTMLDBG_WINTYPE_DEBUGLOG:
        {
            CHtmlSys_dbglogwin *dbgwin;

            /* 
             *   note whether the window is visible - it's visible by
             *   default if there's no configuration setting for it 
             */
            if (config_->getval("log win vis", 0, &vis))
                vis = TRUE;

            /* create the debugger window */
            dbgwin = new CHtmlSys_dbglogwin(parser, prefs_);

            /* load its position */
            if (!config_->getval("log win pos", 0, &hrc))
                SetRect(&pos, hrc.left, hrc.top,
                        hrc.right - hrc.left, hrc.bottom - hrc.top);
            else
                SetRect(&pos, CW_USEDEFAULT, CW_USEDEFAULT,
                        CW_USEDEFAULT, CW_USEDEFAULT);

            /* initialize the HTML panel in the window */
            dbgwin->init_html_panel(formatter);

            /* 
             *   this will show only preformatted text, so don't reformat
             *   on resize 
             */
            dbgwin->get_html_panel()->set_format_on_resize(FALSE);

            /* show a sizing box between the scrollbars */
            dbgwin->get_html_panel()->set_has_sizebox();

            /* create the system window */
            dbgwin->create_system_window(this, vis, win_title, &pos,
                                         w32tdb_new_child_syswin(dbgwin));

            /* 
             *   tell the main game window about the log window, so that
             *   it can send debug messages to the window.  Note that we
             *   do not want to add a menu for the debug log window to the
             *   game window, since the window is conceptually part of the
             *   debugger rather than of the run-time.  
             */
            if (CHtmlSys_mainwin::get_main_win() != 0)
                CHtmlSys_mainwin::get_main_win()->
                    set_debug_win(dbgwin, FALSE);

            /* tell the window to notify the helper on destruction */
            dbgwin->set_destroy_callback(&CHtmlDebugHelper::
                                         on_close_srcwin_cb, helper_);

            /* return the new window's main html panel */
            return dbgwin->get_html_panel();
        }

    case HTMLDBG_WINTYPE_HISTORY:
        win = new CHtmlSys_dbghistwin(parser, prefs_, helper_, this);
        if (!config_->getval("hist win pos", 0, &hrc))
            SetRect(&pos, hrc.left, hrc.top,
                    hrc.right - hrc.left, hrc.bottom - hrc.top);
        else
            SetRect(&pos, 50, 50, 350, 500);
        vis = TRUE;
        break;
    }

    /* initialize the HTML panel in the window */
    win->init_html_panel(formatter);

    /* this will show only preformatted text, so don't reformat on resize */
    win->get_html_panel()->set_format_on_resize(FALSE);

    /* show a sizing box between the scrollbars */
    win->get_html_panel()->set_has_sizebox();

    /* open the system window */
    win->create_system_window(this, vis, win_title, &pos,
                              w32tdb_new_child_syswin(win));

    /* apply the color settings */
    apply_color_options((CHtmlSysWin_win32_dbgsrc *)win->get_html_panel());

    /* return the system window object for the HTML panel in the window */
    return win->get_html_panel();
}

/*
 *   adjust a font for use in source windows 
 */
void CHtmlSys_dbgmain::dbgsys_set_srcfont(CHtmlFontDesc *font_desc)
{
    textchar_t buf[HTMLFONTDESC_FACE_MAX];
    int fontsize;
    
    /* 
     *   use a smaller font size; the base size is too big for monospaced
     *   fonts 
     */
    font_desc->htmlsize = 2;

    /* if there's a configuration setting, use its font */
    if (!config_->getval("src win options", "font name",
                         0, buf, sizeof(buf)))
        strcpy(font_desc->face, buf);

    /* if there's a specific size setting, use it */
    if (!config_->getval("src win options", "font size", &fontsize))
    {
        font_desc->htmlsize = 0;
        font_desc->pointsize = fontsize;
    }
}

/*
 *   compare filenames 
 */
int CHtmlSys_dbgmain::dbgsys_cmp_fnames(const char *fname1,
                                        const char *fname2)
{
    char buf1[256], buf2[256];
    char *fpart1, *fpart2;

    /* 
     *   Get the full names and paths of the files.  This will expand any
     *   short (8.3) names into full names, and will resolve relative
     *   paths to provide the fully-qualified name with drive and absolute
     *   path.  
     */
    GetFullPathName(fname1, sizeof(buf1), buf1, &fpart1);
    GetFullPathName(fname2, sizeof(buf2), buf2, &fpart2);

    /* see if they match, ignoring case */
    return (stricmp(buf1, buf2) == 0);
}

/* ------------------------------------------------------------------------ */
/*
 *   CHtmlSys_srcwin implementation 
 */

CHtmlSys_dbgsrcwin::CHtmlSys_dbgsrcwin(CHtmlParser *parser,
                                       CHtmlPreferences *prefs,
                                       CHtmlDebugHelper *helper,
                                       CHtmlSys_dbgmain *dbgmain)
    : CHtmlSys_dbgwin(parser, prefs)
{
    /* remember my helper object */
    helper_ = helper;

    /* remember the main debugger window */
    dbgmain_ = dbgmain;
}

/*
 *   initialize 
 */
void CHtmlSys_dbgsrcwin::init_html_panel(CHtmlFormatter *formatter)
{
    CHtmlRect margins;

    /* create the main HTML panel as a debugger source HTML panel */
    create_html_panel(formatter);

    /* don't inset from the left margin at all */
    margins = formatter->get_phys_margins();
    margins.left = 0;
    formatter->set_phys_margins(&margins);
}

/*
 *   create the HTML panel 
 */
void CHtmlSys_dbgsrcwin::create_html_panel(CHtmlFormatter *formatter)
{
    /* create the debugger source window panel */
    html_panel_ = new CHtmlSysWin_win32_dbgsrc(formatter, this, TRUE, TRUE,
                                               prefs_, helper_, dbgmain_);
}

void CHtmlSys_dbgsrcwin::do_destroy()
{
    CHtmlDebugHelper *helper;
    CHtmlSysWin *win;

    /* 
     *   tell the main window that I'm gone, in case it's keeping track of
     *   me as a background window for routing commands 
     */
    dbgmain_->on_close_dbg_win(this);

    /* 
     *   note my helper and HTML panel objects for later -- since 'this'
     *   will be destroyed when we inherit our superclass do_destroy(), we
     *   won't be able to access my members after the inherited
     *   do_destroy() returns 
     */
    helper = helper_;
    win = get_html_panel();
    
    /* 
     *   inherit default handling - this will delete 'this', so we can't
     *   access any of my member variables after this call 
     */
    CHtmlSys_dbgwin::do_destroy();

    /* 
     *   Now that the window has been deleted and thus disentangled from
     *   the parser and formatter, tell the helper to delete its objects
     *   associated with the window. 
     */
    helper_->on_close_srcwin(win);
}

/*
 *   Adjust screen coordinates to frame-relative coordinates, if this
 *   window is contained inside another window.  For MDI mode, this will
 *   adjust the given screen coordinates so they're relative to the MDI
 *   client window that contains this window; for SDI mode, this won't
 *   change the coordinates, since the window is not contained inside any
 *   other window. 
 */
void CHtmlSys_dbgsrcwin::adjust_coords_to_frame(CHtmlRect *newpos,
                                                const CHtmlRect *pos)
{
    RECT rc;

    /* set up the RECT structure with the screen coordinates */
    SetRect(&rc, pos->left, pos->top, pos->right, pos->bottom);

    /* adjust the coordinates */
    w32tdb_adjust_client_pos(dbgmain_, &rc);

    /* store the result */
    newpos->set(rc.left, rc.top, rc.right, rc.bottom);
}

/*
 *   store my position in the preferences 
 */
void CHtmlSys_dbgsrcwin::set_winpos_prefs(const CHtmlRect *pos)
{
    char win_title[256];
    CHtmlRect newpos;

    /* get my window title */
    GetWindowText(handle_, win_title, sizeof(win_title));
    
    /* adjust to frame-relative coordinates (for MDI mode) */
    adjust_coords_to_frame(&newpos, pos);

    /* store the position in the debugger configuration */
    dbgmain_->get_config()->setval("src win pos", win_title, &newpos);
}

/*
 *   bring this owner window to the front 
 */
void CHtmlSys_dbgsrcwin::bring_owner_to_front()
{
    /* inherit default behavior */
    CHtmlSys_dbgwin::bring_owner_to_front();

    /* 
     *   update the caret position, since we may have changed it since we
     *   last set it 
     */
    html_panel_->update_caret_pos(TRUE, FALSE);
}

/*
 *   Activate/deactivate.  Notify the debugger main window of the coming
 *   or going of this debugger window as the active debugger window, which
 *   will allow us to receive certain command messages from the main
 *   window.  
 */
int CHtmlSys_dbgsrcwin::do_activate(int flag, int minimized, HWND other_win)
{
    /* register or deregister with the main window */
    if (flag == WA_INACTIVE)
        dbgmain_->clear_active_dbg_win();
    else
        dbgmain_->set_active_dbg_win(this);

    /* inherit default handling */
    return CHtmlSys_dbgwin::do_activate(flag, minimized, other_win);
}

/*
 *   Check a command status.  Pass the command to the main window for
 *   routine.  
 */
TadsCmdStat_t CHtmlSys_dbgsrcwin::check_command(int command_id)
{
    return dbgmain_->check_command(command_id);
}

/*
 *   Process a command.  Pass the command to the main window for routing.  
 */
int CHtmlSys_dbgsrcwin::do_command(int notify_code, int command_id, HWND ctl)
{
    return dbgmain_->do_command(notify_code, command_id, ctl);
}

/*
 *   Check a command status, for routing from the main window.  Pass the
 *   command to our HTML panel for processing.  
 */
TadsCmdStat_t CHtmlSys_dbgsrcwin::active_check_command(int command_id)
{
    return ((CHtmlSysWin_win32_dbgsrc *)html_panel_)
        ->active_check_command(command_id);
}

/*
 *   Process a command, for routing from the main window.  Pass the
 *   command to our HTML panel.  
 */
int CHtmlSys_dbgsrcwin::active_do_command(int notify_code, int command_id,
                                          HWND ctl)
{
    return ((CHtmlSysWin_win32_dbgsrc *)html_panel_)
        ->active_do_command(notify_code, command_id, ctl);
}

/*
 *   Get my selection range 
 */
int CHtmlSys_dbgsrcwin::get_sel_range(CHtmlFormatter **formatter,
                                      unsigned long *sel_start,
                                      unsigned long *sel_end)
{
    /* get the selection range from my html panel's formatter */
    *formatter = ((CHtmlSysWin_win32_dbgsrc *)html_panel_)
                 ->get_formatter();
    (*formatter)->get_sel_range(sel_start, sel_end);

    /* we do have a selection range */
    return TRUE;
}


/* ------------------------------------------------------------------------ */
/*
 *   Debugger source HTML panel window implementation
 */

CHtmlSysWin_win32_dbgsrc::CHtmlSysWin_win32_dbgsrc(
    CHtmlFormatter *formatter, CHtmlSysWin_win32_owner *owner,
    int has_vscroll, int has_hscroll, CHtmlPreferences *prefs,
    CHtmlDebugHelper *helper, CHtmlSys_dbgmain *dbgmain)
    : CHtmlSysWin_win32(formatter, owner, has_vscroll, has_hscroll, prefs)
{
    /* remember the debugger helper object and main window */
    helper_ = helper;
    dbgmain_ = dbgmain;

    /* load the debugger bitmaps */
    ico_dbg_curline_ = LoadImage(CTadsApp::get_app()->get_instance(),
                                 MAKEINTRESOURCE(IDI_DBG_CURLINE),
                                 IMAGE_ICON, 16, 16, 0);
    ico_dbg_curline_ = LoadImage(CTadsApp::get_app()->get_instance(),
                                 MAKEINTRESOURCE(IDI_DBG_CURLINE),
                                 IMAGE_ICON, 16, 16, 0);
    ico_dbg_ctxline_ = LoadImage(CTadsApp::get_app()->get_instance(),
                                 MAKEINTRESOURCE(IDI_DBG_CTXLINE),
                                 IMAGE_ICON, 16, 16, 0);
    ico_dbg_bp_ = LoadImage(CTadsApp::get_app()->get_instance(),
                            MAKEINTRESOURCE(IDI_DBG_BP),
                            IMAGE_ICON, 16, 16, 0);
    ico_dbg_bpdis_ = LoadImage(CTadsApp::get_app()->get_instance(),
                               MAKEINTRESOURCE(IDI_DBG_BPDIS),
                               IMAGE_ICON, 16, 16, 0);
    ico_dbg_curbp_ = LoadImage(CTadsApp::get_app()->get_instance(),
                               MAKEINTRESOURCE(IDI_DBG_CURBP),
                               IMAGE_ICON, 16, 16, 0);
    ico_dbg_curbpdis_ = LoadImage(CTadsApp::get_app()->get_instance(),
                                  MAKEINTRESOURCE(IDI_DBG_CURBPDIS),
                                  IMAGE_ICON, 16, 16, 0);
    ico_dbg_ctxbp_ = LoadImage(CTadsApp::get_app()->get_instance(),
                               MAKEINTRESOURCE(IDI_DBG_CTXBP),
                               IMAGE_ICON, 16, 16, 0);
    ico_dbg_ctxbpdis_ = LoadImage(CTadsApp::get_app()->get_instance(),
                                  MAKEINTRESOURCE(IDI_DBG_CTXBPDIS),
                                  IMAGE_ICON, 16, 16, 0);

    /* enable the caret */
    caret_enabled_ = TRUE;

    /* get rid of the default popup menu and load a new one instead */
    load_context_popup(IDR_DEBUG_POPUP_MENU);

    /* no target column for up/down arrow yet */
    target_col_valid_ = FALSE;
}

CHtmlSysWin_win32_dbgsrc::~CHtmlSysWin_win32_dbgsrc()
{
    /* done with the debugger bitmaps */
    DeleteObject(ico_dbg_curline_);
    DeleteObject(ico_dbg_curline_);
    DeleteObject(ico_dbg_ctxline_);
    DeleteObject(ico_dbg_bp_);
    DeleteObject(ico_dbg_bpdis_);
    DeleteObject(ico_dbg_curbp_);
    DeleteObject(ico_dbg_curbpdis_);
    DeleteObject(ico_dbg_ctxbp_);
    DeleteObject(ico_dbg_ctxbpdis_);
}

/*
 *   create the window 
 */
void CHtmlSysWin_win32_dbgsrc::do_create()
{
    /* inherit default handling */
    CHtmlSysWin_win32::do_create();
}

/*
 *   receive notifications
 */
int CHtmlSysWin_win32_dbgsrc::do_notify(int control_id, int notify_code,
                                        LPNMHDR nmhdr)
{
    switch(notify_code)
    {
    case TTN_NEEDTEXT:
        {
            TOOLTIPTEXT *ttt = (TOOLTIPTEXT *)nmhdr;
            unsigned long sel_start, sel_end;
            unsigned long ofs;
            POINT pt;
            CHtmlPoint hpt;
            CHtmlTextArray *arr;
            unsigned long len;
            static textchar_t buf[512];
            textchar_t valbuf[512];
            textchar_t *p;
            int all_spaces;
            int expr_good;

            /* get the mouse position */
            GetCursorPos(&pt);
            ScreenToClient(handle_, &pt);

            /* get the position of the mouse */
            hpt = screen_to_doc(CHtmlPoint(pt.x, pt.y));
            ofs = formatter_->find_textofs_by_pos(hpt);

            /* get the selection range */
            formatter_->get_sel_range(&sel_start, &sel_end);

            /* 
             *   if the mouse is over the selection, use the whole
             *   selection range; otherwise, just use the current word 
             */
            if (ofs < sel_start || ofs >= sel_end)
                formatter_->get_word_limits(&sel_start, &sel_end, ofs);

            /*
             *   Get text out of the range, up to our buffer size 
             */
            arr = formatter_->get_text_array();
            len = arr->get_char_count(sel_start, sel_end);
            if (len > sizeof(buf) - 1)
                len = sizeof(buf) - 1;

            /* get the selection in chunks */
            for (p = buf ; len > 0 ; )
            {
                const textchar_t *chunk;
                size_t curlen;
                
                /* get the next chunk */
                chunk = arr->get_text_chunk(&sel_start, &curlen, len);
                memcpy(p, chunk, curlen);
                
                /* advance our pointers */
                len -= curlen;
                p += curlen;
            }
            
            /* null-terminate the string */
            *p = '\0';

            /* 
             *   Scan through the text, converting control characters to
             *   spaces.  Also look to see if there are any non-space
             *   characters; if not, we won't bother popping anything up
             *   at this position.  
             */
            for (p = buf, all_spaces = TRUE ; *p != '\0' ; ++p)
            {
                /* if it's not printable, convert it to a space */
                if (!isprint(*p))
                    *p = ' ';

                /* if it's not a space, so note */
                if (!isspace(*p))
                    all_spaces = FALSE;
            }

            /* 
             *   presume this won't work - there are more ways to fail
             *   than to succeed 
             */
            expr_good = FALSE;

            /* if it's not all spaces, try evaluating the expression */
            if (!all_spaces)
            {
                /* 
                 *   try evaluating it - use speculative mode, so that we
                 *   don't accidentally change anything just because the
                 *   user was moving the mouse around 
                 */
                if (!helper_->eval_expr(dbgmain_->get_dbg_ctx(),
                                        valbuf, sizeof(valbuf),
                                        buf, 0, 0, 0, 0, TRUE))
                {
                    size_t rem;
                    size_t copylen;

                    /*
                     *   Suppress tautologies -- if the expression and the
                     *   result are identical, don't bother showing it.
                     *   If they're not identical, show the result.  
                     */
                    if (get_strcmp(buf, valbuf) != 0)
                    {
                        /* 
                         *   the expression is good -- add the result to
                         *   the expression buffer, and show the result as
                         *   the tip text 
                         */
                        rem = sizeof(buf) - (p - buf);
                        if (rem > 3)
                        {
                            strcpy(p, " = ");
                            p += 3;
                            rem -= 3;
                        }
                        
                        /* copy as much as we have room for */
                        copylen = get_strlen(valbuf);
                        if (copylen > rem - 1)
                            copylen = rem - 1;
                        memcpy(p, valbuf, copylen);
                        p[copylen] = '\0';
                        
                        /* use the message */
                        expr_good = TRUE;
                        ttt->lpszText = buf;
                    }
                }
            }

            /*
             *   If we didn't get a valid expression, don't show anything.
             *   To avoid popping up the tip, deactivate the tip, then
             *   immediately reactivate it.  This will suppress this
             *   message display, but will leave the tooltip active for
             *   another try later.  
             */
            if (!expr_good)
            {
                SendMessage(tooltip_, TTM_ACTIVATE, FALSE, 0);
                SendMessage(tooltip_, TTM_ACTIVATE, TRUE, 0);
            }
            return TRUE;
        }

    case TTN_SHOW:
        {
            TOOLINFO ti;
            POINT pt;

            /* get the mouse position */
            GetCursorPos(&pt);
            ScreenToClient(handle_, &pt);

            /*
             *   While it's showing, replace the bounding box for the tool
             *   with the bounding box for the screen area containing the
             *   mouse.  If the user moves out of the box, we'll want to
             *   act as though they've moved on to a new tip.  
             */
            ti.cbSize = sizeof(ti);
            SetRect(&ti.rect, pt.x < 10 ? 0 : pt.x - 10,
                    pt.y < 10 ? 0 : pt.y - 10, pt.x + 10, pt.y + 10);
            ti.hwnd = handle_;
            ti.uId = 1;
            SendMessage(tooltip_, TTM_NEWTOOLRECT, 0, (LPARAM)&ti);
            return TRUE;
        }

    default:
        /* inherit default */
        return CHtmlSysWin_win32::do_notify(control_id, notify_code, nmhdr);
    }
}

int CHtmlSysWin_win32_dbgsrc::do_leftbtn_down(int keys, int x, int y,
                                              int clicks)
{
    /*
     *   When the tooltip control sees a mouse click, it turns itself off
     *   until the mouse moves out of the tool.  We don't want this to
     *   happen, because we're taking over the whole window for the tool.
     *   To get the tooltip to keep popping up after a click, deactivate
     *   it and then immediately reactivate it.  
     */
    SendMessage(tooltip_, TTM_ACTIVATE, FALSE, 0);
    SendMessage(tooltip_, TTM_ACTIVATE, TRUE, 0);

    /* forget any target column we had */
    target_col_valid_ = FALSE;
    
    /* inherit default */
    return CHtmlSysWin_win32::do_leftbtn_down(keys, x, y, clicks);
}

/*
 *   Handle system (ALT) keystrokes.  Send the keystroke to the main
 *   debugger window, so that the user can access the main debugger
 *   window's menu bar from the keyboard while a source window has focus. 
 */
int CHtmlSysWin_win32_dbgsrc::do_syskeydown(int vkey, long keydata)
{
    dbgmain_->run_menu_kb_ifc(vkey, keydata, GetParent(handle_));
    return TRUE;
}


/*
 *   Handle keystrokes 
 */
int CHtmlSysWin_win32_dbgsrc::do_keydown(int vkey, long keydata)
{
    /* check the key */
    switch(vkey)
    {
    case VK_UP:
    case VK_DOWN:
    case VK_LEFT:
    case VK_RIGHT:
    case VK_END:
    case VK_HOME:
    case VK_NEXT:
    case VK_PRIOR:
        /* these keystrokes all move the caret */
        {
            unsigned long sel_start, sel_end;
            unsigned long caret_end, other_end;
            
            /* get current selection range */
            formatter_->get_sel_range(&sel_start, &sel_end);
            
            /* 
             *   if the shift key is down, note the caret end of the range
             *   and the other end of the range; otherwise, treat them as
             *   equivalent 
             */
            if (get_shift_key())
            {
                /* shift key is down - note both ends of the range */
                if (caret_at_right_)
                {
                    caret_end = sel_end;
                    other_end = sel_start;
                }
                else
                {
                    caret_end = sel_start;
                    other_end = sel_end;
                }
            }
            else
            {
                /* no shift key - we only care about the caret position */
                caret_end = other_end =
                    (caret_at_right_ ? sel_end : sel_start);
            }

            switch(vkey)
            {
            case VK_NEXT:
            case VK_PRIOR:
                {
                    RECT rc;
                    int lines;
                    
                    /* 
                     *   if the control key is down, scroll without moving
                     *   the cursor - simply treat this the same way the
                     *   base class does 
                     */
                    if (get_ctl_key())
                        return CHtmlSysWin_win32::do_keydown(vkey, keydata);

                    /* get the window's height */
                    GetClientRect(handle_, &rc);

                    /* 
                     *   figure out how many lines that will hold
                     *   (assuming that all of the lines are of constant
                     *   height, which is the case for our type of window) 
                     */
                    if (formatter_->get_first_disp() != 0)
                    {
                        CHtmlRect lrc;

                        /* get the height of a line */
                        lrc = formatter_->get_first_disp()->get_pos();

                        /* figure out how many fit in the window */
                        lines = (rc.bottom - rc.top) / (lrc.bottom - lrc.top);

                        /* 
                         *   for context, move one line less than the
                         *   window will hold 
                         */
                        if (lines > 1)
                            --lines;
                    }
                    else
                    {
                        /* 
                         *   no lines - use a default window size (it
                         *   doesn't even matter, since if there are no
                         *   lines we can't do any real navigation) 
                         */
                        lines = 25;
                    }
                    
                    /* move the caret up or down by one screenfull */
                    caret_end =
                        do_keydown_arrow(vkey == VK_PRIOR ? VK_UP : VK_DOWN,
                                         caret_end, lines);
                }
                break;
                
            case VK_UP:
            case VK_DOWN:
                caret_end = do_keydown_arrow(vkey, caret_end, 1);
                break;

            case VK_LEFT:
                /* move the caret left a character */
                caret_end = formatter_->get_text_array()->dec_ofs(caret_end);
                break;

            case VK_RIGHT:
                /* move the caret right a character */
                caret_end = formatter_->get_text_array()->inc_ofs(caret_end);
                break;

            case VK_END:
                if (get_ctl_key())
                {
                    /* move to the end of the file */
                    caret_end = formatter_->get_text_array()->get_max_addr();
                }
                else
                {
                    unsigned long lstart, lend;

                    /* move to the end of the current line */
                    formatter_->get_line_limits(&lstart, &lend, caret_end);
                    caret_end = formatter_->get_text_array()->dec_ofs(lend);
                }
                break;

            case VK_HOME:
                if (get_ctl_key())
                {
                    /* move to the start of the file */
                    caret_end = 0;
                }
                else
                {
                    unsigned long lstart, lend;

                    /* move to the start of the current line */
                    formatter_->get_line_limits(&lstart, &lend, caret_end);
                    caret_end = lstart;
                }
                break;
            }

            /* 
             *   if the keystroke wasn't an up/down arrow, forget any
             *   target column we had 
             */
            if (vkey != VK_UP && vkey != VK_DOWN
                && vkey != VK_NEXT && vkey != VK_PRIOR)
                target_col_valid_ = FALSE;

            /*
             *   If we reached the very end, because of the way the final
             *   newlines are laid out, we may find ourselves past the end
             *   of the last item.  If this is the case, decrement our
             *   position so that we're at the start of the last item. 
             */
            if (caret_end == formatter_->get_text_array()->get_max_addr()
                && formatter_->find_by_txtofs(caret_end, FALSE, FALSE) == 0)
            {
                CHtmlDisp *disp;

                /* get the previous item */
                disp = formatter_->find_by_txtofs(caret_end, TRUE, FALSE);

                /* move to the start of that item */
                caret_end = disp->get_text_ofs();
            }

            /* figure out the new selection range */
            if (get_shift_key())
            {
                /* extended range - figure out what direction to go */
                if (caret_end < other_end)
                {
                    /* the caret is at the low (left) end of the range */
                    sel_start = caret_end;
                    sel_end = other_end;
                    caret_at_right_ = FALSE;
                }
                else
                {
                    /* the caret is at the high (right) end */
                    sel_start = other_end;
                    sel_end = caret_end;
                    caret_at_right_ = TRUE;
                }
            }
            else
            {
                /* no shift key - set zero-length range at caret */
                sel_start = sel_end = caret_end;
            }

            /* set the new selection range */
            formatter_->set_sel_range(sel_start, sel_end);
            
            /* make sure the selection is on the screen */
            update_caret_pos(TRUE, FALSE);
        }

        /* handled */
        return TRUE;
        
    default:
        /* inherit default for other keys */
        return CHtmlSysWin_win32::do_keydown(vkey, keydata);
    }
}

/*
 *   process an arrow key; returns new text offset after movement
 */
unsigned long CHtmlSysWin_win32_dbgsrc::
   do_keydown_arrow(int vkey, unsigned long caret_pos, int cnt)
{
    /* keep going for the requested number of lines */
    for ( ; cnt > 0 ; --cnt)
    {
        unsigned long lstart, lend;
        unsigned long ofs;
        CHtmlDisp *nxt;

        /* 
         *   find our current column by examining the limits of the
         *   current line 
         */
        formatter_->get_line_limits(&lstart, &lend, caret_pos);
        
        /* 
         *   if we don't have an existing valid target column, use the
         *   current column as the new target 
         */
        if (!target_col_valid_)
        {
            target_col_ = caret_pos - lstart;
            target_col_valid_ = TRUE;
        }
        
        /* get the next/previous line's display item */
        if (vkey == VK_UP)
        {
            ofs = formatter_->get_text_array()->dec_ofs(lstart);
            nxt = formatter_->find_by_txtofs(ofs, TRUE, FALSE);
        }
        else
        {
            ofs = lend;
            nxt = formatter_->find_by_txtofs(ofs, FALSE, TRUE);
        }

        /* make sure we got a target item */
        if (nxt != 0)
        {
            CHtmlDispTextDebugsrc *disp;
            unsigned int cur_target;
            
            /* cast it to my correct type */
            disp = (CHtmlDispTextDebugsrc *)nxt;
            
            /* see if it contains the target column */
            if (disp->get_text_len() > target_col_)
            {
                /* it's long enough - go to the taret */
                cur_target = target_col_;
            }
            else
            {
                /* it's too short - go to its end */
                cur_target = disp->get_text_len();
            }
            
            /* move to the correct spot */
            caret_pos = disp->get_text_ofs() + cur_target;
        }
    }

    /* return the new caret position */
    return caret_pos;
}

/*
 *   set the colors 
 */
void CHtmlSysWin_win32_dbgsrc::set_src_colors(HTML_color_t bkg_color,
                                              HTML_color_t text_color,
                                              int use_windows_colors)
{
    COLORREF bkg_rgb, text_rgb;
    
    /* if we're using the windows colors, get the defaults */
    if (use_windows_colors)
    {
        bkg_rgb = GetSysColor(COLOR_WINDOW);
        text_rgb = GetSysColor(COLOR_WINDOWTEXT);
    }
    else
    {
        bkg_rgb = HTML_color_to_COLORREF(bkg_color);
        text_rgb = HTML_color_to_COLORREF(text_color);
    }

    /* set the colors */
    internal_set_bg_color(bkg_rgb);
    internal_set_text_color(text_rgb);
}


/*
 *   Measure the debugger source icon 
 */
CHtmlPoint CHtmlSysWin_win32_dbgsrc::measure_dbgsrc_icon()
{
    CHtmlPoint retval;

    /* 
     *   the icon is 16x16, but put in a couple of extra pixels
     *   horizontally to space things out 
     */
    retval.x = 20;
    retval.y = 16;
    return retval;
}

/*
 *   Draw a debugger source line icon 
 */
void CHtmlSysWin_win32_dbgsrc::draw_dbgsrc_icon(const CHtmlRect *pos,
                                                unsigned int stat)
{
    CHtmlRect scrpos;
    RECT rc;
    HICON ico;

    /* set up a windows RECT structure for the screen coordinates */
    scrpos = doc_to_screen(*pos);
    scrpos.right -= 2;
    SetRect(&rc, scrpos.left, scrpos.top, scrpos.right, scrpos.bottom);

    /* draw the margin background */
    draw_margin_bkg(&rc);

    /* offset the area slightly to leave blank space in the margin */
    rc.left += 2;

    /* if there's no breakpoint, lose any related status */
    if ((stat & HTMLTDB_STAT_BP) == 0)
        stat &= ~(HTMLTDB_STAT_BP_DIS | HTMLTDB_STAT_BP_COND);

    /* select the correct icon based on the status */
    switch(stat & (HTMLTDB_STAT_CURLINE | HTMLTDB_STAT_CTXLINE
                   | HTMLTDB_STAT_BP | HTMLTDB_STAT_BP_DIS))
    {
    case HTMLTDB_STAT_CURLINE:
    case HTMLTDB_STAT_CURLINE | HTMLTDB_STAT_CTXLINE:
        ico = ico_dbg_curline_;
        break;

    case HTMLTDB_STAT_CTXLINE:
        ico = ico_dbg_ctxline_;
        break;

    case HTMLTDB_STAT_BP:
        ico = ico_dbg_bp_;
        break;

    case HTMLTDB_STAT_BP | HTMLTDB_STAT_BP_DIS:
        ico = ico_dbg_bpdis_;
        break;

    case HTMLTDB_STAT_CURLINE | HTMLTDB_STAT_BP:
        ico = ico_dbg_curbp_;
        break;

    case HTMLTDB_STAT_CURLINE | HTMLTDB_STAT_BP | HTMLTDB_STAT_BP_DIS:
        ico = ico_dbg_curbpdis_;
        break;

    case HTMLTDB_STAT_CTXLINE | HTMLTDB_STAT_BP:
        ico = ico_dbg_ctxbp_;
        break;

    case HTMLTDB_STAT_CTXLINE | HTMLTDB_STAT_BP | HTMLTDB_STAT_BP_DIS:
        ico = ico_dbg_ctxbpdis_;
        break;

    default:
        ico = 0;
        break;
    }

    /* draw the icon in the given area */
    if (ico != 0)
        DrawIconEx(hdc_, rc.left, rc.top, ico, 16, 16, 0, 0, DI_NORMAL);
}

/*
 *   draw the margin background 
 */
void CHtmlSysWin_win32_dbgsrc::draw_margin_bkg(const RECT *rc)
{
    /* fill in the area with a light gray background */
    FillRect(hdc_, rc, GetSysColorBrush(COLOR_3DFACE));
}

/*
 *   Set the background cursor.  We use the i-beam cursor when the mouse
 *   isn't over a display item.  
 */
int CHtmlSysWin_win32_dbgsrc::do_setcursor_bkg()
{
    /* use the I-beam cursor */
    SetCursor(ibeam_csr_);

    /* indicate that we set a cursor */
    return TRUE;
}

/*
 *   Check a command - route to the main window for processing 
 */
TadsCmdStat_t CHtmlSysWin_win32_dbgsrc::check_command(int command_id)
{
    /* route it to the main window for processing */
    return dbgmain_->check_command(command_id);
}

/*
 *   Process a command - route it to the main window for processing 
 */
int CHtmlSysWin_win32_dbgsrc::do_command(int notify_code,
                                         int command_id, HWND ctl)
{
    /* route it to the main window for processing */
    return dbgmain_->do_command(notify_code, command_id, ctl);
}

/*
 *   Determine command status, routed from the main window
 */
TadsCmdStat_t CHtmlSysWin_win32_dbgsrc::active_check_command(int command_id)
{
    switch(command_id)
    {
    case ID_DEBUG_SETCLEARBREAKPOINT:
    case ID_DEBUG_DISABLEBREAKPOINT:
    case ID_EDIT_FIND:
    case ID_EDIT_FINDNEXT:
    case ID_DEBUG_SETNEXTSTATEMENT:
    case ID_DEBUG_RUNTOCURSOR:
    case ID_DEBUG_ADDTOWATCH:
        /* these are always allowed in a source window */
        return TADSCMD_ENABLED;

    default:
        /* inherit base class behavior for other commands */
        return CHtmlSysWin_win32::check_command(command_id);
    }
}

/*
 *   Handle a command routed from the main window
 */
int CHtmlSysWin_win32_dbgsrc::active_do_command(int notify_code,
                                                int command_id, HWND ctl)
{
    switch (command_id)
    {
    case ID_EDIT_FIND:
    case ID_EDIT_FINDNEXT:
        do_find(command_id);
        return TRUE;
        
    case ID_DEBUG_SETCLEARBREAKPOINT:
        /* set/clear a breakpoint at the current selection */
        helper_->toggle_breakpoint(dbgmain_->get_dbg_ctx(), this);
        return TRUE;

    case ID_DEBUG_DISABLEBREAKPOINT:
        /* enable/disable breakpoint at the current selection */
        helper_->toggle_bp_disable(dbgmain_->get_dbg_ctx(), this);
        return TRUE;

    case ID_DEBUG_RUNTOCURSOR:
        /* try to set a temporary breakpoint here */
        if (helper_->set_temp_bp(dbgmain_->get_dbg_ctx(), this))
        {
            /* got it - tell the main window to resume execution */
            dbgmain_->exec_go();
        }
        else
        {
            /* 
             *   couldn't set a temporary breakpoint - it's not worth a
             *   whole error dialog, but beep to indicate that it failed 
             */
            Beep(1000, 100);
        }
        return TRUE;

    case ID_DEBUG_SETNEXTSTATEMENT:
        /* 
         *   make sure the game is running, and the new address is within
         *   the current function 
         */
        if (dbgmain_->get_game_over())
        {
            Beep(1000, 100);
            return TRUE;
        }
        else if (!helper_->is_in_same_fn(dbgmain_->get_dbg_ctx(),
                                         dbgmain_->get_exec_ofs_ptr(), this))
        {
            MessageBox(0, "This is not a valid location - execution "
                       "cannot be transferred outside of the current "
                       "function or method.", "TADS Debugger",
                       MB_OK | MB_ICONEXCLAMATION | MB_TASKMODAL);
            return TRUE;
        }
        
        /* set the next statement here */
        if (helper_->set_next_statement(dbgmain_->get_dbg_ctx(),
                                        dbgmain_->get_exec_ofs_ptr(), this))
        {
            /* something went wrong - beep an error */
            Beep(1000, 100);
        }
        else
        {
            /* 
             *   success - execute for a single step, so that we enter
             *   into the current line 
             */
            dbgmain_->exec_step_into();
        }

        /* handled */
        return TRUE;
        
    default:
        /* inherit base class behavior */
        return CHtmlSysWin_win32::do_command(notify_code, command_id, ctl);
    }
}

/*
 *   Find text 
 */
void CHtmlSysWin_win32_dbgsrc::do_find(int command_id)
{
    const textchar_t *findstr;
    CHtmlTextArray *txtarr;
    unsigned long sel_start, sel_end, start_ofs;
    HCURSOR oldcsr;
    int found;
    int exact_case;
    int start_at_top;
    int wrap;
    int dir;

    /* get the text to find */
    findstr = dbgmain_->get_find_text(command_id, &exact_case,
                                      &start_at_top, &wrap, &dir);

    /* if they cancelled, there's nothing more to do */
    if (findstr == 0)
        return;

    /* get the text array */
    txtarr = get_formatter()->get_text_array();

    /* 
     *   If we're starting at the top, search from offset zero; otherwise,
     *   get the current selection range and search from the end of the
     *   range. 
     */
    if (start_at_top)
        sel_start = sel_end = 0;
    else
        get_formatter()->get_sel_range(&sel_start, &sel_end);

    /* show the busy cursor while working */
    oldcsr = SetCursor(wait_cursor_);

    /* 
     *   if we're searching forward, start at the end of the current
     *   selection range; otherwise, start at the start of the current
     *   range 
     */
    start_ofs = (dir == 1 ? sel_end : sel_start);

    /* 
     *   search in the text array, starting at the end of the selection
     *   range 
     */
    found = txtarr->search(findstr, get_strlen(findstr), exact_case,
                           wrap, dir, start_ofs, &sel_start, &sel_end);

    /* done working; restore the normal cursor */
    SetCursor(oldcsr);

    /* check to see if the search succeeded */
    if (found)
    {
        /* select the target that we found */
        get_formatter()->set_sel_range(sel_start, sel_end);

        /* make sure the selection is in view */
        update_caret_pos(TRUE, FALSE);
    }
    else
    {
        /* display a message box telling the user that we failed */
        MessageBox(0, "No more matches found.", "TADS Debugger",
                   MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Debugger stack window implementation.  This is a simple
 *   specialization of the debugger source window which provides a
 *   slightly different drawing style. 
 */

/*
 *   create the HTML panel 
 */
void CHtmlSys_dbgstkwin::create_html_panel(CHtmlFormatter *formatter)
{
    /* create the debugger stack window panel */
    html_panel_ = new CHtmlSysWin_win32_dbgstk(formatter, this, TRUE, TRUE,
                                               prefs_, helper_, dbgmain_);
}

/*
 *   store my position in the preferences 
 */
void CHtmlSys_dbgstkwin::set_winpos_prefs(const CHtmlRect *pos)
{
    CHtmlRect newpos;

    /* adjust to frame-relative coordinates (for MDI mode) */
    adjust_coords_to_frame(&newpos, pos);

    /* store the position in the debugger configuration */
    dbgmain_->get_config()->setval("stack win pos", 0, &newpos);
}

/*
 *   process a mouse click 
 */
int CHtmlSysWin_win32_dbgstk::do_leftbtn_down(int keys, int x, int y,
                                              int clicks)
{
    /* if we have a double-click, do some special processing */
    if (clicks == 2)
    {
        CHtmlPoint pos;
        CHtmlDisp *disp;

        /* ignore it if we're not in the debugger */
        if (!dbgmain_->get_in_debugger())
        {
            Beep(1000, 100);
            return TRUE;
        }

        /* 
         *   Get the document coordinates.  Ignore the x position of the
         *   click; act as though they're clicking on the left edge of the
         *   text area.  
         */
        pos = screen_to_doc(CHtmlPoint(x, y));
        pos.x = 0;

        /* find the item we're clicking on */
        disp = formatter_->find_by_pos(pos, TRUE);
        if (disp != 0)
        {
            CHtmlSysWin_win32 *win;
            
            /* tell the helper to select this item */
            win = (CHtmlSysWin_win32 *)helper_->activate_stack_win_item(
                disp, dbgmain_->get_dbg_ctx(), dbgmain_);

            /* 
             *   tell the debugger to refresh its expression windows for
             *   the new context 
             */
            dbgmain_->update_expr_windows();

            /* bring the source window to the front */
            if (win != 0)
                win->bring_to_front();
            
            /* don't track the mouse any further */
            end_mouse_tracking(TRUE);
            
            /* handled */
            return TRUE;
        }
    }
    
    /* inherit the default */
    return CHtmlSysWin_win32_dbgsrc::do_leftbtn_down(keys, x, y, clicks);
}

/* 
 *   check the status of a command 
 */
TadsCmdStat_t CHtmlSysWin_win32_dbgstk::active_check_command(int command_id)
{
    switch(command_id)
    {
    case ID_DEBUG_SETCLEARBREAKPOINT:
    case ID_DEBUG_DISABLEBREAKPOINT:
    case ID_DEBUG_RUNTOCURSOR:
    case ID_DEBUG_SETNEXTSTATEMENT:
    case ID_DEBUG_ADDTOWATCH:
        /* can't do these in a stack window */
        return TADSCMD_DISABLED;

    default:
        /* inherit default */
        return CHtmlSysWin_win32_dbgsrc::active_check_command(command_id);
    }
}

/* 
 *   execute a command 
 */
int CHtmlSysWin_win32_dbgstk::active_do_command(int notify_code,
                                                int command_id, HWND ctl)
{
    switch(command_id)
    {
    case ID_DEBUG_SETCLEARBREAKPOINT:
    case ID_DEBUG_DISABLEBREAKPOINT:
    case ID_DEBUG_RUNTOCURSOR:
    case ID_DEBUG_SETNEXTSTATEMENT:
        /* can't do these in a stack window */
        return FALSE;

    default:
        /* inherit default */
        return CHtmlSysWin_win32_dbgsrc::
            active_do_command(notify_code, command_id, ctl);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   History window implementation 
 */

/*
 *   create the HTML panel 
 */
void CHtmlSys_dbghistwin::create_html_panel(CHtmlFormatter *formatter)
{
    /* create the debugger stack window panel */
    html_panel_ = new CHtmlSysWin_win32_dbghist(formatter, this, TRUE, TRUE,
                                                prefs_, helper_, dbgmain_);
}

/*
 *   store my position in the preferences 
 */
void CHtmlSys_dbghistwin::set_winpos_prefs(const CHtmlRect *pos)
{
    CHtmlRect newpos;

    /* adjust to frame-relative coordinates (for MDI mode) */
    adjust_coords_to_frame(&newpos, pos);

    /* store the position in the debugger configuration */
    dbgmain_->get_config()->setval("hist win pos", 0, &newpos);
}

/* 
 *   check the status of a command 
 */
TadsCmdStat_t CHtmlSysWin_win32_dbghist::active_check_command(int command_id)
{
    switch(command_id)
    {
    case ID_DEBUG_SETCLEARBREAKPOINT:
    case ID_DEBUG_DISABLEBREAKPOINT:
    case ID_DEBUG_RUNTOCURSOR:
    case ID_DEBUG_SETNEXTSTATEMENT:
        /* can't do these in a history window */
        return TADSCMD_DISABLED;

    default:
        /* inherit default */
        return CHtmlSysWin_win32_dbgsrc::active_check_command(command_id);
    }
}

/* 
 *   execute a command 
 */
int CHtmlSysWin_win32_dbghist::active_do_command(
    int notify_code, int command_id, HWND ctl)
{
    switch(command_id)
    {
    case ID_DEBUG_SETCLEARBREAKPOINT:
    case ID_DEBUG_DISABLEBREAKPOINT:
    case ID_DEBUG_RUNTOCURSOR:
    case ID_DEBUG_SETNEXTSTATEMENT:
        /* can't do these in a history window */
        return FALSE;

    default:
        /* inherit default */
        return CHtmlSysWin_win32_dbgsrc::
            active_do_command(notify_code, command_id, ctl);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Debugger about box dialog 
 */

void CTadsDlg_dbgabout::init()
{
    char buf[128];

    /* inherit default */
    CTadsDialog::init();

    /* set up the correct version number string */
    sprintf(buf, "Version %s", TADS_RUNTIME_VERSION);
    SetDlgItemText(handle_, IDC_TXT_ABOUTVSN, buf);
}
