/* $Header: d:/cvsroot/tads/html/win32/w32tdb.h,v 1.1.1.1 1999/05/09 21:32:32 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  w32tdb.h - html tads win32 debugger
Function
  
Notes
  
Modified
  01/31/98 MJRoberts  - Creation
*/

#ifndef W32TDB_H
#define W32TDB_H

#ifndef TADSWIN_H
#include "tadswin.h"
#endif
#ifndef HTMLRES_H
#include "htmlres.h"
#endif
#ifndef HTMLW32_H
#include "htmlw32.h"
#endif
#ifndef HTMLDBG_H
#include "htmldbg.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   MDI/SDI selector.  This block of code encapsulates all of the
 *   differences between running the debugger as an SDI application vs. as
 *   an MDI application.
 *   
 *   This is currently set up so that the MDI/SDI selection is made at
 *   compile time.  The default is MDI.  Define W32TDB_SDI on the compiler
 *   command line to compile an SDI version of the debugger.
 *   
 *   It would be possible to make the MDI/SDI selection at run-time, by
 *   changing the code below so that the MDI and SDI versions of each
 *   function are integrated into a single function that selects the
 *   appropriate behavior based on a global variable.  For now, it seems
 *   sufficient to support only a single mode, since MDI seems clearly
 *   superior to SDI for this application, but this could be changed if
 *   users are divided over which mode they prefer and want to be able to
 *   select a mode at run-time.  
 */

#ifndef W32TDB_SDI
#define W32TDB_MDI
#endif

#ifdef W32TDB_MDI

/* 
 *   MDI mode 
 */

/* use the MDI frame system interface object for the frame */
inline CTadsSyswin *w32tdb_new_frame_syswin(CTadsWin *win, int winmnuidx)
{
    return new CTadsSyswinMdiFrame(win, winmnuidx, ID_WINDOW_FIRST);
}

/* use the MDI child system interface object for children */
inline CTadsSyswin *w32tdb_new_child_syswin(CTadsWin *win)
{
    return new CTadsSyswinMdiChild(win);
}

/*
 *   Set the default frame window rectangle.  In MDI mode, make the frame
 *   window take up most of the desktop window.  
 */
inline void w32tdb_set_default_frame_rc(RECT *rc)
{
    /* get the desktop window area, and inset it a bit */
    SystemParametersInfo(SPI_GETWORKAREA, 0, (void *)rc, 0);
    InflateRect(rc, -16, -16);

    /* adjust the right and bottom elements to width and height */
    rc->right -= rc->left;
    rc->bottom -= rc->top;
}

/*
 *   Get the desktop window area to use for debugger child windows.  For
 *   MDI mode, we'll return the MDI client area.  
 */
inline void w32tdb_get_desktop_rc(CTadsWin *frame_win, RECT *rc)
{
    GetClientRect(frame_win->get_parent_of_children(), rc);
}

/*
 *   Set the default stack window position.  In MDI mode, we'll put the
 *   stack window at the top left of the client area.  
 */
inline void w32tdb_set_default_stk_rc(RECT *rc)
{
    SetRect(rc, 2, 2, 250, 350);
}

/* 
 *   style for main frame window - in MDI mode, make it a normal top-level
 *   window with all of the min/max controls 
 */
inline DWORD w32tdb_main_win_style()
{
    return (WS_OVERLAPPED | WS_BORDER | WS_THICKFRAME
            | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_MAXIMIZEBOX
            | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_SIZEBOX);
}

/*
 *   Bring the debugger window group to the front.  In MDI mode, we simply
 *   bring the main frame to the front, since all other debugger windows
 *   are subwindows in the frame window.  
 */
inline void w32tdb_bring_group_to_top(HWND frame_handle)
{
    BringWindowToTop(frame_handle);
}

/*
 *   Adjust the coordinates of a debugger child window to coordinates
 *   relative to the containing main frame window, if appropriate.  In MDI
 *   mode, this adjusts the coordinates to be relative to the MDI client
 *   window that contains the debugger window.  
 */
inline void w32tdb_adjust_client_pos(CTadsWin *frame_win, RECT *rc)
{
    POINT ofs;
    
    /* get the offset of the MDI client window */
    ofs.x = ofs.y = 0;
    ScreenToClient(frame_win->get_parent_of_children(), &ofs);

    /* adjust the caller's rectangle to be relative to the client window */
    OffsetRect(rc, ofs.x, ofs.y);
}

#else /* W32TDB_MDI */

/*
 *   SDI mode 
 */

/* use the standard system interface object for the frame */
inline CTadsSyswin *w32tdb_new_frame_syswin(CTadsWin *win, int /*winmnuidx*/)
{
    return new CTadsSyswin(win);
}

/* use the standard system interface object for child windows */
inline CTadsSyswin *w32tdb_new_child_syswin(CTadsWin *win)
{
    return new CTadsSyswin(win);
}

/*
 *   Set the default frame window rectangle.  In SDI mode, this window
 *   only contains the main menu and toolbar, so only make it big enough
 *   for the toolbar.  
 */
inline void w32tdb_set_default_frame_rc(RECT *rc)
{
    SetRect(rc, 2, 2, 440, 72);
}

/*
 *   Get the desktop window area to use for debugger child windows.  For
 *   SDI mode, we'll return the entire Windows desktop, since the debugger
 *   child windows are all top-level windows.  
 */
inline void w32tdb_get_desktop_rc(CTadsWin * /*frame_win*/, RECT *rc)
{
    /* get the desktop window working area */
    SystemParametersInfo(SPI_GETWORKAREA, 0, (void *)rc, 0);
}
/*
 *   Set the default stack window position.  In SDI mode, we'll put the
 *   stack window at the left of the screen, a ways below the main toolbar
 *   window.  
 */
inline void w32tdb_set_default_stk_rc(RECT *rc)
{
    SetRect(rc, 2, 120, 250, 350);
}

/* 
 *   style for main frame window - in SDI mode, there's no point in
 *   allowing this window to be maximized, so leave off the maximize
 *   button 
 */
inline DWORD w32tdb_main_win_style()
{
    return (WS_OVERLAPPED | WS_BORDER | WS_THICKFRAME
            | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX
            | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_SIZEBOX);
}


/*
 *   Bring the debugger window group to the front.  In SDI mode, we don't
 *   need to do anything here; we'll automatically bring the correct
 *   source window to the top, which will bring the entire window group to
 *   the top, since all debugger windows are part of a common ownership
 *   group.  
 */
inline void w32tdb_bring_group_to_top(HWND frame_handle)
{
#if 0 // $$$ no longer necessary - debugger windows are an ownership group now
    CHtmlSys_mainwin *mainwin;
    if ((mainwin = CHtmlSys_mainwin::get_main_win()) != 0
        && GetActiveWindow() == mainwin->get_handle())
        SetWindowPos(mainwin->get_handle(),
                     HWND_BOTTOM, 0, 0, 0, 0,
                     SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
#endif
}

/*
 *   Adjust the coordinates of a debugger child window to coordinates
 *   relative to the containing main frame window, if appropriate.  In SDI
 *   mode, the debugger windows are all top-level windows, so this doesn't
 *   do anything.  
 */
inline void w32tdb_adjust_client_pos(HWND /*frame_handle*/, RECT * /*rc*/)
{
    /* leave the coordinates unchanged in SDI mode */
}

#endif /* W32TDB_MDI */


/* ------------------------------------------------------------------------ */
/*
 *   Debugger main control window.  This is a top-level frame window that
 *   we create to hold the debugger menu and toolbar, and to control
 *   operations while the debugger is active.  
 */

class CHtmlSys_dbgmain:
   public CHtmlSys_framewin, public CHtmlDebugSysIfc_win,
   public CHtmlSysWin_win32_owner_null
{
public:
    CHtmlSys_dbgmain(struct dbgcxdef *ctx, const char *game_filename);
    ~CHtmlSys_dbgmain();

    /* 
     *   perform additional initialization required after the .GAM file
     *   has been loaded 
     */
    void post_load_init(struct dbgcxdef *ctx);

    /* 
     *   Debugger event loop.  This is called whenever we enter the
     *   debugger.  We retain control and handle events until the user
     *   resumes execution of the game.  Returns true if execution should
     *   continue, false if the application was terminated from within the
     *   event loop.
     *   
     *   If *restart is set to true on return, the caller should signal a
     *   TADS restart condition to cause the game to start over from the
     *   beginning.  If *abort is set to true on return, the caller should
     *   signal a TADS abort-command condition to cause the game to abort
     *   the current command and start a new command.  
     */
    int dbg_event_loop(struct dbgcxdef *ctx, int bphit, int err,
                       int *restart, int *abort, unsigned int *exec_ofs);

    /*
     *   Terminate the debugger.  Closes all debugger windows. 
     */
    void terminate_debugger();

    /* initialize a popup menu that's about to be displayed */
    void init_menu_popup(HMENU menuhdl, unsigned int pos, int sysmenu);

    /* receive notification that we've closed a menu */
    int menu_close(unsigned int item);
    
    /* process a command */
    int do_command(int notify_code, int command_id, HWND ctl);

    /* process notification */
    int do_notify(int control_id, int notify_code, LPNMHDR nmhdr);

    /* check a command status */
    TadsCmdStat_t check_command(int command_id);

    /*
     *   CHtmlDebugSysIfc_win implementation.  This is an interface that
     *   allows the portable helper code to call back into the
     *   system-dependent user interface code to create new windows. 
     */
    virtual class CHtmlSysWin
        *dbgsys_create_win(class CHtmlParser *parser,
                           class CHtmlFormatter *formatter,
                           const textchar_t *win_title,
                           HtmlDbg_wintype_t win_type);

    /*
     *   Select a source font 
     */
    virtual void dbgsys_set_srcfont(class CHtmlFontDesc *font_desc);

    /* compare filenames */
    virtual int dbgsys_cmp_fnames(const char *fname1, const char *fname2);


    /*
     *   CHtmlSys_win32_owner implementation.  We only override a few
     *   methods, since we implement most of this interface with the null
     *   version. 
     */
    virtual void bring_to_front();

    /* get the TADS debugger context */
    struct dbgcxdef *get_dbg_ctx() const { return dbgctx_; }

    /* get the pointer to the current execution offset in the debugger */
    unsigned int *get_exec_ofs_ptr() const { return exec_ofs_ptr_; }

    /* get the debugger helper object */
    class CHtmlDebugHelper *get_helper() const { return helper_; }

    /* get the configuration object */
    class CHtmlDebugConfig *get_config() const { return config_; }

    /*
     *   Get the active debugger window for command routing 
     */
    class CHtmlDbgSubWin *get_active_dbg_win();
    
    /*
     *   Set the active debugger window.  Source, stack, and other
     *   debugger windows should set themselves active when they are
     *   activated, and should clear the active window when they're
     *   deactivated. 
     */
    void set_active_dbg_win(class CHtmlDbgSubWin *win);
    void clear_active_dbg_win();

    /*
     *   Run the main debugger window's menu, activating it with the given
     *   system keydown message parameters.  This can be used when a
     *   window wants to let the keyboard menu interface in its window run
     *   the main debugger window's menu.  'srcwin' is the top-level
     *   window that's currently active; we'll re-activate this menu when
     *   we're done with the menu.  
     */
    void run_menu_kb_ifc(int vkey, long keydata, HANDLE srcwin);

    /*
     *   Receive notification that a debugger window is closing.  If I'm
     *   keeping track of the window as a background window for command
     *   routing, I'll forget about it now. 
     */
    void on_close_dbg_win(class CHtmlDbgSubWin *win);

    /* determine if the debugger has control */
    int get_in_debugger() const { return in_debugger_; }

    /* update expression windows */
    void update_expr_windows();

    /* 
     *   Add a new watch expression.  This will add the new expression at
     *   the end of the currently selected panel of the watch window.  
     */
    void add_watch_expr(const textchar_t *expr);

    /*
     *   Get the text for a Find command.  If the command is "Find", we'll
     *   run the dialog to get text from the user.  If the command is
     *   "Find Next," we'll simply return the text from the last Find
     *   command, if there was one; if not, we'll run the dialog as though
     *   this were a "Find" command.  If we run the dialog and the user
     *   cancels, we'll return a null pointer. 
     */
    const textchar_t *get_find_text(int command_id, int *exact_case,
                                    int *start_at_top, int *wrap, int *dir);

    /*
     *   Receive notification that source window preferences have changed.
     *   We'll update the source windows accordingly.  
     */
    void on_update_srcwin_options();

    /*
     *   Resume execution.  This is provided as a public method so that
     *   subwindows can tell the main window to start running. 
     */
    void exec_go();

    /*
     *   Resume execution for a single step. 
     */
    void exec_step_into();

    /*
     *   Get/Set the "game over" flag.  This flag is true when we have
     *   control after the game has terminated.  
     */
    int get_game_over() const { return game_over_; }
    void set_game_over(int flag) { game_over_ = flag; }

protected:
    /* window style - make it a document window without a maximize box */
    DWORD get_winstyle() { return w32tdb_main_win_style(); }

    /* window creation */
    void do_create();

    /* window destruction */
    void do_destroy();

    /* close the window */
    int do_close();

    /* erase the background */
    int do_erase_bkg(HDC hdc);

    /* resize */
    void do_resize(int mode, int x, int y);

    /* ask the user if they want to quit - returns true if so */
    int ask_quit();

    /* ask the user if they want to restart - returns true if so */
    int ask_restart();
    
    /* store a position in the appropriate preferences entry */
    virtual void set_winpos_prefs(const class CHtmlRect *pos);

private:
    /* apply color options to a window */
    void apply_color_options(class CHtmlSysWin_win32_dbgsrc *win);

    /* 
     *   Load the file menu - we'll call this on debugger entry if we
     *   haven't yet loaded the file menu.  This sets up the quick-open
     *   file submenu to include the game's source files.  
     */
    void load_file_menu();
    
    /* callback for buiding the file menu by enumerating line sources */
    static void load_menu_cb(void *ctx, const textchar_t *fname,
                             int line_source_id);

    /* callback for enumerating open windows for saving them to config */
    static void enum_win_save_cb(void *ctx0, int idx, CHtmlSysWin *win,
                                 int line_source_id,
                                 HtmlDbg_wintype_t win_type);

    /* callback for enumerating open windows for building the Window menu */
    static void winmenu_win_enum_cb(void *ctx0, int idx, CHtmlSysWin *win,
                                    int line_source_id,
                                    HtmlDbg_wintype_t win_type);

    /* callback for enumerating windows for updating option settings */
    static void options_enum_srcwin_cb(void *ctx, int idx,
                                       class CHtmlSysWin *win,
                                       int line_source_id,
                                       HtmlDbg_wintype_t win_type);
    
    /* 
     *   Context structure from TADS debugger.  We set this on each call
     *   from TADS, so that we have the correct TADS debugger context
     *   whenever we're active. 
     */
    struct dbgcxdef *dbgctx_;

    /*
     *   Pointer to the execution offset in the debugger.  By changing
     *   this, we can change the next instruction that will be executed.
     *   Note that we must not change this so that it's outside of the
     *   object.method or function that contained the original offset upon
     *   the most recent entry into the debugger.  
     */
    unsigned int *exec_ofs_ptr_;

    /*
     *   Debugger helper 
     */
    class CHtmlDebugHelper *helper_;

    /*
     *   Main HTMLTADS preferences object
     */
    class CHtmlPreferences *prefs_;

    /*
     *   The "Find" dialog object 
     */
    class CTadsDialogFind *find_dlg_;

    /* text of current/last find */
    textchar_t find_text_[512];

    /* flag indicating whether exact case should be used in searches */
    int search_exact_case_;

    /* flag indicating whether the search wraps at the end of the file */
    int search_wrap_;

    /* search direction: 1 for forwards, -1 for backwards */
    int search_dir_;

    /* new searches start from top of document */
    int search_from_top_;

    /*
     *   Debugger configuration object, and configuration filename
     */
    class CHtmlDebugConfig *config_;
    CStringBuf config_filename_;

    /* 
     *   Flag indicating that we're stepping execution.  We use this to
     *   exit the debugger event loop when the user steps or runs, so that
     *   control leaves the debugger event loop and returns to TADS to
     *   continue game execution.  
     */
    int go_;

    /*
     *   Flag indicating that we're in the debugger.  This is set to true
     *   whenever the debugger UI takes control, and set to false when the
     *   debugger returns control to TADS. 
     */
    int in_debugger_ : 1;

    /*
     *   Flag indicating that we're quitting.  If a quit command is
     *   entered, we'll set this flag and then exit the event loop as
     *   though we were about to resume execution, aborting the program at
     *   the last moment.  (We can't simply quit, because we need to be
     *   sure to properly exit out of the event loop.)
     */
    int quitting_ : 1;

    /*
     *   Flag indicating that we're restarting.  If a restart command is
     *   entered, we'll ste this flag and then exit the event loop,
     *   flagging to the event loop caller that the game is to be
     *   restarted. 
     */
    int restarting_ : 1;

    /*
     *   Flag indicating that we're aborting the current command.  This
     *   works like the 'quitting' and 'restarting' flags. 
     */
    int aborting_cmd_ : 1;

    /* my accelerator table */
    HACCEL accel_;

    /* 
     *   Current active debugger window.  Each debugger window that wants
     *   to receive commands must register itself here when it becomes
     *   active, and deregister itself on becoming inactive. 
     */
    class CHtmlDbgSubWin *active_dbg_win_;

    /* 
     *   Last active debugger window.  We keep track of this so that we
     *   can route commands that are intended for debugger windows to the
     *   appropriate window when the main window itself is active.  When
     *   the main window is active, we'll route commands to the last
     *   active window if it's the window immediately below the main
     *   window in Z order.  
     */
    class CHtmlDbgSubWin *last_active_dbg_win_;

    /*
     *   Menu delegator window.  If one of the debugger windows sent us a
     *   menu keystroke (with the ALT key) to run our menu, it wants to be
     *   re-activated as soon as the menu is finished.  We'll remember the
     *   window here, and re-activate it again when we're finished. 
     */
    HWND post_menu_active_win_;

    /*
     *   Watch window, local variable window
     */
    class CHtmlDbgWatchWin *watchwin_;
    class CHtmlDbgWatchWin *localwin_;

    /* toolbar handle */
    HWND toolbar_;

    /*
     *   Flag indicating that the game is over.  This is set after the
     *   game has executed the "quit" TADS built-in function, and the
     *   run-time has called dbguquitting() to let us have control one
     *   last time before exiting.
     */
    int game_over_ : 1;
};

/* ------------------------------------------------------------------------ */
/*
 *   Interface that must be provided by a window that can be an active
 *   debugger sub window.  Source and stack windows provide this
 *   interface.  Active windows can have commands routed from the main
 *   window.  
 */
class CHtmlDbgSubWin
{
public:
    /* check the status of a command routed from the main window */
    virtual TadsCmdStat_t active_check_command(int command_id) = 0;

    /* execute a command routed from the main window */
    virtual int active_do_command(int notify_code, int command_id,
                                  HWND ctl) = 0;

    /* get my window handle */
    virtual HWND active_get_handle() = 0;

    /*
     *   Get the current selection, if appropriate.  Returns the true and
     *   fills in the pointers if a selection range was returned, false if
     *   not. 
     */
    virtual int get_sel_range(class CHtmlFormatter **formatter,
                              unsigned long *sel_start,
                              unsigned long *sel_end) = 0;
};


/* ------------------------------------------------------------------------ */
/*
 *   Window frame class for debugger source windows.  This is a simple
 *   window frame that contains an HTML pane. 
 */
class CHtmlSys_dbgsrcwin: public CHtmlSys_dbgwin, public CHtmlDbgSubWin
{
public:
    CHtmlSys_dbgsrcwin(class CHtmlParser *parser,
                       class CHtmlPreferences *prefs,
                       class CHtmlDebugHelper *helper,
                       class CHtmlSys_dbgmain *dbgmain);

    /* initialize the HTML panel */
    virtual void init_html_panel(class CHtmlFormatter *formatter);

    /* store my position in the preferences */
    void set_winpos_prefs(const class CHtmlRect *pos);

    /* destroy the window */
    void do_destroy();

    /* bring the window to the front */
    void bring_owner_to_front();

    /* get the frame handle */
    virtual HWND get_owner_frame_handle() const { return handle_; }

    /* activate or deactivate */
    int do_activate(int flag, int minimized, HWND other_win);

    /* check the status of a command */
    TadsCmdStat_t check_command(int command_id);

    /* execute a command */
    int do_command(int notify_code, int command_id, HWND ctl);

    /* check the status of a command routed from the main window */
    TadsCmdStat_t active_check_command(int command_id);

    /* execute a command routed from the main window */
    int active_do_command(int notify_code, int command_id, HWND ctl);

    /* get my window handle */
    HWND active_get_handle() { return get_handle(); }

    /* get my selection range */
    int get_sel_range(class CHtmlFormatter **formatter,
                      unsigned long *sel_start,
                      unsigned long *sel_end);

protected:
    /* 
     *   adjust screen coordinates to frame-relative coordinates, if this
     *   window is contained in another window
     */
    void adjust_coords_to_frame(CHtmlRect *newpos, const CHtmlRect *pos);

    /* create the HTML panel - called during initialization */
    virtual void create_html_panel(class CHtmlFormatter *formatter);

    /* my debugger helper object */
    class CHtmlDebugHelper *helper_;

    /* the main debugger window */
    class CHtmlSys_dbgmain *dbgmain_;
};

/* ------------------------------------------------------------------------ */
/*
 *   HTML window subclass for debugger source windows.  The debugger
 *   source window displays some extra information in the left margin of
 *   each line to show the status of the line (currently executing line,
 *   breakpoint location, etc).  
 */
class CHtmlSysWin_win32_dbgsrc: public CHtmlSysWin_win32
{
public:
    CHtmlSysWin_win32_dbgsrc(class CHtmlFormatter *formatter,
                             class CHtmlSysWin_win32_owner *owner,
                             int has_vscroll, int has_hscroll,
                             class CHtmlPreferences *prefs,
                             class CHtmlDebugHelper *helper,
                             class CHtmlSys_dbgmain *dbgmain);

    ~CHtmlSysWin_win32_dbgsrc();

    /* draw a debugger source line icon */
    void draw_dbgsrc_icon(const CHtmlRect *pos, unsigned int stat);

    /* measure a debugger source window margin icon */
    CHtmlPoint measure_dbgsrc_icon();

    /* check the status of a command */
    TadsCmdStat_t check_command(int command_id);

    /* execute a command */
    int do_command(int notify_code, int command_id, HWND ctl);

    /* check the status of a command routed from the main window */
    virtual TadsCmdStat_t active_check_command(int command_id);

    /* execute a command routed from the main window */
    virtual int active_do_command(int notify_code, int command_id, HWND ctl);

    /* set the background cursor */
    virtual int do_setcursor_bkg();

    /* set the colors */
    void set_src_colors(HTML_color_t bkg_color, HTML_color_t text_color,
                        int use_windows_colors);

    /* 
     *   ignore HTML-based color changes in source windows - we use our
     *   own direct color settings 
     */
    void set_html_bg_color(HTML_color_t, int) { }
    void set_html_text_color(HTML_color_t, int) { }

    /* handle keystrokes */
    int do_keydown(int virtual_key, long keydata);

    /* handle an arrow keystroke */
    unsigned long do_keydown_arrow(int vkey, unsigned long caret_pos,
                                   int cnt);
    
    /* 
     *   handle "system" keystrokes (with the ALT key down) - we want to
     *   pass these on to the main debug window, so that the main debug
     *   window's menus can be accessed with the keyboard while this
     *   window has focus 
     */
    int do_syskeydown(int vkey, long keydata);

protected:
    /* window creation */
    void do_create();

    /* process a mouse click */
    int do_leftbtn_down(int keys, int x, int y, int clicks);

    /* process a notification */
    int do_notify(int control_id, int notify_code, LPNMHDR nmhdr);

    /* carry out a "Find" or "Find Next" command */
    void do_find(int command_id);

    /* draw the margin background */
    virtual void draw_margin_bkg(const RECT *rc);
    
    /* bitmaps for debugger source windows */
    HICON ico_dbg_curline_;
    HICON ico_dbg_ctxline_;
    HICON ico_dbg_bp_;
    HICON ico_dbg_bpdis_;
    HICON ico_dbg_curbp_;
    HICON ico_dbg_curbpdis_;
    HICON ico_dbg_ctxbp_;
    HICON ico_dbg_ctxbpdis_;

    /* debugger helper */
    class CHtmlDebugHelper *helper_;

    /* the main debugger window */
    class CHtmlSys_dbgmain *dbgmain_;

    /* my tooltip control */
//$$$    HWND tooltip_;

    /* 
     *   target column for up/down arrow operation - this is only valid
     *   after an up/down operation when there has been no invervening
     *   keystroke 
     */
    int target_col_valid_ : 1;
    unsigned int target_col_;
};

/* ------------------------------------------------------------------------ */
/*
 *   Stack window frame.  This is the same as a source window frame,
 *   except that it uses a stack panel. 
 */
class CHtmlSys_dbgstkwin: public CHtmlSys_dbgsrcwin
{
public:
    CHtmlSys_dbgstkwin(class CHtmlParser *parser,
                       class CHtmlPreferences *prefs,
                       class CHtmlDebugHelper *helper,
                       class CHtmlSys_dbgmain *dbgmain)
        : CHtmlSys_dbgsrcwin(parser, prefs, helper, dbgmain)
    {
    }

    /* use the tool window style for the stack window */
    DWORD get_winstyle_ex() { return WS_EX_TOOLWINDOW; }

    /* store my position in the preferences */
    void set_winpos_prefs(const class CHtmlRect *pos);

private:
    /* create the HTML panel - called during initialization */
    virtual void create_html_panel(class CHtmlFormatter *formatter);
};

/*
 *   HTML window subclass for debugger stack window.  This is the same as
 *   a source window, except for a slightly different display style.  
 */
class CHtmlSysWin_win32_dbgstk: public CHtmlSysWin_win32_dbgsrc
{
public:
    CHtmlSysWin_win32_dbgstk(class CHtmlFormatter *formatter,
                             class CHtmlSysWin_win32_owner *owner,
                             int has_vscroll, int has_hscroll,
                             class CHtmlPreferences *prefs,
                             class CHtmlDebugHelper *helper,
                             class CHtmlSys_dbgmain *dbgmain)
        : CHtmlSysWin_win32_dbgsrc(formatter, owner, has_vscroll,
                                   has_hscroll, prefs, helper, dbgmain)
    {
    }

    /* process a mouse click */
    int do_leftbtn_down(int keys, int x, int y, int clicks);

    /* check the status of a command routed from the main window */
    TadsCmdStat_t active_check_command(int command_id);

    /* execute a command routed from the main window */
    int active_do_command(int notify_code, int command_id, HWND ctl);

protected:
    /* 
     *   draw the margin background - we'll just leave it in the window's
     *   background color 
     */
    virtual void draw_margin_bkg(const RECT *) { }
};


/* ------------------------------------------------------------------------ */
/*
 *   History window frame.  This is the based on the source window frame,
 *   but has some differences for call history.  
 */
class CHtmlSys_dbghistwin: public CHtmlSys_dbgsrcwin
{
public:
    CHtmlSys_dbghistwin(class CHtmlParser *parser,
                        class CHtmlPreferences *prefs,
                        class CHtmlDebugHelper *helper,
                        class CHtmlSys_dbgmain *dbgmain)
        : CHtmlSys_dbgsrcwin(parser, prefs, helper, dbgmain)
    {
    }
    
    /* use the tool window style for the history window */
    DWORD get_winstyle_ex() { return WS_EX_TOOLWINDOW; }

    /* store my position in the preferences */
    void set_winpos_prefs(const class CHtmlRect *pos);

private:
    /* create the HTML panel - called during initialization */
    virtual void create_html_panel(class CHtmlFormatter *formatter);
};

/*
 *   HTML window subclass for history trace window.  This is the same as a
 *   source window, except for a slightly different display style.  
 */
class CHtmlSysWin_win32_dbghist: public CHtmlSysWin_win32_dbgsrc
{
public:
    CHtmlSysWin_win32_dbghist(class CHtmlFormatter *formatter,
                              class CHtmlSysWin_win32_owner *owner,
                              int has_vscroll, int has_hscroll,
                              class CHtmlPreferences *prefs,
                              class CHtmlDebugHelper *helper,
                              class CHtmlSys_dbgmain *dbgmain)
        : CHtmlSysWin_win32_dbgsrc(formatter, owner, has_vscroll,
                                   has_hscroll, prefs, helper, dbgmain)
    {
    }
    
    /* check the status of a command routed from the main window */
    TadsCmdStat_t active_check_command(int command_id);
    
    /* execute a command routed from the main window */
    int active_do_command(int notify_code, int command_id, HWND ctl);
    
protected:
    /* 
     *   draw the margin background - we'll just leave it in the window's
     *   background color 
     */
    virtual void draw_margin_bkg(const RECT *) { }
};

/* ------------------------------------------------------------------------ */
/*
 *   Dialog for the debugger about box 
 */

class CTadsDlg_dbgabout: public CTadsDialog
{
public:
    /* initialize */
    void init();
};

#endif /* W32TDB_H */

