#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/htmlw32.cpp,v 1.1.1.1 1999/05/09 21:32:31 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmlsys_w32.cpp - HTML system class implementation for Win32
Function
  
Notes
  
Modified
  09/13/97 MJRoberts  - Creation
*/

#include <windows.h>
#include <dsound.h>
#include <commctrl.h>

#include <string.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/* include TADS OS headers */
#include <os.h>

/* include the TADS interface (for the run-time version number) */
#include <trd.h>

#ifndef HTMLSYS_W32
#include "htmlw32.h"
#endif
#ifndef HTMLFMT_H
#include "htmlfmt.h"
#endif
#ifndef HTMLPRS_H
#include "htmlprs.h"
#endif
#ifndef TADSAPP_H
#include "tadsapp.h"
#endif
#ifndef TADSWIN_H
#include "tadswin.h"
#endif
#ifndef TADSFONT_H
#include "tadsfont.h"
#endif
#ifndef HTMLTAGS_H
#include "htmltags.h"
#endif
#ifndef HTMLINP_H
#include "htmlinp.h"
#endif
#ifndef HTMLDISP_H
#include "htmldisp.h"
#endif
#ifndef HTMLTXAR_H
#include "htmltxar.h"
#endif
#ifndef HTMLURL_H
#include "htmlurl.h"
#endif
#ifndef TADSSTAT_H
#include "tadsstat.h"
#endif
#ifndef HTMLRES_H
#include "htmlres.h"
#endif
#ifndef HTMLPREF_H
#include "htmlpref.h"
#endif
#ifndef HTMLRC_H
#include "htmlrc.h"
#endif
#ifndef W32FONT_H
#include "w32font.h"
#endif
#ifndef W32SND_H
#include "w32snd.h"
#endif
#ifndef TADSDLG_H
#include "tadsdlg.h"
#endif
#ifndef TADSOLE_H
#include "tadsole.h"
#endif


/* flag: allow clicking on links during MORE mode */
#define HTMLW32_MORE_ALLOW_LINKS  TRUE

/*
 *   Our special fake character set ID for bullets.  We distinguish this
 *   from the regular symbol character set, because we use the symbol
 *   character set for the "Symbol" font, but we want to be able to use a
 *   different character set for the bullet font (for which we use
 *   "Wingdings").
 *   
 *   Warning - this MUST NOT OVERLAP with any xxx_CHARSET definition in
 *   wingdi.h.
 */
#define BULLET_CHARSET 10


/* ------------------------------------------------------------------------ */
/*
 *   HTML window panel implementation 
 */

/* ------------------------------------------------------------------------ */
/*
 *   private window hotkey ID's 
 */
static const int HTML_HK_BREAK = 1;

/* ------------------------------------------------------------------------ */

/* MORE prompt text */
char CHtmlSysWin_win32::moreprompt[] = "  *** More ***  ";


CHtmlSysWin_win32::CHtmlSysWin_win32(CHtmlFormatter *formatter,
                                     CHtmlSysWin_win32_owner *owner,
                                     int has_vscroll, int has_hscroll,
                                     CHtmlPreferences *prefs)
    : CHtmlSysWin(formatter), CTadsWinScroll(has_vscroll, has_hscroll)
{
    CHtmlRect margins;
    CHtmlFontDesc font_desc;
    static const textchar_t *wingdings = "Wingdings";

    /* no timers yet */
    timer_list_head_ = 0;

    /* remember my owner */
    owner_ = owner;

    /* remember my preferences source */
    prefs_ = prefs;

    /* no status line yet */
    statusline_ = 0;

    /* formatting and pruning are not yet underway */
    formatting_msg_ = FALSE;
    pruning_msg_ = FALSE;

    /* formatting width is unknown */
    fmt_width_ = 0;

    /* remember the formatter */
    formatter_ = formatter;

    /* no content yet */
    content_height_ = 0;
    last_input_height_ = 0;
    more_mode_ = FALSE;
    game_more_request_ = FALSE;
    clip_ypos_ = 0;

    /* presume MORE mode will not be enabled */
    more_mode_enabled_ = FALSE;

    /* presume we won't ever show a caret */
    caret_enabled_ = FALSE;

    /* the caret isn't visible yet, even if we will be showing it */
    caret_vis_ = FALSE;

    /* set up a default caret shape */
    caret_pos_.set(-100, -100);
    caret_ht_ = 16;
    caret_ascent_ = 0;

    /* not tracking the mouse yet */
    tracking_mouse_ = FALSE;
    track_link_ = 0;
    caret_at_right_ = TRUE;

    /* not hovering over a status line item yet */
    status_link_ = 0;

    /* no command yet */
    cmdtag_ = 0;
    started_reading_cmd_ = FALSE;

    /* create a command buffer */
    cmdbuf_ = new CHtmlInputBuf(100);

    /* set a small inset from the margins */
    margins.set(8, 2, 8, 2);

    /* create the default font */
    font_desc.htmlsize = 3;
    default_font_ = (CHtmlSysFont_win32 *)get_font(&font_desc);

    /* load the I-beam cursor for text areas */
    ibeam_csr_ = LoadCursor(NULL, IDC_IBEAM);

    /* load the hand cursor */
    hand_csr_ = LoadCursor(CTadsApp::get_app()->get_instance(),
                           "HAND_CURSOR");

    /* load the default edit area context menu */
    popup_container_ = 0;
    load_context_popup(IDR_EDIT_POPUP_MENU);

    /* set up the default background and text colors */
    bg_image_ = 0;
    bgbrush_ = 0;
    set_html_bg_color(0, TRUE);
    set_html_text_color(0, TRUE);

    /* 
     *   note whether Wingdings is available; if so, use it for drawing
     *   bullet characters; otherwise, draw all bullets in the prevailing
     *   font at the time of the drawing using the single bullet character
     *   available in most fonts 
     */
    if ((wingdings_avail_ =
        CTadsFont::font_is_present(wingdings, get_strlen(wingdings))) != 0)
    {
        /* use an appropriate character from wingdings for each bullet */
        bullet_square_ = (textchar_t)167;
        bullet_disc_ = (textchar_t)159;
        bullet_circle_ = (textchar_t)250;
    }
    else
    {
        /* don't use wingdings; map all bullets to the standard one */
        bullet_square_ = bullet_disc_ = bullet_circle_ = (char)149;
    }

    /* not pausing before exit or for a keystroke yet */
    exit_pause_ = FALSE;
    waiting_for_key_pause_ = FALSE;

    /* not in an interactive size/move yet */
    in_size_move_ = FALSE;

    /* assume we'll reformat on resizing */
    need_format_on_resize_ = TRUE;

    /* presume we don't have a border */
    has_border_ = FALSE;

    /* presume we're not a banner */
    is_banner_win_ = FALSE;
    banner_alignment_ = HTML_BANNERWIN_POS_TOP;

    /* tell the formatter about us */
    formatter_->set_win(this, &margins);

    /* not in temporary link mode yet */
    temp_show_links_ = FALSE;
    temp_link_timer_id_ = 0;

    /* no timers yet */
    cb_timer_id_ = 0;
    bg_timer_id_ = 0;

    /* no tooltip control yet */
    tooltip_ = 0;
    tooltip_vanished_ = FALSE;

    /* not doing a resize reformat yet */
    doing_resize_reformat_ = FALSE;
}

CHtmlSysWin_win32::~CHtmlSysWin_win32()
{
    /* delete the command buffer */
    delete cmdbuf_;

    /* we're done with the cursors */
    DestroyCursor(ibeam_csr_);
    DestroyCursor(hand_csr_);

    /* done with the context menus */
    if (popup_container_ != 0)
        DestroyMenu(popup_container_);

    /* remove my reference on the background image if I have one */
    if (bg_image_)
        bg_image_->remove_ref();
}

/*
 *   Load an edit popup, destroying any existing one 
 */
void CHtmlSysWin_win32::load_context_popup(int menu_id)
{
    /* destroy any existing edit popup */
    if (popup_container_ != 0)
        DestroyMenu(popup_container_);
    
    /* load the new one */
    popup_container_ = LoadMenu(CTadsApp::get_app()->get_instance(),
                                MAKEINTRESOURCE(menu_id));

    /* extract the context submenu from the menu container */
    edit_popup_ = GetSubMenu(popup_container_, 0);
}


/*
 *   Receive notification that the parent window is being destroyed.  This
 *   means that I'll be destroyed soon, but we need to remove any
 *   references to the parent window before the parent is deleted.  
 */
void CHtmlSysWin_win32::on_destroy_parent()
{
    /* if I have a status line, unregister myself as a source */
    if (statusline_ != 0)
    {
        /* tell the status line to forget about me */
        statusline_->unregister_source(this);

        /* forget about the status line so we don't unregister again later */
        statusline_ = 0;
    }

    /* 
     *   notify my formatter that I'm soon to be deleted - it may need to
     *   do some work to disentangle itself from me 
     */
    formatter_->unset_win();
}


/*
 *   Set the status line 
 */
void CHtmlSysWin_win32::set_status_line(CTadsStatusline *statusline)
{
    /* remember the status line */
    statusline_ = statusline;

    /* register myself as a status source */
    statusline->register_source(this);
}

/*
 *   create a font - used as a callback for CTadsApp::get_font.  We
 *   create and return a CHtmlSysFont_win32 object, rather than a basic
 *   CTadsFont object, since we want the additional interfaces provided by
 *   the Html subclass.  
 */
CTadsFont *CHtmlSysWin_win32::create_font(const CTadsLOGFONT *lf)
{
    return new CHtmlSysFont_win32(lf);
}

/*
 *   Get the system mapping for a parameterized font.  Returns true and
 *   fills in the logical font descriptor with the actual system font name
 *   if the font name is one of the parameterized font names, false if
 *   not.  
 */
int CHtmlSysWin_win32::get_param_font(CTadsLOGFONT *tlf,
                                      const textchar_t *name, size_t len,
                                      CHtmlFontDesc *font_desc)
{
    static const textchar_t nm_serif[] = "tads-serif";
    static const textchar_t nm_sans[] = "tads-sans";
    static const textchar_t nm_script[] = "tads-script";
    static const textchar_t nm_typewriter[] = "tads-typewriter";
    static const textchar_t nm_input[] = "tads-input";

    /* check each name */
    if (len == sizeof(nm_serif) - sizeof(textchar_t)
        && memicmp(nm_serif, name, len) == 0)
    {
        strcpy(tlf->lf.lfFaceName, prefs_->get_font_serif());
    }
    else if (len == sizeof(nm_sans) - sizeof(textchar_t)
             && memicmp(nm_sans, name, len) == 0)
    {
        strcpy(tlf->lf.lfFaceName, prefs_->get_font_sans());
    }
    else if (len == sizeof(nm_script) - sizeof(textchar_t)
             && memicmp(nm_script, name, len) == 0)
    {
        strcpy(tlf->lf.lfFaceName, prefs_->get_font_script());
    }
    else if (len == sizeof(nm_typewriter) - sizeof(textchar_t)
             && memicmp(nm_typewriter, name, len) == 0)
    {
        strcpy(tlf->lf.lfFaceName, prefs_->get_font_typewriter());
    }
    else if (len == sizeof(nm_input) - sizeof(textchar_t)
             && memicmp(nm_input, name, len) == 0)
    {
        HTML_color_t inp_color;

        /* use the selected font name */
        strcpy(tlf->lf.lfFaceName, prefs_->get_inpfont_name());

        /* 
         *   if the face name is empty, it means that we're to use the
         *   standard input font 
         */
        if (tlf->lf.lfFaceName[0] == '\0')
            strcpy(tlf->lf.lfFaceName, prefs_->get_prop_font());

        /* apply the color and style settings as well */
        tlf->lf.lfItalic = prefs_->get_inpfont_italic();
        tlf->lf.lfWeight = (prefs_->get_inpfont_bold()
                            ? FW_BOLD : FW_NORMAL);
        inp_color = prefs_->get_inpfont_color();
        tlf->color = HTML_color_to_COLORREF(inp_color);

        /* 
         *   set the color in the font descriptor, so that we know to use
         *   the explicit color setting 
         */
        font_desc->default_color = FALSE;
        font_desc->color = inp_color;
    }
    else
    {
        /* it's not a parameterized font name */
        return FALSE;
    }

    /* if we got here, we've mapped the font */
    return TRUE;
}

/*
 *   Get a font suitable for the given characteristics 
 */
CHtmlSysFont *CHtmlSysWin_win32::get_font(const CHtmlFontDesc *in_font_desc)
{
    CTadsLOGFONT tlf;
    int pointsize;
    CHtmlSysFont_win32 *new_font;
    int created;
    CHtmlFontDesc font_desc;

    /* make a copy of the original font descriptor, so we can change it */
    font_desc = *in_font_desc;

    /*
     *   Mapping from HTML sizes to point sizes.  Each subarray is for a
     *   default size setting ("smallest", "small", "medium", "large",
     *   "largest"); the elements of each subarray are the point sizes for
     *   absolute SIZE settings of 1 through 7. 
     */
    static const int size_cnt = 7;
    static int size_map[][size_cnt] =
    {
        { 6, 7, 9, 10, 12, 16, 24 },                            /* smallest */
        { 7, 9, 10, 11, 16, 20, 30 },                              /* small */
        { 7, 10, 12, 14, 18, 24, 36 },                            /* medium */
        { 10, 12, 14, 16, 20, 38, 42 },                            /* large */
        { 12, 14, 16, 18, 24, 32, 48 }                           /* largest */
    };

    /* set up the default parts of the LOGFONT that we never change */
    memset(&tlf, 0, sizeof(tlf));
    tlf.lf.lfWidth = 0;
    tlf.lf.lfEscapement = 0;
    tlf.lf.lfOrientation = 0;
    tlf.lf.lfCharSet = (font_desc.default_charset
                        ? owner_->owner_get_default_charset()
                        : font_desc.charset);
    tlf.lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
    tlf.lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    tlf.lf.lfQuality = DEFAULT_QUALITY;
    tlf.lf.lfPitchAndFamily = FF_DONTCARE;

    /*
     *   use the weight they provided (we may change this if a weight
     *   modifier is specified)
     */
    tlf.lf.lfWeight = font_desc.weight;

    /* presume we're not going to specify a face name explicitly */
    tlf.lf.lfFaceName[0] = '\0';

    /* if an HTML size is specified, map it to a point size */
    if (font_desc.htmlsize > 0
        && font_desc.htmlsize <= size_cnt)
    {
        int htmlsize;
        
        /*
         *   an HTML size is specified -- if a BIG or SMALL attribute is
         *   present, bump the HTML size by 1 in the appropriate direction
         */
        htmlsize = font_desc.htmlsize;
        if (font_desc.pe_big && htmlsize < size_cnt)
            ++htmlsize;
        else if (font_desc.pe_small && htmlsize > 1)
            --htmlsize;

        /* get the point size for the HTML size at the current scale */
        pointsize = size_map[prefs_->get_font_scale()][htmlsize - 1];
    }
    else
    {
        /* no HTML size - use the explicit point size provided */
        pointsize = font_desc.pointsize;
    }

    /* map the point size to LOGFONT units */
    tlf.lf.lfHeight = CTadsFont::calc_lfHeight(pointsize);

    /* apply the specific font attributes */
    tlf.lf.lfItalic = (BYTE)font_desc.italic;
    tlf.lf.lfUnderline = (BYTE)font_desc.underline;
    tlf.lf.lfStrikeOut = (BYTE)font_desc.strikeout;

    /* set the extended CTadsLOGFONT attributes */
    tlf.superscript = font_desc.superscript;
    tlf.subscript = font_desc.subscript;
    tlf.color = HTML_color_to_COLORREF(font_desc.color);

    /* 
     *   If a face name is listed, try to find the given face in the
     *   system.  Note that we do not attempt to use the font name if
     *   we're in the symbol character set -- if we're in the symbol
     *   character set, it means that we are explicitly using an HTML
     *   symbol entity, and we map these to the character set "Symbol";
     *   other fonts in this character set do not use the same code point
     *   layout.
     *   
     *   Note that we wait until after setting all of the basic attributes
     *   (in particular, weight, color, and styles) before dealing with
     *   the face name; this is to allow parameterized face names to
     *   override the basic attributes.  For example, the "TADS-Input"
     *   font allows the player to specify color, bold, and italic styles
     *   in addition to the font name.  
     */
    if (font_desc.face[0] != 0
        && tlf.lf.lfCharSet != SYMBOL_CHARSET)
    {
        const char *start;
        const char *end;
        size_t rem;

        /* look at each name in the list */
        for (start = font_desc.face, rem = get_strlen(start) ;
             rem > 0 ; )
        {
            int endrem;
            
            /* find the end of this name */
            for (end = start, endrem = rem ; endrem > 0 && *end != ',' ;
                 ++end, --endrem) ;

            /* 
             *   check to see if this is a parameterized font name, and if
             *   so use it 
             */
            if (get_param_font(&tlf, start, end - start, &font_desc))
                break;

            /* see if this font exists on this system */
            if (CTadsFont::font_is_present(start, end - start))
            {
                size_t len;
                
                /*
                 *   This font is present - use this name.  Note that we
                 *   need to limit the length of the face to the available
                 *   size in the LOGFONT structure's face name buffer.  
                 */
                len = end - start;
                if (len > LF_FACESIZE - 1) len = LF_FACESIZE - 1;
                memcpy(tlf.lf.lfFaceName, start, len);
                tlf.lf.lfFaceName[len] = '\0';
                
                /* no need to keep looking */
                break;
            }

            /* move on to the next name in the list, if there is one */
            start = end + 1;
            rem = endrem;

            /* skip any spaces after the comma */
            while (rem != 0 && is_space(*start))
                ++start, --rem;
        }
    }

    /* 
     *   if we're in the symbol character set, always use "Symbol" as the
     *   font 
     */
    if (tlf.lf.lfCharSet == SYMBOL_CHARSET)
        strcpy(tlf.lf.lfFaceName, "Symbol");

    /* 
     *   if we're in our fake BULLET character set, use the SYMBOL
     *   character set as the actual character set - this is the character
     *   set that the bullet font uses 
     */
    if (tlf.lf.lfCharSet == BULLET_CHARSET)
        tlf.lf.lfCharSet = SYMBOL_CHARSET;

    /*
     *   Apply characteristics only if the face wasn't specified 
     */
    if (tlf.lf.lfFaceName[0] == '\0')
    {
        /* see if fixed-pitch is desired */
        if (font_desc.fixed_pitch)
        {
            /* set the characteristics */
            tlf.lf.lfPitchAndFamily &= ~VARIABLE_PITCH;
            tlf.lf.lfPitchAndFamily |= FIXED_PITCH;

            /* use prefered monospaced font */
            strcpy(tlf.lf.lfFaceName, prefs_->get_mono_font());
        }
        else
        {
            /* use prefered proportional font */
            strcpy(tlf.lf.lfFaceName, prefs_->get_prop_font());
        }

        /* see if serifs are desired for a variable-pitch font */
        if (!font_desc.serif && !font_desc.fixed_pitch)
        {
            /* set the family information */
            tlf.lf.lfPitchAndFamily &= ~FF_ROMAN;
            tlf.lf.lfPitchAndFamily |= FF_SWISS;

            /* ask for Arial */
            strcpy(tlf.lf.lfFaceName, "Arial");
        }

        /* see if emphasis (EM) is desired - render italic if so */
        if (font_desc.pe_em)
            tlf.lf.lfItalic = TRUE;

        /* see if strong emphasis (STRONG) is desired - render bold if so */
        if (font_desc.pe_strong && tlf.lf.lfWeight < 700)
            tlf.lf.lfWeight = 700;

        /* if we're in an address block, render in italics */
        if (font_desc.pe_address)
            tlf.lf.lfItalic = TRUE;

        /* see if this is a defining instance (DFN) - render in italics */
        if (font_desc.pe_dfn)
            tlf.lf.lfItalic = TRUE;

        /*
         *   see if this is sample code (SAMP), keyboard code (KBD), or a
         *   variable (VAR) - render these in a monospaced roman font if
         *   so 
         */
        if (font_desc.pe_samp || font_desc.pe_kbd || font_desc.pe_var)
        {
            /* set the characteristics for a fixed-pitch roman font */
            tlf.lf.lfPitchAndFamily = FIXED_PITCH | FF_ROMAN;

            /* render KBD in bold */
            if (font_desc.pe_kbd && tlf.lf.lfWeight < 700)
                tlf.lf.lfWeight = 700;

            /* ask for Courier New */
            strcpy(tlf.lf.lfFaceName, "Courier New");
        }

        /* see if this is a citation (CITE) - render in italics if so */
        if (font_desc.pe_cite)
            tlf.lf.lfItalic = TRUE;
    }

    /*
     *   We've built up our font record -- create a font
     */
    new_font = (CHtmlSysFont_win32 *)
        CTadsApp::get_app()->get_font(&tlf, &create_font, &created);

    /* remember the font description used to create the font */
    if (created)
        new_font->set_font_desc(&font_desc);

    /* return the new font */
    return new_font;
}

/*
 *   Determine if the MORE prompt is showing.  Returns true if we're
 *   scrolled to the bottom, so that the latest text is showing, false if
 *   not. 
 */
int CHtmlSysWin_win32::moreprompt_is_at_bottom()
{
    RECT rcmore;
    RECT rcwin;
    
    /* get the MORE prompt rectangle and the window area */
    get_moreprompt_rect(&rcmore);
    get_scroll_area(&rcwin, TRUE);
    
    /* return true if the MORE prompt is at the bottom */
    return (rcmore.bottom <= rcwin.bottom);
}

/*
 *   determine if we can copy 
 */
int CHtmlSysWin_win32::can_copy()
{
    unsigned long startofs, endofs;

    /* get the current selection */
    formatter_->get_sel_range(&startofs, &endofs);

    /* we can copy only if there's something selected */
    return (startofs != endofs);
}


/*
 *   Copy text to a new global memory handle 
 */
HGLOBAL CHtmlSysWin_win32::copy_to_new_hglobal(UINT alloc_flags, size_t *lenp)
{
    unsigned long startofs, endofs;
    unsigned long len;
    CHtmlTextArray *arr = formatter_->get_text_array();
    HGLOBAL memhdl;
    textchar_t *memptr;
    unsigned long nlcnt;
    CStringBuf *buf;
    CHtmlDisp *disp;
    textchar_t *p;

    /* get the current selection */
    formatter_->get_sel_range(&startofs, &endofs);

    /* if there's nothing selected, there's nothing to do */
    if (startofs == endofs)
        return FALSE;

    /* 
     *   figure out how much space we need (this is a rough calculation,
     *   which may exceed the amount of space we actually end up using) 
     */
    len = arr->get_char_count(startofs, endofs);

    /* allocate space */
    buf = new CStringBuf(len);

    /* get the text in the internal format */
    for (disp = formatter_->find_by_txtofs(startofs, FALSE, TRUE) ;
         disp != 0 ; disp = disp->get_next_disp())
    {
        /* if we're past the ending offset, we're done */
        if (disp->is_past_text_ofs(endofs))
            break;

        /* add this item's text to the buffer */
        disp->add_text_to_buf_clip(buf, startofs, endofs);
    }

    /* get the actual length of the string */
    len = get_strlen(buf->get());

    /* count newlines, since we'll need to expand them into CR-LF's */
    for (p = buf->get(), nlcnt = 0 ; *p != '\0' ; ++p)
    {
        /* if it's a newline, count it */
        if (*p == '\n')
            ++nlcnt;
    }

    /* allocate and lock a global memory object to hold the text */
    memhdl = GlobalAlloc(alloc_flags, (len + nlcnt + 1)*sizeof(textchar_t));
    memptr = (textchar_t *)GlobalLock(memhdl);

    /* tell the caller how much space we need, if they want to know */
    if (lenp != 0)
        *lenp = (len + nlcnt + 1)*sizeof(textchar_t);

    /* copy the result into the block */
    for (p = buf->get() ; *p != '\0' ; ++p)
    {
        if (*p == '\n')
        {
            *memptr++ = 13;
            *memptr++ = 10;
        }
        else
        {
            /* copy this character directly */
            *memptr++ = *p;
        }
    }

    /* add a terminating null byte */
    *memptr = '\0';

    /* we're done with the buffer */
    delete buf;

    /* done with the handle's memory - unlock it */
    GlobalUnlock(memhdl);

    /* return the new handle */
    return memhdl;
}

/*
 *   Copy text to clipboard 
 */
int CHtmlSysWin_win32::do_copy()
{
    HGLOBAL memhdl;

    /* take over the clipboard */
    if (!OpenClipboard(handle_))
        return FALSE;

    /* delete the current clipboard contents */
    EmptyClipboard();

    /* copy the current selection to a global memory handle */
    memhdl = copy_to_new_hglobal(GHND, 0);

    /* add the text to the clipboard */
    SetClipboardData(CF_TEXT, memhdl);

    /* done with the clipboard */
    CloseClipboard();

    /* success */
    return TRUE;
}

/*
 *   determine if we can cut 
 */
int CHtmlSysWin_win32::can_cut()
{
    unsigned long startofs, endofs;

    /* if there's no command, we can't do any cutting */
    if (cmdtag_ == 0)
        return FALSE;

    /* get the current selection */
    formatter_->get_sel_range(&startofs, &endofs);

    /*
     *   if there's no selection, or the selected text doesn't overlap
     *   the command line, there's nothing to cut; otherwise, cutting is
     *   possible 
     */
    return (startofs != endofs && endofs > cmdtag_->get_text_ofs());
}

/*
 *   Cut text to clipboard 
 */
int CHtmlSysWin_win32::do_cut()
{
    unsigned long startofs, endofs;

    /* if there's no command, we can't do any cutting */
    if (cmdtag_ == 0)
        return FALSE;

    /* get the current selection */
    formatter_->get_sel_range(&startofs, &endofs);

    /*
     *   if there's no selection, or the selected text doesn't overlap
     *   the command line, there's nothing to cut 
     */
    if (startofs == endofs || endofs <= cmdtag_->get_text_ofs())
        return FALSE;

    /* copy the text onto the clipboard */
    if (!do_copy())
        return FALSE;

    /* delete the selected range in the command line */
    if (cmdbuf_->del_right())
        update_input_display();

    /* we did some work */
    return TRUE;
}

/*
 *   determine if we can paste 
 */
int CHtmlSysWin_win32::can_paste()
{
    UINT fmt;
    int found;
    
    /* if we're not editing a command, we can't do any pasting */
    if (cmdtag_ == 0 || !started_reading_cmd_)
        return FALSE;

    /* get access to the clipboard */
    if (!OpenClipboard(handle_))
        return FALSE;

    /* see if "text" is on the clipboard */
    for (fmt = 0, found = FALSE ;; )
    {
        /* get the next format - stop if there are no more */
        fmt = EnumClipboardFormats(fmt);
        if (fmt == 0)
            break;

        /* if it's the one we want, note it */
        if (fmt == CF_TEXT)
        {
            found = TRUE;
            break;
        }
    }

    /* done with the clipboard */
    CloseClipboard();

    /* if we found our format, we can paste */
    return found;
}

/*
 *   Paste text from clipboard 
 */
int CHtmlSysWin_win32::do_paste()
{
    HANDLE hdl;
    const char *startp;
    const char *curp;
    int disp_change = FALSE;
    
    /* if we're not editing a command, we can't do any pasting */
    if (cmdtag_ == 0 || !started_reading_cmd_)
        return FALSE;

    /* get access to the clipboard */
    if (!OpenClipboard(handle_))
        return FALSE;

    /* if there's no text in the clipboard, we can't do any pasting */
    hdl = GetClipboardData(CF_TEXT);
    if (hdl == 0)
        return FALSE;

    /* lock the handle to get the text */
    startp = (const char *)GlobalLock(hdl);
    if (startp == 0)
        return FALSE;

    /*
     *   Go through the characters and insert the ones that we can.  If
     *   we run across any newlines or tabs, insert spaces instead. 
     */
    for (curp = startp ; *curp != '\0' ; ++curp)
    {
        /* see if we have anything we need to convert */
        switch(*curp)
        {
        case 13:
        case 10:
        case 9:
            /* insert what we have so far */
            disp_change |= cmdbuf_->add_string(startp, curp - startp, TRUE);

            /* add a space in place of this character */
            disp_change |= cmdbuf_->add_char(' ');

            /*
             *   if we have a CR-LF sequence, or an LF-CR sequence, treat
             *   the combination as a single character 
             */
            if ((*curp == 13 && *(curp + 1) == 10)
                || (*curp == 10 && *(curp + 1) == 13))
                ++curp;

            /*
             *   remember the next character as the start of the next
             *   substring for later addition 
             */
            startp = curp + 1;
            break;

        default:
            /* let this character accumulate for later string insertion */
            break;
        }
    }

    /* if we still have a substring to add, add it now */
    if (curp > startp)
        disp_change |= cmdbuf_->add_string(startp, curp - startp, TRUE);

    /* done with the clipboard data */
    GlobalUnlock(hdl);
    CloseClipboard();

    /* update the display if necessary */
    if (disp_change)
        update_input_display();

    /* return true if we changed the command buffer */
    return disp_change;
}

/*
 *   Create an IDataObject for dragging from this window.  We'll return a
 *   data object containing the text in the selection range. 
 */
IDataObject *CHtmlSysWin_win32::get_drag_dataobj()
{
    HGLOBAL memhdl;
    size_t len;

    /* copy the selection to a global memory handle */
    memhdl = copy_to_new_hglobal(GHND, &len);

    /* create and return the new text data object */
    return new CTadsDataObjText(memhdl, len);
}

/*
 *   Process a palette change request 
 */
int CHtmlSysWin_win32::do_querynewpalette()
{
    /* set our foreground palette */
    if (recalc_palette_fg())
    {
        InvalidateRect(get_handle(), 0, TRUE);
        return TRUE;
    }
    else
    {
        /* no change, but let secondaries realize */
        SendMessage(handle_, WM_PALETTECHANGED, (WPARAM)handle_, 0);
    }

    /* we didn't change anything */
    return FALSE;
}

/*
 *   Update for a palette change in another window 
 */
void CHtmlSysWin_win32::do_palettechanged(HWND srcwin)
{
    /* if I'm the source of the change, there's no need to do anything */
    if (srcwin == get_handle())
    {
        /* I initiated the change - set our background palettes */
        if (recalc_palette_bg())
            InvalidateRect(get_handle(), 0, TRUE);
        return;
    }
    else
    {
        /* recalculate all of our palettes */
        recalc_palette_fg();
        recalc_palette_bg();

        /* redraw with the new palette */
        InvalidateRect(get_handle(), 0, TRUE);
    }
}

/*
 *   Handle a setcursor message 
 */
int CHtmlSysWin_win32::do_setcursor(HWND hwnd, int /*hittest*/,
                                    int /*mousemsg*/)
{
    POINT pt;
    CHtmlPoint docpt;
    RECT rc;
    CHtmlDisp *disp;
    CHtmlDispLink *link;
    unsigned long txtofs;
    unsigned long sel_start, sel_end;
    
    /* if it's not over my window area, ignore it */
    if (hwnd != handle_)
    {
        /* if I'm tracking a hover link, forget about it */
        
        /* we didn't set a cursor */
        return FALSE;
    }

    /* 
     *   tell other subwindows to forget about any current hovering, since
     *   we're obviously no longer in any other subwindow 
     */
    owner_->clear_other_hover_info(this);

    /* get the mouse position */
    GetCursorPos(&pt);
    ScreenToClient(handle_, &pt);

    /*
     *   if I'm over the vertical scrollbar, ignore it; this can happen
     *   when the vertical scrollbar is disabled, since it won't show up
     *   for setcursor messages 
     */
    get_scroll_area(&rc, TRUE);
    if (pt.x >= rc.right)
        return FALSE;

    /* if I'm over the "MORE" prompt, show the hand cursor */
    get_moreprompt_rect(&rc);
    if (!prefs_->get_alt_more_style() && more_mode_ && PtInRect(&rc, pt))
    {
        SetCursor(hand_csr_);
        return TRUE;
    }

    /* see if we can find an item containing the given position */
    docpt = screen_to_doc(CHtmlPoint(pt.x, pt.y));
    txtofs = formatter_->find_textofs_by_pos(docpt);
    disp = formatter_->find_by_pos(docpt, TRUE);
    if (disp == 0)
    {
        /* if we have a status line link display, remove it */
        set_hover_link(0);
        
        /* we're not over a display list item - use the background cursor */
        return do_setcursor_bkg();
    }

    /* if we're over the selection, use an arrow */
    formatter_->get_sel_range(&sel_start, &sel_end);
    if (txtofs >= sel_start && txtofs < sel_end)
    {
        /* it's over the selection - use an arrow */
        SetCursor(arrow_cursor_);
    }
    else
    {
        /* see what kind of cursor this display item wants to set */
        switch(disp->get_cursor_type(this, formatter_, docpt.x, docpt.y,
                                     !HTMLW32_MORE_ALLOW_LINKS && more_mode_))
        {
        case HTML_CSRTYPE_IBEAM:
            /* set the cursor to the I-Beam */
            SetCursor(ibeam_csr_);
            break;
            
        case HTML_CSRTYPE_HAND:
            /* set the hand cursor */
            SetCursor(hand_csr_);
            break;
            
        case HTML_CSRTYPE_ARROW:
        default:
            /* use default cursor */
            SetCursor(arrow_cursor_);
            break;
        }
    }

    /* 
     *   If this item's link (or lack thereof) is different from the
     *   previous hover link, we need to update hover highlighting and the
     *   status line accordingly.  If we're not allowed to show links,
     *   don't bother even looking for a link.  
     */
    link = ((get_html_show_links()
             && (HTMLW32_MORE_ALLOW_LINKS || !more_mode_))
            ? disp->get_link(formatter_, docpt.x, docpt.y) : 0);
    set_hover_link(link);

    /* handled */
    return TRUE;
}


/*
 *   Update the display after changing the command input buffer contents
 *   or length 
 */
void CHtmlSysWin_win32::update_input_display()
{
    /* update the formatter display */
    if (cmdtag_ != 0)
    {
        cmdtag_->setlen(formatter_->get_text_array(), cmdbuf_->getlen());
        cmdtag_->format(this, formatter_);
    }

    /* update the cursor position */
    update_caret_pos(TRUE, TRUE);
}

/*
 *   Process a mouse click 
 */
int CHtmlSysWin_win32::do_leftbtn_down(int keys, int x, int y, int clicks)
{
    RECT rc;
    CHtmlDisp *disp;
    CHtmlPoint pos;
    unsigned long curofs, oldofs;
    unsigned long startofs, endofs;
    unsigned long startofs2, endofs2;
    CHtmlDispLink *link;

    /*
     *   When the tooltip control sees a mouse click, it turns itself off
     *   until the mouse moves out of the tool.  We don't want this to
     *   happen, because we're taking over the whole window for the tool.
     *   To get the tooltip to keep popping up after a click, deactivate
     *   it and then immediately reactivate it.  
     */
    SendMessage(tooltip_, TTM_ACTIVATE, FALSE, 0);
    SendMessage(tooltip_, TTM_ACTIVATE, TRUE, 0);

    /* presume we won't defer clearing the selection */
    clear_sel_pending_ = FALSE;

    /*
     *   if it's in the vertical scrollbar, ignore it; we can get these
     *   events when the scrollbar is disabled, since it passes them
     *   through to the parent window 
     */
    get_scroll_area(&rc, TRUE);
    if (x >= rc.right)
        return FALSE;

    /* clear the selection ranges in other subwindows */
    owner_->clear_other_sel(this);

    /*
     *   if we're in "MORE" mode, and they clicked on the "MORE" prompt,
     *   show them more text 
     */
    if (!prefs_->get_alt_more_style() && more_mode_)
    {
        RECT rc;
        POINT pt;

        /* get the area of the "MORE" prompt */
        get_moreprompt_rect(&rc);

        /* see if the click is within the "MORE" prompt */
        pt.x = x;
        pt.y = y;
        if (PtInRect(&rc, pt))
        {
            /* it's in the prompt area - release the "MORE" mode */
            release_more_mode(HTMLW32_MORE_PAGE);

            /* we've handled the mouse event */
            return TRUE;
        }
    }

    /* get the position in document coordinates */
    pos = screen_to_doc(CHtmlPoint(x, y));

    /* get the display item containing the position */
    disp = formatter_->find_by_pos(pos, TRUE);

    /*
     *   if we're over a display item with a link, track the link click 
     */
    link = (disp != 0 ? disp->get_link(formatter_, pos.x, pos.y) : 0);
    if (link != 0 && get_html_show_links()
        && (HTMLW32_MORE_ALLOW_LINKS || !more_mode_))
    {
        /* draw all of the items involved in the link in the hilited state */
        link->set_clicked(this, CHtmlDispLink_clicked);
        UpdateWindow(handle_);

        /* remember the link that we're tracking */
        track_link_ = link;

        /* remove any selection we have */
        cmdbuf_->set_sel_range(last_cmd_caret_, last_cmd_caret_,
                               last_cmd_caret_);
        update_caret_pos(FALSE, TRUE);
    }
    else
    {
        /*
         *   We're not clicking on a link - treat this as a text range
         *   selection click 
         */

        /* get the text offset of this position */
        curofs = formatter_->find_textofs_by_pos(pos);
        
        /* see if we're extending the selection */
        if (keys & MK_SHIFT)
        {
            /*
             *   extend the old selection -- if the caret was at the right
             *   of the old selection, extend from the left of the old
             *   selection, otherwise extend from the right of the old
             *   selection 
             */
            formatter_->get_sel_range(&startofs, &endofs);
            if (caret_at_right_)
            {
                oldofs = startofs;
                endofs = curofs;
            }
            else
            {
                oldofs = endofs;
                startofs = curofs;
            }
        }
        else
        {
            /* we're starting a new selection */
            startofs = endofs = oldofs = curofs;

            /* start off with forward drag direction */
            caret_at_right_ = TRUE;
        }

        /*
         *   If this is the first click, and we're inside the selection,
         *   don't immediately reset the selection range, since they may
         *   want to drag or right-click the range.  
         */
        if (clicks <= 1)
        {
            unsigned long cur_start, cur_end;

            /* see if we're inside the selection */
            formatter_->get_sel_range(&cur_start, &cur_end);
            if (curofs >= cur_start && curofs < cur_end)
            {
                /* 
                 *   set 'clicks' to zero to indicate that we don't want
                 *   to clear the selection just yet, but also set the
                 *   clear_sel_pending_ flag so that, on mouse-up, if we
                 *   never ended up doing anything else interesting, we'll
                 *   still end up clearing the selection 
                 */
                clicks = 0;
                clear_sel_pending_ = TRUE;
            }
            else
            {
                /* we're outside the range - reset the selection */
                clicks = 1;
            }
        }

        /* if we have any clicks, select the initial range */
        if (clicks != 0)
        {
            /*
             *   select the range, then read it back to make sure we have
             *   the canonical ordering 
             */
            formatter_->set_sel_range(startofs, endofs);
            formatter_->get_sel_range(&startofs, &endofs);
        }

        /* remember the number of clicks */
        track_clicks_ = clicks;

        /* if we have a multiple click, select a larger region */
        switch(clicks)
        {
        case 0:
            /* we may be starting a drag operation */
            drag_prepare(TRUE);
            break;
            
        case 2:
            /* double click - select a word at each end of the range */
            formatter_->get_word_limits(&startofs2, &endofs2, startofs);
            formatter_->get_word_limits(&startofs, &endofs, endofs);

            /* set the ranges */
            goto set_ranges;

        case 3:
            /* triple click - select a line at each end of the range */
            formatter_->get_line_limits(&startofs2, &endofs2, startofs);
            formatter_->get_line_limits(&startofs, &endofs, endofs);

        set_ranges:
            /* set the maximum extent of the two ranges */
            startofs = (startofs < startofs2 ? startofs : startofs2);
            endofs = (endofs > endofs2 ? endofs : endofs2);
            formatter_->set_sel_range(startofs, endofs);
            break;

        case 4:
            /* quadruple click - select everything */
            formatter_->set_sel_range(0, formatter_->get_text_ofs_max());
            break;
        }

        /* set the appropriate caret direction */
        caret_at_right_ = (curofs >= oldofs);

        /* update the command selection range from the formatter range */
        fmt_sel_to_cmd_sel(caret_at_right_);

        /* remember the tracking offsets */
        track_start_left_ = startofs;
        track_start_right_ = endofs;
    }

    /* remember that we're tracking the mouse, and where we started */
    tracking_mouse_ = TRUE;
    track_start_ = pos;

    /* hide the caret during tracking */
    if (caret_vis_)
        HideCaret(handle_);

    /* capture the mouse while dragging */
    SetCapture(handle_);

    /* prepare for drag scrolling */
    start_drag_scroll();

    /* handled */
    return TRUE;
}


/*
 *   Process mouse movement 
 */
int CHtmlSysWin_win32::do_mousemove(int /*keys*/, int x, int y)
{
    /* 
     *   If the tooltip just disappeared, reset it - we need to do this
     *   because we have only one big tool area covering our whole window,
     *   even though we may have numerous effective tool areas.  When the
     *   tool window times out, the control will normally not show a tip
     *   again until the mouse moves into a new tool; since we don't have
     *   any other tools for it to move into, it would never reappear
     *   until the mouse moved out of the window if we didn't reset it.  
     */
    if (tooltip_vanished_)
    {
        /* power-cycle the control */
        SendMessage(tooltip_, TTM_ACTIVATE, FALSE, 0);
        SendMessage(tooltip_, TTM_ACTIVATE, TRUE, 0);

        /* only do this once per vanishing */
        tooltip_vanished_ = FALSE;
    }
    
    /* if we're tracking the mouse, keep doing so */
    if (tracking_mouse_)
    {
        CHtmlPoint newpos;
        unsigned long curofs;
        unsigned long oldofs;
        unsigned long startofs;
        unsigned long endofs;
        int caret_at_right;

        /* check for a drag/drop operation */
        if (drag_check())
        {
            /* 
             *   don't clear the selection when we're done, since the
             *   click ended up meaning something different from making a
             *   selection 
             */
            clear_sel_pending_ = FALSE;
            
            /* we're done */
            return TRUE;
        }

        /* do drag scrolling if appropriate */
        maybe_drag_scroll(x, y);

        /* get the mouse position in document coordinates */
        newpos = screen_to_doc(CHtmlPoint(x, y));

        /* see what kind of event we're tracking */
        if (track_link_ != 0)
        {
            CHtmlDisp *disp;
            CHtmlDispLink *link;
            int new_clicked;

            /*
             *   tracking a click on a linked item 
             */

            /* 
             *   See if we're still over an object linked to this item,
             *   and update the clicked status of the link accordingly.  
             */
            disp = formatter_->find_by_pos(newpos, TRUE);
            link = (disp != 0 ?
                    disp->get_link(formatter_, newpos.x, newpos.y) : 0);
            new_clicked = (link == track_link_);
            if ((new_clicked != 0) != (track_link_->get_clicked() != 0))
            {
                /* status has changed - update the window immediately */
                track_link_->set_clicked(this, (new_clicked
                                                ? CHtmlDispLink_clicked
                                                : CHtmlDispLink_none));
                UpdateWindow(handle_);
            }
        }
        else
        {
            /*
             *   tracking text selection 
             */

            /* get the text offset of the mouse position */
            curofs = formatter_->find_textofs_by_pos(newpos);
            
            /* extend the selection */
            switch(track_clicks_)
            {
            case 0:
                /* leave the selection alone for now */
                return TRUE;
                
            default:
            case 1:
                /* extend the range to the current point */
                startofs = (caret_at_right_ ? track_start_left_
                          : track_start_right_);
                formatter_->set_sel_range(startofs, curofs);

                /* determine which end has the caret */
                caret_at_right = (curofs >= startofs);
                break;

            case 2:
                /* get the current word limits */
                formatter_->get_word_limits(&startofs, &endofs, curofs);

                /* extend the limits */
                goto extend_limits;

            case 3:
                /* get the current line limits */
                formatter_->get_line_limits(&startofs, &endofs, curofs);

            extend_limits:
                /*
                 *   extend the selection to enclose the lines at both ends
                 *   of the mouse range 
                 */
                oldofs = (caret_at_right_ ? track_start_left_
                          : track_start_right_);
                if (endofs > oldofs)
                {
                    startofs = oldofs;
                    caret_at_right = TRUE;
                }
                else
                {
                    endofs = oldofs;
                    caret_at_right = FALSE;
                }
                formatter_->set_sel_range(startofs, endofs);
                break;

            case 4:
                /* we've selected everything - leave the range alone */
                caret_at_right = TRUE;
                break;
            }

            /* remember the new positions as the new track starts */
            formatter_->get_sel_range(&track_start_left_,
                                      &track_start_right_);

            /* update the command selection range from the formatter range */
            fmt_sel_to_cmd_sel(caret_at_right);
        }

        /* handled */
        return TRUE;
    }

    /* we didn't handle it */
    return FALSE;
}

/*
 *   Process capture loss 
 */
int CHtmlSysWin_win32::do_capture_changed(HWND)
{
    /* terminate any tracking we're doing */
    return end_mouse_tracking(FALSE);
}

/*
 *   Process mouse left key release 
 */
int CHtmlSysWin_win32::do_leftbtn_up(int /*keys*/, int /*x*/, int /*y*/)
{
    /* finish tracking the mouse */
    return end_mouse_tracking(TRUE);
}

/*
 *   Terminate tracking of the mouse 
 */
int CHtmlSysWin_win32::end_mouse_tracking(int release_capture)
{
    if (tracking_mouse_)
    {
        unsigned long startofs, endofs;
        
        /* done with drag scrolling */
        end_drag_scroll();
        
        /* forget that we were tracking */
        tracking_mouse_ = FALSE;

        /* release mouse capture if desired */
        if (release_capture)
            ReleaseCapture();

        /* terminate drag/drop tracking if necessary */
        drag_end(TRUE);

        /* show the caret again */
        if (caret_vis_)
            ShowCaret(handle_);

        /* see what kind of tracking we were doing */
        if (track_link_ != 0)
        {
            const textchar_t *cmd;
            int append, enter;

            /*
             *   tracking a click on a linked item 
             */

            /* presume we won't need to execute any command */
            cmd = 0;

            /* 
             *   If (when last we checked) we were still over an object
             *   linked to this item, activate the item's HREF 
             */
            if (track_link_->get_clicked())
            {
                /* 
                 *   if we're in MORE mode, release the MORE and get ready
                 *   for a new command 
                 */
                if (more_mode_)
                    release_more_mode(HTMLW32_MORE_ALL);

                /* 
                 *   It was still clicked - get the HREF, which we'll
                 *   process as a command.  Note that we do NOT want to
                 *   process it right away, because we want to finish up
                 *   our processing of the link tracking first -
                 *   processing the command could potentially clear the
                 *   screen, which would delete the link object, so we
                 *   must wait until after we've forgotten about the link
                 *   (in just a few lines) to process the command. 
                 */
                cmd = track_link_->href_.get_url();

                /* get the command attributes from the link */
                append = track_link_->get_append();
                enter = !track_link_->get_noenter();
            }

            /* 
             *   the mouse button has been released, so the item is no
             *   longer clicked 
             */
            track_link_->set_clicked(this, CHtmlDispLink_none);

            /* we're done tracking the link - forget about it */
            track_link_ = 0;

            /*
             *   If we found a command to process, process it now.  We
             *   wait until now, because we had to finish our work with
             *   the link object before we process its command, in case
             *   processing the command clears the screen and thereby
             *   deletes the link object.  
             */
            if (cmd != 0)
                process_command(cmd, get_strlen(cmd), FALSE, append, enter);
        }
        else
        {
            /*
             *   Tracking text selection range
             */

            /* 
             *   If we were tracking "zero" clicks, it means that we
             *   clicked over the selection range; if the click never
             *   meant anything else (such as drag/drop or context menu),
             *   clear the selection now.  Note that we ignore this if
             *   capture is being forcibly rested from us, because it
             *   means that whatever was going on was somehow interrupted,
             *   hance the user didn't mean to change the selection.
             */
            if (clear_sel_pending_ && release_capture)
            {
                /* set the new selection range */
                formatter_->set_sel_range(track_start_left_,
                                          track_start_right_);

                /* update the command selection */
                fmt_sel_to_cmd_sel(TRUE);
            }
            
            /*
             *   If we have a zero-length selection, and the selection
             *   isn't in the command line, move the selection back to the
             *   command line.  We don't show the caret outside of the
             *   command line, so there would be no visual feedback for
             *   the selection point in this case; plus, the user can't do
             *   anything with a zero-length selection outside the command
             *   line.  
             */
            if (cmdtag_ != 0)
            {
                /* check the selection to see if it's zero-length */
                formatter_->get_sel_range(&startofs, &endofs);
                if (startofs == endofs && endofs < cmdtag_->get_text_ofs())
                {
                    /* no selection - move back to the command line */
                    cmdbuf_->set_sel_range(last_cmd_caret_, last_cmd_caret_,
                                           last_cmd_caret_);
                    update_caret_pos(FALSE, TRUE);
                }
            }
            else
            {
                /* 
                 *   no command - if we have a caret, set its position to
                 *   the appropriate end of the new selection range 
                 */
                update_caret_pos(FALSE, TRUE);
            }
        }

        /* handled */
        return TRUE;
    }

    /* didn't handle it */
    return FALSE;
}

/*
 *   Pre-drag notification
 */
void CHtmlSysWin_win32::drag_pre()
{
    /* leave the current selection alone when the mouse is released */
    clear_sel_pending_ = FALSE;

    /* hide the caret while we're dragging */
    HideCaret(handle_);
}

/*
 *   Post-drag notification 
 */
void CHtmlSysWin_win32::drag_post()
{
    /* show the caret again */
    ShowCaret(handle_);
}

/*
 *   Right button down event 
 */
int CHtmlSysWin_win32::do_rightbtn_down(int keys, int x, int y,
                                        int /*clicks*/)
{
    /* 
     *   Do normal left button tracking - when the button is released,
     *   we'll display the context menu appropriate for what we're doing
     *   after the tracking.  Note that we pass a click count of zero; the
     *   left-click routine treats this as a single click but doesn't
     *   clear the current selection range until we start moving the
     *   mouse, so if they release the button before moving the mouse
     *   they'll get the current range.  
     */
    return do_leftbtn_down(keys, x, y, 0);
}

/*
 *   Right button up event 
 */
int CHtmlSysWin_win32::do_rightbtn_up(int keys, int x, int y)
{
    /* don't clear the selection */
    clear_sel_pending_ = FALSE;
    
    /* finish the normal button tracking */
    do_leftbtn_up(keys, x, y);

    /* run the edit context menu */
    track_context_menu(edit_popup_, x, y);

    /* handled */
    return TRUE;
}
    
/*
 *   check a command to determine its current state 
 */
TadsCmdStat_t CHtmlSysWin_win32::check_command(int command_id)
{
    switch(command_id)
    {
    case ID_FILE_QUIT:
    case ID_FILE_SAVEGAME:
    case ID_FILE_RESTOREGAME:
        return started_reading_cmd_ ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_EDIT_CUT:
        return can_cut() ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_EDIT_COPY:
        return can_copy() ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_EDIT_PASTE:
        return can_paste() ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_EDIT_DELETE:
        return can_cut() ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_EDIT_SELECTALL:
        return TADSCMD_ENABLED;

    case ID_EDIT_SELECTLINE:
        return started_reading_cmd_ ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_EDIT_UNDO:
        return cmdbuf_->can_undo() ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_VIEW_MUSIC:
        return prefs_->get_music_on() ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_SOUNDS:
        return prefs_->get_sounds_on() ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_LINKS:
        return prefs_->get_links_on() ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_LINKS_HIDE:
        return (!prefs_->get_links_on() && !prefs_->get_links_ctrl())
            ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_LINKS_CTRL:
        return prefs_->get_links_ctrl() ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_LINKS_TOGGLE:
        return TADSCMD_ENABLED;

    case ID_VIEW_GRAPHICS:
        return prefs_->get_graphics_on() ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_FONTS_SMALLEST:
        return prefs_->get_font_scale() == 0
            ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_FONTS_SMALLER:
        return prefs_->get_font_scale() == 1
            ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_FONTS_MEDIUM:
        return prefs_->get_font_scale() == 2
            ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_FONTS_LARGER:
        return prefs_->get_font_scale() == 3
            ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_FONTS_LARGEST:
        return prefs_->get_font_scale() == 4
            ? TADSCMD_CHECKED : TADSCMD_ENABLED;

    case ID_VIEW_OPTIONS:
        return TADSCMD_ENABLED;

    case ID_GO_NEXT:
        return owner_->exists_next_page()
            ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_GO_PREVIOUS:
        return owner_->exists_previous_page()
            ? TADSCMD_ENABLED : TADSCMD_DISABLED;

    case ID_HELP_ABOUT:
    case ID_HELP_COMMAND:
        return TADSCMD_ENABLED;

    default:
        /* not known to us; inherit default handling */
        return CTadsWinScroll::check_command(command_id);
    }
}

/*
 *   Handle a command 
 */
int CHtmlSysWin_win32::do_command(int notify_code,
                                  int command_id, HWND ctl)
{
    /* 
     *   cancel any pending temporary highlighting, in case they activated
     *   the command through a keyboard accelerator 
     */
    set_temp_show_links(FALSE);

    /* check the command code for one we recognize */
    switch(command_id)
    {
    case ID_FILE_QUIT:
        /*
         *   Send a "QUIT" command.  Note that we do NOT set the
         *   quit_command flag, since we're only selecting the menu, not
         *   closing the window.  We only use the quit_command flag when
         *   closing the window, since that's a different sort of
         *   operation than selecting the menu.  
         */
        process_command("quit", 4, FALSE, FALSE, TRUE);
        return TRUE;

    case ID_FILE_SAVEGAME:
        process_command("save", 4, FALSE, FALSE, TRUE);
        return TRUE;
        
    case ID_FILE_RESTOREGAME:
        process_command("restore", 7, FALSE, FALSE, TRUE);
        return TRUE;

    case ID_EDIT_CUT:
        if (can_cut())
            do_cut();
        return TRUE;

    case ID_EDIT_COPY:
        if (can_copy())
            do_copy();
        return TRUE;

    case ID_EDIT_PASTE:
        if (can_paste())
            do_paste();
        return TRUE;

    case ID_EDIT_DELETE:
        if (can_cut() && cmdbuf_->del_right())
            update_input_display();
        return TRUE;

    case ID_EDIT_SELECTALL:
        /* clear selections in other subwindows */
        owner_->clear_other_sel(this);

        /* set the selection in the formatter and in the command buffer */
        formatter_->set_sel_range(0, formatter_->get_text_ofs_max());
        fmt_sel_to_cmd_sel(TRUE);
        return TRUE;

    case ID_EDIT_SELECTLINE:
        /* only proceed if we have a command */
        if (cmdtag_ != 0 && started_reading_cmd_)
        {
            /* clear selections in other subwindows */
            owner_->clear_other_sel(this);
            
            /* select from the start of the command to the end of the text */
            formatter_->set_sel_range(cmdtag_->get_text_ofs(),
                                      formatter_->get_text_ofs_max());

            /* update the selection and show the cursor */
            fmt_sel_to_cmd_sel(TRUE);
            update_caret_pos(TRUE, FALSE);

            /* handled */
            return TRUE;
        }
        break;

    case ID_EDIT_UNDO:
        if (cmdbuf_->can_undo() && cmdbuf_->undo())
            update_input_display();
        return TRUE;

    case ID_VIEW_FONTS_SMALLEST:
        set_font_scale(0);
        return TRUE;

    case ID_VIEW_FONTS_SMALLER:
        set_font_scale(1);
        return TRUE;

    case ID_VIEW_FONTS_MEDIUM:
        set_font_scale(2);
        return TRUE;

    case ID_VIEW_FONTS_LARGER:
        set_font_scale(3);
        return TRUE;

    case ID_VIEW_FONTS_LARGEST:
        set_font_scale(4);
        return TRUE;

    case ID_VIEW_OPTIONS:
        /* run the preferences dialog */
        prefs_->run_preferences_dlg(handle_, this);

        /* set the accelerator table for the Ctrl+V preferences */
        set_current_accel();

        /* handled */
        return TRUE;

    case ID_VIEW_SOUNDS:
        /* 
         *   if turning sounds on, and directx initialization failed, show
         *   a warning 
         */
        if (get_directsound() == 0)
        {
            /* issue the message */
            owner_->owner_show_directx_warning();

            /* nothing more to do here */
            return TRUE;
        }
        
        /* update the preferences */
        prefs_->set_sounds_on(!prefs_->get_sounds_on());

        /* cancel any sounds currently playing if turning sound off */
        if (!prefs_->get_sounds_on())
        {
            formatter_->cancel_sound(HTML_Attrib_ambient);
            formatter_->cancel_sound(HTML_Attrib_bgambient);
            formatter_->cancel_sound(HTML_Attrib_foreground);
        }
        return TRUE;

    case ID_VIEW_MUSIC:
        /* update the preferences */
        prefs_->set_music_on(!prefs_->get_music_on());

        /* cancel any music currently playing if turning music off */
        if (!prefs_->get_music_on())
            formatter_->cancel_sound(HTML_Attrib_background);
        return TRUE;

    case ID_VIEW_GRAPHICS:
        /* invert the preference setting */
        prefs_->set_graphics_on(!prefs_->get_graphics_on());

        /* reformat everything with the new settings */
        owner_->reformat_all_html(TRUE, TRUE, FALSE);
        return TRUE;

    case ID_VIEW_LINKS:
    case ID_VIEW_LINKS_HIDE:
    case ID_VIEW_LINKS_CTRL:
        {
            int old_showing;
            int new_showing;

            /* note whether links are visible right now */
            old_showing = (prefs_->get_links_on() != 0);

            /* note whether we're going to be showing links */
            new_showing = (command_id == ID_VIEW_LINKS);

            /*
             *   If we're changing visibility, note the new toggle status
             *   - if the user hits the toggle button, we want to come
             *   back to the current state.  If we're not changing
             *   visibility, leave the toggle setting alone, since we'll
             *   still want to toggle into the opposite visibility state. 
             */
            if (new_showing != old_showing)
            {
                /* 
                 *   we're changing visibility - the old settings are the
                 *   new toggle destination 
                 */
                toggle_link_on_ = (prefs_->get_links_on() != 0);
                toggle_link_ctrl_ = (prefs_->get_links_ctrl() != 0);
            }

            /* update the accelerator in the menu */
            owner_->set_link_accel(command_id == ID_VIEW_LINKS,
                                   command_id == ID_VIEW_LINKS_CTRL);
            
            /* update the preferences */
            prefs_->set_links_on(command_id == ID_VIEW_LINKS);
            prefs_->set_links_ctrl(command_id == ID_VIEW_LINKS_CTRL);

            /* update links if the visibility changed */
            if (new_showing != old_showing)
                inval_links_on_screen();
        }
        return TRUE;

    case ID_VIEW_LINKS_TOGGLE:
        /* 
         *   look at the toggle links setting and determine which command
         *   to apply
         */
        if (toggle_link_on_)
            command_id = ID_VIEW_LINKS;
        else if (toggle_link_ctrl_)
            command_id = ID_VIEW_LINKS_CTRL;
        else
            command_id = ID_VIEW_LINKS_HIDE;

        /* apply the new command */
        return do_command(notify_code, command_id, ctl);

    case ID_GO_NEXT:
        /* tell the owner to switch to the next page */
        owner_->go_next_page();
        return TRUE;

    case ID_GO_PREVIOUS:
        /* tell the owner to switch to the previous page */
        owner_->go_previous_page();
        return TRUE;

    case ID_HELP_ABOUT:
        {
            /* run our "about" dialog */
            CTadsDialog *dlg = new CTadsDlg_htabout();
            dlg->run_modal(DLG_RT_ABOUT, handle_,
                           CTadsApp::get_app()->get_instance());
            delete dlg;
        }
        return TRUE;

    case ID_HELP_COMMAND:
        /* enter a "help" command to the game */
        process_command("help", 4, FALSE, FALSE, TRUE);
        return TRUE;
    }

    /* not handled */
    return FALSE;
}

/*
 *   Process a keyboard virtual key event 
 */
int CHtmlSysWin_win32::do_keydown(int vkey, long /*keydata*/)
{
    /* 
     *   if it's not a control key, cancel any pending temporary link
     *   highlight mode 
     */
    switch(vkey)
    {
    case VK_CONTROL:
    case VK_LCONTROL:
    case VK_RCONTROL:
        /* turn on temporary links if appropriate */
        if (prefs_->get_links_ctrl() && temp_link_timer_id_ == 0)
        {
            /* 
             *   We want to make sure there's a slight delay before we
             *   show the links, to ensure that the screen doesn't flash
             *   annoyingly if the user only wants to activate an
             *   accelerator command.  Start a timer; if the key is still
             *   down when the timer goes off, we'll show the links at
             *   that point.  
             */
            temp_link_timer_id_ = alloc_timer_id();
            if (temp_link_timer_id_ != 0)
            {
                /* start the timer for a short delay */
                if (SetTimer(handle_, temp_link_timer_id_, 275,
                             (TIMERPROC)0) == 0)
                {
                    free_timer_id(temp_link_timer_id_);
                    temp_link_timer_id_ = 0;
                }
            }
        }
        return TRUE;

    default:
        /* cancel any pending highlighting */
        set_temp_show_links(FALSE);
        break;
    }
    
    /* see what we have */
    switch(vkey)
    {
    case VK_UP:
        /* scroll up a line */
        do_scroll(TRUE, vscroll_, SB_LINEUP, 0, FALSE);
        return TRUE;

    case VK_DOWN:
        /* scroll down a line (releasing more mode if necessary) */
        if (more_mode_ && moreprompt_is_at_bottom())
            release_more_mode(HTMLW32_MORE_LINE);
        else
            do_scroll(TRUE, vscroll_, SB_LINEDOWN, 0, FALSE);
        return TRUE;

    case VK_PRIOR:
        /* scroll up a page */
        do_scroll(TRUE, vscroll_, SB_PAGEUP, 0, FALSE);
        return TRUE;

    case VK_NEXT:
        /*
         *   Scroll down a page.  If we're in more mode and we're showing
         *   the bottom of the current text, show the next page;
         *   otherwise, simply scroll down a page normally.  
         */
        if (more_mode_ && moreprompt_is_at_bottom())
            release_more_mode(HTMLW32_MORE_PAGE);
        else
            do_scroll(TRUE, vscroll_, SB_PAGEDOWN, 0, FALSE);
        return TRUE;

    case VK_LEFT:
        /* scroll left */
        do_scroll(FALSE, hscroll_, SB_LINEUP, 0, FALSE);
        return TRUE;

    case VK_RIGHT:
        /* scroll right */
        do_scroll(FALSE, hscroll_, SB_LINEDOWN, 0, FALSE);
        return TRUE;

    case VK_END:
        /* scroll to bottom */
        do_scroll(TRUE, vscroll_, SB_BOTTOM, 0, FALSE);
        return TRUE;

    case VK_HOME:
        /* scroll to top */
        do_scroll(TRUE, vscroll_, SB_TOP, 0, FALSE);
        return TRUE;

    case VK_HELP:
    case VK_F1:
    case VK_INSERT:
    default:
        /* not handled */
        return FALSE;
    }
}

/*
 *   Process a key up event 
 */
int CHtmlSysWin_win32::do_keyup(int vkey, long keydata)
{
    switch(vkey)
    {
    case VK_CONTROL:
    case VK_LCONTROL:
    case VK_RCONTROL:
        /* turn off temporary links */
        set_temp_show_links(FALSE);
        return TRUE;

    default:
        /* inherit default processing */
        return CTadsWinScroll::do_keyup(vkey, keydata);
    }
}

/*
 *   Process an ALT key 
 */
int CHtmlSysWin_win32::do_syschar(TCHAR c, long /*keydata*/)
{
    /* check the key */
    switch(c)
    {
    case 'V':
    case 'v':
        /* 
         *   Alt-V - emacs-style command to scroll up a page; pass through
         *   to Windows if the emacs alt-V preference is turned off 
         */
        if (!prefs_->get_emacs_alt_v())
            return FALSE;

        /* scroll up a page */
        do_scroll(TRUE, vscroll_, SB_PAGEUP, 0, FALSE);
        break;

    default:
        /* this key was not processed */
        return FALSE;
    }

    /* if we got here, the key was processed */
    return TRUE;
}

/*
 *   user message handler 
 */
int CHtmlSysWin_win32::do_user_message(int msg, WPARAM wpar, LPARAM lpar)
{
    switch(msg)
    {
    case HTMLM_REFORMAT:
        /* reformat all of the windows */
        owner_->reformat_all_html((wpar & HTML_F_SHOW_STAT) != 0,
                                  (wpar & HTML_F_FREEZE_DISP) != 0,
                                  (wpar & HTML_F_RESET_SND) != 0);
        return TRUE;

    case HTMLM_ONRESIZE:
        /* reformat this window to adjust for resizing it */
        format_after_resize((long)lpar);
        return TRUE;

    case HTMLM_WAV_DONE:
        /* call the WAV object and tell it the wavefile is done */
        ((CHtmlSysSoundWav_win32 *)lpar)->on_sound_done((int)wpar);
        return TRUE;

    case HTMLM_MPEG_DONE:
        /* call the MPEG object and tell it the wavefile is done */
        ((CHtmlSysSoundMpeg_win32 *)lpar)->on_sound_done((int)wpar);
        return TRUE;

    default:
        /* message not recognized */
        return FALSE;
    }
}

/*
 *   set the font scale 
 */
void CHtmlSysWin_win32::set_font_scale(int scale)
{
    CHtmlFontDesc font_desc;
    
    /* remember the new scale */
    prefs_->set_font_scale(scale);

    /* set a new default font */
    font_desc.htmlsize = 3;
    default_font_ = (CHtmlSysFont_win32 *)get_font(&font_desc);

    /* reformat all of the windows */
    owner_->reformat_all_html(TRUE, TRUE, FALSE);

    /* reset the caret for the size change */
    set_caret_size(formatter_->get_font());
    update_caret_pos(FALSE, FALSE);
}


/*
 *   Process a keyboard character event 
 */
int CHtmlSysWin_win32::do_char(TCHAR c, long /*keydata*/)
{
    /* if we're in "MORE" mode, allow displaying more data */
    if (more_mode_)
    {
        switch(c)
        {
        case 32:
            /* space - show the next screen */
            release_more_mode(HTMLW32_MORE_PAGE);
            break;

        case 10:
        case 13:
            /* return - show next line */
            release_more_mode(HTMLW32_MORE_LINE);
            break;

        case 22:
            /* ^V - page down */
            do_keydown(VK_NEXT, 0);
            break;

        default:
            /* ignore other characters in more-mode */
            return FALSE;
        }

        /* keystroke handled */
        return TRUE;
    }

    /* see what we have */
    switch(c)
    {
    case 22:
        /* ^V - page down (Emacs mode) or Paste (Windows mode) */
        if (prefs_->get_emacs_ctrl_v())
        {
            /* emacs style - scroll down a page */
            do_scroll(TRUE, vscroll_, SB_PAGEDOWN, 0, FALSE);
            return TRUE;
        }
        break;

    case 3:
        /* ^C - copy */
        return do_copy();
    }

    /* we didn't handle the keystroke */
    return FALSE;
}


/*
 *   Turn "MORE" mode on or off 
 */
void CHtmlSysWin_win32::set_more_mode(int flag)
{
    /* if MORE mode isn't enabled, ignore it */
    if (flag && !more_mode_enabled_)
        return;
    
    /* if it's not changing, ignore the request */
    if (flag == more_mode_)
        return;

    /* set the new mode */
    more_mode_ = flag;

    /* update the status line */
    if (statusline_ != 0)
        statusline_->update();

    /* make sure we redraw the prompt area */
    inval_more_area();
}

/*
 *   Wait for the user to acknowledge a MORE prompt displayed explicitly
 *   by the game. 
 */
void CHtmlSysWin_win32::do_game_more_prompt()
{
    int ret;
    int ht_increment;

    /* make sure the caret is positioned correctly */
    update_caret_pos(FALSE, FALSE);

    /* 
     *   Wait for the user to acknowledge any existing MORE prompt.  
     */
    while (more_mode_)
    {
        /* process events until we're no longer in "MORE" mode */
        released_more_ = FALSE;
        ret = CTadsApp::get_app()->event_loop(&released_more_, 0, 0);

        /* if we're quitting, give up now */
        if (ret == 0)
            return;
    }

    /*
     *   If we're displaying an on-screen more prompt, temporarily
     *   increase the scrollable height of the text by the size of the
     *   more prompt, so that our more prompt doesn't overwrite the last
     *   part of the displayed text.  
     */
    if (!prefs_->get_alt_more_style())
    {
        RECT rc;

        /* get the height adjustment */
        get_moreprompt_rect(&rc);
        ht_increment = rc.bottom - rc.top + 2;

        /* adjust the formatter's total height */
        formatter_->inc_max_y_pos(ht_increment);

        /* make sure we're scrolled all the way down */
        fmt_adjust_vscroll();
    }
    else
        ht_increment = 0;

    /*
     *   Pretend that the last input position was at the top of the
     *   screen.  This is necessary because, when the more prompt is
     *   showing, we assume that we've reached the point where the last
     *   input position is about to scroll off the screen.  Since we're
     *   showing a more prompt manually here, we may have displayed less
     *   text than this, which will be a problem if we display a big pile
     *   of text *after* this manual more prompt, unless we adjust the
     *   input position to compensate. 
     */
    last_input_height_ = screen_to_doc_y(0);

    /* remember that we're processing a game "more" request */
    game_more_request_ = TRUE;

    /* make sure the caret is positioned correctly */
    update_caret_pos(FALSE, FALSE);

    /*
     *   Now display the MORE prompt one more time, and once again process
     *   events until the user acknowledges it. 
     */
    set_more_mode(TRUE);
    while (more_mode_)
    {
        /* process events until we're no longer in "MORE" mode */
        released_more_ = FALSE;
        ret = CTadsApp::get_app()->event_loop(&released_more_, 0, 0);

        /* if we're quitting, give up now */
        if (ret == 0)
            return;
    }

    /* done with the game more request */
    game_more_request_ = FALSE;

    /*
     *   If we temporarily increased the scrolling area size to make room
     *   for the prompt, we must now return things to their prior state. 
     */
    if (ht_increment != 0)
        formatter_->inc_max_y_pos(-ht_increment);
}

/*
 *   Set the exit-pause flag 
 */
void CHtmlSysWin_win32::set_exit_pause(int flag)
{
    /* note the flag */
    exit_pause_ = flag;

    /* update the status line */
    if (statusline_ != 0)
        statusline_->update();
}

/*
 *   Bring the window to the front.  We'll ask our owner to do the work,
 *   since we're meant to run as a child window.  
 */
void CHtmlSysWin_win32::bring_to_front()
{
    owner_->bring_owner_to_front();
}

/*
 *   Get the handle of the frame window.  We'll ask the owner to get the
 *   correct handle, since we're meant to run as a child window. 
 */
HWND CHtmlSysWin_win32::get_frame_handle() const
{
    return owner_->get_owner_frame_handle();
}

/*
 *   Process a command.  This default implementation passes the command to
 *   the owner for processing, since default windows have no native
 *   command processing abilities. 
 */
void CHtmlSysWin_win32::process_command(const textchar_t *cmd, size_t len,
                                        int quit_command, int append,
                                        int enter)
{
    owner_->process_command(cmd, len, quit_command, append, enter);
}

/*
 *   Update the status line 
 */
textchar_t *CHtmlSysWin_win32::get_status_message(int *caller_deletes)
{
    textchar_t *msg;

    /* find out what we should display */
    if (exit_pause_)
    {
        msg = "Press any key to exit.";
    }
    else if (formatting_msg_)
    {
        /* formatting is underway - display "working..." */
        msg = "Working...";
    }
    else if (status_link_ != 0)
    {
        msg = status_link_->title_.get();
        if (msg == 0 || get_strlen(msg) == 0)
            msg = (textchar_t *)status_link_->href_.get_url();
    }
    else if (more_mode_ && prefs_->get_alt_more_style())
    {
        /* we're in "MORE" mode - say so on the status line */
        msg = "*** MORE *** [press the space bar to continue]";
    }
    else if (waiting_for_key_pause_)
    {
        msg = "Please press a key...";
    }
    else if (pruning_msg_)
    {
        msg = "Working (discarding old text)...";
    }
    else
    {
        /* we have nothing to display */
        msg = 0;
    }

    /* the message is static; no need for anyone to delete it */
    *caller_deletes = FALSE;

    /* return our message */
    return msg;
}

/*
 *   Release the "MORE" prompt, showing the next screen of text 
 */
void CHtmlSysWin_win32::release_more_mode(Html_more_t amount)
{
    RECT more_rc;
    RECT win_rc;
    
    /* no longer in "MORE" mode */
    set_more_mode(FALSE);

    /*
     *   Before moving the window's pixels around, remove the "MORE"
     *   prompt and fill in any text that we clipped out.  We need to do
     *   this and actually redraw the window prior to scrolling the window
     *   contents, because Windows will assume that the part on the screen
     *   is all correct after the scroll.  
     */
    inval_more_area();
    UpdateWindow(handle_);

    /* get the size of the "more" prompt area */
    get_moreprompt_rect(&more_rc);

    /* note the new input position */
    switch(amount)
    {
    case HTMLW32_MORE_PAGE:
        /*
         *   advance the input height to just above the last line we
         *   displayed - this will scroll so that the line that's
         *   currently at the bottom will be displayed at the top 
         */
        get_scroll_area(&win_rc, TRUE);
        last_input_height_ += (win_rc.bottom - win_rc.top)
                              - 2*(more_rc.bottom - more_rc.top);
        break;

    case HTMLW32_MORE_LINE:
        /* advance the input height by the size of the more prompt */
        last_input_height_ += (more_rc.bottom - more_rc.top);
        break;

    case HTMLW32_MORE_ALL:
        /* advance to show all pending text */
        last_input_height_ = formatter_->get_max_y_pos();
        break;
    }

    /* adjust the scrollbars and scroll to the bottom */
    fmt_adjust_vscroll();

    /*
     *   If the more prompt is still on the screen, make sure we redraw
     *   everything from the clipped position down, since this can't be
     *   properly reconstructed by a simple scroll redraw operation. 
     */
    if (more_mode_)
        inval_more_area();

    /*
     *   If we're no longer in MORE mode, set the flag to indicate that
     *   MORE mode has ended. 
     */
    if (!more_mode_)
        released_more_ = TRUE;
}

/*
 *   Invalidate the "more" prompt area.  This invalidates everything from
 *   the start of the text clipped out by the "more" prompt to the bottom
 *   of the screen.  This is needed before and after scrolling when the
 *   "more" prompt is visible, since the prompt can't be drawn during
 *   scrolling (as its position is not determined until the scrolling
 *   settings are finalized).  
 */
void CHtmlSysWin_win32::inval_more_area()
{
    RECT more_rc;
    RECT client_rc;

    /* 
     *   if we're not in MORE mode, or we're using the non-client-area
     *   more style, there's nothing to do 
     */
    if (prefs_->get_alt_more_style())
        return;

    /* invalidate from the clip position down */
    get_moreprompt_rect(&more_rc);
    GetClientRect(handle_, &client_rc);
    client_rc.top = doc_to_screen_y(clip_ypos_);
    if (more_rc.top < client_rc.top) client_rc.top = more_rc.top;
    InvalidateRect(handle_, &client_rc, FALSE);
}

/*
 *   Get the area occupied by the MORE prompt 
 */
void CHtmlSysWin_win32::get_moreprompt_rect(RECT *rc)
{
    HDC dc;
    SIZE txtsiz;
    int x, y;
    SCROLLINFO info;
    long vscroll_max;
    
    /* get the text area */
    get_scroll_area(rc, TRUE);

    /*
     *   The absolute position is bottom-aligned at the last scrolling
     *   position.  Figure out where this goes based on our current
     *   scrolling settings.  
     */
    info.cbSize = sizeof(info);
    info.fMask = SIF_ALL;
    GetScrollInfo(vscroll_, SB_CTL, &info);
    vscroll_max = info.nMax - info.nPage + 1;
    y = doc_to_screen_y(vscroll_max*get_vscroll_units() + rc->bottom);
    x = doc_to_screen_x(0);

    /* set the appropriate font */
    dc = GetDC(handle_);
    select_font(dc, get_default_font());

    /* get the size of the prompt */
    GetTextExtentPoint32(dc, moreprompt, sizeof(moreprompt)-1, &txtsiz);

    /*
     *   position it at the bottom of the text area when scrolled all the
     *   way to the bottom 
     */
    SetRect(rc, x, y - txtsiz.cy, rc->right, y);

    /* done with the device context */
    ReleaseDC(handle_, dc);
}

/*
 *   Gain focus - show the caret if appropriate.
 */
void CHtmlSysWin_win32::do_setfocus(HWND)
{
    /* show the caret if it's enabled */
    show_caret();
}

/*
 *   Lose focus.  Hide the caret.
 */
void CHtmlSysWin_win32::do_killfocus(HWND)
{
    /* hide the caret if necessary */
    hide_caret();

    /* forget any temporary link highlighting we're doing */
    set_temp_show_links(FALSE);
}

/*
 *   Create and show the caret 
 */
void CHtmlSysWin_win32::show_caret()
{
    if (!caret_vis_ && caret_enabled_ && GetFocus() == handle_)
    {
        CreateCaret(handle_, 0, 2, caret_ascent_);
        ShowCaret(handle_);
        caret_vis_ = TRUE;
        update_caret_pos(FALSE, FALSE);
    }
}

/*
 *   Hide and destroy the caret 
 */
void CHtmlSysWin_win32::hide_caret()
{
    if (caret_vis_)
    {
        HideCaret(handle_);
        DestroyCaret();
        caret_vis_ = FALSE;
    }
}

/*
 *   Update the caret position 
 */
void CHtmlSysWin_win32::update_caret_pos(int bring_on_screen,
                                         int update_sel_range)
{
    unsigned long sel_start, sel_end;

    /* if there's no caret, ignore it */
    if (!caret_enabled_)
        return;

    /* set the caret to the end of the formatter's last object */
    formatter_->get_sel_range(&sel_start, &sel_end);
    caret_pos_ = formatter_->
                 get_text_pos(caret_at_right_ ? sel_end : sel_start);

    /* convert to screen coordinates */
    caret_pos_ = doc_to_screen(caret_pos_);

    /* don't bring anything on the screen if we're in MORE mode */
    if (more_mode_)
        bring_on_screen = FALSE;

    /* scroll as needed to bring the caret onto the screen if desired */
    if (bring_on_screen)
        scroll_to_show_caret();

    /* move the Windows caret, if we're showing it */
    if (caret_vis_)
    {
        if (more_mode_)
            SetCaretPos(-100, -100);
        else
            SetCaretPos(caret_pos_.x, caret_pos_.y + 2);
    }

    /* update the selection range in the formatter */
    if (update_sel_range)
    {
        /* clear any selections in other subwindows */
        owner_->clear_other_sel(this);
    }
}

/*
 *   Scroll as needed to bring the caret on-screen 
 */
void CHtmlSysWin_win32::scroll_to_show_caret()
{
    RECT rc;
    
    /*
     *   if the caret is off-screen horizontally, and we want to bring it
     *   back on screen, scroll to move the caret back within the window 
     */
    get_scroll_area(&rc, TRUE);
    if (caret_pos_.x < rc.left || caret_pos_.x >= rc.right)
    {
        long target;
        long delta;
        long new_hscroll;

        /* scroll so that the caret is in the middle of the window */
        target = (rc.left + (rc.right - rc.left)/2);
        delta = caret_pos_.x - target;
        new_hscroll = (hscroll_ofs_ * get_hscroll_units() + delta)
                      / get_hscroll_units();
        do_scroll(FALSE, hscroll_, SB_THUMBPOSITION, new_hscroll, TRUE);
    }

    /* likewise, scroll vertically if necessary */
    if (caret_pos_.y < rc.top || caret_pos_.y + caret_ht_ >= rc.bottom)
    {
        long target;
        long delta;
        long new_vscroll;

        /* scroll so that the caret is in the middle of the window */
        target = (rc.top + (rc.bottom - rc.top)/2);
        delta = caret_pos_.y - target;
        new_vscroll = (vscroll_ofs_ * get_vscroll_units() + delta)
                      / get_vscroll_units();
        do_scroll(TRUE, vscroll_, SB_THUMBPOSITION, new_vscroll, TRUE);
    }
}

/*
 *   Handle control notification messages 
 */
int CHtmlSysWin_win32::do_notify(int control_id, int notify_code,
                                 LPNMHDR nmhdr)
{
    TOOLINFO ti;
    POINT pt;
    CHtmlPoint pos;
    static char text_buf[512];
    CHtmlDisp *disp;
    
    switch(notify_code)
    {
    case TTN_NEEDTEXT:
        /* get the mouse position */
        GetCursorPos(&pt);
        ScreenToClient(handle_, &pt);

        /* figure out what we're over */
        pos = screen_to_doc(CHtmlPoint(pt.x, pt.y));
        disp = formatter_->find_by_pos(pos, TRUE);

        /* 
         *   if we're not over anything, or it has no ALT text, display
         *   nothing 
         */
        if (disp == 0 || disp->get_alt_text() == 0)
        {
            /* 
             *   turn the tooltip off and then on again, to keep it from
             *   displaying anything right now but still leave it active
             *   for another try later 
             */
            SendMessage(tooltip_, TTM_ACTIVATE, FALSE, 0);
            SendMessage(tooltip_, TTM_ACTIVATE, TRUE, 0);
        }
        else
        {
            TOOLTIPTEXT *ttt;

            /* use the ALT text as the result */
            ttt = (TOOLTIPTEXT *)nmhdr;
            ttt->lpszText = (char *)disp->get_alt_text();
        }

        /* handled */
        return TRUE;

    case TTN_SHOW:
        /* get the mouse position */
        GetCursorPos(&pt);
        ScreenToClient(handle_, &pt);

        /* figure out what we're over */
        pos = screen_to_doc(CHtmlPoint(pt.x, pt.y));
        disp = formatter_->find_by_pos(pos, TRUE);

        /* if we have an object, get its screen area */
        if (disp != 0)
        {
            CHtmlRect rc;
            
            /* get its screen coordinates */
            rc = doc_to_screen(disp->get_pos());

            /* set the bounds of the object's area */
            ti.cbSize = sizeof(ti);
            SetRect(&ti.rect, rc.left, rc.top, rc.right, rc.bottom);
            ti.hwnd = handle_;
            ti.uId = 1;

            /* set the tooltip area */
            SendMessage(tooltip_, TTM_NEWTOOLRECT, 0, (LPARAM)&ti);
        }

        /* handled */
        return TRUE;

    case TTN_POP:
        /* restore the full-window tool rectangle */
        ti.cbSize = sizeof(ti);
        SetRect(&ti.rect, 0, 0, 32767, 32767);
        ti.hwnd = handle_;
        ti.uId = 1;
        SendMessage(tooltip_, TTM_NEWTOOLRECT, 0, (LPARAM)&ti);

        /* 
         *   note that it just vanished - the next time we move the mouse,
         *   we'll reset the control 
         */
        tooltip_vanished_ = TRUE;

        /* handled */
        return TRUE;

    default:
        /* inherit default handling */
        return CTadsWinScroll::do_notify(control_id, notify_code, nmhdr);
    }
}


/*
 *   Clear the selection range 
 */
void CHtmlSysWin_win32::clear_sel_range()
{
    long start;
    
    /* 
     *   clear the selection range in the formatter by setting both ends
     *   of the range to the maximum text offset in the formatter's
     *   display list 
     */
    start = formatter_->get_text_ofs_max();
    formatter_->set_sel_range(start, start);
    fmt_sel_to_cmd_sel(TRUE);
}

/*
 *   Clear the hover information 
 */
void CHtmlSysWin_win32::clear_hover_info()
{
    /* forget any link tracking information */
    set_hover_link(0);
}

/*
 *   Clear any memory of the link we're hovering over 
 */
void CHtmlSysWin_win32::set_hover_link(CHtmlDispLink *link)
{
    /* if we're changing it, update everything */
    if (link != status_link_)
    {
        /* remove highlighting from the previous hovering item */
        if (status_link_ != 0 && prefs_->get_hover_hilite())
            status_link_->set_clicked(this, CHtmlDispLink_none);

        /* set the new link */
        status_link_ = link;

        /* show highlighting on the new link if appropriate */
        if (status_link_ != 0 && prefs_->get_hover_hilite())
            status_link_->set_clicked(this, CHtmlDispLink_hover);

        /* update the status line for the change */
        if (statusline_ != 0)
            statusline_->update();
    }
}

/*
 *   enter an interactive size/move operation 
 */
void CHtmlSysWin_win32::do_entersizemove()
{
    /* remember that we're doing an interactive size/move */
    in_size_move_ = TRUE;
}

/*
 *   finish an interactive size/move operation 
 */
void CHtmlSysWin_win32::do_exitsizemove()
{
    RECT rc;
    long new_width;
    
    /* forget about the operation */
    in_size_move_ = FALSE;

    /* now we can reformat, if necessary */
    GetClientRect(handle_, &rc);
    new_width = rc.right - rc.left;
    if (new_width != fmt_width_ && formatter_ != 0)
        format_after_resize(new_width);
}

/*
 *   timer events 
 */
int CHtmlSysWin_win32::do_timer(int timer_id)
{
    /* see if we recognize the timer */
    if (timer_id == cb_timer_id_)
    {
        htmlw32_timer_t *cur;
        htmlw32_timer_t *nxt;
        
        /* 
         *   it's the callback timer - go through our list of registered
         *   callbacks and invoke each one 
         */
        for (cur = timer_list_head_ ; cur != 0 ; cur = nxt)
        {
            /* 
             *   remember the next timer in the list, in case the current
             *   one is deregistered within the callback (assume that one
             *   callback won't deregister another, though) 
             */
            nxt = cur->nxt_;

            /* invoke the callback */
            (*cur->func_)(cur->ctx_);
        }

        /* handled */
        return TRUE;
    }
    else if (timer_id == temp_link_timer_id_)
    {
        /* turn on temporary links */
        set_temp_show_links(TRUE);

        /* turn off the timer - it's done its job */
        KillTimer(handle_, temp_link_timer_id_);
        free_timer_id(temp_link_timer_id_);
        temp_link_timer_id_ = 0;

        /* handled */
        return TRUE;
    }
    else if (timer_id == bg_timer_id_)
    {
        /*
         *   It's our miscellaneous background event timer - call our idle
         *   routine to do any housekeeping chores that we need to perform
         *   periodically. 
         */
        do_idle();
        return TRUE;
    }
    else
    {
        /* not one of ours - inherit default handling */
        return CTadsWinScroll::do_timer(timer_id);
    }
}

/*
 *   Perform background tasks.  This is called periodically on a timer. 
 */
void CHtmlSysWin_win32::do_idle()
{
    POINT pos;
    
    /*
     *   Check the mouse location.  If the mouse is no longer in our
     *   window, remove any highlighted link. 
     */
    GetCursorPos(&pos);
    if (WindowFromPoint(pos) != handle_)
        set_hover_link(0);
}

/*
 *   resize 
 */
void CHtmlSysWin_win32::do_resize(int mode, int x, int /*y*/)
{
    RECT rc;

    /* move the scrollbars to their new positions */
    adjust_scrollbar_positions();

    /* update our memory of the display width */
    get_scroll_area(&rc, TRUE);
    disp_width_ = rc.right - rc.left;
    disp_height_ = rc.bottom - rc.top;

    /* see what we're doing */
    switch(mode)
    {
    case SIZE_MAXHIDE:
    case SIZE_MAXSHOW:
    case SIZE_MINIMIZED:
        /* ignore these modes */
        break;

    case SIZE_MAXIMIZED:
    case SIZE_RESTORED:
    default:
        /*
         *   If the width changed, reformat the text to account for the
         *   new size.  Only schedule the reformatting right now, since we
         *   can get a flurry of these messages if the system is set to
         *   show window contents while moving or resizing windows.
         *   
         *   If we're in the process of formatting after a resize, ignore
         *   the change entirely -- we're just fitting everything to the
         *   final size, and we don't want to change the size again.  It's
         *   important that we don't resize again just because we're doing
         *   a reformat, since this would prevent us from converging --
         *   we'd just keep resizing over and over on each pass.  
         */
        if (x != fmt_width_ && formatter_ != 0
            && !doing_resize_reformat_)
        {
            MSG msg;

            /* if we haven't already scheduled reformatting, do so now */
            if (!PeekMessage(&msg, handle_, HTMLM_ONRESIZE, HTMLM_ONRESIZE,
                             PM_NOREMOVE))
            {
                /* haven't scheduled it yet - post a message to myself */
                PostMessage(handle_, HTMLM_ONRESIZE, 0, (LPARAM)x);
            }
        }

        /* done */
        break;
    }
}

/*
 *   Reformat after a resize 
 */
void CHtmlSysWin_win32::format_after_resize(long new_width)
{
    int drawn;
    SCROLLINFO info;

    /* 
     *   if an interactive size/move is in progress, ignore the
     *   notification for now, and defer it until the size/move is done 
     */
    if (in_size_move_)
        return;

    /* 
     *   if we're currently scrolled to the bottom of the text, we'll want
     *   to show the bottom of the text after reformatting, so note this
     *   now 
     */
    get_scroll_info(TRUE, &info);
        
    /* reformat everything if necessary */
    if (need_format_on_resize_)
    {
        /* remember that we're doing a post-resize format pass */
        doing_resize_reformat_ = TRUE;
        
        /* forget any tracking links */
        status_link_ = track_link_ = 0;

        /* discard all current formatting */
        formatter_->start_at_top(FALSE);

        /* run the formatter */
        drawn = do_formatting(TRUE, TRUE, TRUE);

        /* if we didn't redraw the window already, do so now */
        if (!drawn)
            InvalidateRect(handle_, 0, FALSE);

        /* remember the new format width */
        fmt_width_ = new_width;

        /* we're done with the resize pass */
        doing_resize_reformat_ = FALSE;
    }

    /* adjust the caret and scrollbar settings */
    adjust_for_reformat();

    /* scroll to the bottom of the text if we were there previously */
    if (need_format_on_resize_
        && info.nPos + (int)info.nPage - 1 >= info.nMax)
    {
        /* scroll to the bottom */
        do_scroll(TRUE, vscroll_, SB_BOTTOM, 0, 0);
    }
}

/*
 *   Adjust the caret and scrollbars after reformatting the window 
 */
void CHtmlSysWin_win32::adjust_for_reformat()
{
    /* refigure the scrollbar ranges */
    if (!more_mode_ || prefs_->get_alt_more_style())
        content_height_ = formatter_->get_max_y_pos();
    adjust_scrollbar_ranges();

    /* figure the new caret position */
    update_caret_pos(FALSE, FALSE);
}

/*
 *   Get the basic scroll settings 
 */
int CHtmlSysWin_win32::get_scroll_info(int vert, SCROLLINFO *info)
{
    RECT rc;
    
    /* set up the basic elements of the structure */
    CTadsWinScroll::get_scroll_info(vert, info);
    
    /* get our scrolling area */
    get_scroll_area(&rc, TRUE);

    if (vert)
    {
        /* set the limits based on the current display size */
        info->nMin = 0;
        info->nMax = content_height_ / get_vscroll_units();
        info->nPage = rc.bottom / get_vscroll_units();
        info->nPos = vscroll_ofs_;

        /*
         *   disable the scrollbar if scrolling is impossible, rather
         *   than making it invisible 
         */
        info->fMask |= SIF_DISABLENOSCROLL;

        /* the vertical scroll is always visible */
        return TRUE;
    }
    else
    {
        int new_hscroll_vis;
        
        /* presume we won't show the horizontal scrollbar */
        new_hscroll_vis = FALSE;

        if (formatter_ != 0)
        {
            /*
             *   show the horizontal scrollbar if we're smaller than the
             *   longest line 
             */
            if (rc.right < formatter_->get_max_line_width()
                || hscroll_ofs_ != 0)
            {
                /* show the scrollbar */
                new_hscroll_vis = TRUE;

                /* set its size */
                info->nMin = 0;
                info->nMax = formatter_->get_max_line_width()
                             / get_hscroll_units();
                info->nPos = hscroll_ofs_;
                info->nPage = rc.right / get_hscroll_units();

                /*
                 *   if the max is below the current position, raise the
                 *   max temporarily 
                 */
                if (info->nMax < info->nPos + (int)info->nPage - 1)
                    info->nMax = info->nPos + info->nPage - 1;
            }
        }

        /* return visibility indication */
        return new_hscroll_vis;
    }
}

/*
 *   Adjust the scrollbar ranges
 */
void CHtmlSysWin_win32::adjust_scrollbar_ranges()
{
    int new_hscroll_vis;
    SCROLLINFO info;

    /* vertical scrollbar is always visible */
    vscroll_vis_ = TRUE;

    /* get the vertical scrollbar info and set it in the scrollbar */
    get_scroll_info(TRUE, &info);
    SetScrollInfo(vscroll_, SB_CTL, &info, TRUE);

    /* adjust our offset if it changed */
    GetScrollInfo(vscroll_, SB_CTL, &info);
    if (vscroll_ofs_ != info.nPos)
        do_scroll(TRUE, vscroll_, SB_THUMBPOSITION, info.nPos, TRUE);

    /* get horizontal scrollbar info and set it in the scrollbar */
    new_hscroll_vis = get_scroll_info(FALSE, &info);
    SetScrollInfo(hscroll_, SB_CTL, &info, TRUE);

    /* adjust our offset if it changed */
    if (new_hscroll_vis)
    {
        GetScrollInfo(hscroll_, SB_CTL, &info);
        if (hscroll_ofs_ != info.nPos)
            do_scroll(FALSE, hscroll_, SB_THUMBPOSITION, info.nPos, TRUE);
    }

    /* see if we're changing the horizontal scrollbar's visibility */
    if ((hscroll_vis_ != 0) != (new_hscroll_vis != 0))
    {
        /* note the new visibility */
        hscroll_vis_ = new_hscroll_vis;
        
        /* show the scrollbars as appropriate */
        ShowWindow(hscroll_, hscroll_vis_ ? SW_SHOW : SW_HIDE);

        /* fix up the scrollbar positions */
        adjust_scrollbar_positions();
    }
}


/*
 *   Receive notification that scrolling has occurred 
 */
void CHtmlSysWin_win32::notify_scroll(HWND, long /*oldpos*/, long /*newpos*/)
{
    /* turn off alternate more mode if we're at the bottom */
    if (prefs_->get_alt_more_style() && more_mode_)
    {
        SCROLLINFO info;

        /* 
         *   if we're at the bottom, cancel more mode; don't do this if
         *   we're processing an explicit game morePrompt request, since
         *   we want to make sure they hit a key for those 
         */
        get_scroll_info(TRUE, &info);
        if (info.nPos + (int)info.nPage - 1 >= info.nMax
            && !game_more_request_)
        {
            /* we've scrolled to the bottom - cancel more mode */
            set_more_mode(FALSE);

            /* flag that we've released the more prompt */
            released_more_ = TRUE;

            /* check to see if we can start reading a command now */
            check_cmd_start();
        }
        else
        {
            RECT rc;

            /*
             *   set the input position to the current position, so that
             *   we'll continue scrolling from here if we scroll via the
             *   keyboard (with space or return) 
             */
            get_scroll_area(&rc, TRUE);
            last_input_height_ = screen_to_doc_y(0);
        }
    }

    /* move the caret */
    update_caret_pos(FALSE, FALSE);
}


/*
 *   process creation event 
 */
void CHtmlSysWin_win32::do_create()
{
    TOOLINFO ti;

    /* do default creation */
    CTadsWinScroll::do_create();

    /* set the default caret height according to the default font height */
    set_caret_size(default_font_);

    /* start the system timer if appropriate */
    start_sys_timer();

    /* 
     *   set up an idle timer to take care of some miscellaneous
     *   background tasks 
     */
    bg_timer_id_ = alloc_timer_id();
    if (bg_timer_id_ != 0
        && SetTimer(handle_, bg_timer_id_, 500, (TIMERPROC)0) == 0)
    {
        /* that failed - free our timer ID */
        free_timer_id(bg_timer_id_);
        bg_timer_id_ = 0;
    }

    /* 
     *   Determine the link toggle settings.  If links are currently on,
     *   toggle to off.  If they're currently off or in ctrl mode, toggle
     *   to on. 
     */
    toggle_link_ctrl_ = FALSE;
    toggle_link_on_ = !prefs_->get_links_on();

    /* create the tooltip control */
    tooltip_ = CreateWindowEx(0, TOOLTIPS_CLASS, 0, 0,
                              CW_USEDEFAULT, CW_USEDEFAULT,
                              CW_USEDEFAULT, CW_USEDEFAULT,
                              handle_, 0,
                              CTadsApp::get_app()->get_instance(), 0);
    
    /* 
     *   add the "tool" - the entire display area; we'll figure out where
     *   we actually are when we need to get tooltip text 
     */
    ti.cbSize = sizeof(ti);
    ti.uFlags = TTF_SUBCLASS;
    ti.hwnd = handle_;
    ti.uId = 1;
    ti.hinst = CTadsApp::get_app()->get_instance();
    ti.lpszText = LPSTR_TEXTCALLBACK;
    SetRect(&ti.rect, 0, 0, 32767, 32767);
    SendMessage(tooltip_, TTM_ADDTOOL, 0, (LPARAM)&ti);
}

/*
 *   process window deletion 
 */
void CHtmlSysWin_win32::do_destroy()
{
    /* if we have a status line, unregister for notifications */
    if (statusline_ != 0)
        statusline_->unregister_source(this);

    /* remove our miscellaneous background timer if we have one */
    if (bg_timer_id_ != 0)
    {
        KillTimer(handle_, bg_timer_id_);
        free_timer_id(bg_timer_id_);
    }
    
    /* inherit default */
    CTadsWinScroll::do_destroy();
}

/*
 *   erase the background 
 */
int CHtmlSysWin_win32::do_erase_bkg(HDC)
{
    /* 
     *   We'll erase the background in the main paint routine - do nothing
     *   here.  We can do the erasing more efficiently in the main paint
     *   routine, since we can limit the drawing to the clipping area.  
     */
    return TRUE;
}

/*
 *   paint the window contents
 */
void CHtmlSysWin_win32::do_paint_content(HDC dc, const RECT *rc)
{
    CHtmlRect area;
    RECT more_rc;
    int clip_lines;
    int bg_done = FALSE;

    /* adjust the paint area to local coordinates */
    area.set(rc->left, rc->top, rc->right, rc->bottom);

    /* 
     *   First, draw the background.  If I have an image, draw it;
     *   otherwise, use the background color 
     */
    if (bg_image_ != 0)
    {
        CHtmlSysImage *img;

        /* figure out where the image goes, taking scrolling into account */
        img = bg_image_->get_image();
        if (img != 0)
        {
            unsigned long ht, wid;

            /* get the height and width of the image */
            ht = img->get_height();
            wid = img->get_width();
            if (ht != 0 && wid != 0)
            {
                CHtmlRect docrc;
                long x1, y1;
                long x, y;
                
                /* compute where we're drawing, in document coordinates */
                docrc = screen_to_doc(area);

                /* 
                 *   compute, in document coordinates, the location of the
                 *   tile that overlaps the rectangle, assuming that the
                 *   image is tiled from the upper left of the document 
                 */
                x1 = (docrc.left / wid) * wid;
                y1 = (docrc.top / ht) * ht;

                /* tile the image until we've filled up the drawing area */
                for (y = y1 ; y < docrc.bottom ; y += ht)
                {
                    /* tile across the current row */
                    for (x = x1 ; x < docrc.right ; x += wid)
                    {
                        CHtmlRect pos;
                        
                        /* draw this copy of the image */
                        pos.set(x, y, x + wid, y + ht);
                        img->draw_image(this, &pos);
                    }
                }
                
                /* we've finished with the background */
                bg_done = TRUE;
            }
        }
    }

    /* if we didn't draw an image, use the background color */
    if (!bg_done)
    {
        /* fill the redraw area with the background brush */
        FillRect(dc, rc, bgbrush_);
    }

    /* assume we'll allow the last line to show */
    clip_lines = FALSE;

    /*
     *   if we're in "MORE" mode, clip the drawing area to the line
     *   before the prompt 
     */
    if (more_mode_)
    {
        /* get the prompt area */
        get_moreprompt_rect(&more_rc);

        /*
         *   if the drawing area intersects the prompt area, clip out the
         *   prompt area and tell the formatter not to draw lines that are
         *   partially clipped 
         */
        if (area.bottom >= more_rc.top)
        {
            area.bottom = more_rc.top - 1;
            clip_lines = TRUE;
        }
    }

    /* offset by scrolling position */
    area = screen_to_doc(area);

    /* draw everything in the client area */
    if (formatter_ != 0)
        formatter_->draw(&area, clip_lines, &clip_ypos_);

    /* if we're in "MORE" mode, draw a prompt at the bottom */
    if (more_mode_ && !prefs_->get_alt_more_style())
    {
        COLORREF oldclr;
        int oldbkmode;

        /* write the prompt in white text in the default font */
        select_font(dc, get_default_font());
        oldclr = SetTextColor(dc, GetSysColor(COLOR_HIGHLIGHTTEXT));
        oldbkmode = SetBkMode(dc, TRANSPARENT);

        /* write the MORE prompt over a dark gray background */
        FillRect(dc, &more_rc, GetSysColorBrush(COLOR_HIGHLIGHT));
        ExtTextOut(dc, more_rc.left, more_rc.top, 0, 0,
                   moreprompt, sizeof(moreprompt)-1, 0);

        /* restore the old text color and mode */
        SetBkMode(dc, oldbkmode);
        SetTextColor(dc, oldclr);
    }

    /*
     *   If my 'border' flag is set, or I'm a banner window and game
     *   colors have been overridden in the preferences, draw a border at
     *   the bottom of the window to visually separate it from the
     *   adjacent window below. 
     */
    if (has_border_
        || (is_banner_win_ && prefs_->get_override_colors()))
    {
        RECT rc;
        HPEN pen;
        HPEN oldpen;

        /* create a pen for the border */
        pen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DSHADOW));
        oldpen = SelectObject(dc, pen);

        /* get the window area */
        GetClientRect(handle_, &rc);

        /* figure out where it goes based on alignment */
        switch(get_banner_alignment())
        {
        case HTML_BANNERWIN_POS_TOP:
            /* draw a line across the bottom of the window */
            MoveToEx(dc, rc.left, rc.bottom - 1, 0);
            LineTo(dc, rc.right, rc.bottom - 1);
            break;

        case HTML_BANNERWIN_POS_BOTTOM:
            /* draw a line across the top of the window */
            MoveToEx(dc, rc.left, rc.top, 0);
            LineTo(dc, rc.right, rc.top);
            break;

        case HTML_BANNERWIN_POS_LEFT:
            /* draw a line down the right edge of the window */
            MoveToEx(dc, rc.right - 1, rc.top, 0);
            LineTo(dc, rc.right - 1, rc.bottom);
            break;

        case HTML_BANNERWIN_POS_RIGHT:
            /* draw a line down the left edge of the window */
            MoveToEx(dc, rc.left, rc.top, 0);
            LineTo(dc, rc.left, rc.bottom);
            break;
        }

        /* restore the original pen */
        SelectObject(dc, oldpen);
    }
}


/*
 *   Close the window 
 */
int CHtmlSysWin_win32::close_window(int force)
{
    /* ask the owner to close */
    return owner_->close_owner_window(force);
}

/*
 *   get the width of the display area 
 */
long CHtmlSysWin_win32::get_disp_width()
{
    return disp_width_;
}

/*
 *   get the height of the display area 
 */
long CHtmlSysWin_win32::get_disp_height()
{
    return disp_height_;
}

/*
 *   get number of pixels per inch 
 */
long CHtmlSysWin_win32::get_pix_per_inch()
{
    HDC deskdc;
    long ret;

    /* get the system horizontal pixels per inch for the desktop window */
    deskdc = GetDC(GetDesktopWindow());
    ret = GetDeviceCaps(deskdc, LOGPIXELSX);
    ReleaseDC(GetDesktopWindow(), deskdc);

    /* return the value we got from the system */
    return ret;
}

/*
 *   measure the size of a string of text in a given font
 */
CHtmlPoint CHtmlSysWin_win32::measure_text(CHtmlSysFont *font,
                                           const textchar_t *txt, size_t len,
                                           int *ascent)
{
    SIZE txtsiz;
    HDC dc;
    TEXTMETRIC tm;

    /* get my device context */
    dc = GetDC(handle_);

    /* select the font */
    select_font(dc, font);

    /* get the text width */
    GetTextExtentPoint32(dc, txt, len, &txtsiz);

    /* get the height from the text metrics */
    GetTextMetrics(dc, &tm);
    txtsiz.cy = tm.tmHeight;

    /* return the ascender height if the caller wants it */
    if (ascent != 0)
        *ascent = tm.tmAscent;

    /* done with the DC */
    ReleaseDC(handle_, dc);

    /* return the value, reinterpreted into our own Point type */
    return CHtmlPoint(txtsiz.cx, txtsiz.cy);
}

/*
 *   Get the maximum number of characters that fit in a given width 
 */
size_t CHtmlSysWin_win32::get_max_chars_in_width(
    class CHtmlSysFont *font, const textchar_t *str, size_t len, long wid)
{
    HDC dc;
    INT fit_cnt;
    SIZE siz;

    /* get the device context */
    dc = GetDC(handle_);
    
    /* select the font */
    select_font(dc, font);

    /* measure the text */
    fit_cnt = 0;
    GetTextExtentExPoint(dc, str, len, wid, &fit_cnt, 0, &siz);

    /* done with the DC */
    ReleaseDC(handle_, dc);

    /* return the number that fit */
    return (size_t)fit_cnt;
}

/*
 *   select a font 
 */
HGDIOBJ CHtmlSysWin_win32::select_font(HDC dc, class CHtmlSysFont *font)
{
    return ((CHtmlSysFont_win32 *)font)->select(dc);
}

/*
 *   restore old font 
 */
void CHtmlSysWin_win32::unselect_font(HDC dc, class CHtmlSysFont *font,
                                      HGDIOBJ oldfont)
{
    ((CHtmlSysFont_win32 *)font)->unselect(dc, oldfont);
}


/*
 *   draw text 
 */
void CHtmlSysWin_win32::draw_text(int hilite, long x, long y,
                                  CHtmlSysFont *font,
                                  const textchar_t *str, size_t len)
{
    COLORREF oldbkcolor;
    COLORREF oldtxtcolor;
    int oldbkmode;
    
    /* select the font */
    select_font(hdc_, font);

    /* adjust for scrolling */
    x = doc_to_screen_x(x);
    y = doc_to_screen_y(y);

    /* if highlighting, turn on text highlight colors */
    if (hilite)
    {
        /* set highlight text color and background color, and opaque mode */
        oldtxtcolor = SetTextColor(hdc_, GetSysColor(COLOR_HIGHLIGHTTEXT));
        oldbkcolor = SetBkColor(hdc_, GetSysColor(COLOR_HIGHLIGHT));
        oldbkmode = SetBkMode(hdc_, OPAQUE);
    }
    else
    {
        /* set transparent mode */
        oldbkmode = SetBkMode(hdc_, TRANSPARENT);

        /*
         *   if the font specifies a color, use the font color, otherwise
         *   use the default text color 
         */
        if (font->use_font_color())
            oldtxtcolor =
                SetTextColor(hdc_,
                             HTML_color_to_COLORREF(font->get_font_color()));
        else
            oldtxtcolor = SetTextColor(hdc_, text_color_);
    }

    /* draw the text */
    ExtTextOut(hdc_, x, y, 0, 0, str, len, 0);

    /* restore original background color scheme if we changed it */
    if (hilite)
        SetBkColor(hdc_, oldbkcolor);

    /* restore original text color */
    SetTextColor(hdc_, oldtxtcolor);

    /* restore original background mode */
    SetBkMode(hdc_, oldbkmode);
}

/*
 *   draw a bullet 
 */
void CHtmlSysWin_win32::draw_bullet(int hilite, long x, long y,
                                    CHtmlSysFont *font,
                                    HTML_SysWin_Bullet_t style)
{
    textchar_t ch;

    /* get the appropriate bullet character */
    switch(style)
    {
    case HTML_SYSWIN_BULLET_SQUARE:
        ch = bullet_square_;
        break;
        
    case HTML_SYSWIN_BULLET_CIRCLE:
        ch = bullet_circle_;
        break;

    default:
    case HTML_SYSWIN_BULLET_DISC:
        ch = bullet_disc_;
        break;

    case HTML_SYSWIN_BULLET_PLAIN:
        /* nothing to draw */
        return;
    }

    /* draw the bullet */
    draw_text(hilite, x, y, font, &ch, 1);
}

/*
 *   Get the font to use to draw bullets.  If wingdings is present, we'll
 *   use wingdings at the current font size; otherwise, we'll just use the
 *   current font. 
 */
CHtmlSysFont *CHtmlSysWin_win32::get_bullet_font(CHtmlSysFont *font)
{
    CHtmlFontDesc desc;
    static const textchar_t symfont[] = "Wingdings";
    
    /* if wingdings isn't available, use the current font */
    if (!wingdings_avail_)
        return font;

    /* get Wingdings in the same size as the current font */
    font->get_font_desc(&desc);
    memcpy(desc.face, symfont, sizeof(symfont));
    desc.default_charset = FALSE;
    desc.charset = BULLET_CHARSET;
    return get_font(&desc);
}

/*
 *   Get the DirectSound interface 
 */
IDirectSound *CHtmlSysWin_win32::get_directsound()
{
    return owner_->get_directsound();
}

/*
 *   Register a timer function 
 */
void CHtmlSysWin_win32::register_timer_func(void (*timer_func)(void *),
                                            void *func_ctx)
{
    htmlw32_timer_t *rec;
    int need_sys_timer = FALSE;

    /* if we don't have a system timer, note that we need to start one */
    if (timer_list_head_ == 0)
        need_sys_timer = TRUE;
    
    /* allocate a new link record */
    rec = new htmlw32_timer_t(timer_func, func_ctx);

    /* add the link record to our list */
    rec->nxt_ = timer_list_head_;
    timer_list_head_ = rec;

    /* if necessary, start the system timer */
    if (need_sys_timer)
        start_sys_timer();
}

/*
 *   Start the system timer for functions registered via
 *   register_time_func().  Since we can't start a timer until we have a
 *   window handle, we may need to start a timer when we create our
 *   window. 
 */
void CHtmlSysWin_win32::start_sys_timer()
{
    /* 
     *   if we don't have a window yet, we can't start the timer yet --
     *   we'll have to do this later when we actually create our window
     *   handle 
     */
    if (handle_ == 0)
        return;

    /* if there are no registered timer callbacks, ignore the request */
    if (timer_list_head_ == 0)
        return;
    
    /* allocate a system timer ID - if we can't, abort registration */
    cb_timer_id_ = alloc_timer_id();
    if (cb_timer_id_ == 0)
        return;
    
    /* start a system timer to go off once per second */
    if (SetTimer(handle_, cb_timer_id_, 1000, (TIMERPROC)0) == 0)
    {
        /* that failed - free our timer ID and give up */
        free_timer_id(cb_timer_id_);
        cb_timer_id_ = 0;
    }
}

/*
 *   Unregister a timer function 
 */
void CHtmlSysWin_win32::unregister_timer_func(void (*timer_func)(void *),
                                              void *func_ctx)
{
    htmlw32_timer_t *cur;
    htmlw32_timer_t *prv;
    
    /* find the given timer in our list of timer functions */
    for (prv = 0, cur = timer_list_head_ ; cur != 0 ;
         prv = cur, cur = cur->nxt_)
    {
        /* if this is the one, remove it */
        if (cur->func_ == timer_func && cur->ctx_ == func_ctx)
        {
            /* unlink the item from our list */
            if (prv == 0)
                timer_list_head_ = cur->nxt_;
            else
                prv->nxt_ = cur->nxt_;

            /* delete the record */
            delete cur;

            /* if the list is empty, kill the system timer */
            if (timer_list_head_ == 0)
            {
                /* remove the system timer */
                KillTimer(handle_, cb_timer_id_);

                /* release the timer identifier */
                free_timer_id(cb_timer_id_);
                cb_timer_id_ = 0;
            }

            /* done */
            return;
        }
    }
}

/*
 *   draw a horizontal rule 
 */
void CHtmlSysWin_win32::draw_hrule(const CHtmlRect *pos, int shade)
{
    HBRUSH shadow;
    HBRUSH hilite;
    RECT rc;
    CHtmlRect scrpos;

    /* convert the position to absolute coordinates */
    scrpos = doc_to_screen(*pos);

    /* set up a windows RECT structure for the position */
    SetRect(&rc, scrpos.left, scrpos.top, scrpos.right, scrpos.bottom);

    /* get system brushes for shadow and highlight colors */
    shadow = GetSysColorBrush(COLOR_3DSHADOW);
    hilite = GetSysColorBrush(COLOR_3DHILIGHT);

    /* see if we're shading */
    if (shade)
    {
        RECT lrc;
        
        /* fill the top line with shadow color */
        SetRect(&lrc, rc.left, rc.top, rc.right, rc.top + 1);
        FillRect(hdc_, &lrc, shadow);

        /* draw the left edge with shadow color */
        SetRect(&lrc, rc.left, rc.top, rc.left + 1, rc.bottom - 1);
        FillRect(hdc_, &lrc, shadow);

        /* fill the bottom line with hilite color */
        SetRect(&lrc, rc.left, rc.bottom - 1, rc.right, rc.bottom);
        FillRect(hdc_, &lrc, hilite);

        /* fill the right line with hilite color */
        SetRect(&lrc, rc.right - 1, rc.top, rc.right, rc.bottom);
        FillRect(hdc_, &lrc, hilite);
    }
    else
    {
        /* not shading - fill in the rectangle with the shadow color */
        FillRect(hdc_, &rc, shadow);
    }
}

/*
 *   Draw a table/cell background 
 */
void CHtmlSysWin_win32::draw_table_bkg(const CHtmlRect *pos,
                                       HTML_color_t bgcolor)
{
    HBRUSH hbr;
    LOGBRUSH lbr;
    RECT rc;
    CHtmlRect scrpos;

    /* create a brush with the given color */
    lbr.lbStyle = BS_SOLID;
    lbr.lbColor = HTML_color_to_COLORREF(bgcolor);
    lbr.lbHatch = 0;
    hbr = CreateBrushIndirect(&lbr);

    /* set up a Windows RECT structure in screen coordinates */
    scrpos = doc_to_screen(*pos);
    SetRect(&rc, scrpos.left, scrpos.top, scrpos.right, scrpos.bottom);

    /* fill the area with the given brush */
    FillRect(hdc_, &rc, hbr);

    /* we're done with the brush */
    DeleteObject(hbr);
}

/*
 *   Draw a table/cell border 
 */
void CHtmlSysWin_win32::draw_table_border(const CHtmlRect *pos, int width,
                                          int cell)
{
    HBRUSH shadow;
    HBRUSH hilite;
    RECT rc;
    RECT lrc;
    CHtmlRect scrpos;

    /* convert the position to absolute coordinates */
    scrpos = doc_to_screen(*pos);

    /* set up a windows RECT structure for the position */
    SetRect(&rc, scrpos.left, scrpos.top, scrpos.right, scrpos.bottom);

    /* get system brushes for shadow and highlight colors */
    shadow = GetSysColorBrush(COLOR_3DSHADOW);
    hilite = GetSysColorBrush(COLOR_3DHILIGHT);

    /* draw the top and left lines in highlight color */
    SetRect(&lrc, rc.left, rc.top, rc.right - width, rc.top + width);
    FillRect(hdc_, &lrc, cell ? shadow : hilite);
    SetRect(&lrc, rc.left, rc.top, rc.left + width, rc.bottom - width);
    FillRect(hdc_, &lrc, cell ? shadow : hilite);

    /* draw the right and bottom lines in shadow color */
    SetRect(&lrc, rc.left + width, rc.bottom - width, rc.right, rc.bottom);
    FillRect(hdc_, &lrc, cell ? hilite : shadow);
    SetRect(&lrc, rc.right - width, rc.top + width, rc.right, rc.bottom);
    FillRect(hdc_, &lrc, cell ? hilite : shadow);
}

/*
 *   get the default font for the window 
 */
CHtmlSysFont *CHtmlSysWin_win32::get_default_font()
{
    CHtmlFontDesc desc;

    /* get a font matching the current defaults */
    desc.htmlsize = 3;
    return get_font(&desc);
}

/*
 *   Adjust the vertical scrollbar.  This is called whenever the
 *   formatter adds a new line. 
 */
void CHtmlSysWin_win32::fmt_adjust_vscroll()
{
    long max_y_pos;
    RECT rc;
    int moreht;
    int winht;
    long next_stop;
    long old_content_ht = content_height_;
    int scroll_cmd;
    long scroll_dst;
    int scroll_use_pos;

    /* presume we'll scroll to the bottom of the text */
    scroll_cmd = SB_BOTTOM;
    scroll_dst = 0;
    scroll_use_pos = 0;
    
    /* get the current vertical extent */
    max_y_pos = formatter_->get_max_y_pos();

    /*
     *   if we're currently reading a command, and our content height is
     *   already greater than the formatter's maximum y position, we must
     *   have deleted some text on the command line, causing the command
     *   line to take up fewer lines than it did before; ignore this type
     *   of shrinkage, because it would cause undesirable jumpiness by
     *   scrolling excessively during command line editing 
     */
    if (started_reading_cmd_ && (max_y_pos < content_height_))
        max_y_pos = content_height_;

    /* get the size of the "MORE" prompt */
    get_moreprompt_rect(&rc);
    moreht = rc.bottom - rc.top;

    /* get the vertical size of the window */
    get_scroll_area(&rc, TRUE);
    winht = rc.bottom - rc.top;

    /*
     *   Figure out how much room we have, minus the "MORE" prompt, from
     *   the last input position.  We don't want to show more information
     *   than this at once, to make sure the user has a chance to
     *   acknowledge the information we've displayed so far.
     */
    next_stop = last_input_height_ + winht - moreht;

    /*
     *   If there's room for the added information, extend our screen
     *   height and adjust scrollbars.  There's room only if we can fit
     *   this information onto the screen without scrolling the last input
     *   position off the screen.  If we've started reading a command, do
     *   not go into "MORE" mode, since the user has already seen
     *   everything that we've displayed so far, and must simply be
     *   scrolling back or resizing the window.  
     */
    if (max_y_pos < next_stop || started_reading_cmd_ || !more_mode_enabled_)
    {
        /* it'll fit - include it in our content height */
        content_height_ = max_y_pos;

        /* we don't need to wait for user input to display more */
        set_more_mode(FALSE);
    }
    else if (prefs_->get_alt_more_style())
    {
        /* scroll to the next available position */
        content_height_ = max_y_pos;
        scroll_cmd = SB_THUMBPOSITION;
        scroll_dst = last_input_height_ / get_vscroll_units();
        scroll_use_pos = TRUE;

        /* we're in more mode */
        set_more_mode(TRUE);
    }
    else
    {
        /* go only up to the next stop */
        content_height_ = next_stop;
        
        /*
         *   there's more information than will fit on the screen --
         *   pause for user input before displaying more 
         */
        set_more_mode(TRUE);
    }

    /*
     *   Invalidate the window from the old content height to the new
     *   content height, since this area has changed 
     */
    get_scroll_area(&rc, TRUE);
    rc.top = doc_to_screen_y(old_content_ht);
    rc.bottom = doc_to_screen_y(content_height_);
    InvalidateRect(handle_, &rc, FALSE);

    /* adjust scrollbars accordingly */
    adjust_scrollbar_ranges();

    /* scroll to the bottom of the available range */
    do_scroll(TRUE, vscroll_, scroll_cmd, scroll_dst, scroll_use_pos);

    /* check to see if we have started the command */
    check_cmd_start();
}

/*
 *   Load the command selection range from the formatter selection range.
 *   This should be called whenever the formatter range is updated
 *   directly, as when the user selects text using the mouse.  
 */
void CHtmlSysWin_win32::fmt_sel_to_cmd_sel(int caret_at_high_end)
{
    unsigned long inp_txt_ofs;
    unsigned long startofs, endofs;

    /* if we don't have a command tag yet, do nothing */
    if (cmdtag_ == 0)
    {
        /* 
         *   there's no command -- remember the selection direction, but
         *   otherwise there's nothing to do 
         */
        caret_at_right_ = caret_at_high_end;
        return;
    }

    /* get the current tag's text offset */
    inp_txt_ofs = cmdtag_->get_text_ofs();

    /* get the formatter selection range */
    formatter_->get_sel_range(&startofs, &endofs);

    /* check to see if the selection overlaps the command */
    if (endofs < inp_txt_ofs)
    {
        /*
         *   no overlap - move the caret to the end of the command, but
         *   hide it 
         */
        cmdbuf_->end_of_line(FALSE);
        cmdbuf_->hide_caret();
    }
    else if (startofs >= inp_txt_ofs)
    {
        /*
         *   The entire selection is in the command.  Select the range
         *   and set the caret to the appropriate end. 
         */
        cmdbuf_->set_sel_range(startofs - inp_txt_ofs, endofs - inp_txt_ofs,
                               caret_at_high_end ? endofs - inp_txt_ofs
                               : startofs - inp_txt_ofs);
    }
    else
    {
        /*
         *   The selection includes the command, but includes other text
         *   as well.  Select up to the end of the selection, and show the
         *   caret only if it's at the high end (which includes the
         *   command). 
         */
        cmdbuf_->set_sel_range(0, endofs - inp_txt_ofs, endofs - inp_txt_ofs);
        if (!caret_at_high_end)
            cmdbuf_->hide_caret();
    }

    /* remember the caret positioning */
    caret_at_right_ = caret_at_high_end;

    /* set the caret's displayed position */
    update_caret_pos(FALSE, FALSE);
}


/*
 *   Scroll a document position into view 
 */
void CHtmlSysWin_win32::scroll_to_doc_coords(const CHtmlRect *pos)
{
    RECT rc;
    CHtmlRect scrpos;

    /* get the current on-screen area */
    get_scroll_area(&rc, TRUE);

    /* get the desired position in screen coordinates */
    scrpos = doc_to_screen(*pos);
    
    /* scroll vertically if necessary */
    if (scrpos.top < rc.top || scrpos.bottom > rc.bottom)
    {
        long target;
        long delta;
        long new_vscroll;

        /* scroll so that the top is in the middle of the window */
        target = (rc.top + (rc.bottom - rc.top)/2);

        /* make sure we display as much of the rectangle as possible */
        if (target + scrpos.bottom - scrpos.top > rc.bottom)
        {
            target = scrpos.bottom - scrpos.top - rc.bottom;
            if (target < rc.top)
                target = rc.top;
        }

        /* figure out where we're going */
        delta = scrpos.top - target;
        new_vscroll = (vscroll_ofs_ * get_vscroll_units() + delta)
                      / get_vscroll_units();
        do_scroll(TRUE, vscroll_, SB_THUMBPOSITION, new_vscroll, TRUE);
    }

    /* do the same calculation horizontally */
    if (scrpos.left < rc.left || scrpos.right > rc.right)
    {
        long target;
        long delta;
        long new_hscroll;

        /* scroll so that the left is in the middle of the window */
        target = (rc.left + (rc.right - rc.left)/2);

        /* make sure we display as much of the rectangle as possible */
        if (target + scrpos.right - scrpos.left > rc.right)
        {
            target = scrpos.right - scrpos.left - rc.right;
            if (target < rc.left)
                target = rc.left;
        }

        /* figure out where we're going */
        delta = scrpos.left - target;
        new_hscroll = (hscroll_ofs_ * get_hscroll_units() + delta)
                      / get_hscroll_units();
        do_scroll(FALSE, hscroll_, SB_THUMBPOSITION, new_hscroll, TRUE);
    }
}

/*
 *   invalidate an area given in document coordinates 
 */
void CHtmlSysWin_win32::inval_doc_coords(const CHtmlRect *doc_area)
{
    CHtmlRect screen_area;
    RECT rc;

    /* adjust to window coordinates */
    screen_area = doc_to_screen(*doc_area);

    /* check for max values and adjust accordingly */
    GetClientRect(handle_, &rc);
    if (doc_area->right == HTMLSYSWIN_MAX_RIGHT)
        screen_area.right = rc.right;
    if (doc_area->bottom == HTMLSYSWIN_MAX_BOTTOM)
        screen_area.bottom = rc.bottom;

    /* set the system rectangle */
    SetRect(&rc, screen_area.left, screen_area.top,
            screen_area.right, screen_area.bottom);

    /* invalidate the area on the screen */
    InvalidateRect(handle_, &rc, FALSE);
}

/*
 *   Receive notification that the formatter is clearing the display list.
 *   We must forget any display list items that we're referencing (such as
 *   the link the mouse is hovering over). 
 */
void CHtmlSysWin_win32::advise_clearing_disp_list()
{
    /* forget any link we're tracking */
    track_link_ = 0;

    /* 
     *   Forget any link we're hovering over.  Note that it shouldn't be
     *   necessary to update the window at this point -- the formatter is
     *   about to rebuild the display list, so the formatter should take
     *   care of any updating that needs to be done. 
     */
    status_link_ = 0;
}


/*
 *   Remove all banners 
 */
void CHtmlSysWin_win32::remove_all_banners()
{
    /* tell the formatter to remove the banners */
    formatter_->remove_all_banners();
}

/*
 *   Do a complete reformatting of the window 
 */
void CHtmlSysWin_win32::do_reformat(int show_status, int freeze_display,
                                    int reset_sounds)
{
    /* forget any tracking links */
    status_link_ = track_link_ = 0;

    /* restart the formatter */
    formatter_->start_at_top(reset_sounds);

    /* make sure we redraw everything */
    InvalidateRect(handle_, 0, FALSE);

    /* format the window contents */
    do_formatting(show_status, TRUE, freeze_display);

    /* adjust everything for the reformatting */
    adjust_for_reformat();
}

/*
 *   Schedule reformatting 
 */
void CHtmlSysWin_win32::schedule_reformat(int show_status, int freeze_display,
                                          int reset_sounds)
{
    MSG msg;
    WPARAM wpar;
    
    /* 
     *   if we already have a reformat message pending, ignore this
     *   request
     */
    if (PeekMessage(&msg, handle_, HTMLM_REFORMAT, HTMLM_REFORMAT,
                    PM_NOREMOVE))
        return;

    /* build the WPARAM from the flags */
    wpar = 0;
    if (show_status) wpar |= HTML_F_SHOW_STAT;
    if (freeze_display) wpar |= HTML_F_FREEZE_DISP;
    if (reset_sounds) wpar |= HTML_F_RESET_SND;

    /* post a REFORMAT message to myself */
    PostMessage(handle_, HTMLM_REFORMAT, wpar, 0);
}


/*
 *   Run the formatter
 */
int CHtmlSysWin_win32::do_formatting(int show_status, int update_win,
                                     int freeze_display)
{
    RECT rc;
    int drawn;
    unsigned long win_bottom;
    HCURSOR old_cursor;

    /* if desired, show "wait" cursor and status message while working */
    if (show_status)
    {
        old_cursor = SetCursor(wait_cursor_);
        formatting_msg_ = TRUE;
        if (statusline_ != 0)
            statusline_->source_to_front(this);
    }

    /* if desired, freeze display updating while we're working */
    if (freeze_display)
        formatter_->freeze_display(TRUE);

    /* 
     *   redraw the window before we start, so that something happens
     *   immediately and the window doesn't look frozen 
     */
    if (update_win)
        UpdateWindow(handle_);

    /* 
     *   Get the window area in document coordinates, so we'll know when
     *   we've formatted past the bottom of the current display area.
     *   Note that we'll ignore the presence or absence of a horizontal
     *   scrollbar, since it could come or go while we're formatting; by
     *   assuming that it won't be there, we'll be maximally conservative
     *   about redrawing the whole area.  
     */
    GetClientRect(handle_, &rc);
    win_bottom = screen_to_doc_y(rc.bottom);
    
    /* we don't have enough formatting done yet to draw the window */
    drawn = FALSE;

    /* reformat everything, drawing as soon as possible, if desired */
    if (update_win)
    {
        while (formatter_->more_to_do() && !drawn)
        {
            /* format another line */
            formatter_->do_formatting();
            
            /* 
             *   if we have enough content to do so, redraw the window;
             *   we're not really done with the formatting yet, but at
             *   least we'll update the window as soon as we can, so the
             *   user isn't staring at a blank window longer than
             *   necessary 
             */
            if (formatter_->get_max_y_pos() >= win_bottom)
            {
                /* 
                 *   if the formatter's updating is frozen, invalidate the
                 *   window -- the formatter won't have been invalidating
                 *   it as it goes 
                 */
                if (freeze_display)
                    InvalidateRect(handle_, 0, FALSE);
                
                /* redraw the window immediately */
                UpdateWindow(handle_);
                
                /* we've now drawn it */
                drawn = TRUE;
            }
        }
    }
    
    /* after drawing, keep going until we've formatted everything */
    while (formatter_->more_to_do())
        formatter_->do_formatting();

    /* restore the original cursor and status line, if we updated it */
    if (show_status)
    {
        SetCursor(old_cursor);
        formatting_msg_ = FALSE;
        if (statusline_ != 0)
            statusline_->update();
    }

    /* unfreeze the display if we froze it */
    if (freeze_display)
        formatter_->freeze_display(FALSE);

    /* 
     *   if we didn't do any drawing, and we updated the window coming in,
     *   invalidate the window - the initial redraw will have validated
     *   everything, but we've just made it all invalid again and didn't
     *   get around to drawing it 
     */
    if (update_win && !drawn)
        InvalidateRect(handle_, 0, FALSE);

    /* return an indication of whether we did any updating */
    return drawn;
}

/*
 *   Recalculate the palette.  The formatter calls this routine whenever
 *   it adds an image to the initial images list.  We'll post ourselves a
 *   palette recalculation message, so that when we're all done with the
 *   current set of changes we'll update the palette, accounting for all
 *   images that are in the list at that point.
 *   
 *   Note that we'll post the message to our owner frame, so that the
 *   message is processed at the application frame level rather than
 *   within an individual subwindow.  
 */
void CHtmlSysWin_win32::recalc_palette()
{
    MSG msg;
    HWND handle;

    /* get our owner frame handle */
    handle = owner_->get_owner_frame_handle();

    /* post a palette update message if there isn't one pending */
    if (!PeekMessage(&msg, handle, WM_QUERYNEWPALETTE, WM_QUERYNEWPALETTE,
                     PM_NOREMOVE))
        PostMessage(handle, WM_QUERYNEWPALETTE, 0, 0);
}

/*
 *   recalculate the palette for the foreground object 
 */
int CHtmlSysWin_win32::recalc_palette_fg()
{
    CHtmlSysImage *image;

    /* 
     *   if we're not in a palette color mode, or we have no initial
     *   images in the formatter's list, there's nothing to do 
     */
    if (get_bits_per_pixel() > 8
        || formatter_->get_initial_image_count() == 0)
        return FALSE;

    /* get the foreground image */
    image = formatter_->get_initial_image(0)->get_image();

    /* map its palette into the foreground */
    return image->map_palette(this, TRUE);
}

/*
 *   recalculate the palette for the background objects 
 */
int CHtmlSysWin_win32::recalc_palette_bg()
{
    int changed = FALSE;
    int i;

    /* if we're not in a palette color mode, there's nothing to do */
    if (get_bits_per_pixel() > 8)
        return FALSE;

    /* go through all of the images in the list after the first one */
    for (i = 1 ; i < formatter_->get_initial_image_count() ; ++i)
    {
        CHtmlSysImage *image;
        
        /* get the current image from the initial image list */
        image = formatter_->get_initial_image(i)->get_image();

        /* map its palette in the background */
        changed |= image->map_palette(this, FALSE);
    }

    /* return indication of change in the palette */
    return changed;
}


/*
 *   Set the window's title 
 */
void CHtmlSysWin_win32::set_window_title(const textchar_t *title)
{
    /* set the caption of my parent, which is the main window */
    SetWindowText(parent_->get_handle(), title);
}

/*
 *   Set the background image 
 */
void CHtmlSysWin_win32::set_html_bg_image(CHtmlResCacheObject *image)
{
    /* if we're changing anything, schedule a redraw */
    if (image != bg_image_)
        InvalidateRect(handle_, 0, FALSE);

    /* if we currently have a background image, forget it */
    if (bg_image_ != 0)
        bg_image_->remove_ref();

    /* remember the new image */
    bg_image_ = image;

    /* add a reference to the new image, if there is a new image */
    if (image != 0)
        image->add_ref();
}

/*
 *   Set the default background color 
 */
void CHtmlSysWin_win32::set_html_bg_color(HTML_color_t color, int use_default)
{
    COLORREF rgb;

    if (use_default || prefs_->get_override_colors())
    {
        /* ignore the provided color and use the default setting */
        if (prefs_->get_use_win_colors())
            rgb = GetSysColor(COLOR_WINDOW);
        else
            rgb = HTML_color_to_COLORREF(prefs_->get_bg_color());
    }
    else
    {
        /* use the color they provided */
        rgb = HTML_color_to_COLORREF(color);
    }

    /* set the color */
    internal_set_bg_color(rgb);
}

/*
 *   Internal routine to set the background color 
 */
void CHtmlSysWin_win32::internal_set_bg_color(COLORREF rgb)
{
    LOGBRUSH lbr;

    /* if the color isn't changing, there's nothing more to do */
    if (bgbrush_ != 0 && rgb == bgcolor_)
        return;

    /* remember the new color */
    bgcolor_ = rgb;

    /* discard the old background brush */
    if (bgbrush_)
        DeleteObject(bgbrush_);

    /* create a new background brush */
    lbr.lbStyle = BS_SOLID;
    lbr.lbColor = rgb;
    lbr.lbHatch = 0;
    bgbrush_ = CreateBrushIndirect(&lbr);

    /* redraw the whole window so we draw in the new color */
    if (handle_ != 0)
        InvalidateRect(handle_, 0, FALSE);
}

/*
 *   Set the default text color 
 */
void CHtmlSysWin_win32::set_html_text_color(HTML_color_t color,
                                            int use_default)
{
    COLORREF new_color;
    
    if (use_default || prefs_->get_override_colors())
    {
        /* ignore the provided color and use the default setting */
        if (prefs_->get_use_win_colors())
            new_color = GetSysColor(COLOR_WINDOWTEXT);
        else
            new_color = HTML_color_to_COLORREF(prefs_->get_text_color());
    }
    else
    {
        /* use the color they provided */
        new_color = HTML_color_to_COLORREF(color);
    }

    /* set the color */
    internal_set_text_color(new_color);
}

/*
 *   Internal routine to set the text color 
 */
void CHtmlSysWin_win32::internal_set_text_color(COLORREF new_color)
{
    /* if we're not changing the color, there's nothing to do */
    if (new_color == text_color_)
        return;

    /* set the new color */
    text_color_ = new_color;

    /* invalidate the whole window so we draw in the new color */
    if (handle_ != 0)
        InvalidateRect(handle_, 0, FALSE);
}

/*
 *   Set the link colors 
 */
void CHtmlSysWin_win32::set_html_link_colors(HTML_color_t link_color,
                                             int link_use_default,
                                             HTML_color_t vlink_color,
                                             int vlink_use_default,
                                             HTML_color_t alink_color,
                                             int alink_use_default)
{
    COLORREF new_color;
    int changed;

    /* presume nothing will change */
    changed = FALSE;

    /* set the new colors */
    if (link_use_default)
    {
        /* ignore the provided color and use the default setting */
        new_color = HTML_color_to_COLORREF(prefs_->get_link_color());
    }
    else
    {
        new_color = HTML_color_to_COLORREF(link_color);
    }
    if (new_color != link_color_)
    {
        changed = TRUE;
        link_color_ = new_color;
    }

    if (vlink_use_default)
    {
        /* ignore the provided color and use the default setting */
        new_color = HTML_color_to_COLORREF(prefs_->get_vlink_color());
    }
    else
    {
        new_color = HTML_color_to_COLORREF(vlink_color);
    }
    if (new_color != vlink_color_)
    {
        changed = TRUE;
        vlink_color_ = new_color;
    }

    if (alink_use_default)
    {
        /* ignore the provided color and use the default setting */
        new_color = HTML_color_to_COLORREF(prefs_->get_alink_color());
    }
    else
    {
        new_color = HTML_color_to_COLORREF(alink_color);
    }
    if (new_color != alink_color_)
    {
        changed = TRUE;
        alink_color_ = new_color;
    }

    new_color = HTML_color_to_COLORREF(prefs_->get_hlink_color());
    if (hlink_color_ != new_color)
    {
        changed = TRUE;
        hlink_color_ = new_color;
    }

    /* update the window if needed */
    if (handle_ != 0 && changed)
        InvalidateRect(handle_, 0, FALSE);
}

/*
 *   Map a parameterized system color 
 */
HTML_color_t CHtmlSysWin_win32::map_system_color(HTML_color_t color)
{
    switch(color)
    {
    case HTML_COLOR_STATUSBG:
        return prefs_->get_color_status_bg();
        
    case HTML_COLOR_STATUSTEXT:
        return prefs_->get_color_status_text();

    default:
        /* return black for other colors */
        return 0x00000000;
    }
}

/*
 *   determine if links should be underlined 
 */
int CHtmlSysWin_win32::get_html_link_underline() const
{
    return prefs_->get_underline_links();
}

/*
 *   Set the temporary link visibility 
 */
void CHtmlSysWin_win32::set_temp_show_links(int flag)
{
    /* 
     *   if we're cancelling visibility, and we have a timer that hasn't
     *   elapsed yet, kill the timer, since we don't need to receive the
     *   notification after all 
     */
    if (!flag && temp_link_timer_id_ != 0)
    {
        KillTimer(handle_, temp_link_timer_id_);
        free_timer_id(temp_link_timer_id_);
        temp_link_timer_id_ = 0;
    }
    
    /* update the window if the visibility has changed */
    if ((flag != 0) != (temp_show_links_ != 0))
    {
        /* set the new visibility */
        temp_show_links_ = flag;

        /* redraw links, since they'll look different now */
        inval_links_on_screen();
    }
}

/*
 *   invalidate all visible links - this should be called whenever link
 *   visibility changes 
 */
void CHtmlSysWin_win32::inval_links_on_screen()
{
    RECT rc;
    CHtmlRect area;

    /* get the visible area of the window in document coordinates */
    get_scroll_area(&rc, TRUE);
    area = screen_to_doc(CHtmlRect(rc.left, rc.top, rc.right, rc.bottom));

    /* invalidate all links within the screen area */
    formatter_->inval_links_on_screen(&area);
}

/*
 *   determine if links are to be shown as links (if not, they're shown as
 *   ordinary text) 
 */
int CHtmlSysWin_win32::get_html_show_links() const
{
    return prefs_->get_links_on() || temp_show_links_;
}

/*
 *   determine if we're allowed to show graphics 
 */
int CHtmlSysWin_win32::get_html_show_graphics() const
{
    return prefs_->get_graphics_on();
}

/*
 *   Create a subwindow for a displaying banner 
 */
CHtmlSysWin *CHtmlSysWin_win32::
   create_banner_subwin(CHtmlFormatter *formatter, int has_border,
                        HTML_BannerWin_Pos_t pos)
{
    /* ask our owner to set up a banner subwindow */
    return owner_->create_banner_subwin(formatter, has_border, pos);
}

/*
 *   Create a subwindow for the "about" box 
 */
CHtmlSysWin *CHtmlSysWin_win32::
   create_aboutbox_subwin(CHtmlFormatter *formatter)
{
    /* ask our owner to set up a banner subwindow */
    return owner_->create_aboutbox_subwin(formatter);
}


/*
 *   Remove a banner subwindow 
 */
void CHtmlSysWin_win32::remove_banner_subwin(CHtmlSysWin *subwin)
{
    /* ask our owner to remove the banner */
    owner_->remove_banner_subwin(subwin);
}

/*
 *   Set the size of this banner window 
 */
void CHtmlSysWin_win32::set_banner_size(long width, int use_width,
                                        long height, int use_height)
{
    owner_->set_banner_size(this, width, use_width, height, use_height);
}

/*
 *   Set the border and alignment settings of this banner window 
 */
void CHtmlSysWin_win32::set_banner_info(int has_border,
                                        HTML_BannerWin_Pos_t pos)
{
    /* set the border flag */
    has_border_ = has_border;

    /* set the alignemnt */
    banner_alignment_ = pos;
}


/* ------------------------------------------------------------------------ */
/*
 *   HTML Input window implementation 
 */

CHtmlSysWin_win32_Input::CHtmlSysWin_win32_Input(
    CHtmlFormatterInput *formatter, CHtmlSysWin_win32_owner *owner,
    int has_vscroll, int has_hscroll, CHtmlPreferences *prefs)
    : CHtmlSysWin_win32(formatter, owner, has_vscroll, has_hscroll,
                        prefs)
{
    /* enable MORE mode in an input window */
    more_mode_enabled_ = TRUE;

    /* load accelerator tables */
    accel_emacs_ = LoadAccelerators(CTadsApp::get_app()->get_instance(),
                                    MAKEINTRESOURCE(IDR_ACCEL_EMACS));
    accel_win_ = LoadAccelerators(CTadsApp::get_app()->get_instance(),
                                  MAKEINTRESOURCE(IDR_ACCEL_WIN));

    /* set the correct accelerator for the current preferences */
    set_current_accel();

    /* use the caret in this window */
    caret_enabled_ = TRUE;

    /* we're not yet in single-key mode */
    waiting_for_key_ = FALSE;
    last_keystroke_ = 0;
    last_keystroke_cmd_ = 0;

    /* not waiting for an event */
    waiting_for_evt_ = FALSE;
    last_evt_info_ = 0;

    /* no input timeout timer yet */
    input_timer_id_ = 0;
}

/*
 *   notification of system window creation 
 */
void CHtmlSysWin_win32_Input::do_create()
{
    /* inherit default handling */
    CHtmlSysWin_win32::do_create();

    /* 
     *   Set up to receive WM_HOTKEY messages for Ctrl+Break.  This
     *   ensures that Ctrl+Break will go to the front of the message
     *   queue, so if we enter an infinite loop, and the user types a few
     *   keys then realizes that the machine's stuck, and only then hits
     *   Ctrl+Break, we'll still see the Ctrl+Break, because it won't be
     *   stuck behind all the other keystrokes in the queue. 
     */
    RegisterHotKey(handle_, HTML_HK_BREAK, MOD_CONTROL, VK_CANCEL);
}

/* 
 *   notification that the window is being destroyed 
 */
void CHtmlSysWin_win32_Input::do_destroy()
{
    /* remove our hotkey registration */
    UnregisterHotKey(handle_, HTML_HK_BREAK);
    
    /* inherit default handling */
    CHtmlSysWin_win32::do_destroy();
}

/*
 *   handle timer events 
 */
int CHtmlSysWin_win32_Input::do_timer(int timer_id)
{
    if (timer_id == input_timer_id_)
    {
        /*
         *   it's the input timer - if we're waiting for input, set a flag
         *   indicating that we're finished with that, and set the flag
         *   indicating that the input timed out 
         */
        waiting_for_evt_timeout_ = TRUE;
        if (cmdtag_ != 0)
            command_read_ = TRUE;
        if (waiting_for_key_)
            got_keystroke_ = TRUE;
        if (waiting_for_evt_)
            got_event_ = TRUE;

        /* handled */
        return TRUE;
    }
    else
    {
        /* not one of ours - inherit default handling */
        return CHtmlSysWin_win32::do_timer(timer_id);
    }
}


/*
 *   Set the correct accelerator table for the current preferences 
 */
void CHtmlSysWin_win32_Input::set_current_accel()
{
    textchar_t paste_key;
    
    /* check the Ctrl+V preferences */
    if (prefs_->get_emacs_ctrl_v())
    {
        /* emacs-style Ctrl+V */
        current_accel_ = accel_emacs_;
        paste_key = 'Y';
    }
    else
    {
        /* Windows-style Ctrl+V */
        current_accel_ = accel_win_;
        paste_key = 'V';
    }

    /* set the Edit/Paste menu item text for the change */
    owner_->set_paste_accel(paste_key);
}

/*
 *   Set the caret size for a given font 
 */
void CHtmlSysWin_win32::set_caret_size(CHtmlSysFont *font)
{
    CHtmlFontMetrics fm;

    /* hide the caret */
    hide_caret();

    /* get the font metrics */
    font->get_font_metrics(&fm);

    /* set the total size of the caret to the total size of the font */
    caret_ht_ = fm.total_height;

    /* set the visible portion of the caret to the ascender height */
    caret_ascent_ = fm.ascender_height;

    /* show the caret */
    show_caret();
}

/*
 *   Update the caret position.  This should be used when starting a new
 *   command, scrolling, or otherwise altering the caret position.  
 */
void CHtmlSysWin_win32_Input::update_caret_pos(int bring_on_screen,
                                               int update_sel_range)
{
    unsigned long inp_txt_ofs;

    /* 
     *   if we're bringing the cursor on the screen, make sure we're on
     *   the active page 
     */
    if (bring_on_screen)
        owner_->go_to_active_page();

    /*
     *   if we don't have a command tag yet, or the caret is set invisible
     *   in the input buffer, or if we're not showing the active command
     *   input page, put the caret off-screen 
     */
    if (cmdtag_ == 0
        || (!cmdbuf_->is_caret_visible() && !update_sel_range)
        || !owner_->is_active_page())
    {
        SetCaretPos(-100, -100);
        return;
    }

    /* get the current tag's text offset */
    inp_txt_ofs = cmdtag_->get_text_ofs();

    /* set the caret to the end of the formatter's last object */
    caret_pos_ = formatter_->get_text_pos(inp_txt_ofs + cmdbuf_->get_caret());

    /* convert to screen coordinates */
    caret_pos_ = doc_to_screen(caret_pos_);

    /* don't bring anything on the screen if we're in MORE mode */
    if (more_mode_)
        bring_on_screen = FALSE;

    /*
     *   if the caret is off-screen horizontally, and we want to bring it
     *   back on screen, scroll to move the caret back within the window 
     */
    if (bring_on_screen)
        scroll_to_show_caret();

    /* move the Windows caret, if we're showing it */
    if (caret_vis_)
    {
        if (more_mode_)
            SetCaretPos(-100, -100);
        else
            SetCaretPos(caret_pos_.x, caret_pos_.y + 2);
    }

    /* update the selection range in the formatter */
    if (update_sel_range)
    {
        size_t start, end, caret;

        /* get the command buffer selection range */
        cmdbuf_->get_sel_range(&start, &end, &caret);

        /* remember which end has the caret */
        caret_at_right_ = (caret == end);

        /* set this range in the formatter for the input item */
        formatter_->set_sel_range(start + inp_txt_ofs, end + inp_txt_ofs);

        /* clear any selections in other subwindows */
        owner_->clear_other_sel(this);
    }

    /*
     *   remember this as the most recent position that we've displayed
     *   the caret in the command line 
     */
    last_cmd_caret_ = cmdbuf_->get_caret();
}

/*
 *   Wait for a single keystroke 
 */
htmlsys_input_stat_t CHtmlSysWin_win32_Input::
   wait_for_keystroke(textchar_t *ch, int pause_only,
                      time_t timeout, int use_timeout)
{
    int ret;

    /* if we have a keystroke command to return, return it */
    if (last_keystroke_cmd_ != 0)
    {
        /* set the return value */
        *ch = last_keystroke_cmd_;

        /* forget it now that it's been consumed */
        last_keystroke_cmd_ = 0;

        /* return it */
        return HTMLSYS_INPUT_STAT_OK;
    }
    
    /* note that we're waiting for a single key */
    waiting_for_key_ = TRUE;
    waiting_for_key_pause_ = TRUE /*pause_only*/;

    /* update the status line */
    if (statusline_ != 0)
        statusline_->update();

    /* we haven't timed out yet */
    waiting_for_evt_timeout_ = FALSE;

    /* 
     *   bring this window to the front (this is useful mostly when we're
     *   running in the debugger, in which case a debugger window will
     *   probably be on top of the game window) 
     */
    owner_->bring_owner_to_front();

    /* make sure the caret is positioned correctly */
    update_caret_pos(FALSE, FALSE);

    /* if there's a timeout, start a timer for the timeout period */
    if (use_timeout)
    {
        /* start a timer for the timeout period */
        input_timer_id_ = alloc_timer_id();
        SetTimer(handle_, input_timer_id_, (DWORD)(timeout * 1000L),
                 (TIMERPROC)0);
    }

    /* wait for a valid keystroke */
    for (;;)
    {
        /* wait for a keystroke */
        ret = CTadsApp::get_app()->event_loop(&got_keystroke_, 0, 0);

        /* 
         *   if the keystroke is valid, or we're quitting, or we've timed
         *   out, stop looping 
         */
        if (last_keystroke_ != 0 || last_keystroke_cmd_ != 0
            || ret == 0 || waiting_for_evt_timeout_)
            break;
    }

    /* we're no longer in single-key mode */
    waiting_for_key_ = FALSE;
    waiting_for_key_pause_ = FALSE;
    if (statusline_ != 0)
        statusline_->update();

    /* 
     *   reset the MORE prompt position to this point, since the user has
     *   obviously seen everything up to here 
     */
    last_input_height_ = formatter_->get_max_y_pos();

    /* if we're quitting, we didn't get a keystroke */
    if (ret == 0)
        return HTMLSYS_INPUT_STAT_EOF;

    /* if we timed out, so indicate */
    if (waiting_for_evt_timeout_)
        return HTMLSYS_INPUT_STAT_TIMEOUT;

    /* if we're merely pausing, discard any extended keystroke code */
    if (pause_only)
        last_keystroke_cmd_ = 0;

    /* return the keystroke we got */
    *ch = last_keystroke_;
    return HTMLSYS_INPUT_STAT_OK;
}

/*
 *   Read an event 
 */
int CHtmlSysWin_win32_Input::get_input_event(unsigned long timeout,
                                             int use_timeout,
                                             os_event_info_t *info)
{
    int ret;

    /* 
     *   note that we're waiting for an event, and remember the return
     *   info structure so that we'll fill it in when we get a suitable
     *   event 
     */
    waiting_for_evt_ = TRUE;
    last_evt_info_ = info;

    /* update the status line */
    if (statusline_ != 0)
        statusline_->update();

    /* we haven't timed out yet */
    waiting_for_evt_timeout_ = FALSE;

    /* 
     *   bring this window to the front (this is useful mostly when we're
     *   running in the debugger, in which case a debugger window will
     *   probably be on top of the game window) 
     */
    owner_->bring_owner_to_front();

    /* make sure the caret is positioned correctly */
    update_caret_pos(FALSE, FALSE);

    /* if there's a timeout, start a timer for the timeout period */
    if (use_timeout)
    {
        /* start a timer for the timeout period */
        input_timer_id_ = alloc_timer_id();
        SetTimer(handle_, input_timer_id_, (DWORD)timeout, (TIMERPROC)0);
    }

    /* wait for an event */
    ret = CTadsApp::get_app()->event_loop(&got_event_, 0, 0);

    /* if we started a timer, stop it */
    if (use_timeout)
    {
        KillTimer(handle_, input_timer_id_);
        free_timer_id(input_timer_id_);
        input_timer_id_ = 0;
    }

    /* 
     *   reset the MORE prompt position to this point, since the user has
     *   obviously seen everything up to here 
     */
    last_input_height_ = formatter_->get_max_y_pos();

    /* no longer waiting for an event */
    waiting_for_evt_ = FALSE;

    /* forget the event info - it won't be valid after we return */
    last_evt_info_ = 0;

    /* if we're quitting, so indicate */
    if (ret == 0)
        return OS_EVT_EOF;

    /* if we timed out, so indicate */
    if (waiting_for_evt_timeout_)
        return OS_EVT_TIMEOUT;

    /* return the event type */
    return last_evt_type_;
}

/*
 *   Read a command 
 */
htmlsys_input_stat_t
   CHtmlSysWin_win32_Input::get_input(textchar_t *buf, size_t bufsiz,
                                      int keep_buf,
                                      time_t timeout, int use_timeout)
{
    size_t initlen;
    
    /* 
     *   calculate the initial length of the text - if we're keeping the
     *   old buffer's contents, use the current string length of the
     *   buffer's contents, otherwise start off with nothing displayed 
     */
    if (keep_buf)
        initlen = get_strlen(buf);
    else
        initlen = 0;

    /* make sure we've formatted all available input */
    while (formatter_->more_to_do())
        formatter_->do_formatting();

    /* set up to read into the caller's buffer */
    cmdbuf_->setbuf(buf, bufsiz - 1, initlen);

    /* tell the formatter to begin a new input */
    cmdtag_ = ((CHtmlFormatterInput *)formatter_)->begin_input(buf, initlen);
    cmdtag_->format(this, formatter_);

    /* set the caret size for the input font */
    set_caret_size(formatter_->get_font());

    /* we haven't read a command yet */
    command_read_ = FALSE;

    /* set the caret position */
    cmdbuf_->set_sel_range(initlen, initlen, initlen);
    update_caret_pos(TRUE, TRUE);

    /* make sure everything's adjusted */
    fmt_adjust_vscroll();

    /* we're not in single-key mode */
    waiting_for_key_ = FALSE;
    waiting_for_key_pause_ = FALSE;

    /* we haven't timed out yet */
    waiting_for_evt_timeout_ = FALSE;

    /* 
     *   bring this window to the front (this is useful mostly when we're
     *   running in the debugger, in which case a debugger window will
     *   probably be on top of the game window) 
     */
    owner_->bring_owner_to_front();

    /* if there's a timeout, set up a timer for the timeout period */
    if (use_timeout)
    {
        /* start a timer for the timeout period */
        input_timer_id_ = alloc_timer_id();
        SetTimer(handle_, input_timer_id_, (DWORD)(timeout * 1000L),
                 (TIMERPROC)0);
    }

    /* read events until we've read a command */
    if (!CTadsApp::get_app()->event_loop(&command_read_, &current_accel_,
                                         handle_))
    {
        /* notify the caller that the application is terminating */
        return HTMLSYS_INPUT_STAT_EOF;
    }

    /* if we started an event timer, we can kill it now */
    if (use_timeout)
    {
        KillTimer(handle_, input_timer_id_);
        free_timer_id(input_timer_id_);
        input_timer_id_ = 0;
    }

    /* tell the formatter that the input is finished */
    ((CHtmlFormatterInput *)formatter_)->end_input();

    /* add the command to the history list if we didn't time out */
    if (cmdbuf_->getlen() != 0 && !waiting_for_evt_timeout_)
        cmdbuf_->add_hist();

    /* set horizontal scrolling back to left edge */
    if (hscroll_ofs_ != 0)
        do_scroll(FALSE, hscroll_, SB_TOP, 0, FALSE);

    /* make sure the command is null-terminated */
    buf[cmdbuf_->getlen()] = '\0';

    /* forget about the command */
    cmdbuf_->setbuf(0, 0);
    cmdtag_ = 0;
    started_reading_cmd_ = FALSE;

    /* don't prompt with "MORE" again until we scroll the command away */
    last_input_height_ = screen_to_doc_y(caret_pos_.y + caret_ht_);

    /* get rid of the caret by moving it off-screen */
    caret_pos_.set(-100, -100);

    /* 
     *   check the timeout flag - if we timed out, return a timeout
     *   indication, otherwise return success 
     */
    if (waiting_for_evt_timeout_)
        return HTMLSYS_INPUT_STAT_TIMEOUT;
    else
        return HTMLSYS_INPUT_STAT_OK;
}

/*
 *   Check for a break key.  We'll look for a Ctrl+Break hotkey message
 *   in our message queue.  
 */
int CHtmlSysWin_win32_Input::check_break_key()
{
    MSG msg;
    
    /* check for a hotkey message */
    while (PeekMessage(&msg, handle_, WM_HOTKEY, WM_HOTKEY, PM_REMOVE))
    {
        switch(msg.wParam)
        {
        case HTML_HK_BREAK:
            /* this is the Ctrl+Break hot key - return BREAK indication */
            return TRUE;

            /* 
             *   note - if we add more hot keys in the future, they'll
             *   need to be handled here in additional cases 
             */

        default:
            /* not a recognized hotkey - ignore it */
            break;
        }
    }

    /* didn't find a break key */
    return FALSE;
}

/*
 *   Check to see if we've started reading a command.  Once we've started
 *   reading a command, we turn off MORE mode until the command has been
 *   entered.  
 */
void CHtmlSysWin_win32_Input::check_cmd_start()
{
    /*
     *   If we're not in "MORE" mode, and we have shown everything in the
     *   formatter's display list, and we have a command buffer, we must
     *   be ready to read a command.  
     */
    if (!started_reading_cmd_ && !more_mode_
        && cmdbuf_->getbuf() != 0 && cmdtag_ != 0
        && formatter_ != 0
        && !((CHtmlFormatterInput *)formatter_)->more_before_input())
    {
        /* this is it - we've started reading a command */
        started_reading_cmd_ = TRUE;

        /* move the caret to the input point */
        cmdbuf_->set_sel_range(0, cmdbuf_->getlen(), cmdbuf_->getlen());
        update_caret_pos(TRUE, TRUE);
    }
}


/*
 *   Process a system (ALT) key 
 */
int CHtmlSysWin_win32_Input::do_syschar(TCHAR c, long keydata)
{
    /* if we're waiting for a keystroke, return the ALT key directly */
    if (isalpha(c) && is_waiting_for_key())
    {
        /* store the key event */
        set_key_event(0, CMD_ALT + c - (isupper(c) ? 'A' : 'a'));

        /* processed */
        return TRUE;
    }

    /* inherit default */
    return CHtmlSysWin_win32::do_syschar(c, keydata);
}

/*
 *   Process a keyboard virtual key event 
 */
int CHtmlSysWin_win32_Input::do_keydown(int vkey, long keydata)
{
    /* 
     *   if it's not a control key, cancel any pending temporary link
     *   highlight mode 
     */
    switch(vkey)
    {
    case VK_CONTROL:
    case VK_LCONTROL:
    case VK_RCONTROL:
        /* handle normally */
        break;

    default:
        /* cancel any pending highlighting */
        set_temp_show_links(FALSE);
        break;
    }

    /*
     *   If we're waiting for a keystroke, check to see if we can map this
     *   key to one of the extended command codes. 
     */
    if (is_waiting_for_key())
    {
        static struct
        {
            int vkey;
            int cmd;
            int ctl;
        } cmdmap[] =
        {
            { VK_UP,     CMD_UP,         FALSE },
            { VK_DOWN,   CMD_DOWN,       FALSE },
            { VK_RIGHT,  CMD_RIGHT,      FALSE },
            { VK_LEFT,   CMD_LEFT,       FALSE },
            { VK_END,    CMD_END,        FALSE },
            { VK_HOME,   CMD_HOME,       FALSE },
            { VK_ESCAPE, CMD_KILL,       FALSE },
            { VK_DELETE, CMD_DEL,        FALSE },
            { VK_SCROLL, CMD_SCR,        FALSE },
            { VK_PRIOR,  CMD_PGUP,       FALSE },
            { VK_NEXT,   CMD_PGDN,       FALSE },
            { VK_HOME,   CMD_TOP,        TRUE  },
            { VK_END,    CMD_BOT,        TRUE  },
            { VK_F1,     CMD_F1,         FALSE },
            { VK_F2,     CMD_F2,         FALSE },
            { VK_F3,     CMD_F3,         FALSE },
            { VK_F4,     CMD_F4,         FALSE },
            { VK_F5,     CMD_F5,         FALSE },
            { VK_F6,     CMD_F6,         FALSE },
            { VK_F7,     CMD_F7,         FALSE },
            { VK_F8,     CMD_F8,         FALSE },
            { VK_F9,     CMD_F9,         FALSE },
            { VK_F10,    CMD_F10,        FALSE },
            { VK_LEFT,   CMD_WORD_LEFT,  TRUE  },
            { VK_RIGHT,  CMD_WORD_RIGHT, TRUE  }
        };
        int i;
        int ctl = get_ctl_key();
        
        /* find the keystroke in our table */
        for (i = 0 ; i < sizeof(cmdmap)/sizeof(cmdmap[0]) ; ++i)
        {
            /* if this one matches, use it */
            if (cmdmap[i].vkey == vkey && cmdmap[i].ctl == ctl)
            {
                /* store the keystroke */
                set_key_event(0, cmdmap[i].cmd);

                /* do not process this keystroke any further */
                return TRUE;
            }
        }
    }
    
    /* see what we have */
    switch(vkey)
    {
    case VK_UP:
        /*
         *   if we're in more-mode, or using arrow keys exclusively for
         *   scrolling, or on an inactive page, treat it as a scrolling
         *   key; otherwise, treat it as a command history selector 
         */
        if (more_mode_ || prefs_->get_arrow_keys_always_scroll()
            || !owner_->is_active_page())
            do_scroll(TRUE, vscroll_, SB_LINEUP, 0, FALSE);
        else if (cmdbuf_->select_prev_hist())
            update_input_display();

        /* handled */
        return TRUE;

    case VK_DOWN:
        /*
         *   If we're in more-mode, or using arrow keys exclusively for
         *   scrolling, treat it as a scrolling key; otherwise, treat it
         *   as a command history selector.  
         */
        if (more_mode_ || prefs_->get_arrow_keys_always_scroll()
            || !owner_->is_active_page())
        {
            /*
             *   if the MORE prompt is at the bottom of the window, show
             *   the next line; otherwise, scroll down normally 
             */
            if (moreprompt_is_at_bottom())
                release_more_mode(HTMLW32_MORE_LINE);
            else
                do_scroll(TRUE, vscroll_, SB_LINEDOWN, 0, FALSE);
        }
        else if (cmdbuf_->select_next_hist())
            update_input_display();

        /* handled */
        return TRUE;

    case VK_PRIOR:
        /* scroll up a page */
        do_scroll(TRUE, vscroll_, SB_PAGEUP, 0, FALSE);
        return TRUE;

    case VK_NEXT:
        /*
         *   Scroll down a page.  If we're in more mode and we're showing
         *   the bottom of the current text, show the next page;
         *   otherwise, simply scroll down a page normally.  
         */
        if (more_mode_ && moreprompt_is_at_bottom())
            release_more_mode(HTMLW32_MORE_PAGE);
        else
            do_scroll(TRUE, vscroll_, SB_PAGEDOWN, 0, FALSE);
        return TRUE;

    case VK_LEFT:
        /* if in a command, move insertion point one character */
        if (!more_mode_)
        {
            cmdbuf_->move_left(get_shift_key(), get_ctl_key());
            update_caret_pos(TRUE, TRUE);
        }

        /* handled */
        return TRUE;

    case VK_RIGHT:
        /* if in a command, move insertion point one character */
        if (!more_mode_)
        {
            cmdbuf_->move_right(get_shift_key(), get_ctl_key());
            update_caret_pos(TRUE, TRUE);
        };

        /* handled */
        return TRUE;

    case VK_END:
        if (more_mode_ || !owner_->is_active_page())
        {
            /* "more" mode - scroll to bottom */
            do_scroll(TRUE, vscroll_, SB_BOTTOM, 0, FALSE);
        }
        else if (get_ctl_key())
        {
            /* delete to end of command */
            if (cmdbuf_->del_eol())
                update_input_display();
        }
        else
        {
            /* move to end of command */
            cmdbuf_->end_of_line(get_shift_key());
            update_caret_pos(TRUE, TRUE);
        }
        return TRUE;

    case VK_HOME:
        if (more_mode_ || !owner_->is_active_page())
        {
            /* "more" mode - scroll to top */
            do_scroll(TRUE, vscroll_, SB_TOP, 0, FALSE);
        }
        else
        {
            /* move to start of comamnd */
            cmdbuf_->start_of_line(get_shift_key());
            update_caret_pos(TRUE, TRUE);
        }
        return TRUE;

    case VK_DELETE:
        /*
         *   if not in more mode, delete selection or character to right
         *   of cursor 
         */
        if (!more_mode_)
        {
            if (cmdbuf_->del_right())
                update_input_display();
        }
        return TRUE;

    case VK_ESCAPE:
        /* escape - delete entire line if editing a command */
        if (!more_mode_)
        {
            if (cmdbuf_->del_line())
                update_input_display();
        }
        return TRUE;

    case VK_HELP:
    case VK_F1:
    case VK_INSERT:
        /* not handled */
        return FALSE;

    default:
        /* inherit default handling for other keys */
        return CHtmlSysWin_win32::do_keydown(vkey, keydata);
    }
}


/*
 *   Process a keyboard character event 
 */
int CHtmlSysWin_win32_Input::do_char(TCHAR c, long /*keydata*/)
{
    /* 
     *   if we're just waiting for a key, return the key event without
     *   further processing 
     */
    if (is_waiting_for_key())
    {
        /* set the keystroke */
        set_key_event(c, 0);

        /* no further processing is required */
        return TRUE;
    }
    
    /* if we're in "MORE" mode, allow displaying more data */
    if (more_mode_)
    {
        switch(c)
        {
        case 32:
            /* space - show the next screen */
            release_more_mode(HTMLW32_MORE_PAGE);
            break;

        case 10:
        case 13:
            /* return - show next line */
            release_more_mode(HTMLW32_MORE_LINE);
            break;

        case 22:
            /* ^V - page down */
            do_keydown(VK_NEXT, 0);
            break;

        default:
            /* ignore other characters in more-mode */
            return FALSE;
        }

        /* keystroke handled */
        return TRUE;
    }

    /* see what we have */
    switch(c)
    {
    case 1:
        /* ^A - start of line */
        cmdbuf_->start_of_line(FALSE);
        update_caret_pos(TRUE, TRUE);
        return TRUE;

    case 5:
        /* ^E - end of line */
        cmdbuf_->end_of_line(FALSE);
        update_caret_pos(TRUE, TRUE);
        return TRUE;

    case 4:
        /* ^D - delete character to right of cursor */
        if (cmdbuf_->del_right())
            update_input_display();
        return TRUE;

    case 6:
        /* ^F - forward a character */
        cmdbuf_->move_right(FALSE, FALSE);
        update_caret_pos(TRUE, TRUE);
        return TRUE;

    case 2:
        /* ^B - back a character */
        cmdbuf_->move_left(FALSE, FALSE);
        update_caret_pos(TRUE, TRUE);
        return TRUE;

    case 11:
        /* ^K - delete to end of line */
        if (cmdbuf_->del_eol())
            update_input_display();
        break;

    case 14:
        /* ^N - next line in history */
        if (cmdbuf_->select_next_hist())
            update_input_display();
        break;

    case 16:
        /* ^P - previous line in history or paste */
        if (cmdbuf_->select_prev_hist())
            update_input_display();
        break;

    case 21:
        /* ^U - delete entire line */
        if (cmdbuf_->del_line())
            update_input_display();
        return TRUE;

    case 22:
        /* ^V - page down (Emacs mode) or Paste (Windows mode) */
        if (prefs_->get_emacs_ctrl_v())
        {
            /* emacs style - scroll down a page */
            do_scroll(TRUE, vscroll_, SB_PAGEDOWN, 0, FALSE);
            return TRUE;
        }
        else
        {
            /* windows style - paste clipboard */
            return do_paste();
        }

    case 25:
        /* ^Y - yank - paste clipboard */
        return do_paste();

    case 24:
        /* ^X - cut */
        return do_cut();

    case 3:
        /* ^C - copy */
        return do_copy();

    case 26:
        /* ^Z - undo */
        if (cmdbuf_->undo())
            update_input_display();
        return TRUE;

    case 13:
    case 10:
        /* 
         *   return/enter key - end the command.  Make sure we're on the
         *   active page, then set the flag indicating that the current
         *   command has been successfully read. 
         */
        owner_->go_to_active_page();
        command_read_ = TRUE;
        return TRUE;

    case 8:
        /* process backspace key, and update the display if necessary */
        if (cmdbuf_->backspace())
            update_input_display();
        return TRUE;

    default:
        /* if it's a normal character, insert it into the buffer */
        if ((unsigned char)c >= 32 && (unsigned char)c < 255)
        {
            /* insert it, and update the display if necessary */
            if (cmdbuf_->add_char(c))
                update_input_display();

            /* keystroke handled */
            return TRUE;
        }
        break;
    }

    /* we didn't handle the keystroke */
    return FALSE;
}

void CHtmlSysWin_win32_Input::process_command(const textchar_t *cmd,
                                              size_t len, int quit_cmd,
                                              int append, int enter)
{
    /* if we're waiting for an event, fill in the event information */
    if (waiting_for_evt_ && last_evt_info_ != 0)
    {
        size_t copylen;

        /* limit the length of our copy to the available space */
        copylen = len;
        if (copylen > sizeof(last_evt_info_->href) - sizeof(textchar_t))
            copylen = sizeof(last_evt_info_->href) - sizeof(textchar_t);

        /* copy the href */
        memcpy(last_evt_info_->href, cmd, copylen);

        /* fill in a null terminator */
        last_evt_info_->href[copylen / sizeof(textchar_t)] = '\0';

        /* 
         *   Note that we got an HREF event.  If quit_cmd is true, though,
         *   treat it as an EOF event, since we're closing the main window
         *   and thus terminating the application. 
         */
        last_evt_type_ = (quit_cmd ? OS_EVT_EOF : OS_EVT_HREF);
        got_event_ = TRUE;

        /* 
         *   we're done - do not process it as a command, since we're
         *   returning the raw event 
         */
        return;
    }
    
    /* if we're waiting for a single keystroke, ignore this */
    if (waiting_for_key_)
        return;

    /* 
     *   if we're in MORE mode, release the MORE and get ready for a new
     *   command 
     */
    if (more_mode_)
        release_more_mode(HTMLW32_MORE_ALL);

    /* if we're not reading a command, ignore this */
    if (!started_reading_cmd_)
        return;

    /* 
     *   if we're not in APPEND mode, clear out the current command;
     *   otherwise, make sure we're at the end of the current text 
     */
    if (!append)
        cmdbuf_->del_line();
    else
        cmdbuf_->end_of_line(FALSE);

    /* add this string */
    cmdbuf_->add_string(cmd, len, TRUE);

    /* update the display */
    update_input_display();

    /* 
     *   If 'enter' is true, indicate that we've finished reading the
     *   command, so that get_input() will return the new command as its
     *   result; otherwise, let the player continue editing this command.  
     */
    if (enter)
        command_read_ = TRUE;
}

/*
 *   Abort the current command 
 */
void CHtmlSysWin_win32_Input::abort_command()
{
    /* clear out anything currently in the command buffer */
    cmdbuf_->del_line();
    update_input_display();

    /* set all the flags to exit out of any input wait we're in */
    command_read_ = TRUE;
    got_keystroke_ = TRUE;
    got_event_ = TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   Generic frame window base class implementation 
 */

/*
 *   remember my position in the preferences - this should be called any
 *   time the window is moved or resized 
 */
void CHtmlSys_framewin::set_winpos_prefs()
{
    RECT rc;
    CHtmlRect winpos;

    GetWindowRect(handle_, &rc);
    winpos.set(rc.left, rc.top, rc.right, rc.bottom);
    set_winpos_prefs(&winpos);
}


/*
 *   enter an interactive size/move operation 
 */
void CHtmlSys_framewin::do_entersizemove()
{
    HWND hdl;

    /* tell all my subwindows about this */
    for (hdl = GetWindow(handle_, GW_CHILD) ; hdl != 0 ;
         hdl = GetWindow(hdl, GW_HWNDNEXT))
        SendMessage(hdl, WM_ENTERSIZEMOVE, 0, 0);
}

/*
 *   finish an interactive size/move operation 
 */
void CHtmlSys_framewin::do_exitsizemove()
{
    HWND hdl;

    /* tell all my subwindows about this */
    for (hdl = GetWindow(handle_, GW_CHILD) ; hdl != 0 ;
         hdl = GetWindow(hdl, GW_HWNDNEXT))
        SendMessage(hdl, WM_EXITSIZEMOVE, 0, 0);
}

/*
 *   move the window 
 */
void CHtmlSys_framewin::do_move(int /*x*/, int /*y*/)
{
    /* remember my position in the preferences */
    set_winpos_prefs();
}

/* 
 *   resize the window 
 */
void CHtmlSys_framewin::do_resize(int mode, int /*x*/, int /*y*/)
{
    /* remember my position in the preferences */
    set_winpos_prefs();
}


/* ------------------------------------------------------------------------ */
/*
 *   HTML main window implementation 
 */


/* 
 *   The global main window.  While the application is running and a
 *   window is open, we keep track of the primary window for the
 *   application.  Since TADS doesn't keep track of an active window, we
 *   need a global window that we can use for TADS input/output
 *   operations. 
 */
CHtmlSys_mainwin *CHtmlSys_mainwin::main_win_ = 0;

/*
 *   create the main window 
 */
CHtmlSys_mainwin::CHtmlSys_mainwin(CHtmlFormatterInput *formatter,
                                   CHtmlParser *parser)
{
    int code_page;
    CHARSETINFO csinfo;
    HDC hdc;
    TEXTMETRIC tm;
    
    /* create a text buffer */
    txtbuf_ = new CHtmlTextBuffer();
    
    /* remember my parser */
    parser_ = parser;

    /* create preferences object */
    prefs_ = new CHtmlPreferences();

    /* read stored preference settings */
    prefs_->restore();
    
    /* no main panel offset yet */
    main_panel_yoffset_ = 0;

    /* no banners yet */
    banners_ = 0;
    banner_cnt_ = 0;
    banners_max_ = 0;

    /* 
     *   tell the oshtml layer about me, since I provide the application
     *   frame services to the TADS output system 
     */
    oshtml_set_frame(this);

    /* remember me as the main window */
    if (main_win_ == 0)
        main_win_ = this;

    /* initialize the first page's state structure */
    first_page_state_ = last_page_state_ = cur_page_state_ = 0;
    add_new_page_state();

    /* no DirectSound interface yet - create it when we create the window */
    directsound_ = 0;
    directsound_init_err_ = HTMLW32_DIRECTX_OK;

    /* no debug window yet */
    dbgwin_ = 0;

    /* create the main HTML display panel window */
    main_panel_ = new CHtmlSysWin_win32_Input(formatter, this, TRUE, TRUE,
                                              prefs_);

    /* no "about" box window yet */
    aboutbox_ = 0;

    /* 
     *   Find the character set that corresponds to the current system
     *   ANSI code page, and use this as the default character set. 
     */
    code_page = GetACP();
    TranslateCharsetInfo((DWORD *)code_page, &csinfo, TCI_SRCCODEPAGE);
    default_charset_ = csinfo.ciCharset;

    /* the game character set has not been set yet */
    game_internal_charset_.set("Unknown");

    /* 
     *   Get the default character in the default character set -- we'll
     *   use this whenever the game tries to display a character for which
     *   we can provide no character mapping.  Use the DC for the desktop
     *   window, which presumably has a font selected that uses the system
     *   default character set.  
     */
    hdc = GetDC(GetDesktopWindow());
    GetTextMetrics(hdc, &tm);
    invalid_char_val_ = tm.tmDefaultChar;
    ReleaseDC(GetDesktopWindow(), hdc);
}

CHtmlSys_mainwin::~CHtmlSys_mainwin()
{
    CHtmlSys_mainwin_state_link *link;
    CHtmlSys_mainwin_state_link *nxt;
    
    /* save preference settings */
    prefs_->save();

    /* delete the banner list */
    if (banners_ != 0)
        th_free(banners_);

    /* delete the preferences object */
    delete prefs_;

    /* delete the status line */
    if (statusline_ != 0)
        delete statusline_;

    /* tell the oshtml layer that I'm no longer callable */
    oshtml_set_frame(0);

    /* forget me if I'm the main window */
    if (main_win_ == this)
        main_win_ = 0;

    /* done with our text buffer */
    delete txtbuf_;

    /* delete all of the saved parse trees */
    for (link = first_page_state_ ; link != 0 ; link = nxt)
    {
        /* remember the next link, since we're going to delete this one */
        nxt = link->next_;

        /* delete this one */
        delete link;
    }

    /* release our DirectSound interface */
    if (directsound_ != 0)
        directsound_->Release();
}

/*
 *   Set my debug window 
 */
void CHtmlSys_mainwin::set_debug_win(CHtmlSys_dbglogwin *dbgwin,
                                     int add_menu_item)
{
    /* 
     *   if I don't have a debug window already, and we now have one, add
     *   a menu item to show it should it become hidden 
     */
    if (dbgwin_ == 0 && dbgwin != 0 && add_menu_item)
    {
        HMENU hmenu;
        int cnt;
        MENUITEMINFO info;
        static char item_name[] = "Show &Debug Window";

        /* get the View menu */
        hmenu = GetMenu(handle_);
        hmenu = GetSubMenu(hmenu, 2);

        /* get the number of items in the View menu */
        cnt = GetMenuItemCount(hmenu);

        /* insert a separator at the end of the View menu */
        info.cbSize = sizeof(info);
        info.fMask = MIIM_TYPE;
        info.fType = MFT_SEPARATOR;
        InsertMenuItem(hmenu, cnt++, TRUE, &info);

        /* insert a new item for View Debug Window */
        info.fMask = MIIM_TYPE | MIIM_DATA | MIIM_ID;
        info.fType = MFT_STRING;
        info.wID = ID_SHOW_DEBUG_WIN;
        info.dwTypeData = item_name;
        info.cch = get_strlen(item_name);
        InsertMenuItem(hmenu, cnt, TRUE, &info);
    }

    /* remember the new setting */
    dbgwin_ = dbgwin;
}

/*
 *   Abort the current command 
 */
void CHtmlSys_mainwin::abort_command()
{
    /* tell the main window to abort its command */
    main_panel_->abort_command();
}

/*
 *   set my window position in the preferences 
 */
void CHtmlSys_mainwin::set_winpos_prefs(const CHtmlRect *winpos)
{
    prefs_->set_win_pos(winpos);
}

/*
 *   process creation event 
 */
void CHtmlSys_mainwin::do_create()
{
    RECT panel_pos;
    HINSTANCE dsound;
    int warned;

    /* inherit default creation */
    CHtmlSys_framewin::do_create();

    /* create and show the status line */
    statusline_ = new CTadsStatusline(this, TRUE, HTML_MAINWIN_ID_STATUS);

    /* tell the main panel about the status line */
    main_panel_->set_status_line(statusline_);

    /* create the main panel's system window */
    GetClientRect(handle_, &panel_pos);
    main_panel_->create_system_window(this, TRUE, "", &panel_pos);

    /* load my menu */
    SetMenu(handle_, LoadMenu(CTadsApp::get_app()->get_instance(),
                              MAKEINTRESOURCE(IDR_MAIN_MENU)));

    /*
     *   Adjust the menu so that the View Links and Font Size menu
     *   subitems use radio button checkmarks, since they're all mutually
     *   exclusive settings 
     */
    set_menu_radiocheck(GetSubMenu(GetSubMenu(GetMenu(handle_), 2), 0));
    set_menu_radiocheck(GetSubMenu(GetSubMenu(GetMenu(handle_), 2), 1));

    /* set up the correct accelerators the preferences */
    main_panel_->set_current_accel();

    /* set the appropriate accelerators for the link menu */
    set_link_accel(prefs_->get_links_on(), prefs_->get_links_ctrl());

    /* presume we won't have DirectSound access */
    directsound_ = 0;

    /* we haven't issued a warning yet */
    warned = FALSE;

    /* load the DirectSound DLL and get the DirectSoundCreate function */
    dsound = LoadLibrary("DSOUND.DLL");
    if (dsound != 0)
    {
        HRESULT (WINAPI *dscproc)(LPGUID, LPDIRECTSOUND *, LPUNKNOWN);

        dscproc = (HRESULT (WINAPI *)(LPGUID, LPDIRECTSOUND *, LPUNKNOWN))
                  GetProcAddress(dsound, "DirectSoundCreate");
        if (dscproc != 0)
        {
            /* create the Windows DirectSound interface */
            if ((*dscproc)(0, &directsound_, 0) == DS_OK)
            {
                WAVEFORMATEX fmt;
                DSBUFFERDESC bufdesc;
                IDirectSoundBuffer *buf;
                IDirectSoundNotify *inotify;
                int vsnok;

                /* 
                 *   set our DirectSound cooperative level to "normal",
                 *   since we don't need anything fancy 
                 */
                directsound_->SetCooperativeLevel(handle_, DSSCL_NORMAL);

                /* presume the version will be incorrect */
                vsnok = FALSE;

                /*
                 *   We've created the DirectSound interface object
                 *   successfully, but there's still the matter of the
                 *   DirectX version.  We need to be able to use the
                 *   IDirectSoundNotify interface, which only exists in
                 *   DirectX version 5 and higher.  Try creating a sound
                 *   buffer and creating a Notify interface; if that
                 *   fails, we can't use this DirectSound system because
                 *   it's too old a version.  Note that we're just making
                 *   up a fake sound descriptor here so that we can create
                 *   a sound buffer; this is a simple sound descriptor
                 *   that should be compatible with any directx
                 *   installation.  
                 */
                memset(&bufdesc, 0, sizeof(bufdesc));
                bufdesc.dwSize = sizeof(bufdesc);
                bufdesc.dwFlags = DSBCAPS_CTRLPOSITIONNOTIFY
                                  | DSBCAPS_CTRLDEFAULT
                                  | DSBCAPS_GETCURRENTPOSITION2;
                bufdesc.dwBufferBytes = 10000;
                fmt.wFormatTag = WAVE_FORMAT_PCM;
                fmt.nChannels = 1;
                fmt.nSamplesPerSec = 8000;
                fmt.nAvgBytesPerSec = 8000;
                fmt.nBlockAlign = 1;
                fmt.wBitsPerSample = 8;
                fmt.cbSize = 0;
                bufdesc.lpwfxFormat = &fmt;
                if (directsound_->CreateSoundBuffer(&bufdesc, &buf, 0)
                    == DS_OK)
                {
                    /* 
                     *   we have a sound buffer - ask it for the
                     *   IDirectSoundNotify interface 
                     */
                    if (buf->QueryInterface(IID_IDirectSoundNotify,
                                            (void **)&inotify) == DS_OK)
                    {
                        /* we're done with the notifier - release it */
                        inotify->Release();

                        /* success */
                        vsnok = TRUE;

                        /*
                         *   If DirectX initialization failed in the past,
                         *   note that it's working now, and turn on
                         *   sounds.  If there wasn't an error last time,
                         *   respect the sound settings from last time.  
                         */
                        if (prefs_->get_directx_error_code()
                            != HTMLW32_DIRECTX_OK)
                        {
                            /* remember that there's no error */
                            prefs_->
                                set_directx_error_code(HTMLW32_DIRECTX_OK);

                            /* turn on sounds */
                            prefs_->set_sounds_on(TRUE);
                        }
                    }
                    else
                    {
                        OSVERSIONINFO os_info;
                        
                        /*
                         *   We couldn't create the IDirectSoundNotify
                         *   interface.  As a special case, if we're
                         *   running on NT 4, allow this (since we
                         *   evidently have a DirectSound installation
                         *   that we can otherwise use) without even
                         *   warning about it, since NT 4 users can't have
                         *   DirectX 5 even if they want it (Microsoft
                         *   doesn't support it). 
                         */
                        os_info.dwOSVersionInfoSize = sizeof(os_info);
                        GetVersionEx(&os_info);
                        if (os_info.dwPlatformId == VER_PLATFORM_WIN32_NT
                            && os_info.dwMajorVersion == 4)
                            vsnok = TRUE;
                    }

                    /* we're done with the buffer - release it */
                    buf->Release();
                }

                /*
                 *   If the version check failed, and notify the user of
                 *   the problem.  However, do keep the IDirectSound
                 *   interface, since we can use it in DirectX 3 mode.
                 *   This doesn't work as well as DirectX 5 mode, which is
                 *   why we take the trouble to warn the user about it,
                 *   but we'll still be able to operate correctly anyway.  
                 */
                if (!vsnok)
                {
                    /* warn the user */
                    show_directx_warning(TRUE, HTMLW32_DIRECTX_BADVSN);
                    
                    /* we've warned the user */
                    warned = TRUE;
                }
            }
        }
    }

    /*
     *   check to see if direct sound initialization failed 
     */
    if (directsound_ == 0)
    {
        /* if we haven't warned yet, do so now */
        if (!warned)
            show_directx_warning(TRUE, HTMLW32_DIRECTX_MISSING);

        /* turn off sound in the preferences */
        prefs_->set_sounds_on(FALSE);

        /* if we successfully loaded the library, free it */
        if (dsound != 0)
            FreeLibrary(dsound);
    }

    /* 
     *   create the "about" box window -- we'll just create the window and
     *   keep it hidden for now 
     */
    aboutbox_ = new CHtmlSys_aboutwin(prefs_);
    SetRect(&panel_pos, 0, 0, 500, 300);
    aboutbox_->create_system_window(this, FALSE,
                                    "About This Game", &panel_pos);
}


/*
 *   Show a DirectX warning 
 */
void CHtmlSys_mainwin::show_directx_warning(int init,
                                            htmlw32_directx_err_t warn_type)
{
    CTadsDlg_dxwarn *dlg;
    
    /*
     *   Check to see if the user has previously told us to suppress this
     *   type of warning.  We'll only suppress the warnings on startup,
     *   since at other times, the user has explicitly done something
     *   requesting sound. 
     */
    if (init)
    {
        /* 
         *   remember the cause of the failure, in case we need to
         *   generate another message about this later 
         */
        directsound_init_err_ = warn_type;

        /* 
         *   if the error code is different than last time, turn off the
         *   "hide this warning" checkbox, since the user didn't tell us
         *   to hide this new warning 
         */
        if (warn_type != prefs_->get_directx_error_code())
            prefs_->set_directx_hidewarn(FALSE);

        /*
         *   if the user told us to hide this warning last time, skip the
         *   dialog 
         */
        if (prefs_->get_directx_hidewarn())
            return;
    }

    /* run the dialog */
    dlg = new CTadsDlg_dxwarn(prefs_, warn_type);
    dlg->run_modal(DLG_DIRECTX_WARNING, handle_,
                   CTadsApp::get_app()->get_instance());
    delete dlg;

    /* 
     *   Remember the error type persistently for the next startup.  Note
     *   that it's important to wait until *after* we've run the dialog,
     *   because the dialog needs to be able to check the *old* value of
     *   the error code in the preferences - this way, it can determine if
     *   the checkbox setting is still applicable, since it only applies
     *   when the current error is the same as the old error. 
     */
    prefs_->set_directx_error_code(warn_type);
}


/*
 *   Receive notification that a character mapping file has been loaded.
 *   For Windows character mapping table files, the extra system
 *   information is the Windows code page that should be used with the
 *   game's internal character set.  We'll find the Windows xxx_CHARSET
 *   value that corresponds to the Windows code page, and set that as the
 *   default character set. 
 */
void CHtmlSys_mainwin::advise_load_charmap(const char *ldesc,
                                           const char *sysinfo)
{
    int code_page;
    CHARSETINFO csinfo;

    /* remember the internal character set name */
    game_internal_charset_.set(ldesc);
    
    /* if the system information is not a valid code page, ignore it */
    code_page = atoi(sysinfo);
    // if (!IsValidCodePage(code_page)) // for now, allow uninstalled cp's
    if (code_page == 0)
        return;

    /* translate the code page to a character set identifier */
    TranslateCharsetInfo((DWORD *)code_page, &csinfo, TCI_SRCCODEPAGE);

    /* remember the default character set identifier */
    default_charset_ = csinfo.ciCharset;
}


/*
 *   process a close-window request 
 */
int CHtmlSys_mainwin::do_close()
{
    /* send a QUIT command to the main command window, if possible */
    if (main_panel_ != 0)
    {
        /* if the main panel is waiting for exit, let it exit */
        if (main_panel_->get_exit_pause())
            PostMessage(main_panel_->get_handle(), WM_CHAR, (WPARAM)' ', 0);

        /* process a QUIT command */
        process_command("quit", 4, TRUE, FALSE, TRUE);

        /* don't actually close the window yet */
        return FALSE;
    }
    else
    {
        /* no main panel - just close normally */
        return TRUE;
    }
}


/*
 *   destroy the window 
 */
void CHtmlSys_mainwin::do_destroy()
{
    int i;
    CHtmlSysWin_win32 **banner;
    
    /* tell my child windows to disengage from their parent */
    main_panel_->on_destroy_parent();
    for (i = 0, banner = banners_ ; i < banner_cnt_ ; ++i, ++banner)
        (*banner)->on_destroy_parent();
    
    /* terminate the application */
    PostQuitMessage(0);
    
    /* forget about our subwindow */
    main_panel_ = 0;

    /* inherit default */
    CTadsWin::do_destroy();
}

/*
 *   paint the window contents 
 */
void CHtmlSys_mainwin::do_paint_content(HDC hdc, const RECT *paintrc)
{
    HGDIOBJ oldpen;
    RECT clientrc;
    RECT panelrc;

    /* erase the background in white */
    FillRect(hdc, paintrc, GetStockObject(WHITE_BRUSH));

    /* select a black pen */
    oldpen = SelectObject(hdc, GetStockObject(BLACK_PEN));

    /* get the area left over between panel and status line */
    GetClientRect(handle_, &clientrc);
    GetClientRect(statusline_->get_handle(), &panelrc);
    clientrc.top = clientrc.bottom - panelrc.bottom - 3;
    clientrc.bottom = clientrc.top + 2;

    /* paint a black line across the top and left edges of the main panel */
    MoveToEx(hdc, 0, clientrc.bottom - 1, 0);
    LineTo(hdc, 0, main_panel_yoffset_);
    LineTo(hdc, clientrc.right, main_panel_yoffset_);

    /* put a gray line just below the main panel */
    FillRect(hdc, &clientrc, GetStockObject(LTGRAY_BRUSH));

    /* paint a white line down the right side */
    SelectObject(hdc, GetStockObject(WHITE_PEN));
    MoveToEx(hdc, clientrc.right, main_panel_yoffset_, 0);
    LineTo(hdc, clientrc.right, clientrc.bottom + 1);

    /* done with the pen */
    SelectObject(hdc, oldpen);
}

/*
 *   resize 
 */
void CHtmlSys_mainwin::do_resize(int mode, int x, int y)
{
    switch(mode)
    {
    case SIZE_MAXHIDE:
    case SIZE_MAXSHOW:
    case SIZE_MINIMIZED:
        /* don't bother resizing the subwindows on these changes */
        break;
        
    case SIZE_MAXIMIZED:
    case SIZE_RESTORED:
    default:
        {
            RECT statrc;
            RECT rc;
            int i;
            CHtmlSysWin_win32 **banner;

            /* notify the status line of the resize */
            statusline_->notify_parent_resize();
            
            /* 
             *   Figure out the total area available to the subwindows.
             *   The first pixel on the left and top are taken up by
             *   highlighting, so start at (1,1).  We need two pixels at
             *   the right edge and three at the bottom edge for
             *   highlighting, and we also need to remove the area taken
             *   up by the status line.  
             */
            GetClientRect(statusline_->get_handle(), &statrc);
            SetRect(&rc, 1, main_panel_yoffset_ + 1,
                    x - 2, y - statrc.bottom - 3);
            
            /* place the banners */
            for (i = 0, banner = banners_ ; i < banner_cnt_ ; ++i, ++banner)
            {
                RECT banrc;
                RECT oldbanrc;
                int banwid, banht;
                
                /* 
                 *   get the current size of this banner window - we'll
                 *   leave the unconstrained dimension the same as it was 
                 */
                GetClientRect((*banner)->get_handle(), &banrc);
                banwid = banrc.right - banrc.left;
                banht = banrc.bottom - banrc.top;

                /* remember the original banner position */
                oldbanrc = banrc;

                /* 
                 *   start off assuming the banner will consume the entire
                 *   input window - since it's aligned at three edges with
                 *   the current input window size, we'll only have to
                 *   undo one of these settings, but which one depends on
                 *   the banner alignment 
                 */
                banrc = rc;

                /* 
                 *   figure out where the banner goes, according to its
                 *   alignment type 
                 */
                switch((*banner)->get_banner_alignment())
                {
                case HTML_BANNERWIN_POS_TOP:
                    /* align the banner at the top of the window */
                    banrc.bottom = rc.top + banht;

                    /* take the height out of hte top of the input window */
                    rc.top += banht;
                    break;

                case HTML_BANNERWIN_POS_BOTTOM:
                    /* align the banner at the bottom of the window */
                    banrc.top = rc.bottom - banht;

                    /* take the space out of the bottom of the input area */
                    rc.bottom -= banht;
                    break;

                case HTML_BANNERWIN_POS_LEFT:
                    /* align the banner at the left of the window */
                    banrc.right = rc.left + banwid;

                    /* remove the space from the left of the input window */
                    rc.left += banwid;
                    break;

                case HTML_BANNERWIN_POS_RIGHT:
                    /* align the banner at the right of the window */
                    banrc.left = rc.right - banwid;

                    /* deduct the space from the right of the input window */
                    rc.right -= banwid;
                    break;
                }

                /* 
                 *   Set the new banner window position.  Avoid moving the
                 *   window if the position isn't actually going to
                 *   change, since doing so can cause convergence problems
                 *   as we keep shifting windows around unnecessarily. 
                 */
                if (memcmp(&banrc, &oldbanrc, sizeof(banrc)) != 0)
                    MoveWindow((*banner)->get_handle(),
                               banrc.left, banrc.top,
                               banrc.right - banrc.left,
                               banrc.bottom - banrc.top, TRUE);
            }
        
            /* size the panel, based on what's left over */
            if (main_panel_ != 0)
                MoveWindow(main_panel_->get_handle(),
                           rc.left, rc.top,
                           rc.right - rc.left, rc.bottom - rc.top, TRUE);
            
            /* inherit default handling to save position in prefs */
            CHtmlSys_framewin::do_resize(mode, x, y);
        }
        break;
    }
}

/*
 *   Gain focus.  We simply move focus to the main panel window. 
 */
void CHtmlSys_mainwin::do_setfocus(HWND /*previous_focus*/)
{
    SetFocus(main_panel_->get_handle());
}

/*
 *   user message handler 
 */
int CHtmlSys_mainwin::do_user_message(int /*msg*/, WPARAM, LPARAM)
{
    /* not handled */
    return FALSE;
}

/*
 *   Handle a command 
 */
int CHtmlSys_mainwin::do_command(int notify_code,
                                 int command_id, HWND ctl)
{
    /* check to see if it's one of our own commands */
    switch(command_id)
    {
    case ID_SHOW_DEBUG_WIN:
        /* show the debug window and bring it to the top */
        if (dbgwin_ != 0)
        {
            /* show it */
            ShowWindow(dbgwin_->get_handle(), SW_SHOW);

            /* bring it to the top */
            SetWindowPos(dbgwin_->get_handle(), HWND_TOP, 0, 0, 0, 0,
                         SWP_NOMOVE | SWP_NOSIZE);
        }
        return TRUE;

#if 0 // transcript viewing - not yet implemented
    case ID_FILE_VIEW_SCRIPT:
        view_script();
        return TRUE;
#endif

    case ID_HELP_ABOUT_GAME:
        aboutbox_->run_aboutbox(this);
        return TRUE;

    default:
        /* pass the command to the main panel for processing */
        if (main_panel_ != 0)
            return main_panel_->do_command(notify_code, command_id, ctl);
        break;
    }

    /* didn't handle it */
    return FALSE;
}

/*
 *   Check the status of a command 
 */
TadsCmdStat_t CHtmlSys_mainwin::check_command(int command_id)
{
    /* check to see if it's one of our own commands */
    switch(command_id)
    {
    case ID_SHOW_DEBUG_WIN:
        return TADSCMD_ENABLED;

    case ID_HELP_ABOUT_GAME:
        return (aboutbox_->get_html_subwin() != 0
                ? TADSCMD_ENABLED : TADSCMD_DISABLED);

    case ID_FILE_VIEW_SCRIPT:
        return TADSCMD_ENABLED;

    default:
        /* not one of ours - pass it to the main panel for processing */
        if (main_panel_ != 0)
            return main_panel_->check_command(command_id);
        break;
    }

    /* if we got here, we didn't handle it */
    return TADSCMD_UNKNOWN;
}

/*
 *   Process a command.  We'll send the command to our main command input
 *   panel for processing.  This allows banner panels to set up links, and
 *   have them easily processed by the main command window.  
 */
void CHtmlSys_mainwin::process_command(const textchar_t *cmd, size_t len,
                                       int quit_command, int append,
                                       int enter)
{
    /* send the command to the main panel for processing */
    main_panel_->process_command(cmd, len, quit_command, append, enter);
}

/*
 *   Realize my palette.  We'll call the subwindows to do the real work.  
 */
int CHtmlSys_mainwin::do_querynewpalette()
{
    /* let the main panel set its palette */
    return main_panel_->do_querynewpalette();
}

/*
 *   Update the window for a system palette change.  We'll call the
 *   subwindows to do the real work.  
 */
void CHtmlSys_mainwin::do_palettechanged(HWND initiating_window)
{
    int i;
    CHtmlSysWin_win32 **banner;
    
    /* tell the main panel to adjust for the new palette */
    main_panel_->do_palettechanged(initiating_window);

    /* tell each banner subwindow to adjust */
    for (i = 0, banner = banners_ ; i < banner_cnt_ ; ++i, ++banner)
        (*banner)->do_palettechanged(initiating_window);
}


/*
 *   Prune the parse tree that we're displaying if we're using too much
 *   memory. 
 */
void CHtmlSys_mainwin::maybe_prune_parse_tree()
{
    static int check_count = 0;
    unsigned long max_mem = (unsigned long)prefs_->get_mem_text_limit();
    HCURSOR old_cursor;

    /* 
     *   skip this entirely most of the time; only check every so often,
     *   so that we don't waste a lot of time doing this too frequently 
     */
    ++check_count;
    if (check_count < 10)
        return;

    /* 
     *   reset the counter to zero, so that we don't check again until the
     *   next interval has elapsed 
     */
    check_count = 0;

    /* never do any pruning unless we're showing the active page */
    if (!is_active_page())
        return;

    /* 
     *   check to see if we're consuming too much memory - if not, there's
     *   nothing we need to do here 
     */
    if (get_parse_tree_mem() <= max_mem)
        return;

    /* this may take a while - provide a status display */
    old_cursor = SetCursor(wait_cursor_);
    main_panel_->set_pruning_msg(TRUE);
    if (statusline_ != 0)
        statusline_->update();

    /*
     *   First, delete old pages until we run out of pages or we free up
     *   enough memory.  Never delete the last page, since that's the
     *   active page.  
     */
    while (get_parse_tree_mem() > max_mem
           && first_page_state_ != 0 && first_page_state_ != last_page_state_)
    {
        CHtmlSys_mainwin_state_link *cur;

        /* remember the oldest page */
        cur = first_page_state_;

        /* unlink it from the list */
        first_page_state_ = cur->next_;
        first_page_state_->prev_ = 0;
        if (first_page_state_->next_ != 0)
            first_page_state_->next_->prev_ = 0;

        /* delete the page */
        delete cur;
    }

    /* if we're still over budget, prune the active page */
    if (get_parse_tree_mem() > max_mem)
    {
        /* prune the active page's parse tree */
        parser_->prune_tree(max_mem);
        
        /* reformat the active page, since we've changed its contents */
        reformat_all_html(TRUE, TRUE, TRUE);
    }

    /* remove status display */
    SetCursor(old_cursor);
    main_panel_->set_pruning_msg(FALSE);
    if (statusline_ != 0)
        statusline_->update();
}

/*
 *   Get the total memory used by parse tree text arrays 
 */
unsigned long CHtmlSys_mainwin::get_parse_tree_mem() const
{
    CHtmlSys_mainwin_state_link *pg;
    unsigned long sum;
    
    /* add up the memory used by old pages */
    for (sum = 0, pg = first_page_state_ ; pg != 0 ; pg = pg->next_)
    {
        /* only consider inactive pages - we'll get the active page later */
        if (pg->state_->get_text_array() != 0)
            sum += pg->state_->get_text_array()->get_mem_in_use();
    }

    /* add the memory used by the active page */
    sum += parser_->get_text_array()->get_mem_in_use();

    /* return the total */
    return sum;
}


/*
 *   Read a command line
 */
htmlsys_input_stat_t
   CHtmlSys_mainwin::get_input(textchar_t *buf, size_t bufsiz, int keep_buf,
                               time_t timeout, int use_timeout)
{
    static const textchar_t brk[] = "<br>";
    htmlsys_input_stat_t stat;

    /* flush output and make sure it's displayed */
    flush_txtbuf(TRUE);

    /* do some pruning if memory is too full */
    maybe_prune_parse_tree();

    /* let the main panel do the work */
    stat = main_panel_->get_input(buf, bufsiz, keep_buf,
                                  timeout, use_timeout);

    /* check to see if the user quit the app */
    if (stat != HTMLSYS_INPUT_STAT_EOF)
    {
        /* add a break after the command */
        if (parser_->get_obey_markups())
            txtbuf_->append(brk, sizeof(brk) - sizeof(textchar_t));
        else
            txtbuf_->append("\n", 1);

        /* 
         *   flush the newline, and update the window immediately, in case
         *   the operation takes a while to complete 
         */
        flush_txtbuf(TRUE);
        UpdateWindow(main_panel_->get_handle());

        /* return the status */
        return stat;
    }
    else
    {
        /* 
         *   quitting - send another quit message, to make sure the main
         *   event loop receives it (since the one that was already posted
         *   was probably eaten by the nested event loop we just exited) 
         */
        PostQuitMessage(0);

        /* indicate that the user terminated the program */
        return HTMLSYS_INPUT_STAT_EOF;
    }
}

/*
 *   Read an event.  Follows the semantics of the TADS os_get_event()
 *   interface.  
 */
int CHtmlSys_mainwin::get_input_event(unsigned long timeout, int use_timeout,
                                      os_event_info_t *info)
{
    int evt;
    
    /* flush output and make sure it's displayed */
    flush_txtbuf(TRUE);

    /* let the main panel do the work */
    evt = main_panel_->get_input_event(timeout, use_timeout, info);

    /* 
     *   if it returned EOF, we must be quitting - send another quit
     *   message, to make sure the main event loop receives it (since the
     *   one that was already posted was probably eaten by the nested
     *   event loop that we just exited) 
     */
    if (evt == OS_EVT_EOF)
        PostQuitMessage(0);

    /* return the event code */
    return evt;
}

/*
 *   Check for a break key 
 */
int CHtmlSys_mainwin::check_break_key()
{
    /* ask the main panel to check for a break key */
    return main_panel_->check_break_key();
}

/*
 *   Flush buffered output 
 */
void CHtmlSys_mainwin::flush_txtbuf(int fmt)
{
    /* parse what's in the buffer */
    parser_->parse(txtbuf_, this);

    /* we're done with the text in the buffer - clear it out */
    txtbuf_->clear();

    /* 
     *   if desired, run the parsed source through the formatter and
     *   display it 
     */
    if (fmt)
        main_panel_->do_formatting(FALSE, FALSE, FALSE);
}

#if 0 // transcript viewing is not yet implemented
/*
 *   View a transcript 
 */
void CHtmlSys_mainwin::view_script()
{
    OPENFILENAME info;
    char fullname[MAX_PATH];

    /* set up to get the filename */
    info.lStructSize = sizeof(info);
    info.hwndOwner = handle_;
    info.hInstance = CTadsApp::get_app()->get_instance();
    info.lpstrFilter = "Text Files\0*.TXT\0"
                       "Log Files\0*.LOG\0"
                       "All Files\0*.*\0\0";
    info.lpstrCustomFilter = 0;
    info.nFilterIndex = 0;
    info.lpstrFile = fullname;
    fullname[0] = '\0';
    info.nMaxFile = sizeof(fullname);
    info.lpstrFileTitle = 0;
    info.nMaxFileTitle = 0;
    info.lpstrInitialDir = CTadsApp::get_app()->get_openfile_dir();
    info.lpstrTitle = "Choose a transcript file to view";
    info.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
    info.nFileOffset = 0;
    info.nFileExtension = 0;
    info.lpstrDefExt = 0;
    info.lCustData = 0;
    info.lpfnHook = 0;
    info.lpTemplateName = 0;
    CTadsDialog::set_filedlg_center_hook(&info);

    /* get the filename - if they cancel, simply return with nothing done */
    if (!GetOpenFileName(&info))
        return;

    /* start a new page */
    start_new_page();
}
#endif /* 0 */

/*
 *   Clear the screen and start a new page 
 */
void CHtmlSys_mainwin::start_new_page()
{
    /* make sure the current page is completely flushed */
    flush_txtbuf(TRUE);
    
    /* 
     *   Export the page to our current save structure.  This not only
     *   saves the current page, but clears out the parser and resets it
     *   to a new page. 
     */
    parser_->export_parse_tree(get_cur_page_state());
    
    /* save the current scrolling position */
    cur_page_state_->scrollpos_.x = main_panel_->get_hscroll_pos();
    cur_page_state_->scrollpos_.y = main_panel_->get_vscroll_pos();

    /* add a new page to our page list */
    add_new_page_state();

    /*
     *   we haven't seen any of the text on the new page yet, so we'll
     *   need a "more" prompt if it goes off the bottom of the screen
     *   before we get to the next command line 
     */
    main_panel_->reset_last_input_height();
    
    /* reformat the window for the new blank page */
    main_panel_->remove_all_banners();
    reformat_all_html(FALSE, FALSE, TRUE);

    /* reset the selection range in all panels */
    clear_other_sel(0);
}

/*
 *   Clear all pages
 */
void CHtmlSys_mainwin::clear_all_pages()
{
    CHtmlSys_mainwin_state_link *cur;
    CHtmlSys_mainwin_state_link *nxt;
    
    /* delete any about box */
    aboutbox_->delete_html_subwin();

    /* start a new blank page */
    start_new_page();

    /* delete all previous pages */
    for (cur = first_page_state_ ; cur != 0 && cur != cur_page_state_ ;
         cur = nxt)
    {
        /* remember the next page */
        nxt = cur->next_;

        /* delete this one */
        delete cur;

        /* advance the head pointer */
        first_page_state_ = nxt;

        /* the new head has no back pointer now */
        nxt->prev_ = 0;
    }
}

/*
 *   Get the current page's state structure 
 */
CHtmlParserState *CHtmlSys_mainwin::get_cur_page_state()
{
    return cur_page_state_->state_;
}

/*
 *   Add a new page state structure, and switch to the new page 
 */
void CHtmlSys_mainwin::add_new_page_state()
{
    CHtmlSys_mainwin_state_link *link;
    CHtmlParserState *new_state;
    
    /* create a new state structure for the new page */
    new_state = new CHtmlParserState();

    /* add it to our list */
    link = new CHtmlSys_mainwin_state_link(new_state);
    link->prev_ = last_page_state_;
    if (last_page_state_)
        last_page_state_->next_ = link;
    else
        first_page_state_ = link;
    last_page_state_ = link;

    /* make it the current page */
    cur_page_state_ = link;
}

/*
 *   Check to see if there's a next/previous page in our list 
 */
int CHtmlSys_mainwin::exists_next_page()
{
    return cur_page_state_->next_ != 0;
}

int CHtmlSys_mainwin::exists_previous_page()
{
    return cur_page_state_->prev_ != 0;
}

/*
 *   Switch to the next/previous page 
 */
void CHtmlSys_mainwin::go_next_page()
{
    go_to_page(cur_page_state_->next_);
}

void CHtmlSys_mainwin::go_previous_page()
{
    go_to_page(cur_page_state_->prev_);
}

/*
 *   go to the active page 
 */
void CHtmlSys_mainwin::go_to_active_page()
{
    /* if we're already on the active page, there's nothing to do */
    if (is_active_page())
        return;

    /* switch to the active page - this is always the last page */
    go_to_page(last_page_state_);
}

/*
 *   Switch to a page 
 */
void CHtmlSys_mainwin::go_to_page(CHtmlSys_mainwin_state_link *new_page)
{
    /* if there's no new page, ignore the request */
    if (new_page == 0)
        return;

    /* 
     *   make sure the current page is fully flushed, if we're on the
     *   active page 
     */
    if (cur_page_state_->next_ == 0)
        flush_txtbuf(TRUE);

    /* save the current scrolling position */
    cur_page_state_->scrollpos_.x = main_panel_->get_hscroll_pos();
    cur_page_state_->scrollpos_.y = main_panel_->get_vscroll_pos();

    /* save the current page */
    parser_->export_parse_tree(get_cur_page_state());

    /* move to the new page */
    cur_page_state_ = new_page;

    /* import the new page into the parser */
    parser_->import_parse_tree(get_cur_page_state());

    /* make sure the main panel is scrolled to the top */
    main_panel_->do_scroll(TRUE, main_panel_->get_vscroll_handle(),
                           SB_TOP, 0, FALSE);

    /* reset the "more" prompt location on the main page */
    main_panel_->reset_last_input_height();

    /* reformat the window for the new page */
    main_panel_->remove_all_banners();
    reformat_all_html(TRUE, TRUE, TRUE);

    /* if this is the active page, show the caret */
    if (is_active_page())
        main_panel_->update_caret_pos(TRUE, TRUE);

    /* reset the selection range in all panels */
    clear_other_sel(0);

    /*
     *   Now scroll the main panel to the scrolling position that was in
     *   effect when the page was last viewed.
     */
    main_panel_->do_scroll(TRUE, main_panel_->get_vscroll_handle(),
                           SB_THUMBPOSITION,
                           cur_page_state_->scrollpos_.y, TRUE);
    main_panel_->do_scroll(FALSE, main_panel_->get_hscroll_handle(),
                           SB_THUMBPOSITION,
                           cur_page_state_->scrollpos_.x, TRUE);
}

/*
 *   Clear the selection ranges in all subwindows; optionally excludes a
 *   given subwindow. 
 */
void CHtmlSys_mainwin::clear_other_sel(CHtmlSysWin_win32 *exclude_win)
{
    int i;
    CHtmlSysWin_win32 **banner;
    
    /* clear the selection in the main panel if it's not excluded */
    if (main_panel_ != 0 && main_panel_ != exclude_win)
        main_panel_->clear_sel_range();

    /* clear the selection in each banner subwindow */
    for (i = 0, banner = banners_ ; i < banner_cnt_ ; ++i, ++banner)
    {
        /* if this subwindow isn't excluded, clear its selection */
        if (*banner != exclude_win)
            (*banner)->clear_sel_range();
    }
}

/*
 *   Clear hover information from all subwindows; optionally excludes a
 *   given subwindow. 
 */
void CHtmlSys_mainwin::clear_other_hover_info(CHtmlSysWin_win32 *exclude_win)
{
    int i;
    CHtmlSysWin_win32 **banner;

    /* clear the selection in the main panel if it's not excluded */
    if (main_panel_ != 0 && main_panel_ != exclude_win)
        main_panel_->clear_hover_info();

    /* clear the selection in each banner subwindow */
    for (i = 0, banner = banners_ ; i < banner_cnt_ ; ++i, ++banner)
    {
        /* if this subwindow isn't excluded, clear its selection */
        if (*banner != exclude_win)
            (*banner)->clear_hover_info();
    }
}

/*
 *   Display output 
 */
void CHtmlSys_mainwin::display_output(const textchar_t *buf)
{
    /* add the text to the buffer we're accumulating */
    txtbuf_->append(buf, get_strlen(buf));
}


/*
 *   Close the owner window 
 */
int CHtmlSys_mainwin::close_owner_window(int force)
{
    /* 
     *   if forcing, destroy the window without asking the user anything;
     *   otherwise, send a normal close message so that the user has a
     *   chance to object 
     */
    if (force)
        return !DestroyWindow(handle_);
    else
        return !SendMessage(handle_, WM_CLOSE, 0, 0);
}

/*
 *   Bring the window to the front 
 */
void CHtmlSys_mainwin::bring_owner_to_front()
{
    BringWindowToTop(handle_);
}

/*
 *   Create a banner subwindow.  We'll create a peer of the original main
 *   panel, and add it to our panel list.  
 */
CHtmlSysWin_win32 *CHtmlSys_mainwin::
   create_banner_subwin(CHtmlFormatter *formatter, int has_border,
                        HTML_BannerWin_Pos_t pos)
{
    CHtmlSysWin_win32 *subwin;
    RECT rc;
    
    /* create a new HTML window */
    subwin = new CHtmlSysWin_win32(formatter, this, FALSE, FALSE, prefs_);

    /* let the window know it's being used as a banner */
    subwin->set_is_banner_win(TRUE, pos);

    /* set the border as desired by the caller */
    subwin->set_has_border(has_border);

    /* 
     *   create the system window - give it an arbitrary initial height,
     *   since this will be changed by the formatter anyway, but give it
     *   the correct initial width 
     */
    GetClientRect(handle_, &rc);
    rc.right -= 2;
    rc.bottom = rc.top + 20;
    subwin->create_system_window(this, TRUE, "", &rc);

    /* add it to our panel list */
    add_banner_win(subwin);

    /* set the banner to use our status line */
    subwin->set_status_line(statusline_);

    /* return it */
    return subwin;
}

/*
 *   Create a subwindow for the "about" box.
 */
CHtmlSysWin_win32 *CHtmlSys_mainwin::
   create_aboutbox_subwin(CHtmlFormatter *formatter)
{
    /* create (or replace) the HTML subwindow in the dialog window */
    aboutbox_->create_html_subwin(formatter);
    
    /* return the subwindow */
    return aboutbox_->get_html_subwin();
}


void CHtmlSys_mainwin::remove_banner_subwin(CHtmlSysWin *subwin)
{
    RECT mainrc;
    CHtmlSysWin_win32 *syswin = (CHtmlSysWin_win32 *)subwin;

    /* if this is the about box, handle it specially */
    if (syswin == aboutbox_->get_html_subwin())
    {
        aboutbox_->delete_html_subwin();
        return;
    }
    
    /* remove the window from our panel list */
    remove_banner_win(syswin);

    /* close the window */
    SendMessage(syswin->get_handle(), WM_CLOSE, 0, 0);

    /* refigure all of our panel sizes */
    GetClientRect(handle_, &mainrc);
    do_resize(SIZE_RESTORED, mainrc.right, mainrc.bottom);
}

/*
 *   Set the size of a banner subwindow.  If the window isn't a banner,
 *   we'll ignore this call; otherwise, we'll set its width or height as
 *   appropriate for the banner's alignment, and adjust the other window
 *   positions accordingly.  
 */
void CHtmlSys_mainwin::set_banner_size(CHtmlSysWin_win32 *banner,
                                       long width, int use_width,
                                       long height, int use_height)
{
    CHtmlSysWin_win32 **cur;
    int i;

    /* find it in our list */
    for (cur = banners_, i = 0 ; i < banner_cnt_ ; ++i, ++cur)
    {
        /* see if this is the one we're looking for */
        if (*cur == banner)
        {
            RECT mainrc;

            /* get the size of the main window, for the width */
            GetClientRect(handle_, &mainrc);

            /* check the banner's alignment */
            switch(banner->get_banner_alignment())
            {
            case HTML_BANNERWIN_POS_TOP:
            case HTML_BANNERWIN_POS_BOTTOM:
                /* if we're to ignore the height, there's nothing to do */
                if (!use_height)
                    return;
                
                /* 
                 *   horizontal banner - set only its height, since its
                 *   width is determined by the size of the input window 
                 */
                SetWindowPos(banner->get_handle(), 0, 0, 0,
                             mainrc.right - 2, height,
                             SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
                break;
                
            case HTML_BANNERWIN_POS_LEFT:
            case HTML_BANNERWIN_POS_RIGHT:
                /* if we're to ignore the width, there's nothing to do */
                if (!use_width)
                    return;
                
                /* 
                 *   vertical banner - set only its width, since its
                 *   height is determined by the size of the input window 
                 */
                SetWindowPos(banner->get_handle(), 0, 0, 0,
                             width, mainrc.bottom - 2,
                             SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
                break;
            }

            /* tell the main panel to reformat for the resize */
            do_resize(SIZE_RESTORED, mainrc.right, mainrc.bottom);
            
            /* done */
            return;
        }
    }
}

/*
 *   Reformat all HTML windows 
 */
void CHtmlSys_mainwin::reformat_all_html(int show_status, int freeze_display,
                                         int reset_sounds)
{
    CHtmlSysWin_win32 **cur;
    int i;

    /* reformat the main panel window */
    main_panel_->do_reformat(show_status, freeze_display, reset_sounds);

    /* reformat each banner window */
    for (i = 0, cur = banners_ ; i < banner_cnt_ ; ++i, ++cur)
        (*cur)->do_reformat(show_status, freeze_display, FALSE);
}

/*
 *   Set the accelerator key for the Edit/Paste menu item 
 */
void CHtmlSys_mainwin::set_paste_accel(textchar_t paste_key)
{
    HMENU edit_menu;
    textchar_t buf[128];
    textchar_t *p;
    
    /* get the old menu label */
    edit_menu = GetSubMenu(GetMenu(handle_), 1);
    GetMenuString(edit_menu, ID_EDIT_PASTE, buf, sizeof(buf), MF_BYCOMMAND);

    /* find the tab */
    for (p = buf ; *p != '\0' && *p != '\t' ; ++p) ;
    if (*p == '\t')
    {
        /* find the '+' */
        for ( ; *p != '\0' && *p != '+' ; ++p);

        /* if we found the '+', set the next character */
        if (*p == '+')
        {
            /* set the character */
            *++p = paste_key;

            /* update the menu label */
            ModifyMenu(edit_menu, ID_EDIT_PASTE,
                       MF_BYCOMMAND | MF_STRING, ID_EDIT_PASTE, buf);
        }
    }
}

/*
 *   Set the accelerator key for the View Links submenu.  The F5 key moves
 *   around between turning links on and off.  Call this before updating
 *   the preferences.
 */
void CHtmlSys_mainwin::set_link_accel(int links_on, int links_ctrl)
{
    HMENU link_menu;
    int new_item;
    int i;

    /* get the link menu */
    link_menu = GetSubMenu(GetSubMenu(GetMenu(handle_), 2), 1);

    /*
     *   Determine which item now has the accelerator.  If links are to be
     *   off or in ctrl-mode, the accelerator will be on links-on, since
     *   we want the accelerator to toggle links back on.  If links are to
     *   be on, the accelerator goes in whichever one was previously set,
     *   since we want to toggle back to that item (if links were
     *   previously on, put the accelerator in the off item). 
     */
    if (!links_on)
    {
        /* links are off or in ctrl-mode - put the accelerator in links-on */
        new_item = 0;
    }
    else if (prefs_->get_links_on())
    {
        /* 
         *   links are to be on, but they were already on - put the
         *   accelerator in links-off
         */
        new_item = 2;
    }
    else if (prefs_->get_links_ctrl())
    {
        /* 
         *   links are to be on, and we're currently in ctrl mode, so put
         *   the accelerator in ctrl-mode to toggle back here 
         */
        new_item = 1;
    }
    else
    {
        /* 
         *   links are to be on, and they're currently off - put the
         *   accelerator in links-off 
         */
        new_item = 2;
    }

    /* 
     *   scan the items, removing the accelerator from the old one and
     *   adding it to the new one 
     */
    for (i = 0 ; i < 3 ; ++i)
    {
        textchar_t buf[128];
        textchar_t *p;
        int modify;
        MENUITEMINFO info;

        /* get this item's string */
        info.cbSize = sizeof(info);
        info.fMask = MIIM_TYPE;
        info.dwTypeData = buf;
        info.cch = sizeof(buf);
        GetMenuItemInfo(link_menu, i, TRUE, &info);

        /* find the tab or the end of the string */
        for (p = buf ; *p != '\0' && *p != '\t' ; ++p) ;

        /* presume we won't need to modify this item */
        modify = FALSE;

        /* 
         *   if we found a tab and this isn't the new destination, remove
         *   the accelerator; if it's the new item, add the accelerator 
         */
        if (new_item != i && *p == '\t')
        {
            /* remove the accelerator from this item */
            *p = '\0';
            modify = TRUE;
        }
        else if (new_item == i)
        {
            /* add the accelerator to this item */
            strcpy(p, "\tF5");
            modify = TRUE;
        }

        /* update the menu item if necessary */
        if (modify)
        {
            info.cch = get_strlen(buf);
            SetMenuItemInfo(link_menu, i, TRUE, &info);
        }
    }
}

/*
 *   Add a window to our list of banner windows 
 */
void CHtmlSys_mainwin::add_banner_win(CHtmlSysWin_win32 *win)
{
    /* reallocate our list if it's not big enough */
    if (banners_ == 0)
    {
        banners_max_ = 5;
        banners_ = (CHtmlSysWin_win32 **)
                   th_malloc(banners_max_ * sizeof(*banners_));
    }
    else if (banner_cnt_ == banners_max_)
    {
        banners_max_ += 5;
        banners_ = (CHtmlSysWin_win32 **)
                   th_realloc(banners_, banners_max_ * sizeof(*banners_));
    }

    /* add it to our list */
    banners_[banner_cnt_++] = win;
}

/*
 *   Remove a window from our banner list 
 */
void CHtmlSys_mainwin::remove_banner_win(CHtmlSysWin_win32 *win)
{
    CHtmlSysWin_win32 **cur;
    int i;
    
    /* find it in our list */
    for (cur = banners_, i = 0 ; i < banner_cnt_ ; ++i, ++cur)
    {
        /* see if this is the one we're looking for */
        if (*cur == win)
        {
            /* move the remaining ones down a notch */
            for ( ; i + 1 < banner_cnt_ ; ++i, ++cur)
                *cur = *(cur + 1);

            /* that's one less banner */
            --banner_cnt_;

            /* no need to look further */
            return;
        }
    }
}

/*
 *   Pause before exiting 
 */
void CHtmlSys_mainwin::pause_for_exit()
{
    MSG msg;
    textchar_t ch;
    
    /* turn on the exit flag in the main window */
    main_panel_->set_exit_pause(TRUE);

    /* flush output */
    flush_txtbuf(TRUE);

    /* clear out any keystrokes currently in the message queue */
    while (PeekMessage(&msg, 0, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE))
        ;

    /* handle events until we get a keystroke */
    main_panel_->wait_for_keystroke(&ch, TRUE, 0, FALSE);
}

/*
 *   Display the MORE prompt 
 */
void CHtmlSys_mainwin::pause_for_more()
{
    /* flush output */
    flush_txtbuf(TRUE);
    
    /* display the MORE prompt in the main window */
    main_panel_->do_game_more_prompt();
}

/*
 *   Display a debug message 
 */
void CHtmlSys_mainwin::dbg_print(const char *msg)
{
    /* if we have a debug window, display the message there */
    if (dbgwin_ != 0)
    {
        /*
         *   Send our private DBG_PRINT message to the debug window.  Use
         *   a message rather than calling the debug window's disp_text()
         *   method directly -- this allows us to print debug messages
         *   from background threads without fear of any concurrency
         *   problems.  (A sound playback thread, for example, may want to
         *   display a message if it encounters an error decoding its
         *   sound file.)  
         */
        SendMessage(dbgwin_->get_handle(), HTMLM_DBG_PRINT,
                    0, (LPARAM)msg);
    }
}

/*
 *   Get a keystroke 
 */
textchar_t CHtmlSys_mainwin::wait_for_keystroke(int pause_only)
{
    textchar_t ch;
    
    /* flush output */
    flush_txtbuf(TRUE);

    /* get a keystroke */
    main_panel_->wait_for_keystroke(&ch, pause_only, 0, FALSE);
    return ch;
}

/*
 *   Get a keystroke with a timeout 
 */
htmlsys_input_stat_t CHtmlSys_mainwin::
   wait_for_keystroke_timeout(textchar_t *ch, int pause_only, time_t timeout)
{
    /* flush output */
    flush_txtbuf(TRUE);

    /* get the keystroke */
    return main_panel_->wait_for_keystroke(ch, pause_only, timeout, TRUE);
}

/*
 *   Map an HTML 4 code point value 
 */
size_t CHtmlSys_mainwin::xlat_html4_entity(textchar_t *result,
                                           size_t result_size,
                                           unsigned int charval,
                                           oshtml_charset_id_t *charset,
                                           int *changed_charset)
{
    oshtml_charset_id_t new_charset;
    size_t len;
    
    /* presume we're going to return a single byte */
    result[1] = '\0';
    len = 1;

    /* check the character to see what range it's in */
    if (charval < 128)
    {
        /* 
         *   it's in the ASCII range, so assume that it's part of the
         *   ASCII subset of every Windows character set -- return the
         *   value unchanged and in the current default character set 
         */
        result[0] = (textchar_t)charval;
        new_charset = default_charset_;
    }
    else if (charval >= 160 && charval <= 255)
    {
        /*
         *   It's in the 160-255 range, which is the ISO Latin-1 range for
         *   HTML 4.  These map trivially to Windows code page 1252
         *   characters, so leave the value unchanged and use the ANSI
         *   character set.  
         *   
         *   Make sure we can return a value in the right code page - if
         *   we can set the character set, or we're already using
         *   ANSI_CHARSET by default, we can return the value 
         */
        if (charset != 0 || default_charset_ == ANSI_CHARSET)
        {
            /* set the result */
            result[0] = (textchar_t)charval;
            new_charset = ANSI_CHARSET;
        }
        else
        {
            /* other mappings are not possible */
            result[0] = invalid_char_val_;
            new_charset = default_charset_;
        }

        return 1;
    }
    else
    {
        static struct
        {
            unsigned int html_val;
            unsigned int win_val;
            int          win_charset;
        } mapping[] =
        {
            /*
             *   IMPORTANT: This table must be sorted by html_val, because
             *   we perform a binary search on this key.  
             */

            /* 
             *   Latin-2 characters and additional alphabetic characters.
             *   Most of these characters exist only in the Windows
             *   Eastern European code page, but a few are in the common
             *   Windows extended code page set (code points 128-159) and
             *   a few are only in the ANSI code page.  
             */
            { 258, 0xC3, EASTEUROPE_CHARSET },                    /* Abreve */
            { 259, 0xE3, EASTEUROPE_CHARSET },                    /* abreve */
            { 260, 0xA5, EASTEUROPE_CHARSET },                     /* Aogon */
            { 261, 0xB9, EASTEUROPE_CHARSET },                     /* aogon */
            { 262, 0xC6, EASTEUROPE_CHARSET },                    /* Cacute */
            { 263, 0xE6, EASTEUROPE_CHARSET },                    /* cacute */
            { 268, 0xC8, EASTEUROPE_CHARSET },                    /* Ccaron */
            { 269, 0xE8, EASTEUROPE_CHARSET },                    /* ccaron */
            { 270, 0xCF, EASTEUROPE_CHARSET },                    /* Dcaron */
            { 271, 0xEF, EASTEUROPE_CHARSET },                    /* dcaron */
            { 272, 0xD0, EASTEUROPE_CHARSET },                    /* Dstrok */
            { 273, 0xF0, EASTEUROPE_CHARSET },                    /* dstrok */
            { 280, 0xCA, EASTEUROPE_CHARSET },                     /* Eogon */
            { 281, 0xEA, EASTEUROPE_CHARSET },                     /* eogon */
            { 282, 0xCC, EASTEUROPE_CHARSET },                    /* Ecaron */
            { 283, 0xEC, EASTEUROPE_CHARSET },                    /* ecaron */
            { 313, 0xC5, EASTEUROPE_CHARSET },                    /* Lacute */
            { 314, 0xE5, EASTEUROPE_CHARSET },                    /* lacute */
            { 317, 0xBC, EASTEUROPE_CHARSET },                    /* Lcaron */
            { 318, 0xBE, EASTEUROPE_CHARSET },                    /* lcaron */
            { 321, 0xA3, EASTEUROPE_CHARSET },                    /* Lstrok */
            { 322, 0xB3, EASTEUROPE_CHARSET },                    /* lstrok */
            { 323, 0xD1, EASTEUROPE_CHARSET },                    /* Nacute */
            { 324, 0xF1, EASTEUROPE_CHARSET },                    /* nacute */
            { 327, 0xD2, EASTEUROPE_CHARSET },                    /* Ncaron */
            { 328, 0xF2, EASTEUROPE_CHARSET },                    /* ncaron */
            { 336, 0xD5, EASTEUROPE_CHARSET },                    /* Odblac */
            { 337, 0xF5, EASTEUROPE_CHARSET },                    /* odblac */
            { 338,  140, ANSI_CHARSET },                           /* OElig */
            { 339,  156, ANSI_CHARSET },                           /* oelig */
            { 340, 0xC0, EASTEUROPE_CHARSET },                    /* Racute */
            { 341, 0xE0, EASTEUROPE_CHARSET },                    /* racute */
            { 344, 0xD8, EASTEUROPE_CHARSET },                    /* Rcaron */
            { 345, 0xF8, EASTEUROPE_CHARSET },                    /* rcaron */
            { 346, 0x8C, EASTEUROPE_CHARSET },                    /* Sacute */
            { 347, 0x9C, EASTEUROPE_CHARSET },                    /* sacute */
            { 350, 0xAA, EASTEUROPE_CHARSET },                    /* Scedil */
            { 351, 0xBA, EASTEUROPE_CHARSET },                    /* scedil */
            { 352,  138, DEFAULT_CHARSET },                       /* Scaron */
            { 353,  154, DEFAULT_CHARSET },                       /* scaron */
            { 354, 0xDE, EASTEUROPE_CHARSET },                    /* Tcedil */
            { 355, 0xFE, EASTEUROPE_CHARSET },                    /* tcedil */
            { 356, 0x8D, EASTEUROPE_CHARSET },                    /* Tcaron */
            { 357, 0x9D, EASTEUROPE_CHARSET },                    /* tcaron */
            { 366, 0xD9, EASTEUROPE_CHARSET },                     /* Uring */
            { 367, 0xF9, EASTEUROPE_CHARSET },                     /* uring */
            { 368, 0xDB, EASTEUROPE_CHARSET },                    /* Udblac */
            { 369, 0xFB, EASTEUROPE_CHARSET },                    /* udblac */
            { 376,  159, ANSI_CHARSET },                            /* Yuml */
            { 377, 0x8F, EASTEUROPE_CHARSET },                    /* Zacute */
            { 378, 0x9F, EASTEUROPE_CHARSET },                    /* zacute */
            { 379, 0xAF, EASTEUROPE_CHARSET },                      /* Zdot */
            { 380, 0xBF, EASTEUROPE_CHARSET },                      /* zdot */
            { 381, 0x8E, EASTEUROPE_CHARSET },                    /* Zcaron */
            { 382, 0x9E, EASTEUROPE_CHARSET },                    /* zcaron */

            /* Latin-extended B */
            { 402, 166, SYMBOL_CHARSET },                           /* fnof */

            /* Latin-2 accents */
            { 710, 136, ANSI_CHARSET },                             /* circ */
            { 711, 0xA1, EASTEUROPE_CHARSET },                     /* caron */
            { 728, 0xA2, EASTEUROPE_CHARSET },                     /* breve */
            { 729, 0xFF, EASTEUROPE_CHARSET },                       /* dot */
            { 731, 0xB2, EASTEUROPE_CHARSET },                      /* ogon */
            { 732, 152, ANSI_CHARSET },                            /* tilde */
            { 733, 0xBD, EASTEUROPE_CHARSET },                     /* dblac */

            /* Greek letters */
            { 913, 'A', SYMBOL_CHARSET },                          /* Alpha */
            { 914, 'B', SYMBOL_CHARSET },                           /* Beta */
            { 915, 'G', SYMBOL_CHARSET },                          /* Gamma */
            { 916, 'D', SYMBOL_CHARSET },                          /* Delta */
            { 917, 'E', SYMBOL_CHARSET },                        /* Epsilon */
            { 918, 'Z', SYMBOL_CHARSET },                           /* Zeta */
            { 919, 'H', SYMBOL_CHARSET },                            /* Eta */
            { 920, 'Q', SYMBOL_CHARSET },                          /* Theta */
            { 921, 'I', SYMBOL_CHARSET },                           /* Iota */
            { 922, 'K', SYMBOL_CHARSET },                          /* Kappa */
            { 923, 'L', SYMBOL_CHARSET },                         /* Lambda */
            { 924, 'M', SYMBOL_CHARSET },                             /* Mu */
            { 925, 'N', SYMBOL_CHARSET },                             /* Nu */
            { 926, 'X', SYMBOL_CHARSET },                             /* Xi */
            { 927, 'O', SYMBOL_CHARSET },                        /* Omicron */
            { 928, 'P', SYMBOL_CHARSET },                             /* Pi */
            { 929, 'R', SYMBOL_CHARSET },                            /* Rho */
            { 931, 'S', SYMBOL_CHARSET },                          /* Sigma */
            { 932, 'T', SYMBOL_CHARSET },                            /* Tau */
            { 933, 'U', SYMBOL_CHARSET },                        /* Upsilon */
            { 934, 'F', SYMBOL_CHARSET },                            /* Phi */
            { 935, 'C', SYMBOL_CHARSET },                            /* Chi */
            { 936, 'Y', SYMBOL_CHARSET },                            /* Psi */
            { 937, 'W', SYMBOL_CHARSET },                          /* Omega */
            { 945, 'a', SYMBOL_CHARSET },                          /* alpha */
            { 946, 'b', SYMBOL_CHARSET },                           /* beta */
            { 947, 'g', SYMBOL_CHARSET },                          /* gamma */
            { 948, 'd', SYMBOL_CHARSET },                          /* delta */
            { 949, 'e', SYMBOL_CHARSET },                        /* epsilon */
            { 950, 'z', SYMBOL_CHARSET },                           /* zeta */
            { 951, 'h', SYMBOL_CHARSET },                            /* eta */
            { 952, 'q', SYMBOL_CHARSET },                          /* theta */
            { 953, 'i', SYMBOL_CHARSET },                           /* iota */
            { 954, 'k', SYMBOL_CHARSET },                          /* kappa */
            { 955, 'l', SYMBOL_CHARSET },                         /* lambda */
            { 956, 'm', SYMBOL_CHARSET },                             /* mu */
            { 957, 'n', SYMBOL_CHARSET },                             /* nu */
            { 958, 'x', SYMBOL_CHARSET },                             /* xi */
            { 959, 'o', SYMBOL_CHARSET },                        /* omicron */
            { 960, 'p', SYMBOL_CHARSET },                             /* pi */
            { 961, 'r', SYMBOL_CHARSET },                            /* rho */
            { 962, 'V', SYMBOL_CHARSET },                         /* sigmaf */
            { 963, 's', SYMBOL_CHARSET },                          /* sigma */
            { 964, 't', SYMBOL_CHARSET },                            /* tau */
            { 965, 'u', SYMBOL_CHARSET },                        /* upsilon */
            { 966, 'f', SYMBOL_CHARSET },                            /* phi */
            { 967, 'c', SYMBOL_CHARSET },                            /* chi */
            { 968, 'y', SYMBOL_CHARSET },                            /* psi */
            { 969, 'w', SYMBOL_CHARSET },                          /* omega */
            { 977, 'J', SYMBOL_CHARSET },                       /* thetasym */
            { 978, 161, SYMBOL_CHARSET },                          /* upsih */
            { 982, 'v', SYMBOL_CHARSET },                            /* piv */

            /* additional punctuation */
            { 8211, 150, DEFAULT_CHARSET },                        /* ndash */
            { 8212, 151, DEFAULT_CHARSET },                        /* mdash */
            { 8216, 145, DEFAULT_CHARSET },                        /* lsquo */
            { 8217, 146, DEFAULT_CHARSET },                        /* rsquo */
            { 8218, 130, DEFAULT_CHARSET },                        /* sbquo */
            { 8220, 147, DEFAULT_CHARSET },                        /* ldquo */
            { 8221, 148, DEFAULT_CHARSET },                        /* rdquo */
            { 8222, 132, DEFAULT_CHARSET },                        /* bdquo */
            { 8224, 134, DEFAULT_CHARSET },                       /* dagger */
            { 8225, 135, DEFAULT_CHARSET },                       /* Dagger */
            { 8226, 149, DEFAULT_CHARSET },                         /* bull */
            { 8230, 133, DEFAULT_CHARSET },                       /* hellip */
            { 8240, 137, DEFAULT_CHARSET },                       /* permil */
            { 8242, 162, SYMBOL_CHARSET },                         /* prime */
            { 8243, 178, SYMBOL_CHARSET },                         /* Prime */
            { 8249, 139, DEFAULT_CHARSET },                       /* lsaquo */
            { 8250, 155, DEFAULT_CHARSET },                       /* rsaquo */
            { 8254, '`', SYMBOL_CHARSET },                         /* oline */
            { 8260, 164, SYMBOL_CHARSET },                         /* frasl */

            /* letter-like symbols */
            { 8465, 193, SYMBOL_CHARSET },                         /* image */
            { 8472, 195, SYMBOL_CHARSET },                        /* weierp */
            { 8476, 194, SYMBOL_CHARSET },                          /* real */
            { 8482, 153, DEFAULT_CHARSET },                        /* trade */
            { 8501, 192, SYMBOL_CHARSET },                       /* alefsym */
            
            /* arrows */
            { 8592, 172, SYMBOL_CHARSET },                          /* larr */
            { 8593, 173, SYMBOL_CHARSET },                          /* uarr */
            { 8594, 174, SYMBOL_CHARSET },                          /* rarr */
            { 8595, 175, SYMBOL_CHARSET },                          /* darr */
            { 8596, 171, SYMBOL_CHARSET },                          /* harr */
            { 8629, 191, SYMBOL_CHARSET },                         /* crarr */
            { 8656, 220, SYMBOL_CHARSET },                          /* lArr */
            { 8657, 221, SYMBOL_CHARSET },                          /* uArr */
            { 8658, 222, SYMBOL_CHARSET },                          /* rArr */
            { 8659, 223, SYMBOL_CHARSET },                          /* dArr */
            { 8660, 219, SYMBOL_CHARSET },                          /* hArr */
            
            /* mathematical operators */
            { 8704, 34, SYMBOL_CHARSET },                         /* forall */
            { 8706, 182, SYMBOL_CHARSET },                          /* part */
            { 8707, '$', SYMBOL_CHARSET },                         /* exist */
            { 8709, 198, SYMBOL_CHARSET },                         /* empty */
            { 8711, 209, SYMBOL_CHARSET },                         /* nabla */
            { 8712, 206, SYMBOL_CHARSET },                          /* isin */
            { 8713, 207, SYMBOL_CHARSET },                         /* notin */
            { 8715, '\'', SYMBOL_CHARSET },                           /* ni */
            { 8719, 213, SYMBOL_CHARSET },                          /* prod */
            { 8721, 229, SYMBOL_CHARSET },                           /* sum */
            { 8722, '-', SYMBOL_CHARSET },                         /* minus */
            { 8727, '*', SYMBOL_CHARSET },                        /* lowast */
            { 8730, 214, SYMBOL_CHARSET },                         /* radic */
            { 8733, 181, SYMBOL_CHARSET },                          /* prop */
            { 8734, 165, SYMBOL_CHARSET },                         /* infin */
            { 8736, 208, SYMBOL_CHARSET },                           /* ang */
            { 8743, 217, SYMBOL_CHARSET },                           /* and */
            { 8744, 218, SYMBOL_CHARSET },                            /* or */
            { 8745, 199, SYMBOL_CHARSET },                           /* cap */
            { 8746, 200, SYMBOL_CHARSET },                           /* cup */
            { 8747, 242, SYMBOL_CHARSET },                           /* int */
            { 8756, '\\', SYMBOL_CHARSET },                       /* there4 */
            { 8764, '~', SYMBOL_CHARSET },                           /* sim */
            { 8773, '@', SYMBOL_CHARSET },                          /* cong */
            { 8776, 187, SYMBOL_CHARSET },                         /* asymp */
            { 8800, 185, SYMBOL_CHARSET },                            /* ne */
            { 8801, 186, SYMBOL_CHARSET },                         /* equiv */
            { 8804, 163, SYMBOL_CHARSET },                            /* le */
            { 8805, 179, SYMBOL_CHARSET },                            /* ge */
            { 8834, 204, SYMBOL_CHARSET },                           /* sub */
            { 8835, 201, SYMBOL_CHARSET },                           /* sup */
            { 8836, 203, SYMBOL_CHARSET },                          /* nsub */
            { 8838, 205, SYMBOL_CHARSET },                          /* sube */
            { 8839, 202, SYMBOL_CHARSET },                          /* supe */
            { 8853, 197, SYMBOL_CHARSET },                         /* oplus */
            { 8855, 196, SYMBOL_CHARSET },                        /* otimes */
            { 8869, '^', SYMBOL_CHARSET },                          /* perp */
            { 8901, 215, SYMBOL_CHARSET },                          /* sdot */
            { 8968, 233, SYMBOL_CHARSET },                         /* lceil */
            { 8969, 249, SYMBOL_CHARSET },                         /* rceil */
            { 8970, 235, SYMBOL_CHARSET },                        /* lfloor */
            { 8971, 251, SYMBOL_CHARSET },                        /* rfloor */
            { 9001, 225, SYMBOL_CHARSET },                          /* lang */
            { 9002, 241, SYMBOL_CHARSET },                          /* rang */

            /* geometric shapes */
            { 9674, 224, SYMBOL_CHARSET },                           /* loz */

            /* misceallaneous symbols */
            { 9824, 170, SYMBOL_CHARSET },                        /* spades */
            { 9827, 167, SYMBOL_CHARSET },                         /* clubs */
            { 9829, 169, SYMBOL_CHARSET },                        /* hearts */
            { 9830, 168, SYMBOL_CHARSET }                       /* diamonds */
        };
        int hi, lo;

        /*
         *   Find the character in the mapping table 
         */
        lo = 0;
        hi = sizeof(mapping)/sizeof(mapping[0]) - 1;
        for (;;)
        {
            int cur;
            
            /* if there's nothing left in the table, we didn't find it */
            if (lo > hi)
            {
                /* use the invalid character in the default character set */
                result[0] = invalid_char_val_;
                new_charset = default_charset_;
                break;
            }

            /* split the difference */
            cur = lo + (hi - lo)/2;

            /* is this the one we're looking for? */
            if (mapping[cur].html_val == charval)
            {
                /* this is it */
                result[0] = (textchar_t)mapping[cur].win_val;
                new_charset = mapping[cur].win_charset;

                /* 
                 *   if the character set to use is the default, use the
                 *   current system default 
                 */
                if (new_charset == DEFAULT_CHARSET)
                    new_charset = default_charset_;

                /*
                 *   If the character set is not the default, and we're
                 *   not allowed to change character sets, return the
                 *   invalid character. 
                 */
                if (new_charset != DEFAULT_CHARSET && charset == 0)
                {
                    result[0] = invalid_char_val_;
                    new_charset = default_charset_;
                }

                /* done */
                break;
            }
            else if (mapping[cur].html_val < charval)
            {
                /* we want something higher - look in the upper half */
                lo = (cur == lo ? lo + 1 : cur);
            }
            else
            {
                /* we want something smaller - look in the lower half */
                hi = (cur == hi ? hi - 1 : cur);
            }
        }
    }

    /* if they want to know about the character set, return the info */
    if (charset != 0)
    {
        /* tell them the new character set */
        *charset = new_charset;

        /* tell them whether it's the same as the default character set */
        *changed_charset = (new_charset != default_charset_);
    }

    /* return the number of bytes in the result */
    return len;
}

/* ------------------------------------------------------------------------ */
/*
 *   Parser state object link structure implementation
 */

CHtmlSys_mainwin_state_link::~CHtmlSys_mainwin_state_link()
{
    /* delete our parser state object */
    delete state_;
}

/* ------------------------------------------------------------------------ */
/*
 *   Debug window implementation 
 */

CHtmlSys_dbgwin::CHtmlSys_dbgwin(CHtmlParser *parser,
                                 CHtmlPreferences *prefs)
{
    /* remember my parser */
    parser_ = parser;

    /* remember my preferences */
    prefs_ = prefs;
}

/*
 *   initialize the HTML panel
 */
void CHtmlSys_dbgwin::init_html_panel(CHtmlFormatter *formatter)
{
    /* initialize the parser */
    init_parser();

    /* create the HTML display window */
    html_panel_ = new CHtmlSysWin_win32(formatter, this, TRUE, TRUE, prefs_);
}

CHtmlSys_dbgwin::~CHtmlSys_dbgwin()
{
}

/*
 *   process creation event 
 */
void CHtmlSys_dbgwin::do_create()
{
    RECT pos;

    /* inherit default */
    CHtmlSys_framewin::do_create();
    
    /* create the main panel's system window */
    GetClientRect(handle_, &pos);
    html_panel_->create_system_window(this, TRUE, "", &pos);

    /* load my menu */
    load_menu();
}

/*
 *   destroy the window 
 */
void CHtmlSys_dbgwin::do_destroy()
{
    /* tell my child window to disengage from its parent */
    html_panel_->on_destroy_parent();

    /* forget my subwindow */
    html_panel_ = 0;

    /* inherit default */
    CTadsWin::do_destroy();
}

void CHtmlSys_dbgwin::do_resize(int mode, int x, int y)
{
    switch(mode)
    {
    case SIZE_MAXHIDE:
    case SIZE_MAXSHOW:
    case SIZE_MINIMIZED:
        /* don't bother resizing the subwindow on these changes */
        break;

    case SIZE_MAXIMIZED:
    case SIZE_RESTORED:
    default:
        {
            RECT rc;

            /* compute the area available for the subwindow */
            SetRect(&rc, 1, 1, x, y);

            /* place the main panel */
            if (html_panel_ != 0)
                MoveWindow(html_panel_->get_handle(),
                           rc.left, rc.top,
                           rc.right - rc.left, rc.bottom - rc.top, TRUE);
            
            /* inherit default handling to save position in prefs */
            CHtmlSys_framewin::do_resize(mode, x, y);
        }
        break;
    }
}

/*
 *   Close the owner window 
 */
int CHtmlSys_dbgwin::close_owner_window(int force)
{
    /* 
     *   if forcing, destroy the window without asking the user anything;
     *   otherwise, send a normal close message so that the user has a
     *   chance to object 
     */
    if (force)
        return !DestroyWindow(handle_);
    else
        return !SendMessage(handle_, WM_CLOSE, 0, 0);
}

/*
 *   Bring the window to the front 
 */
void CHtmlSys_dbgwin::bring_owner_to_front()
{
    BringWindowToTop(handle_);
}

/*
 *   Gain focus - set focus to our main panel 
 */
void CHtmlSys_dbgwin::do_setfocus(HWND)
{
    /* move focus to our HTML panel */
    SetFocus(html_panel_->get_handle());
}


/* ------------------------------------------------------------------------ */
/*
 *   Debug log window implementation 
 */

CHtmlSys_dbglogwin::~CHtmlSys_dbglogwin()
{
    /* done with my text buffer */
    delete txtbuf_;
}

/*
 *   process destruction of the window 
 */
void CHtmlSys_dbglogwin::do_destroy()
{
    void (*cb)(void *, CHtmlSysWin *);
    void *ctx;
    CHtmlSysWin *subwin;

    /* 
     *   remember in local variables the information we'll need for the
     *   destruction callback (see below) 
     */
    cb = destroy_cb_;
    ctx = destroy_cb_ctx_;
    subwin = get_html_panel();
    
    /* inherit default */
    CHtmlSys_dbgwin::do_destroy();

    /* 
     *   if I'm the current debug log window for the main game window,
     *   tell the game window that I'm gone 
     */
    if (CHtmlSys_mainwin::get_main_win() != 0
        && CHtmlSys_mainwin::get_main_win()->get_debug_win() == this)
        CHtmlSys_mainwin::get_main_win()->set_debug_win(0, TRUE);
    
    /* 
     *   Invoke the destruction callback, if present.  Note that we cached
     *   the member variable values related to the callback in locals
     *   prior to inheriting the superclass do_destroy, since 'this' will
     *   probably have been deleted by the time that returns.  It's
     *   necessary to let 'this' be deleted prior to invoking the
     *   callback, since the callback will probably delete other objects
     *   (such as the parser and formatter) that we depend on as long as
     *   this object exists.  
     */
    if (cb != 0)
        (*cb)(ctx, subwin);
}

/*
 *   load my menu 
 */
void CHtmlSys_dbglogwin::load_menu()
{
    SetMenu(handle_, LoadMenu(CTadsApp::get_app()->get_instance(),
                              MAKEINTRESOURCE(IDR_DEBUGWIN_MENU)));
}

/*
 *   Process a close-window request.  Never actually close the debug
 *   window; merely hide it.  
 */
int CHtmlSys_dbglogwin::do_close()
{
    /* hide the window */
    ShowWindow(handle_, SW_HIDE);

    /* tell the caller not to really close the window */
    return FALSE;
}

/*
 *   Set up the parser 
 */
void CHtmlSys_dbglogwin::init_parser()
{
    static const textchar_t htmltxt[] = "<basefont size=4><listing>";

    /* create a text buffer */
    txtbuf_ = new CHtmlTextBuffer();

    /* 
     *   put the parser into <listing> mode, so that it displays in a
     *   monospaced font 
     */
    parser_->obey_markups(TRUE);
    txtbuf_->append(htmltxt, get_strlen(htmltxt));
    parser_->parse(txtbuf_, this);
    txtbuf_->clear();
    parser_->obey_markups(FALSE);
    parser_->obey_end_markups(FALSE);
}

/*
 *   process a user event 
 */
int CHtmlSys_dbglogwin::do_user_message(int msg, WPARAM wpar, LPARAM lpar)
{
    switch(msg)
    {
    case HTMLM_DBG_PRINT:
        /* display the text */
        disp_text((const textchar_t *)lpar);

        /* make sure the window is showing, and bring it to the front */
        ShowWindow(get_handle(), SW_SHOW);
        BringWindowToTop(get_handle());

        /* processed */
        return TRUE;

    default:
        /* inherit default behavior */
        return CHtmlSys_dbgwin::do_user_message(msg, wpar, lpar);
    }
}

/*
 *   Display text 
 */
void CHtmlSys_dbglogwin::disp_text(const textchar_t *msg)
{
    /* put the text in our buffer */
    txtbuf_->append(msg, get_strlen(msg));

    /* parse the text and display it */
    parser_->parse(txtbuf_, this);

    /* we're done with this text - clear the buffer */
    txtbuf_->clear();

    /* update the display */
    html_panel_->do_formatting(FALSE, FALSE, FALSE);
}

/*
 *   set my window position in the preferences 
 */
void CHtmlSys_dbglogwin::set_winpos_prefs(const CHtmlRect *winpos)
{
    prefs_->set_dbgwin_pos(winpos);
}

/*
 *   Handle a command 
 */
int CHtmlSys_dbglogwin::do_command(int notify_code,
                                   int command_id, HWND ctl)
{
    switch(command_id)
    {
    case ID_FILE_HIDE:
        /* hide the window */
        ShowWindow(get_handle(), SW_HIDE);
        return TRUE;

    default:
        /* pass the command to the main panel for processing */
        if (html_panel_ != 0)
            return html_panel_->do_command(notify_code, command_id, ctl);
        break;
    }

    /* didn't handle it */
    return FALSE;
}

/*
 *   Check the status of a command 
 */
TadsCmdStat_t CHtmlSys_dbglogwin::check_command(int command_id)
{
    switch(command_id)
    {
    case ID_FILE_HIDE:
        /* we can always hide the debug window */
        return TADSCMD_ENABLED;

    default:
        /* pass it to the main panel for processing */
        if (html_panel_ != 0)
            return html_panel_->check_command(command_id);
        break;
    }

    /* didn't handle it */
    return TADSCMD_UNKNOWN;
}

/* ------------------------------------------------------------------------ */
/*
 *   Dialog class for the main about box 
 */

void CTadsDlg_htabout::init()
{
    char buf[128];
    
    /* inherit default */
    CTadsDialog::init();

    /* set up the correct version number string */
    sprintf(buf, "Version %s", TADS_RUNTIME_VERSION);
    SetDlgItemText(handle_, IDC_TXT_ABOUTVSN, buf);
}


/* ------------------------------------------------------------------------ */
/*
 *   Window class for the "about this game" dialog 
 */

/* dimensions of the "OK" button in the dialog */
const int okbtn_wid = 70;
const int okbtn_ht = 20;

void CHtmlSys_aboutwin::do_create()
{
    RECT rc;
    CTadsLOGFONT tlf;
    CTadsFont *font;

    /* inherit default */
    CTadsWin::do_create();

    /* create the dismiss button */
    GetClientRect(handle_, &rc);
    okbtn_ = CreateWindow("BUTTON", "OK",
                          WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                          (rc.right - okbtn_wid)/2,
                          rc.bottom - okbtn_ht - 2,
                          okbtn_wid, okbtn_ht, handle_, (HMENU)IDOK,
                          CTadsApp::get_app()->get_instance(), 0);

    /* set the font on the button to MS Sans Serif 8 */
    tlf.lf.lfWidth = 0;
    tlf.lf.lfEscapement = 0;
    tlf.lf.lfOrientation = 0;
    tlf.lf.lfCharSet = ANSI_CHARSET;
    tlf.lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
    tlf.lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    tlf.lf.lfQuality = DEFAULT_QUALITY;
    strcpy(tlf.lf.lfFaceName, "MS Sans Serif");
    tlf.lf.lfHeight = CTadsFont::calc_lfHeight(8);
    tlf.lf.lfItalic = 0;
    tlf.lf.lfUnderline = 0;
    tlf.lf.lfStrikeOut = 0;
    tlf.superscript = 0;
    tlf.subscript = 0;
    tlf.color = RGB(0, 0, 0);
    tlf.lf.lfWeight = FW_NORMAL;
    tlf.lf.lfPitchAndFamily = FF_SWISS | VARIABLE_PITCH;
    font = CTadsApp::get_app()->get_font(&tlf, 0, 0);
    PostMessage(okbtn_, WM_SETFONT, (WPARAM)font->get_handle(),
                MAKELPARAM(TRUE, 0));

    /* no subpanel yet */
    html_subwin_ = 0;
}

void CHtmlSys_aboutwin::do_destroy()
{
    /* notify the subwindow that I'm being destroyed */
    if (html_subwin_ != 0)
        html_subwin_->on_destroy_parent();

    /* forget the subwindow */
    html_subwin_ = 0;
    
    /* inherit default */
    CTadsWin::do_destroy();
}

/*
 *   Create the panel subwindow 
 */
void CHtmlSys_aboutwin::create_html_subwin(CHtmlFormatter *formatter)
{
    RECT rc;

    /* if we already have a subpanel window, destroy it */
    if (html_subwin_ != 0)
    {
        html_subwin_->on_destroy_parent();
        DestroyWindow(html_subwin_->get_handle());
    }
    
    /* create the HTML subpanel */
    html_subwin_ = new CHtmlSysWin_win32(formatter, this,
                                         FALSE, FALSE, prefs_);

    /* create the panel's system window */
    GetClientRect(handle_, &rc);
    rc.bottom -= okbtn_ht + 2;
    html_subwin_->create_system_window(this, TRUE, "", &rc);
}

/*
 *   Delete the subwindow 
 */
void CHtmlSys_aboutwin::delete_html_subwin()
{
    /* if we have a subpanel window, destroy it */
    if (html_subwin_ != 0)
    {
        /* notify it of descruction */
        html_subwin_->on_destroy_parent();

        /* destroy it */
        DestroyWindow(html_subwin_->get_handle());

        /* forget it */
        html_subwin_ = 0;
    }
}


/*
 *   Run the dialog 
 */
void CHtmlSys_aboutwin::run_aboutbox(CTadsWin *owner)
{
    void *ctx;
    RECT drc;
    RECT rc;
    RECT wrc;
    int wid, ht;
    CHtmlRect text_area;

    /* 
     *   disable windows owned by the main window while the dialog is
     *   running 
     */
    ctx = CTadsDialog::modal_dlg_pre(owner->get_handle(), TRUE);

    /* get my size */
    GetClientRect(handle_, &rc);
    wid = rc.right - rc.left;
    ht = rc.bottom - rc.top;

    /* 
     *   increase the window's height if necessary to make room for its
     *   contents 
     */
    if (ht < (int)html_subwin_->get_formatter()->get_max_y_pos() + 30)
        ht = (int)html_subwin_->get_formatter()->get_max_y_pos() + 30;

    /* adjust the height and width to include the non-client areas */
    GetWindowRect(handle_, &wrc);
    wid += (wrc.right - wrc.left) - (rc.right - rc.left);
    ht += (wrc.bottom - wrc.top) - (rc.bottom - rc.top);

    /* center my window on the screen */
    GetWindowRect(GetDesktopWindow(), &drc);
    MoveWindow(handle_, drc.left + (drc.right - drc.left - wid)/2,
               drc.top + (drc.bottom - drc.top - ht)/2, wid, ht, TRUE);

    /* 
     *   show the window and enable it - it will have been disabled by the
     *   modal_dlg_pre setup above, so we need to re-enable it now 
     */
    ShowWindow(handle_, SW_SHOW);
    EnableWindow(handle_, TRUE);

    /* enter a recursive event loop until the window is closed */
    CTadsApp::get_app()->event_loop(&closing_, 0, 0);

    /* re-enable the windows we disabled before running the dialog */
    CTadsDialog::modal_dlg_post(ctx);

    /* hide myself again */
    ShowWindow(handle_, SW_HIDE);
}

/*
 *   close the window - we won't do anything except set the "closing"
 *   flag, which will dismiss the window; we won't allow the window to
 *   actually close at this point 
 */
int CHtmlSys_aboutwin::do_close()
{
    /* set the flag to indicate the dialog has been dismissed */
    closing_ = TRUE;

    /* don't really close yet */
    return FALSE;
}

/*
 *   resize the window - move all of the controls to compensate 
 */
void CHtmlSys_aboutwin::do_resize(int mode, int x, int y)
{
    switch(mode)
    {
    case SIZE_MAXHIDE:
    case SIZE_MAXSHOW:
    case SIZE_MINIMIZED:
        /* don't bother resizing the subwindows on these changes */
        break;

    case SIZE_MAXIMIZED:
    case SIZE_RESTORED:
    default:
        {
            RECT rc;
            
            /* inherit default handling */
            CTadsWin::do_resize(mode, x, y);
            
            /* move the "OK" button */
            GetClientRect(handle_, &rc);
            MoveWindow(okbtn_, (rc.right - okbtn_wid) / 2,
                       rc.bottom - okbtn_ht - 2,
                       okbtn_wid, okbtn_ht, TRUE);

            /* move the main panel */
            MoveWindow(html_subwin_->get_handle(),
                       rc.left, rc.top, rc.right,
                       rc.bottom - okbtn_ht - 4, TRUE);
        }
    }
}

/*
 *   handle a keystroke 
 */
int CHtmlSys_aboutwin::do_char(TCHAR ch, long /*keydata*/)
{
    switch(ch)
    {
    case 10:
    case 13:
    case 27:
        /* close on return/enter/escape */
        closing_ = TRUE;
        return TRUE;

    default:
        /* eat other keys */
        return TRUE;
    }
}

/*
 *   handle a command 
 */
int CHtmlSys_aboutwin::do_command(int notify_code, int cmd, HWND ctl)
{
    /* if it's the "OK" button, set the closing flag */
    if (ctl == okbtn_)
    {
        closing_ = TRUE;
        return TRUE;
    }

    /* ignore other commands */
    return FALSE;
}

/*
 *   erase the background 
 */
void CHtmlSys_aboutwin::do_paint_content(HDC hdc, const RECT *paintrc)
{
    /* fill the background with gray */
    FillRect(hdc, paintrc, GetStockObject(LTGRAY_BRUSH));
}

/* ------------------------------------------------------------------------ */
/*
 *   DirectX warning dialog 
 */

void CTadsDlg_dxwarn::init()
{
    int ck;
    
    /* inherit default */
    CTadsDialog::init();

    /*
     *   Determine if the checkbox should be checked - it should be if the
     *   error code is the same as last time, and the "hide this warning"
     *   item was checked last time. 
     */
    ck = (warn_type_ == prefs_->get_directx_error_code()
          && prefs_->get_directx_hidewarn());
    CheckDlgButton(handle_, IDC_CK_SUPPRESS,
                   ck ? BST_CHECKED : BST_UNCHECKED);

    /* 
     *   Show the appropriate message, depending on the type of error that
     *   caused us to display the dialog 
     */
    switch(warn_type_)
    {
    default:
    case HTMLW32_DIRECTX_MISSING:
        ShowWindow(GetDlgItem(handle_, IDC_STATIC_DXINST), SW_SHOW);
        ShowWindow(GetDlgItem(handle_, IDC_STATIC_DXINST2), SW_SHOW);
        ShowWindow(GetDlgItem(handle_, IDC_STATIC_DXINST3), SW_SHOW);
        break;

    case HTMLW32_DIRECTX_BADVSN:
        ShowWindow(GetDlgItem(handle_, IDC_STATIC_DXVSN), SW_SHOW);
        ShowWindow(GetDlgItem(handle_, IDC_STATIC_DXVSN2), SW_SHOW);
        ShowWindow(GetDlgItem(handle_, IDC_STATIC_DXVSN3), SW_SHOW);
        break;
    }
}

/*
 *   process a command 
 */
int CTadsDlg_dxwarn::do_command(WPARAM id, HWND ctl)
{
    int ck;
    
    switch(id)
    {
    case IDC_CK_SUPPRESS:
        /* note the new value */
        ck = (IsDlgButtonChecked(handle_, IDC_CK_SUPPRESS) == BST_CHECKED);
        
        /* save the new setting in the preferences */
        prefs_->set_directx_hidewarn(ck);

        /* handled */
        return TRUE;

    default:
        /* inherit default handling */
        return CTadsDialog::do_command(id, ctl);
    }
}

